@@ -1,12 +1,12 @@
 /*
  * binary.h - Binary I/O routines
  *
- *   Copyright (c) 2004-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2004-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -32,7 +32,7 @@
  */
 
 #include <gauche.h>
-#include <gauche/builtin-syms.h>
+#include <gauche/priv/builtin-syms.h>
 
 extern ScmObj Scm_ReadBinaryU8(ScmPort *iport, ScmSymbol *e);
 extern ScmObj Scm_ReadBinaryU16(ScmPort *iport, ScmSymbol *e);
