@@ -1,7 +1,7 @@
 /*
  * charconv.c - character code conversion library
  *
- *   Copyright (c) 2000-2012  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -35,6 +35,8 @@
 #include <errno.h>
 #include <gauche.h>
 #include <gauche/extend.h>
+#include "gauche/priv/charP.h"
+#include "gauche/priv/portP.h"
 #include "charconv.h"
 
 #define DEFAULT_CONVERSION_BUFFER_SIZE 1024
@@ -49,18 +51,16 @@ typedef struct conv_guess_rec {
 
 /* anchor of the chain of conversion guessing procedures */
 static struct {
-    int dummy;                  /* trick to place this in .data section */
     conv_guess *procs;
     ScmInternalMutex mutex;
-} guess = { 1 };
+} guess;
 
 /* anchor of the conversion context used for UCS -> internal char routine */
 static struct {
-    int dummy;                  /* trick to place this in .data section */
     ScmConvInfo *ucs2char;
     ScmConvInfo *char2ucs;
     ScmInternalMutex mutex;
-} ucsconv = { 1 };
+} ucsconv;
 
 /*------------------------------------------------------------
  * Query
@@ -144,11 +144,9 @@ static ScmObj conv_name(int dir, ScmPort *remote, const char *from, const char *
  *  <-- Buffered port <--- filler <--(info->buf)--- getz(remote)
  */
 
-static int conv_input_filler(ScmPort *port, int mincnt)
+static ScmSize conv_input_filler(ScmPort *port, ScmSize mincnt SCM_UNUSED)
 {
     ScmConvInfo *info = (ScmConvInfo*)port->src.buf.data;
-    size_t insize, inroom, outroom, result;
-    int nread;
     const char *inbuf = info->buf;
     char *outbuf = port->src.buf.end;
 
@@ -157,14 +155,14 @@ static int conv_input_filler(ScmPort *port, int mincnt)
     /* Fill the input buffer.  There may be some remaining bytes in the
        inbuf from the last conversion (insize), so we try to fill the
        rest. */
-    insize = info->ptr - info->buf;
-    nread = Scm_Getz(info->ptr, info->bufsiz - (int)insize, info->remote);
+    ScmSize insize = info->ptr - info->buf;
+    ScmSize nread = Scm_Getz(info->ptr, info->bufsiz - insize, info->remote);
     if (nread <= 0) {
         /* input reached EOF.  finish the output state */
         if (insize == 0) {
-            outroom = SCM_PORT_BUFFER_ROOM(port);
-            result = jconv_reset(info, outbuf, outroom);
-            if (result < 0) {
+            ScmSize outroom = SCM_PORT_BUFFER_ROOM(port);
+            ScmSize result = jconv_reset(info, outbuf, outroom);
+            if (result == OUTPUT_NOT_ENOUGH) {
                 /* The port buffer doesn't have enough space to contain the
                    finishing sequence.  Its unusual, for the port buffer
                    must be almost empty at this time, and the finishing
@@ -181,20 +179,20 @@ static int conv_input_filler(ScmPort *port, int mincnt)
             fprintf(stderr, "<= r=%d (reset), out(%p)%d\n",
                     result, outbuf, outroom);
 #endif
-            return (int)result;
+            return (ScmSize)result;
         }
     } else {
         insize += nread;
     }
 
     /* Conversion. */
-    inroom = insize;
-    outroom = SCM_PORT_BUFFER_ROOM(port);
+    ScmSize inroom = insize;
+    ScmSize outroom = SCM_PORT_BUFFER_ROOM(port);
 
 #ifdef JCONV_DEBUG
     fprintf(stderr, "=> in(%p)%d out(%p)%d\n", inbuf, insize, outbuf, outroom);
 #endif
-    result = jconv(info, &inbuf, &inroom, &outbuf, &outroom);
+    ScmSize result = jconv(info, &inbuf, &inroom, &outbuf, &outroom);
 #ifdef JCONV_DEBUG
     fprintf(stderr, "<= r=%d, in(%p)%d out(%p)%d\n",
             result, inbuf, inroom, outbuf, outroom);
@@ -207,10 +205,10 @@ static int conv_input_filler(ScmPort *port, int mincnt)
            buffer. */
         memmove(info->buf, info->buf+insize-inroom, inroom);
         info->ptr = info->buf + inroom;
-        return info->bufsiz - (int)outroom;
+        return info->bufsiz - (ScmSize)outroom;
     } else if (result == ILLEGAL_SEQUENCE) {
         /* it's likely that the input contains invalid sequence. */
-        int cnt = inroom >= 6 ? 6 : (int)inroom;
+        ScmSize cnt = inroom >= 6 ? 6 : (ScmSize)inroom;
         ScmObj s = Scm_MakeString(info->buf+insize-inroom, cnt, cnt,
                                   SCM_STRING_COPYING|SCM_STRING_INCOMPLETE);
         Scm_Error("invalid character sequence in the input stream: %S ...", s);
@@ -222,10 +220,10 @@ static int conv_input_filler(ScmPort *port, int mincnt)
     if (inroom > 0) {
         memmove(info->buf, info->buf+insize-inroom, inroom);
         info->ptr = info->buf + inroom;
-        return info->bufsiz - (int)outroom;
+        return info->bufsiz - (ScmSize)outroom;
     } else {
         info->ptr = info->buf;
-        return info->bufsiz - (int)outroom;
+        return info->bufsiz - (ScmSize)outroom;
     }
 }
 
@@ -242,16 +240,12 @@ static void conv_input_closer(ScmPort *p)
 ScmObj Scm_MakeInputConversionPort(ScmPort *fromPort,
                                    const char *fromCode,
                                    const char *toCode,
-                                   ScmObj handler,
-                                   int bufsiz,
+                                   ScmObj handler SCM_UNUSED, /* for now */
+                                   ScmSize bufsiz,
                                    int ownerp)
 {
-    ScmConvInfo *cinfo;
-    conv_guess *guess;
     char *inbuf = NULL;
-    int preread = 0;
-    ScmPortBuffer bufrec;
-    ScmObj name;
+    ScmSize preread = 0;
 
     if (!SCM_IPORTP(fromPort))
         Scm_Error("input port required, but got %S", fromPort);
@@ -260,7 +254,7 @@ ScmObj Scm_MakeInputConversionPort(ScmPort *fromPort,
     if (bufsiz <= MINIMUM_CONVERSION_BUFFER_SIZE) {
         bufsiz = MINIMUM_CONVERSION_BUFFER_SIZE;
     }
-    guess = findGuessingProc(fromCode);
+    conv_guess *guess = findGuessingProc(fromCode);
     if (guess) {
         const char *guessed;
 
@@ -278,7 +272,7 @@ ScmObj Scm_MakeInputConversionPort(ScmPort *fromPort,
         fromCode = guessed;
     }
 
-    cinfo = jconv_open(toCode, fromCode);
+    ScmConvInfo *cinfo = jconv_open(toCode, fromCode);
     if (cinfo == NULL) {
         Scm_Error("conversion from code %s to code %s is not supported",
                   fromCode, toCode);
@@ -295,6 +289,7 @@ ScmObj Scm_MakeInputConversionPort(ScmPort *fromPort,
         cinfo->ptr = cinfo->buf;
     }
 
+    ScmPortBuffer bufrec;
     memset(&bufrec, 0, sizeof(bufrec));
     bufrec.size = cinfo->bufsiz;
     bufrec.buffer = SCM_NEW_ATOMIC2(char *, cinfo->bufsiz);
@@ -306,7 +301,7 @@ ScmObj Scm_MakeInputConversionPort(ScmPort *fromPort,
     bufrec.filenum = conv_fileno;
     bufrec.data = (void*)cinfo;
 
-    name = conv_name(SCM_PORT_INPUT, fromPort, fromCode, toCode);
+    ScmObj name = conv_name(SCM_PORT_INPUT, fromPort, fromCode, toCode);
     return Scm_MakeBufferedPort(SCM_CLASS_PORT, name, SCM_PORT_INPUT, TRUE, &bufrec);
 }
 
@@ -347,7 +342,6 @@ static ScmPort *coding_aware_conv(ScmPort *src, const char *encoding)
 static void conv_output_closer(ScmPort *port)
 {
     ScmConvInfo *info = (ScmConvInfo*)port->src.buf.data;
-    int r;
 
     /* if there's remaining bytes in buf, send them to the remote port. */
     if (info->ptr > info->buf) {
@@ -355,7 +349,7 @@ static void conv_output_closer(ScmPort *port)
         info->ptr = info->buf;
     }
     /* sends out the closing sequence, if any */
-    r = (int)jconv_reset(info, info->buf, info->bufsiz);
+    ScmSize r = jconv_reset(info, info->buf, info->bufsiz);
 #ifdef JCONV_DEBUG
     fprintf(stderr, "<= r=%d(reset), buf(%p)\n",
             r, info->buf);
@@ -365,7 +359,7 @@ static void conv_output_closer(ScmPort *port)
                   info->fromCode, info->toCode);
     }
     if (r > 0) {
-        Scm_Putz(info->buf, r, info->remote);
+        Scm_Putz(info->buf, (ScmSize)r, info->remote);
     }
     /* flush remove port */
     Scm_Flush(info->remote);
@@ -376,26 +370,23 @@ static void conv_output_closer(ScmPort *port)
     jconv_close(info);
 }
 
-static int conv_output_flusher(ScmPort *port, int cnt, int forcep)
+static ScmSize conv_output_flusher(ScmPort *port, ScmSize cnt, int forcep)
 {
     ScmConvInfo *info = (ScmConvInfo*)port->src.buf.data;
-    size_t outsize, inroom, outroom, result, len;
-    const char *inbuf;
-    char *outbuf;
+    ScmSize inroom = SCM_PORT_BUFFER_AVAIL(port);
+    ScmSize len = inroom;
+    const char *inbuf = port->src.buf.buffer;
 
-    inbuf = port->src.buf.buffer;
-    inroom = len = SCM_PORT_BUFFER_AVAIL(port);
     for (;;) {
         /* Conversion. */
-        outbuf = info->ptr;
-        outsize = info->bufsiz - (info->ptr - info->buf);
-        outroom = outsize;
+        char *outbuf = info->ptr;
+        ScmSize outroom = info->bufsiz - (info->ptr - info->buf);
 #ifdef JCONV_DEBUG
         fprintf(stderr, "=> in(%p,%p)%d out(%p,%p)%d\n",
                 inbuf, len, inroom,
                 info->buf, info->ptr, outroom);
 #endif
-        result = jconv(info, &inbuf, &inroom, &outbuf, &outroom);
+        ScmSize result = jconv(info, &inbuf, &inroom, &outbuf, &outroom);
 #ifdef JCONV_DEBUG
         fprintf(stderr, "<= r=%d, in(%p)%d out(%p)%d\n",
                 result, inbuf, inroom, outbuf, outroom);
@@ -410,14 +401,14 @@ static int conv_output_flusher(ScmPort *port, int cnt, int forcep)
 #else
             /* See the above notes.  We always flush the output buffer
                here, so that we can avoid output buffer overrun. */
-            Scm_Putz(info->buf, (int)(outbuf - info->buf), info->remote);
+            Scm_Putz(info->buf, (ScmSize)(outbuf - info->buf), info->remote);
             info->ptr = info->buf;
 #endif
-            return (int)(len - inroom);
+            return (ScmSize)(len - inroom);
         } else if (result == OUTPUT_NOT_ENOUGH) {
             /* Output buffer got full.  Flush it, and continue
                conversion. */
-            Scm_Putz(info->buf, (int)(outbuf - info->buf), info->remote);
+            Scm_Putz(info->buf, (ScmSize)(outbuf - info->buf), info->remote);
             info->ptr = info->buf;
             continue;
         } else if (result == ILLEGAL_SEQUENCE) {
@@ -432,11 +423,11 @@ static int conv_output_flusher(ScmPort *port, int cnt, int forcep)
 #else
             /* See the above notes.  We always flush the output buffer here,
                so that we can avoid output buffer overrun. */
-            Scm_Putz(info->buf, (int)(outbuf - info->buf), info->remote);
+            Scm_Putz(info->buf, (ScmSize)(outbuf - info->buf), info->remote);
             info->ptr = info->buf;
 #endif
             if (forcep && len - inroom != cnt) continue;
-            return (int)(len - inroom);
+            return len - inroom;
         }
     }
 }
@@ -444,12 +435,8 @@ static int conv_output_flusher(ScmPort *port, int cnt, int forcep)
 ScmObj Scm_MakeOutputConversionPort(ScmPort *toPort,
                                     const char *toCode,
                                     const char *fromCode,
-                                    int bufsiz, int ownerp)
+                                    ScmSize bufsiz, int ownerp)
 {
-    ScmConvInfo *cinfo;
-    ScmPortBuffer bufrec;
-    ScmObj name;
-
     if (!SCM_OPORTP(toPort))
         Scm_Error("output port required, but got %S", toPort);
 
@@ -458,7 +445,7 @@ ScmObj Scm_MakeOutputConversionPort(ScmPort *toPort,
         bufsiz = MINIMUM_CONVERSION_BUFFER_SIZE;
     }
 
-    cinfo = jconv_open(toCode, fromCode);
+    ScmConvInfo *cinfo = jconv_open(toCode, fromCode);
     if (cinfo == NULL) {
         Scm_Error("conversion from code %s to code %s is not supported",
                   fromCode, toCode);
@@ -470,6 +457,7 @@ ScmObj Scm_MakeOutputConversionPort(ScmPort *toPort,
     cinfo->buf = SCM_NEW_ATOMIC2(char *, cinfo->bufsiz);
     cinfo->ptr = cinfo->buf;
 
+    ScmPortBuffer bufrec;
     memset(&bufrec, 0, sizeof(bufrec));
     bufrec.size = cinfo->bufsiz;
     bufrec.buffer = SCM_NEW_ATOMIC2(char *, cinfo->bufsiz);
@@ -481,14 +469,14 @@ ScmObj Scm_MakeOutputConversionPort(ScmPort *toPort,
     bufrec.filenum = conv_fileno;
     bufrec.data = (void*)cinfo;
 
-    name = conv_name(SCM_PORT_OUTPUT, toPort, fromCode, toCode);
+    ScmObj name = conv_name(SCM_PORT_OUTPUT, toPort, fromCode, toCode);
     return Scm_MakeBufferedPort(SCM_CLASS_PORT, name, SCM_PORT_OUTPUT, TRUE, &bufrec);
 }
 
 /*------------------------------------------------------------
  * Direct interface for code guessing
  */
-const char *Scm_GuessCES(const char *code, const char *buf, int buflen)
+const char *Scm_GuessCES(const char *code, const char *buf, ScmSize buflen)
 {
     conv_guess *guess = findGuessingProc(code);
     if (guess == NULL)
@@ -515,14 +503,13 @@ static ScmChar ucstochar(int ucs4)
     char inbuf[6], outbuf[6];
     const char *inb = inbuf;
     char *outb = outbuf;
-    size_t inroom, outroom, r;
 
     if (ucsconv.ucs2char == NULL) return SCM_CHAR_INVALID;
-    inroom = UCS2UTF_NBYTES(ucs4);
-    outroom = 6;
+    ScmSize inroom = UCS2UTF_NBYTES(ucs4);
+    ScmSize outroom = 6;
     jconv_ucs4_to_utf8(ucs4, inbuf);
     (void)SCM_INTERNAL_MUTEX_LOCK(ucsconv.mutex);
-    r = jconv(ucsconv.ucs2char, &inb, &inroom, &outb, &outroom);
+    ScmSize r = jconv(ucsconv.ucs2char, &inb, &inroom, &outb, &outroom);
     (void)SCM_INTERNAL_MUTEX_UNLOCK(ucsconv.mutex);
     if (r == INPUT_NOT_ENOUGH || r == OUTPUT_NOT_ENOUGH) {
         Scm_Error("can't convert UCS4 code %d to a character: implementation problem?", ucs4);
@@ -546,15 +533,14 @@ static int chartoucs(ScmChar ch)
     char inbuf[6], outbuf[6];
     const char *inb = inbuf;
     char *outb = outbuf;
-    size_t inroom, outroom, r;
 
     if (ch == SCM_CHAR_INVALID) return -1;
     if (ucsconv.char2ucs == NULL) return -1;
-    inroom = SCM_CHAR_NBYTES(ch);
-    outroom = 6;
+    ScmSize inroom = SCM_CHAR_NBYTES(ch);
+    ScmSize outroom = 6;
     SCM_CHAR_PUT(inbuf, ch);
     (void)SCM_INTERNAL_MUTEX_LOCK(ucsconv.mutex);
-    r = jconv(ucsconv.char2ucs, &inb, &inroom, &outb, &outroom);
+    ScmSize r = jconv(ucsconv.char2ucs, &inb, &inroom, &outb, &outroom);
     (void)SCM_INTERNAL_MUTEX_UNLOCK(ucsconv.mutex);
     if (r == INPUT_NOT_ENOUGH || r == OUTPUT_NOT_ENOUGH) {
         Scm_Error("can't convert character %u to UCS4 code: implementation problem?", ch);
@@ -606,9 +592,7 @@ extern void Scm_Init_convguess(void);
 
 SCM_EXTENSION_ENTRY void Scm_Init_gauche__charconv(void)
 {
-    ScmModule *mod;
     SCM_INIT_EXTENSION(gauche__charconv);
-    mod = SCM_FIND_MODULE("gauche.charconv", SCM_FIND_MODULE_CREATE);
     guess.procs = NULL;
     (void)SCM_INTERNAL_MUTEX_INIT(guess.mutex);
 #if   defined(GAUCHE_CHAR_ENCODING_UTF_8)
