@@ -1,23 +1,23 @@
 ;;;
 ;;; convaux - auxiliary charconv routines
-;;;  
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;;   
+;;;
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;;   
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
-;;;  
+;;;
 ;;;   2. Redistributions in binary form must reproduce the above copyright
 ;;;      notice, this list of conditions and the following disclaimer in the
 ;;;      documentation and/or other materials provided with the distribution.
-;;;  
+;;;
 ;;;   3. Neither the name of the authors nor the names of its contributors
 ;;;      may be used to endorse or promote products derived from this
 ;;;      software without specific prior written permission.
-;;;  
+;;;
 ;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 ;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 ;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -29,11 +29,14 @@
 ;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-;;;  
+;;;
 
 (select-module gauche.charconv)
 (use srfi-1)
 (use srfi-13)
+(use gauche.sequence)
+(autoload gauche.vport
+          open-input-uvector open-output-uvector get-output-uvector)
 
 ;; Determine charset compatibility.  (ces-equivalent? a b) is true if CES a and
 ;; CES b refer to the same CES.
@@ -85,84 +88,105 @@
 
     ;; find ces entry
     (define (find-entry ces)
-      (find (lambda (e) (memq ces (car e))) ces-compatibility-table))
+      (find (^e (memq ces (car e))) ces-compatibility-table))
 
-    (define (ces-equivalent? a b . unknown-fallback)
-      (let* ((ces-a   (canon-name a))
-             (ces-b   (canon-name b))
-             (entry-a (find-entry ces-a))
-             (entry-b (find-entry ces-b)))
-        (cond ((or (eq? ces-a 'none) (eq? ces-b 'none)) #t)
-              ((or (not entry-a) (not entry-b))
-               (get-optional unknown-fallback #f))
-              (else (eq? entry-a entry-b)))))
+    (define (ces-equivalent? a b :optional (unknown-fallback #f))
+      (let* ([ces-a   (canon-name a)]
+             [ces-b   (canon-name b)]
+             [entry-a (find-entry ces-a)]
+             [entry-b (find-entry ces-b)])
+        (cond [(or (eq? ces-a 'none) (eq? ces-b 'none)) #t]
+              [(or (not entry-a) (not entry-b))
+               unknown-fallback]
+              [else (eq? entry-a entry-b)])))
+
+    (define (ces-upper-compatible? a b :optional (unknown-fallback #f))
+      (let* ([ces-a   (canon-name a)]
+             [ces-b   (canon-name b)]
+             [entry-a (find-entry ces-a)])
+        (cond [(or (eq? ces-a 'none) (eq? ces-b 'none)) #t]
+              [(or (not entry-a) (not (find-entry ces-b)))
+	       unknown-fallback]
+              [else (let loop ([entry entry-a])
+                      (or (boolean (memq ces-b (car entry)))
+                          (any loop (map find-entry (cdr entry)))))])))
 
-    (define (ces-upper-compatible? a b . unknown-fallback)
-      (let* ((ces-a   (canon-name a))
-             (ces-b   (canon-name b))
-             (entry-a (find-entry ces-a)))
-        (cond ((or (eq? ces-a 'none) (eq? ces-b 'none)) #t)
-              ((or (not entry-a) (not (find-entry ces-b)))
-               (get-optional unknown-fallback #f))
-              (else 
-               (let loop ((entry entry-a))
-                 (if (memq ces-b (car entry))
-                     #t
-                     (any loop (map find-entry (cdr entry)))))))))
-    
     (values ces-equivalent? ces-upper-compatible?)))
 
-;; Convert string
-(define (ces-convert string fromcode . args)
-  (let-optionals* args ((tocode #f))
-    (let ((out (open-output-string :private? #t))
-          (in  (open-input-conversion-port
-                (open-input-string string :private? #t)
-                fromcode
-                :to-code tocode :buffer-size (string-size string) :owner? #t)))
-      (copy-port in out :unit 'byte)
-      (close-input-port in)
-      (begin0
-       (get-output-string out)
-       (close-output-port out)))))
+;; returns appropriate input port and size
+(define (%ces-input input)
+  (cond [(string? input)
+         (values (open-input-string input :private? #t) (string-length input))]
+        ;; avoid using u8vector? so that we don't depend on gauche.uvector
+        [(is-a? input <u8vector>)
+         (values (open-input-uvector input) (uvector-length input))]
+        [else (error "string or u8vector required, but got:" input)]))
+
+;; returns appropriate output port and get-output-*
+(define (%ces-output class)
+  (cond
+   [(eq? class <string>)
+    (values (open-output-string :private? #t) get-output-string)]
+   [(eq? class <u8vector>)
+    (values (open-output-uvector) get-output-uvector)]
+   [else (error "Only <string> or <u8vector> is supported, but got:" class)]))
+
+;; Convert string or uvector -> string or uvector
+(define (ces-convert-to class input fromcode :optional (tocode #f))
+  (receive (inp isize) (%ces-input input)
+    (receive (out get) (%ces-output class)
+      (let1 in ($ open-input-conversion-port inp fromcode
+                  :to-code tocode :buffer-size isize :owner? #t)
+        (copy-port in out :unit 'byte)
+        (close-input-port in)
+        (flush out)
+        (begin0 (get out)
+          (close-output-port out))))))
+
+(define (ces-convert input fromcode :optional (tocode #f))
+  (ces-convert-to <string> input fromcode tocode))
 
 ;; "Wrap" the given port for convering to/from native encoding if needed.
 ;; Unlike open-*-conversion-port, these return port itself if the conversion
 ;; is not required.
-(define (wrap-with-input-conversion port from-code . opts)
-  (let1 to-code (get-keyword :to-code opts (gauche-character-encoding))
-    (if (ces-upper-compatible? to-code from-code)
-      port
-      (apply open-input-conversion-port port from-code :owner? #t opts))))
+(define (wrap-with-input-conversion port from-code
+                                    :key (to-code (gauche-character-encoding))
+                                    :allow-other-keys
+                                    :rest opts)
+  (if (ces-upper-compatible? to-code from-code)
+    port
+    (apply open-input-conversion-port port from-code :owner? #t opts)))
 
-(define (wrap-with-output-conversion port to-code . opts)
-  (let1 from-code (get-keyword :from-code opts (gauche-character-encoding))
-    (if (ces-upper-compatible? from-code to-code)
-      port
-      (apply open-output-conversion-port port to-code :owner? #t opts))))
+(define (wrap-with-output-conversion port to-code
+                                     :key (from-code (gauche-character-encoding))
+                                     :allow-other-keys
+                                     :rest opts)
+  (if (ces-upper-compatible? from-code to-code)
+    port
+    (apply open-output-conversion-port port to-code :owner? #t opts)))
 
 ;; Call with conversion port
-(define (call-with-input-conversion port proc . opts)
-  (let-keywords opts ((from-code :encoding (gauche-character-encoding))
-                      (bufsiz    :conversion-buffer-size 0))
-    (if (ces-upper-compatible? (gauche-character-encoding) from-code)
-      (proc port)
-      (let1 cvp (open-input-conversion-port port from-code
-                                            :owner? #f :buffer-size bufsiz)
-        (unwind-protect
-            (proc cvp)
-          (close-input-port cvp))))))
+(define (call-with-input-conversion port proc
+                                    :key ((:encoding from-code)
+                                          (gauche-character-encoding))
+                                         ((:conversion-buffer-size bufsiz) 0))
+  (if (ces-upper-compatible? (gauche-character-encoding) from-code)
+    (proc port)
+    (let1 cvp (open-input-conversion-port port from-code
+                                          :owner? #f :buffer-size bufsiz)
+      (unwind-protect (proc cvp)
+        (close-input-port cvp)))))
 
-(define (call-with-output-conversion port proc . opts)
-  (let-keywords opts ((to-code :encoding (gauche-character-encoding))
-                      (bufsiz  :conversion-buffer-size 0))
-    (if (ces-upper-compatible? (gauche-character-encoding) to-code)
-      (proc port)
-      (let1 cvp (open-output-conversion-port port to-code
-                                             :owner? #f :buffer-size bufsiz)
-        (unwind-protect
-            (proc cvp)
-          (close-output-port cvp))))))
+(define (call-with-output-conversion port proc
+                                     :key ((:encoding to-code)
+                                           (gauche-character-encoding))
+                                          ((:conversion-buffer-size bufsiz) 0))
+  (if (ces-upper-compatible? (gauche-character-encoding) to-code)
+    (proc port)
+    (let1 cvp (open-output-conversion-port port to-code
+                                           :owner? #f :buffer-size bufsiz)
+      (unwind-protect (proc cvp)
+        (close-output-port cvp)))))
 
 (define (with-input-conversion port thunk . opts)
   (apply call-with-input-conversion port
@@ -176,35 +200,34 @@
 
 ;; Inserts conversion port.  These are called from system's
 ;; open-{input|output}-port when :encoding argument is given.
-(define (%open-input-file/conv name . args)
-  (let-keywords args ((encoding #f)
-                      (bufsiz :conversion-buffer-size 0)
-                      . rest)
-    (and-let* ((port (apply %open-input-file name rest)))
-      (wrap-with-input-conversion port encoding
-       :buffer-size bufsiz
-       :owner? #t))))
+(define (%open-input-file/conv name :key (encoding #f)
+                                         ((:conversion-buffer-size bufsiz) 0)
+                                    :allow-other-keys rest)
+  (and-let* ([port (apply (with-module gauche.internal %open-input-file)
+                          name rest)])
+    (wrap-with-input-conversion port encoding
+                                :buffer-size bufsiz :owner? #t)))
 
-(define (%open-output-file/conv name . args)
-  (let-keywords args ((encoding #f)
-                      (bufsiz :conversion-buffer-size 0)
-                      . rest)
-    (and-let* ((port (apply %open-output-file name rest)))
-      (wrap-with-output-conversion port encoding
-       :buffer-size bufsiz
-       :owner? #t))))
+(define (%open-output-file/conv name :key (encoding #f)
+                                          ((:conversion-buffer-size bufsiz) 0)
+                                     :allow-other-keys rest)
+  (and-let* ([port (apply (with-module gauche.internal %open-output-file)
+                          name rest)])
+    (wrap-with-output-conversion port encoding
+                                 :buffer-size bufsiz :owner? #t)))
 
 ;;
 ;; Low-level API
 ;;
 
 (inline-stub
- "#include \"charconv.h\""
- 
+ (declcode
+  (.include "charconv.h"))
+
  (define-cproc ces-conversion-supported? (from to) ::<boolean>
    (let* ([cfrom::(const char*) (Scm_GetCESName from "from-code")]
           [cto  ::(const char*) (Scm_GetCESName to "to-code")])
-     (result (Scm_ConversionSupportedP cfrom cto))))
+     (return (Scm_ConversionSupportedP cfrom cto))))
 
  ;; NB: :handler interface is experimental.  Do not use it.
  (define-cproc open-input-conversion-port (source::<input-port>
@@ -215,7 +238,7 @@
                                                 (handler #f))
    (let* ([fc::(const char*) (Scm_GetCESName from_code "from-code")]
           [tc::(const char*) (Scm_GetCESName to_code "to-code")])
-     (result (Scm_MakeInputConversionPort source fc tc handler buffer_size
+     (return (Scm_MakeInputConversionPort source fc tc handler buffer_size
                                           (not (SCM_FALSEP ownerP))))))
 
  (define-cproc open-output-conversion-port (sink::<output-port>
@@ -225,17 +248,17 @@
                                                  (owner? #f))
    (let* ([fc::(const char*) (Scm_GetCESName from_code "from-code")]
           [tc::(const char*) (Scm_GetCESName to_code "to-code")])
-     (result (Scm_MakeOutputConversionPort sink tc fc buffer_size
+     (return (Scm_MakeOutputConversionPort sink tc fc buffer_size
                                            (not (SCM_FALSEP ownerP))))))
 
  (define-cproc ces-guess-from-string (string::<string> scheme::<string>)
-   (let* ([size::u_int]
+   (let* ([size::ScmSmallInt]
           [s::(const char*) (Scm_GetStringContent string (& size) NULL NULL)]
           [guessed::(const char*)
                     (Scm_GuessCES (Scm_GetStringConst scheme) s size)])
      (if guessed
-       (result (SCM_MAKE_STR guessed))
-       (result '#f))))
+       (return (SCM_MAKE_STR guessed))
+       (return '#f))))
  )
 
 
