@@ -1,12 +1,12 @@
 /*
  * jconv.c - alternative japanese code conversion routines
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -93,7 +93,7 @@
  */
 
 /* Shift_JISX0213 -> EUC-JP
- * 
+ *
  * Mapping anormalities
  *
  *   0x5c, 0x7e : Shift_JISX0213 mapping table maps 0x5c to U+00A5
@@ -144,13 +144,14 @@
  *     JIS X 0213 to EUC-JP is a straightfoward conversion.
  */
 
-static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                        char *outptr, size_t outroom, size_t *outchars)
+static ScmSize sjis2eucj(ScmConvInfo *cinfo SCM_UNUSED,
+                         const char *inptr, ScmSize inroom,
+                         char *outptr, ScmSize outroom, 
+                         ScmSize *outchars)
 {
-    unsigned char s1, s2;
-    static unsigned char cvt[] = { 0xa1, 0xa8, 0xa3, 0xa4, 0xa5, 0xac, 0xae, 0xad, 0xaf, 0xee };
+    static const unsigned char cvt[] = { 0xa1, 0xa8, 0xa3, 0xa4, 0xa5, 0xac, 0xae, 0xad, 0xaf, 0xee };
 
-    s1 = inptr[0];
+    unsigned char s1 = inptr[0];
     if (s1 < 0x7f) {
         *outptr = s1;
         *outchars = 1;
@@ -160,7 +161,7 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         /* Double byte char */
         unsigned char m, e1, e2;
         INCHK(2);
-        s2 = inptr[1];
+        unsigned char s2 = inptr[1];
         if (s2 < 0x40 || s2 > 0xfc) {
             EUCJ_SUBST;
             return 2;
@@ -183,7 +184,7 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
             m = 2;
             e1 = cvt[(s1-0xf0)*2+((s2 < 0x9f)? 1 : 0)];
         }
-        
+
         if (s2 < 0x7f) {
             e2 = s2 - 0x3f + 0xa0;
         } else if (s2 < 0x9f) {
@@ -236,7 +237,7 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         *outchars = 2;
         return 1;
     }
-    
+
     /* s1 == 0x80 or 0xa0 */
     outptr[0] = SUBST1_CHAR;
     *outchars = 1;
@@ -244,7 +245,7 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
 }
 
 /* EUC_JISX0213 -> Shift_JIS
- * 
+ *
  * Mapping anormalities
  *
  *   0x80--0xa0 except 0x8e and 0x8f : C1 region.
@@ -264,7 +265,7 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
  *   For double or trible-byte character, subsequent byte has to be in
  *   the range between 0xa1 and 0xfe inclusive.  If not, it is replaced
  *   for the substitution character.
- *   
+ *
  *   If the first byte is in the range of 0xa1--0xfe, two bytes (e1, e2)
  *   is mapped to SJIS (s1, s2) by:
  *
@@ -287,11 +288,12 @@ static size_t sjis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
  *     s2 is mapped with the same rule above.
  */
 
-static size_t eucj2sjis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                        char *outptr, size_t outroom, size_t *outchars)
+static ScmSize eucj2sjis(ScmConvInfo *cinfo SCM_UNUSED,
+                         const char *inptr, ScmSize inroom,
+                         char *outptr, ScmSize outroom,
+                         ScmSize *outchars)
 {
-    unsigned char e1, e2;
-    e1 = inptr[0];
+    unsigned char e1 = inptr[0];
     if (e1 <= 0x7f) {
         outptr[0] = e1;
         *outchars = 1;
@@ -301,7 +303,7 @@ static size_t eucj2sjis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         /* double byte char (JISX 0213 plane 1) */
         unsigned char s1, s2;
         INCHK(2);
-        e2 = inptr[1];
+        unsigned char e2 = inptr[1];
         if (e2 < 0xa1 || e2 == 0xff) {
             SJIS_SUBST;
             return 2;
@@ -323,7 +325,7 @@ static size_t eucj2sjis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
     if (e1 == 0x8e) {
         /* JISX 0201 kana */
         INCHK(2);
-        e2 = inptr[1];
+        unsigned char e2 = inptr[1];
         if (e2 < 0xa1 || e2 == 0xff) {
             outptr[0] = SUBST1_CHAR;
         } else {
@@ -335,12 +337,12 @@ static size_t eucj2sjis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
     if (e1 == 0x8f) {
         /* triple byte char */
         unsigned char s1, s2;
-        unsigned char cvt[] = { 0xf0, 0, 0xf1, 0xf1, 0xf2, 0, 0, 0xf0, 0, 0, 0, 0xf2, 0xf3, 0xf3, 0xf4 };
-        
+        static const unsigned char cvt[] = { 0xf0, 0, 0xf1, 0xf1, 0xf2, 0, 0, 0xf0, 0, 0, 0, 0xf2, 0xf3, 0xf3, 0xf4 };
+
         INCHK(3);
         OUTCHK(2);
         e1 = inptr[1];
-        e2 = inptr[2];
+        unsigned char e2 = inptr[2];
         if (e1 < 0xa1 || e1 == 0xff || e2 < 0xa1 || e2 == 0xff) {
             SJIS_SUBST;
             return 3;
@@ -441,7 +443,11 @@ static size_t eucj2sjis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
 #include "ucs2eucj.c"
 
 /* Emit given euc char */
-static inline size_t utf2euc_emit_euc(unsigned short euc, size_t inchars, char *outptr, size_t outroom, size_t *outchars)
+static inline ScmSize utf2euc_emit_euc(unsigned short euc, 
+                                       ScmSize inchars,
+                                       char *outptr, 
+                                       ScmSize outroom,
+                                       ScmSize *outchars)
 {
     if (euc == 0) {
         EUCJ_SUBST;
@@ -461,15 +467,15 @@ static inline size_t utf2euc_emit_euc(unsigned short euc, size_t inchars, char *
 }
 
 /* handle 2-byte UTF8 sequence.  0xc0 <= u0 <= 0xdf */
-static inline size_t utf2euc_2(ScmConvInfo *cinfo, unsigned char u0,
-                               const char *inptr, size_t inroom,
-                               char *outptr, size_t outroom, size_t *outchars)
+static inline ScmSize utf2euc_2(ScmConvInfo *cinfo SCM_UNUSED, unsigned char u0,
+                                const char *inptr, ScmSize inroom,
+                                char *outptr, ScmSize outroom, 
+                                ScmSize *outchars)
 {
-    unsigned char u1;
-    unsigned short *etab = NULL;
-    
+    const unsigned short *etab = NULL;
+
     INCHK(2);
-    u1 = (unsigned char)inptr[1];
+    unsigned char u1 = (unsigned char)inptr[1];
     if (u1 < 0x80 || u1 >= 0xc0) return ILLEGAL_SEQUENCE;
 
     switch (u0) {
@@ -506,18 +512,18 @@ static inline size_t utf2euc_2(ScmConvInfo *cinfo, unsigned char u0,
 }
 
 /* handle 3-byte UTF8 sequence.  0xe0 <= u0 <= 0xef */
-static inline size_t utf2euc_3(ScmConvInfo *cinfo, unsigned char u0,
-                               const char *inptr, size_t inroom,
-                               char *outptr, size_t outroom, size_t *outchars)
+static inline ScmSize utf2euc_3(ScmConvInfo *cinfo SCM_UNUSED, unsigned char u0,
+                                const char *inptr, ScmSize inroom,
+                                char *outptr, ScmSize outroom, 
+                                ScmSize *outchars)
 {
-    unsigned char u1, u2;
-    unsigned char *tab1 = NULL;
-    unsigned short (*tab2)[64] = NULL;
+    const unsigned char *tab1 = NULL;
+    const unsigned short (*tab2)[64] = NULL;
 
     INCHK(3);
-    u1 = (unsigned char)inptr[1];
-    u2 = (unsigned char)inptr[2];
-    
+    unsigned char u1 = (unsigned char)inptr[1];
+    unsigned char u2 = (unsigned char)inptr[2];
+
     switch (u0) {
     case 0xe1: /* special case : there's only 6 chars */
         {
@@ -556,22 +562,22 @@ static inline size_t utf2euc_3(ScmConvInfo *cinfo, unsigned char u0,
 }
 
 /* handle 4-byte UTF8 sequence.  u0 == 0xf0, 0xa0 <= u1 <= 0xaa */
-static inline size_t utf2euc_4(ScmConvInfo *cinfo, unsigned char u0,
-                               const char *inptr, size_t inroom,
-                               char *outptr, size_t outroom, size_t *outchars)
+static inline ScmSize utf2euc_4(ScmConvInfo *cinfo SCM_UNUSED, unsigned char u0,
+                                const char *inptr, ScmSize inroom,
+                                char *outptr, ScmSize outroom,
+                                ScmSize *outchars)
 {
-    unsigned char u1, u2, u3;
-    unsigned short *tab = NULL;
+    const unsigned short *tab = NULL;
 
     INCHK(4);
     if (u0 != 0xf0) {
         EUCJ_SUBST;
         return 4;
     }
-    u1 = (unsigned char)inptr[1];
-    u2 = (unsigned char)inptr[2];
-    u3 = (unsigned char)inptr[3];
-    
+    unsigned char u1 = (unsigned char)inptr[1];
+    unsigned char u2 = (unsigned char)inptr[2];
+    unsigned char u3 = (unsigned char)inptr[3];
+
     switch (u1) {
     case 0xa0: tab = utf2euc_f0_a0; break;
     case 0xa1: tab = utf2euc_f0_a1; break;
@@ -588,9 +594,8 @@ static inline size_t utf2euc_4(ScmConvInfo *cinfo, unsigned char u0,
         break;
     }
     if (tab != NULL) {
-        int i;
         unsigned short u2u3 = u2*256 + u3;
-        for (i=0; tab[i]; i+=2) {
+        for (int i=0; tab[i]; i+=2) {
             if (tab[i] == u2u3) {
                 return utf2euc_emit_euc(tab[i+1], 4, outptr, outroom, outchars);
             }
@@ -601,11 +606,12 @@ static inline size_t utf2euc_4(ScmConvInfo *cinfo, unsigned char u0,
 }
 
 /* Body of UTF8 -> EUC_JP conversion */
-static size_t utf2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                       char *outptr, size_t outroom, size_t *outchars)
+static ScmSize utf2eucj(ScmConvInfo *cinfo,     
+                        const char *inptr, ScmSize inroom,
+                        char *outptr, ScmSize outroom,
+                        ScmSize *outchars)
 {
-    unsigned char u0;
-    u0 = (unsigned char)inptr[0];
+    unsigned char u0 = (unsigned char)inptr[0];
 
     if (u0 <= 0x7f) {
         *outptr = u0;
@@ -703,9 +709,9 @@ void jconv_ucs4_to_utf8(unsigned int ucs, char *cp)
 /* Given 'encoded' ucs, emit utf8.  'Encoded' ucs is the entry of the
    conversion table.  If ucs >= 0x100000, it is composed by two UCS2
    character.  Otherwise, it is one UCS4 character. */
-static inline size_t eucj2utf_emit_utf(unsigned int ucs, size_t inchars,
-                                       char *outptr, size_t outroom,
-                                       size_t *outchars)
+static inline ScmSize eucj2utf_emit_utf(unsigned int ucs, ScmSize inchars,
+                                        char *outptr, ScmSize outroom,
+                                        ScmSize *outchars)
 {
     if (ucs == 0) {
         UTF8_SUBST;
@@ -728,29 +734,27 @@ static inline size_t eucj2utf_emit_utf(unsigned int ucs, size_t inchars,
     return inchars;
 }
 
-static size_t eucj2utf(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                       char *outptr, size_t outroom, size_t *outchars)
+static ScmSize eucj2utf(ScmConvInfo *cinfo SCM_UNUSED,
+                        const char *inptr, ScmSize inroom,
+                        char *outptr, ScmSize outroom, ScmSize *outchars)
 {
-    unsigned char e0, e1, e2;
-    unsigned int ucs;
-    
-    e0 = (unsigned char)inptr[0];
+    unsigned char e0 = (unsigned char)inptr[0];
     if (e0 < 0xa0) {
         if (e0 == 0x8e) {
             /* JIS X 0201 KANA */
             INCHK(2);
-            e1 = (unsigned char)inptr[1];
+            unsigned char e1 = (unsigned char)inptr[1];
             if (e1 < 0xa1 || e1 > 0xdf) return ILLEGAL_SEQUENCE;
-            ucs = 0xff61 + (e1 - 0xa1);
+            unsigned int ucs = 0xff61 + (e1 - 0xa1);
             return eucj2utf_emit_utf(ucs, 2, outptr, outroom, outchars);
         }
         else if (e0 == 0x8f) {
             /* JIS X 0213 plane 2 */
             int index;
-            
+
             INCHK(3);
-            e1 = (unsigned char)inptr[1];
-            e2 = (unsigned char)inptr[2];
+            unsigned char e1 = (unsigned char)inptr[1];
+            unsigned char e2 = (unsigned char)inptr[2];
             if (e1 < 0xa1 || e1 > 0xfe || e2 < 0xa1 || e2 > 0xfe) {
                 return ILLEGAL_SEQUENCE;
             }
@@ -759,7 +763,7 @@ static size_t eucj2utf(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
                 UTF8_SUBST;
                 return 3;
             }
-            ucs = euc_jisx0213_2_to_ucs2[index][e2 - 0xa1];
+            unsigned int ucs = euc_jisx0213_2_to_ucs2[index][e2 - 0xa1];
             return eucj2utf_emit_utf(ucs, 3, outptr, outroom, outchars);
         }
         else {
@@ -772,9 +776,9 @@ static size_t eucj2utf(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
     if (e0 > 0xa0 && e0 < 0xff) {
         /* JIS X 0213 plane 1 */
         INCHK(2);
-        e1 = (unsigned char)inptr[1];
+        unsigned char e1 = (unsigned char)inptr[1];
         if (e1 < 0xa1 || e1 > 0xfe) return ILLEGAL_SEQUENCE;
-        ucs = euc_jisx0213_1_to_ucs2[e0 - 0xa1][e1 - 0xa1];
+        unsigned int ucs = euc_jisx0213_1_to_ucs2[e0 - 0xa1][e1 - 0xa1];
         return eucj2utf_emit_utf(ucs, 2, outptr, outroom, outchars);
     }
     return ILLEGAL_SEQUENCE;
@@ -811,7 +815,7 @@ static size_t eucj2utf(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
  *  <ESC> $ ( C   (KS X 1001:1992) unsupported
  *  <ESC> . A     (ISO8859-1:1998) unsupported
  *  <ESC> . F     (ISO8859-7:1998) unsupported
- * 
+ *
  * If other escape sequence is seen, the converter returns ILLEGAL_SEQUENCE.
  *
  * JIS8 kana is allowed.
@@ -832,12 +836,11 @@ enum {
 /* deal with escape sequence.  escape byte itself is already consumed.
    returns # of input bytes consumed by the escape sequence,
    or an error code.  cinfo->istate is updated accordingly. */
-static size_t jis_esc(ScmConvInfo *cinfo, const char *inptr, size_t inroom)
+static ScmSize jis_esc(ScmConvInfo *cinfo, const char *inptr, ScmSize inroom)
 {
-    unsigned char j1, j2;
     INCHK(2);
-    j1 = inptr[0];
-    j2 = inptr[1];
+    unsigned char j1 = inptr[0];
+    unsigned char j2 = inptr[1];
     switch (j1) {
     case '(':
         switch (j2) {
@@ -892,17 +895,16 @@ static size_t jis_esc(ScmConvInfo *cinfo, const char *inptr, size_t inroom)
 }
 
 /* main routine for iso2022-jp -> euc_jp */
-static size_t jis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                       char *outptr, size_t outroom, size_t *outchars)
+static ScmSize jis2eucj(ScmConvInfo *cinfo, const char *inptr, ScmSize inroom,
+                        char *outptr, ScmSize outroom, ScmSize *outchars)
 {
-    unsigned char j0, j1;
-    size_t inoffset = 0, r;
-    
-    j0 = inptr[inoffset];
+    ScmSize inoffset = 0;
+
+    unsigned char j0 = inptr[inoffset];
     /* skip escape sequence */
     while (j0 == 0x1b) {
         inoffset++;
-        r = jis_esc(cinfo, inptr+inoffset, inroom-inoffset);
+        ScmSize r = jis_esc(cinfo, inptr+inoffset, inroom-inoffset);
         if (ERRP(r)) return r;
         inoffset += r;
         if (inoffset >= inroom) {
@@ -911,7 +913,7 @@ static size_t jis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         }
         j0 = inptr[inoffset];
     }
-    
+
     if (j0 == '\n' || j0 == '\r') {
         cinfo->istate = JIS_ASCII;
         outptr[0] = j0;
@@ -947,27 +949,29 @@ static size_t jis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         case JIS_78:
             /* for now, I ignore the difference between JIS78 and JIS83 */
             /* FALLTHROUGH */
-        case JIS_0213_1:
+        case JIS_0213_1: {
             INCHK(inoffset+2);
             OUTCHK(2);
-            j1 = inptr[inoffset+1];
+            unsigned char j1 = inptr[inoffset+1];
             outptr[0] = j0 + 0x80;
             outptr[1] = j1 + 0x80;
             *outchars = 2;
             return 2+inoffset;
+        }
         case JIS_0212:
             /* jis x 0212 and jis x 0213 plane 2 are different character sets,
                but uses the same conversion scheme. */
             /* FALLTHROUGH */
-        case JIS_0213_2:
+        case JIS_0213_2: {
             INCHK(inoffset+2);
             OUTCHK(3);
-            j1 = inptr[inoffset+1];
+            unsigned char j1 = inptr[inoffset+1];
             outptr[0] = 0x8f;
             outptr[1] = j0 + 0x80;
             outptr[2] = j1 + 0x80;
             *outchars = 3;
             return 2+inoffset;
+        }
         case JIS_UNKNOWN:
             outptr[0] = SUBST1_CHAR;
             *outchars = 1;
@@ -987,11 +991,12 @@ static size_t jis2eucj(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
 
 /* ensure the current state is newstate.  returns # of output chars.
    may return OUTPUT_NOT_ENOUGH. */
-static size_t jis_ensure_state(ScmConvInfo *cinfo, int newstate, size_t outbytes,
-                               char *outptr, size_t outroom)
+static ScmSize jis_ensure_state(ScmConvInfo *cinfo, int newstate, 
+                                ScmSize outbytes, 
+                                char *outptr, ScmSize outroom)
 {
     const char *escseq = NULL;
-    size_t esclen = 0;
+    ScmSize esclen = 0;
 
     if (cinfo->ostate == newstate) {
         OUTCHK(outbytes);
@@ -1019,23 +1024,21 @@ static size_t jis_ensure_state(ScmConvInfo *cinfo, int newstate, size_t outbytes
     return esclen;
 }
 
-static size_t eucj2jis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                       char *outptr, size_t outroom, size_t *outchars)
+static ScmSize eucj2jis(ScmConvInfo *cinfo, const char *inptr, ScmSize inroom,
+                        char *outptr, ScmSize outroom, ScmSize *outchars)
 {
-    unsigned char e0, e1;
-    size_t outoffset = 0;
-    e0 = inptr[0];
+    unsigned char e0 = inptr[0];
     if (e0 < 0x80) {
-        outoffset = jis_ensure_state(cinfo, JIS_ASCII, 1, outptr, outroom);
+        ScmSize outoffset = jis_ensure_state(cinfo, JIS_ASCII, 1, outptr, outroom);
         if (ERRP(outoffset)) return outoffset;
         outptr[outoffset] = e0;
         *outchars = outoffset+1;
         return 1;
     } else if (e0 == 0x8e) {
         INCHK(2);
-        e1 = inptr[1];
+        unsigned char e1 = inptr[1];
         if (e1 > 0xa0 && e1 < 0xff) {
-            outoffset = jis_ensure_state(cinfo, JIS_KANA, 1, outptr, outroom);
+            ScmSize outoffset = jis_ensure_state(cinfo, JIS_KANA, 1, outptr, outroom);
             if (ERRP(outoffset)) return outoffset;
             outptr[outoffset] = e1 - 0x80;
             *outchars = outoffset+1;
@@ -1044,7 +1047,7 @@ static size_t eucj2jis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
     } else if (e0 == 0x8f) {
         INCHK(3);
         e0 = inptr[1];
-        e1 = inptr[2];
+        unsigned char e1 = inptr[2];
         if (e0 > 0xa0 && e0 < 0xff && e1 > 0xa0 && e1 < 0xff) {
             int newstate = JIS_0212;
             switch (e0) {
@@ -1054,7 +1057,7 @@ static size_t eucj2jis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
             default:
                 if (e0 >= 0xee) newstate = JIS_0213_2;
             }
-            outoffset = jis_ensure_state(cinfo, newstate, 2, outptr, outroom);
+            ScmSize outoffset = jis_ensure_state(cinfo, newstate, 2, outptr, outroom);
             outptr[outoffset] = e0 - 0x80;
             outptr[outoffset+1] = e1 - 0x80;
             *outchars = outoffset+1;
@@ -1062,9 +1065,9 @@ static size_t eucj2jis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
         }
     } else if (e0 > 0xa0 && e0 < 0xff) {
         INCHK(2);
-        e1 = inptr[1];
+        unsigned char e1 = inptr[1];
         if (e1 > 0xa0 && e1 < 0xff) {
-            outoffset = jis_ensure_state(cinfo, JIS_0213_1, 2, outptr, outroom);
+            ScmSize outoffset = jis_ensure_state(cinfo, JIS_0213_1, 2, outptr, outroom);
             if (ERRP(outoffset)) return outoffset;
             outptr[outoffset] = e0 - 0x80;
             outptr[outoffset+1] = e1 - 0x80;
@@ -1076,7 +1079,7 @@ static size_t eucj2jis(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
 }
 
 /* reset proc */
-static size_t jis_reset(ScmConvInfo *cinfo, char *outptr, size_t outroom)
+static ScmSize jis_reset(ScmConvInfo *cinfo, char *outptr, ScmSize outroom)
 {
     if (outptr == NULL) {
         /* just reset */
@@ -1099,8 +1102,12 @@ static size_t jis_reset(ScmConvInfo *cinfo, char *outptr, size_t outroom)
 
 /* EUC_JP is a pivot code, so we don't need to convert.  This function
    is just a placeholder. */
-static size_t pivot(ScmConvInfo *cinfo, const char *inptr, size_t inroom,
-                    char *outptr, size_t outroom, size_t *outchars)
+static ScmSize pivot(ScmConvInfo *cinfo SCM_UNUSED,
+                     const char *inptr SCM_UNUSED,
+                     ScmSize inroom SCM_UNUSED,
+                     char *outptr SCM_UNUSED,
+                     ScmSize outroom SCM_UNUSED,
+                     ScmSize *outchars SCM_UNUSED)
 {
     return 0;
 }
@@ -1204,10 +1211,10 @@ static int conv_name_find(const char *name)
 */
 
 /* case (1) */
-static size_t jconv_ident(ScmConvInfo *info, const char **iptr,
-                          size_t *iroom, char **optr, size_t *oroom)
+static ScmSize jconv_ident(ScmConvInfo *info SCM_UNUSED, const char **iptr,
+                           ScmSize *iroom, char **optr, ScmSize *oroom)
 {
-    size_t inroom = *iroom, outroom = *oroom;
+    ScmSize inroom = *iroom, outroom = *oroom;
 #ifdef JCONV_DEBUG
     fprintf(stderr, "jconv_ident %s->%s\n", info->fromCode, info->toCode);
 #endif
@@ -1227,23 +1234,24 @@ static size_t jconv_ident(ScmConvInfo *info, const char **iptr,
         return OUTPUT_NOT_ENOUGH;
     }
 }
-   
+
 /* case (2) or (3) */
-static size_t jconv_1tier(ScmConvInfo *info, const char **iptr,
-                          size_t *iroom, char **optr, size_t *oroom)
+static ScmSize jconv_1tier(ScmConvInfo *info, const char **iptr,
+                           ScmSize *iroom, char **optr, ScmSize *oroom)
 {
     ScmConvProc cvt = info->convproc[0];
     const char *inp = *iptr;
     char *outp = *optr;
-    int inr = (int)*iroom, outr = (int)*oroom; 
-    size_t outchars, inchars, converted = 0;
+    int inr = (int)*iroom, outr = (int)*oroom;
+    ScmSize converted = 0;
 
 #ifdef JCONV_DEBUG
     fprintf(stderr, "jconv_1tier %s->%s\n", info->fromCode, info->toCode);
 #endif
     SCM_ASSERT(cvt != NULL);
     while (inr > 0 && outr > 0) {
-        inchars = cvt(info, inp, inr, outp, outr, &outchars);
+        ScmSize outchars;
+        ScmSize inchars = cvt(info, inp, inr, outp, outr, &outchars);
         if (ERRP(inchars)) {
             converted = inchars;
             break;
@@ -1261,11 +1269,11 @@ static size_t jconv_1tier(ScmConvInfo *info, const char **iptr,
     *oroom = outr;
     return converted;
 }
-   
+
 /* case (4) */
 #define INTBUFSIZ 20            /* intermediate buffer size */
-static size_t jconv_2tier(ScmConvInfo *info, const char **iptr, size_t *iroom,
-                          char **optr, size_t *oroom)
+static ScmSize jconv_2tier(ScmConvInfo *info, const char **iptr, ScmSize *iroom,
+                           char **optr, ScmSize *oroom)
 {
     char buf[INTBUFSIZ];
     ScmConvProc icvt = info->convproc[0];
@@ -1273,13 +1281,14 @@ static size_t jconv_2tier(ScmConvInfo *info, const char **iptr, size_t *iroom,
     const char *inp = *iptr;
     char *outp = *optr;
     int inr = (int)*iroom, outr = (int)*oroom;
-    size_t outchars, inchars, bufchars, converted = 0;
+    ScmSize converted = 0;
 
 #ifdef JCONV_DEBUG
     fprintf(stderr, "jconv_2tier %s->%s\n", info->fromCode, info->toCode);
 #endif
     while (inr > 0 && outr > 0) {
-        inchars  = icvt(info, inp, inr, buf, INTBUFSIZ, &bufchars);
+        ScmSize outchars, bufchars;
+        ScmSize inchars = icvt(info, inp, inr, buf, INTBUFSIZ, &bufchars);
         if (ERRP(inchars)) {
             converted = inchars;
             break;
@@ -1314,37 +1323,39 @@ static size_t jconv_2tier(ScmConvInfo *info, const char **iptr, size_t *iroom,
  * reset sequence; the first call should emit the sequence, but the second
  * call shouldn't.
  */
-static size_t jconv_iconv(ScmConvInfo *info, const char **iptr, size_t *iroom,
-                          char **optr, size_t *oroom)
+static ScmSize jconv_iconv(ScmConvInfo *info, const char **iptr, ScmSize *iroom,
+                           char **optr, ScmSize *oroom)
 {
-    size_t r;
 #ifdef JCONV_DEBUG
     fprintf(stderr, "jconv_iconv %s->%s\n", info->fromCode, info->toCode);
 #endif
-    r = iconv(info->handle, (char **)iptr, iroom, optr, oroom);
+    size_t ir = *iroom, or = *oroom;
+    size_t r = iconv(info->handle, (char **)iptr, &ir, optr, &or);
+    *iroom = ir;
+    *oroom = or;
     info->ostate = JIS_UNKNOWN;
     if (r == (size_t)-1) {
         if (errno == EINVAL) return INPUT_NOT_ENOUGH;
         if (errno == E2BIG)  return OUTPUT_NOT_ENOUGH;
-	return ILLEGAL_SEQUENCE;
+        return ILLEGAL_SEQUENCE;
     } else {
-        return (int)r;
+        return (ScmSize)r;
     }
 }
 
 /* reset routine for iconv */
-static size_t jconv_iconv_reset(ScmConvInfo *info, char *optr, size_t oroom)
+static ScmSize jconv_iconv_reset(ScmConvInfo *info, char *optr, ScmSize oroom)
 {
-    size_t oroom_prev = oroom;
-    size_t r;
+    ScmSize oroom_prev = oroom;
     if (info->ostate == JIS_ASCII) return 0;
-    r = iconv(info->handle, NULL, 0, &optr, &oroom);
+    size_t or = oroom;
+    size_t r = iconv(info->handle, NULL, 0, &optr, &or);
     if (r == (size_t)-1) {
         if (errno == E2BIG)  return OUTPUT_NOT_ENOUGH;
         Scm_Panic("jconv_iconv_reset: unknown error number %d\n", errno);
     }
     info->ostate = JIS_ASCII;
-    return oroom_prev - oroom;
+    return oroom_prev - (ScmSize)or;
 }
 #endif /*HAVE_ICONV_H*/
 
@@ -1355,15 +1366,13 @@ static size_t jconv_iconv_reset(ScmConvInfo *info, char *optr, size_t oroom)
  */
 ScmConvInfo *jconv_open(const char *toCode, const char *fromCode)
 {
-    ScmConvInfo *info;
     ScmConvHandler handler = NULL;
-    int incode, outcode;
     ScmConvProc convproc[2];
     ScmConvReset reset;
     iconv_t handle = (iconv_t)-1;
 
-    incode  = conv_name_find(fromCode);
-    outcode = conv_name_find(toCode);
+    int incode  = conv_name_find(fromCode);
+    int outcode = conv_name_find(toCode);
 
     if (incode == JCODE_NONE || outcode == JCODE_NONE) {
         /* conversion to/from none means no conversion */
@@ -1371,7 +1380,7 @@ ScmConvInfo *jconv_open(const char *toCode, const char *fromCode)
         convproc[0] = convproc[1] = NULL;
         reset = NULL;
     } else if (incode < 0 || outcode < 0) {
-#ifdef HAVE_ICONV_H        
+#ifdef HAVE_ICONV_H
         /* try iconv */
         handle = iconv_open(toCode, fromCode);
         if (handle == (iconv_t)-1) return NULL;
@@ -1405,6 +1414,7 @@ ScmConvInfo *jconv_open(const char *toCode, const char *fromCode)
         convproc[1] = conv_converter[outcode].outconv;
         reset = conv_converter[outcode].reset;
     }
+    ScmConvInfo *info;
     info = SCM_NEW(ScmConvInfo);
     info->jconv = handler;
     info->convproc[0] = convproc[0];
@@ -1435,9 +1445,9 @@ int jconv_close(ScmConvInfo *info)
 /*------------------------------------------------------------------
  * JCONV - main conversion routine
  */
-size_t jconv(ScmConvInfo *info,
-             const char **inptr, size_t *inroom,
-             char **outptr, size_t *outroom)
+ScmSize jconv(ScmConvInfo *info,
+              const char **inptr, ScmSize *inroom,
+              char **outptr, ScmSize *outroom)
 {
     SCM_ASSERT(info->jconv != NULL);
     return info->jconv(info, inptr, inroom, outptr, outroom);
@@ -1446,7 +1456,7 @@ size_t jconv(ScmConvInfo *info,
 /*------------------------------------------------------------------
  * JCONV_RESET - reset
  */
-size_t jconv_reset(ScmConvInfo *info, char *outptr, size_t outroom)
+ScmSize jconv_reset(ScmConvInfo *info, char *outptr, ScmSize outroom)
 {
     if (info->reset) {
         return info->reset(info, outptr, outroom);
