@@ -1,23 +1,23 @@
 ;;;
 ;;; gdbm - gdbm interface
-;;;  
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;;   
+;;;
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;;   
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
-;;;  
+;;;
 ;;;   2. Redistributions in binary form must reproduce the above copyright
 ;;;      notice, this list of conditions and the following disclaimer in the
 ;;;      documentation and/or other materials provided with the distribution.
-;;;  
+;;;
 ;;;   3. Neither the name of the authors nor the names of its contributors
 ;;;      may be used to endorse or promote products derived from this
 ;;;      software without specific prior written permission.
-;;;  
+;;;
 ;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 ;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 ;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -29,7 +29,7 @@
 ;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-;;;  
+;;;
 
 (define-module dbm.gdbm
   (extend dbm)
@@ -70,22 +70,22 @@
     (error "path must be set to open gdbm database"))
   (when (gdbm-file-of self)
     (errorf "gdbm ~S already opened" self))
-  (let* ((path   (slot-ref self 'path))
-         (rwmode (slot-ref self 'rw-mode))
-         (sync   (slot-ref self 'sync))
-         (nolock (slot-ref self 'nolock))
-         (rwopt  (case rwmode
-                   ((:read) GDBM_READER)
-                   ((:write) (+ GDBM_WRCREAT
+  (let* ([path   (slot-ref self 'path)]
+         [rwmode (slot-ref self 'rw-mode)]
+         [sync   (slot-ref self 'sync)]
+         [nolock (slot-ref self 'nolock)]
+         [rwopt  (case rwmode
+                   [(:read) GDBM_READER]
+                   [(:write) (+ GDBM_WRCREAT
                                  (if sync GDBM_SYNC 0)
-                                 (if nolock GDBM_NOLOCK 0)))
-                   ((:create) (+ GDBM_NEWDB
+                                 (if nolock GDBM_NOLOCK 0))]
+                   [(:create) (+ GDBM_NEWDB
                                  (if sync GDBM_SYNC 0)
-                                 (if nolock GDBM_NOLOCK 0)))))
-         (fp     (gdbm-open path
+                                 (if nolock GDBM_NOLOCK 0))])]
+         [fp     (gdbm-open path
                             (slot-ref self 'bsize)
                             rwopt
-                            (slot-ref self 'file-mode))))
+                            (slot-ref self 'file-mode))])
     (slot-set! self 'gdbm-file fp)
     self))
 
@@ -94,11 +94,11 @@
 ;;
 
 (define-method dbm-close ((self <gdbm>))
-  (let ((gdbm (gdbm-file-of self)))
+  (let1 gdbm (gdbm-file-of self)
     (and gdbm (gdbm-close gdbm))))
 
 (define-method dbm-closed? ((self <gdbm>))
-  (let ((gdbm (gdbm-file-of self)))
+  (let1 gdbm (gdbm-file-of self)
     (or (not gdbm) (gdbm-closed? gdbm))))
 
 ;;
@@ -115,11 +115,11 @@
 
 (define-method dbm-get ((self <gdbm>) key . args)
   (next-method)
-  (cond ((gdbm-fetch (gdbm-file-of self) (%dbm-k2s self key))
-         => (lambda (v) (%dbm-s2v self v)))
-        ((pair? args) (car args))     ;fall-back value
-        (else  (errorf "gdbm: no data for key ~s in database ~s"
-                       key (gdbm-file-of self)))))
+  (cond [(gdbm-fetch (gdbm-file-of self) (%dbm-k2s self key))
+         => (cut %dbm-s2v self <>)]
+        [(pair? args) (car args)]     ;fall-back value
+        [else  (errorf "gdbm: no data for key ~s in database ~s"
+                       key (gdbm-file-of self))]))
 
 (define-method dbm-exists? ((self <gdbm>) key)
   (next-method)
@@ -135,14 +135,13 @@
 ;;
 
 (define-method dbm-fold ((self <gdbm>) proc knil)
-  (let ((gdbm (gdbm-file-of self)))
-    (let loop ((key (gdbm-firstkey gdbm))
-               (r   knil))
+  (let1 gdbm (gdbm-file-of self)
+    (let loop ([key (gdbm-firstkey gdbm)] [r knil])
       (if key
-          (let ((val (gdbm-fetch gdbm key)))
-            (loop (gdbm-nextkey gdbm key)
-                  (proc (%dbm-s2k self key) (%dbm-s2v self val) r)))
-          r))))
+        (let1 val (gdbm-fetch gdbm key)
+          (loop (gdbm-nextkey gdbm key)
+                (proc (%dbm-s2k self key) (%dbm-s2v self val) r)))
+        r))))
 
 ;;
 ;; Metaoperations
@@ -153,7 +152,7 @@
 (define (%with-gdbm-locking path thunk)
   (let1 db (gdbm-open path 0 GDBM_READER #o664) ;; put read-lock
     (unwind-protect (thunk) (gdbm-close db))))
-        
+
 (define-method dbm-db-exists? ((class <gdbm-meta>) name)
   (file-exists? name))
 
@@ -162,25 +161,26 @@
 
 (define-method dbm-db-copy ((class <gdbm-meta>) from to . keys)
   (%with-gdbm-locking from
-   (lambda () (apply copy-file from to :safe #t keys))))
+   (^[] (apply copy-file from to :safe #t keys))))
 
 (define-method dbm-db-move ((class <gdbm-meta>) from to . keys)
   (%with-gdbm-locking from
-   (lambda () (apply move-file from to :safe #t keys))))
+   (^[] (apply move-file from to :safe #t keys))))
 
 ;;;
 ;;; Low-level bindings
 ;;;
 
 (inline-stub
- "#include <gdbm.h>"
- "#include <stdlib.h>"
+ (declcode
+  (.include <gdbm.h>)
+  (.include <stdlib.h>))
 
- "typedef struct ScmGdbmFileRec {
-    SCM_HEADER;
-    ScmObj name;
-    GDBM_FILE dbf;              /* NULL if closed */
-  } ScmGdbmFile;"
+ (define-ctype ScmGdbmFile::(.struct
+                             (SCM_HEADER :: ""
+                              name
+                              dbf::GDBM_FILE ; NULL if closed
+                              )))
 
  (define-cclass <gdbm-file> :private ScmGdbmFile* "Scm_GdbmFileClass" ()
    ()
@@ -188,6 +188,7 @@
     (Scm_Printf port "#<gdbm-file %S>" (-> (SCM_GDBM_FILE obj) name))])
 
  (define-cfn gdbm_finalize (obj data::void*) ::void :static
+   (cast void data)                     ; suppress unused var warning
    (let* ((g :: ScmGdbmFile* (SCM_GDBM_FILE obj)))
      (when (-> g dbf)
        (gdbm_close (-> g dbf))
@@ -215,21 +216,11 @@
     `(unless (-> ,g dbf) (Scm_Error "gdbm file already closed: %S" ,g))])
 
  ;; Those symbols may not be defined in the older gdbm
- "#ifndef GDBM_SYNC"
- "#define GDBM_SYNC 0"
- "#endif"
- "#ifndef GDBM_NOLOCK"
- "#define GDBM_NOLOCK 0"
- "#endif"
- "#ifndef GDBM_SYNCMODE"
- "#define GDBM_SYNCMODE 0"
- "#endif"
- "#ifndef GDBM_CENTFREE"
- "#define GDBM_CENTFREE 0"
- "#endif"
- "#ifndef GDBM_COALESCEBLKS"
- "#define GDBM_COALESCEBLKS 0"
- "#endif"
+ (.unless (defined GDBM_SYNC) (.define GDBM_SYNC 0))
+ (.unless (defined GDBM_NOLOCK) (.define GDBM_NOLOCK 0))
+ (.unless (defined GDBM_SYNCMODE) (.define GDBM_SYNCMODE 0))
+ (.unless (defined GDBM_CENTFREE) (.define GDBM_CENTFREE 0))
+ (.unless (defined GDBM_COALESCEBLKS) (.define GDBM_COALESCEBLKS 0))
 
  (define-cproc gdbm-open
    (name::<string> :optional (size::<fixnum> 0)
@@ -242,7 +233,7 @@
      (set! (-> z dbf) (gdbm_open (Scm_GetString name) size rwmode fmode NULL))
      (when (== (-> z dbf) NULL)
        (Scm_Error "couldn't open gdbm file %S (gdbm_errno=%d)" name gdbm_errno))
-     (result (SCM_OBJ z))))
+     (return (SCM_OBJ z))))
 
  (define-cproc gdbm-close (gdbm::<gdbm-file>) ::<void>
    (when (-> gdbm dbf)
@@ -250,7 +241,7 @@
      (set! (-> gdbm dbf) NULL)))
 
  (define-cproc gdbm-closed? (gdbm::<gdbm-file>) ::<boolean>
-   (result (== (-> gdbm dbf) NULL)))
+   (return (== (-> gdbm dbf) NULL)))
 
  (define-cproc gdbm-store (gdbm::<gdbm-file>
                            key::<string> val::<string>
@@ -260,24 +251,24 @@
      (CHECK_GDBM gdbm)
      (TO_DATUM dkey key)
      (TO_DATUM dval val)
-     (result (gdbm_store (-> gdbm dbf) dkey dval flags))))
+     (return (gdbm_store (-> gdbm dbf) dkey dval flags))))
 
-  (define-cproc gdbm-fetch (gdbm::<gdbm-file> key::<string>)
-    (let* ([dkey::datum] [dval::datum])
-      (CHECK_GDBM gdbm)
-      (TO_DATUM dkey key)
-      (set! dval (gdbm_fetch (-> gdbm dbf) dkey))
-      (FROM_DATUM SCM_RESULT dval)))
+ (define-cproc gdbm-fetch (gdbm::<gdbm-file> key::<string>)
+   (let* ([dkey::datum] [dval::datum])
+     (CHECK_GDBM gdbm)
+     (TO_DATUM dkey key)
+     (set! dval (gdbm_fetch (-> gdbm dbf) dkey))
+     (FROM_DATUM SCM_RESULT dval)))
 
-  (define-cproc gdbm-delete (gdbm::<gdbm-file> key::<string>) ::<int>
-    (let* ([dkey::datum])
-      (CHECK_GDBM gdbm)
-      (TO_DATUM dkey key)
-      (result (gdbm_delete (-> gdbm dbf) dkey))))
+ (define-cproc gdbm-delete (gdbm::<gdbm-file> key::<string>) ::<int>
+   (let* ([dkey::datum])
+     (CHECK_GDBM gdbm)
+     (TO_DATUM dkey key)
+     (return (gdbm_delete (-> gdbm dbf) dkey))))
 
-  (define-cproc gdbm-firstkey (gdbm::<gdbm-file>)
-    (let* ([dkey::datum (gdbm_firstkey (-> gdbm dbf))])
-      (FROM_DATUM SCM_RESULT dkey)))
+ (define-cproc gdbm-firstkey (gdbm::<gdbm-file>)
+   (let* ([dkey::datum (gdbm_firstkey (-> gdbm dbf))])
+     (FROM_DATUM SCM_RESULT dkey)))
 
  (define-cproc gdbm-nextkey (gdbm::<gdbm-file> key::<string>)
    (let* ([dkey::datum] [dnkey::datum])
@@ -288,7 +279,7 @@
 
  (define-cproc gdbm-reorganize (gdbm::<gdbm-file>) ::<int>
    (CHECK_GDBM gdbm)
-   (result (gdbm_reorganize (-> gdbm dbf))))
+   (return (gdbm_reorganize (-> gdbm dbf))))
 
  (define-cproc gdbm-sync (gdbm::<gdbm-file>) ::<void>
    (CHECK_GDBM gdbm)
@@ -298,10 +289,10 @@
    (let* ([dkey::datum])
      (CHECK_GDBM gdbm)
      (TO_DATUM dkey key)
-     (result (gdbm_exists (-> gdbm dbf) dkey))))
+     (return (gdbm_exists (-> gdbm dbf) dkey))))
 
- (define-cproc gdbm-strerror (errno::<fixnum>)
-   (result (SCM_MAKE_STR_IMMUTABLE (gdbm_strerror errno))))
+ (define-cproc gdbm-strerror (err_num::<fixnum>)
+   (return (SCM_MAKE_STR_IMMUTABLE (gdbm_strerror err_num))))
 
  (define-cproc gdbm-setopt (gdbm::<gdbm-file> option::<fixnum> val) ::<int>
    (let* ([ival::int])
@@ -309,14 +300,15 @@
      (if (SCM_EXACTP val)
        (set! ival (Scm_GetUInteger val))
        (set! ival (not (SCM_FALSEP val))))
-     (result (gdbm_setopt (-> gdbm dbf) option (& ival) (sizeof (int))))))
+     (return (gdbm_setopt (-> gdbm dbf) option (& ival) (sizeof (int))))))
 
  (define-cproc gdbm-version ()
-   (result (SCM_MAKE_STR_IMMUTABLE gdbm_version)))
+   (return (SCM_MAKE_STR_IMMUTABLE gdbm_version)))
 
  (define-cproc gdbm-errno () ::<int>
-   (result gdbm_errno)
-   (set! gdbm_errno 0))
+   (let* ([r::int gdbm_errno])
+     (set! gdbm_errno 0)
+     (return r)))
 
  (define-enum GDBM_READER)
  (define-enum GDBM_WRITER)
