@@ -9,11 +9,15 @@
 
 (test-start "file.util")
 (use file.util)
-(test-module 'file.util)
+(cond-expand
+ [gauche.os.windows
+  (test-module 'file.util :allow-undefined '(sys-symlink))]
+ [else
+  (test-module 'file.util)])
 
 ;; shorthand of normalizing pathname.  this doesn't do anything on
 ;; unix, but on Windows the separator in PATHNAME is replaced.
-;; NB: we remove the drive letter from DOS path to accomodate
+;; NB: we remove the drive letter from DOS path to accommodate
 ;; different installations.
 (define (n pathname . pathnames)
   (define (normalize s)
@@ -24,7 +28,7 @@
     (map normalize (cons pathname pathnames))))
 
 ;; mingw doesn't have fully compatible permissions as unix.
-;; this procedure compensates it.  
+;; this procedure compensates it.
 (define (P perm)
   (cond-expand
    [gauche.os.windows (let1 p (ash perm -6)
@@ -47,21 +51,21 @@
 (test* "build-path" (n "foo") (build-path "" "foo" ""))
 
 (test* "decompose-path" '("/a/b/c" "d" "e")
-       (receive r (decompose-path "/a/b/c/d.e") r))
+       (values->list (decompose-path "/a/b/c/d.e")))
 (test* "decompose-path" '("." "d" "e")
-       (receive r (decompose-path "d.e") r))
+       (values->list (decompose-path "d.e")))
 (test* "decompose-path" '("." "d" "")
-       (receive r (decompose-path "d.") r))
+       (values->list (decompose-path "d.")))
 (test* "decompose-path" '("." "d" #f)
-       (receive r (decompose-path "d") r))
+       (values->list (decompose-path "d")))
 (test* "decompose-path" '("/a.b" "c" #f)
-       (receive r (decompose-path "/a.b/c") r))
+       (values->list (decompose-path "/a.b/c")))
 (test* "decompose-path" '("/a.b" #f #f)
-       (receive r (decompose-path "/a.b/") r))
+       (values->list (decompose-path "/a.b/")))
 (test* "decompose-path" '("/a.b" "c.c" "d")
-       (receive r (decompose-path "/a.b/c.c.d") r))
+       (values->list (decompose-path "/a.b/c.c.d")))
 (test* "decompose-path" '("/a.b" ".d" #f)
-       (receive r (decompose-path "/a.b/.d") r))
+       (values->list (decompose-path "/a.b/.d")))
 
 (test* "path-extension" "c" (path-extension "/a.b/c.d/e.c"))
 (test* "path-extension" "" (path-extension "/a.b/c.d/e.c."))
@@ -98,20 +102,20 @@
 (test* "path-swap-extension" ".a.o" (path-swap-extension ".a" "o"))
 (test* "path-swap-extension" ".a.o" (path-swap-extension ".a.c" "o"))
 
+;; find-file-in-paths is tested after diretory util tests
 
 ;;=====================================================================
 (test-section "directories")
 
-(test "current-directory" (list (n "/") (n "/") #t #t)
-      (lambda ()
-        (let* ((cur   (sys-getcwd))
-               (root  (begin (current-directory "/")
-                             (n (sys-getcwd))))
-               (root* (n (current-directory)))
-               (cur*  (begin (current-directory cur)
-                             (sys-getcwd)))
-               (cur** (current-directory)))
-          (list root root* (string=? cur cur*) (string=? cur* cur**)))))
+(test* "current-directory" (list (n "/") (n "/") #t #t)
+       (let* ([cur   (sys-getcwd)]
+              [root  (begin (current-directory "/")
+                            (n (sys-getcwd)))]
+              [root* (n (current-directory))]
+              [cur*  (begin (current-directory cur)
+                            (sys-getcwd))]
+              [cur** (current-directory)])
+         (list root root* (string=? cur cur*) (string=? cur* cur**))))
 
 ;; rm -rf
 (define (cmd-rmrf dir)
@@ -121,10 +125,10 @@
 
   (cond-expand
    [gauche.os.windows
-    (sys-system #`"rmdir /q /s ,(n dir) > NUL 2>&1")
-    (sys-system #`"del /q ,(n dir) > NUL 2>&1")]
+    (sys-system #"rmdir /q /s ~(n dir) > NUL 2>&1")
+    (sys-system #"del /q ~(n dir) > NUL 2>&1")]
    [else
-    (sys-system #`"rm -rf ,dir > /dev/null")]))
+    (sys-system #"rm -rf ~dir > /dev/null")]))
 
 ;; prepare test data set
 (define *test-tree*
@@ -143,8 +147,8 @@
          `((test6.o ,(make-string 100 #\o))
            (test7.o ,(make-string 100 #\o))
            (test2.d :mode #o777
-		    ((test10.o ,(make-string 100 #\o))
-		     (test11.o ,(make-string 100 #\o))
+                    ((test10.o ,(make-string 100 #\o))
+                     (test11.o ,(make-string 100 #\o))
                      test12.o)))])
      (test.d ((test10.o ,(make-string 100 #\o))
               ,(cond-expand
@@ -171,7 +175,7 @@
 
 (test* "check-directory-tree" #t
        (check-directory-tree "." *test-tree*))
-           
+
 (test* "directory-list"
        '("." ".." "test.d" "test1.o" "test2.d" "test2.o"
          "test3.o" "test4.o" "test5.o" "test6.o" "test7.o" )
@@ -194,7 +198,7 @@
           "test.out/test2.o"  "test.out/test3.o" "test.out/test4.o"
           "test.out/test5.o" "test.out/test6.o" "test.out/test7.o" )
        (directory-list "test.out" :add-path? #t
-                       :filter (lambda (p) (string-suffix? "o" p))))
+                       :filter (^p (string-suffix? "o" p))))
 
 (test* "directory-list :filter"
        (n "test.out/test1.o"
@@ -207,30 +211,29 @@
        '(("." ".." "test.d" "test2.d")
          ("test1.o" "test2.o" "test3.o" "test4.o"
           "test5.o" "test6.o" "test7.o" ))
-       (receive x (directory-list2 "test.out") x))
+       (values->list (directory-list2 "test.out")))
 
 (test* "directory-list2 :add-path"
        `(,(n "test.out/." "test.out/.." "test.out/test.d" "test.out/test2.d")
          ,(n "test.out/test1.o" "test.out/test2.o"  "test.out/test3.o"
              "test.out/test4.o" "test.out/test5.o" "test.out/test6.o"
              "test.out/test7.o"))
-       (receive x (directory-list2 "test.out" :add-path? #t) x))
+       (values->list (directory-list2 "test.out" :add-path? #t)))
 
 (test* "directory-list2 :children"
        `(,(n "test.out/test.d" "test.out/test2.d")
          ,(n "test.out/test1.o" "test.out/test2.o"  "test.out/test3.o"
-          "test.out/test4.o" "test.out/test5.o" "test.out/test6.o"
+             "test.out/test4.o" "test.out/test5.o" "test.out/test6.o"
          "test.out/test7.o"))
-       (receive x (directory-list2 "test.out" :add-path? #t :children? #t) x))
+       (values->list (directory-list2 "test.out" :add-path? #t :children? #t)))
 
 (test* "directory-list2 :filter"
        '(("test.d" "test2.d")
          ("test1.o" "test2.o" "test3.o" "test4.o"
           "test5.o" "test6.o" "test7.o" ))
-       (receive x 
-           (directory-list2 "test.out"
-                            :filter (lambda (p) (string-contains p "test")))
-         x))
+       (values->list
+        (directory-list2 "test.out"
+                         :filter (^p (string-contains p "test")))))
 
 (cond-expand
  [gauche.sys.symlink
@@ -238,9 +241,8 @@
          '(("test.d")
            ("test1.o" "test2.d" "test2.o" "test3.o" "test4.o"
             "test5.o" "test6.o" "test7.o" ))
-         (receive x (directory-list2 "test.out" :follow-link? #f :children? #t)
-           x)
-         )]
+         (values->list
+          (directory-list2 "test.out" :follow-link? #f :children? #t)))]
  [else])
 
 (test* "directory-fold"
@@ -251,10 +253,10 @@
           "test.out/test6.o" "test.out/test7.o")
        (reverse
         (directory-fold "test.out"
-                        (lambda (path result)
+                        (^[path result]
                           (if (= (file-size path) 100)
-                              (cons path result)
-                              result))
+                            (cons path result)
+                            result))
                         '()))
        )
 
@@ -274,7 +276,7 @@
           "test.out/test6.o" "test.out/test7.o")
        (reverse
         (directory-fold "test.out" cons '()
-                        :lister (lambda (path seed)
+                        :lister (^[path seed]
                                   (values
                                    (directory-list path
                                                    :add-path? #t
@@ -290,10 +292,11 @@
             "test.out/test1.o"
             "test.out/test2.o" "test.out/test3.o"
             "test.out/test6.o" "test.out/test7.o")
-         (reverse 
+         (reverse
           (directory-fold "test.out"
-                          (lambda (path result)
-                            (if (= (file-size path) 100)
+                          (^[path result]
+                            (if (and (file-is-regular? path)
+                                     (= (file-size path) 100))
                                 (cons path result)
                                 result))
                           '()
@@ -327,7 +330,7 @@
        (reverse
         (directory-fold "test.out" cons '()
                         :follow-link? #f
-                        :lister (lambda (dir knil)
+                        :lister (^[dir knil]
                                   (receive (dirs files)
                                       (directory-list2 dir :add-path? #t :children? #t :follow-link? #f)
                                     (append dirs
@@ -393,19 +396,44 @@
  [else])
 (test* "file-eqv?" #f
        (file-eqv? "test.out/test1.o" "test.out/test2.o"))
-(test* "file-equal?" #t      
+(test* "file-equal?" #t
        (file-equal? "test.out/test1.o" "test.out/test1.o"))
-(test* "file-equal?" #t      
+(test* "file-equal?" #t
        (file-equal? "test.out/test1.o" "test.out/test2.o"))
 (test* "file-equal?" #f
        (file-equal? "test.out/test1.o" "test.out/test4.o"))
 (test* "file-equal?" #t
        (file-equal? "test.out/test4.o" "test.out/test5.o"))
 
-(test* "touch-file" #t
-       (and (not (file-exists? "test.out/touched"))
-            (begin (touch-file "test.out/touched")
-                   (file-exists? "test.out/touched"))))
+(let1 touched "test.out/touched"
+  (define (times file)
+    (let1 s (sys-stat file)
+      (list (~ s'atime) (~ s'mtime))))
+  (test* "touch-file :create #f" #t
+         (and (not (file-exists? touched))
+              (begin (touch-file touched :create #f)
+                     (not (file-exists? touched)))))
+  (test* "touch-file" #t
+         (and (not (file-exists? touched))
+              (begin (touch-file touched)
+                     (file-exists? touched))))
+  ;; NB: win32 has a problem of reading file's timestamp which can be
+  ;; before 1970/1/1 _on_local_time_.  For example, if it is 100 in UTC
+  ;; and we call _stat() in the timezone UTC-8, we get UINT_MAX.  Here
+  ;; we use a timestamp value large enough to avoid the problem.
+  (test* "touch-file :time" '(50001 50001)
+         (begin (touch-file touched :time 50001) (times touched)))
+  (test* "touch-file :time :type" '(60000 50001)
+         (begin (touch-file touched :time 60000 :type 'atime) (times touched)))
+  (test* "touch-file :time :type" '(60000 70000)
+         (begin (touch-file touched :time 70000 :type 'mtime) (times touched)))
+
+  (test* "touch-files" (n "test.out/touched" "test.out/touched1"
+                          "test.out/touched2" "test.out/touched3")
+         (begin (touch-files (n "test.out/touched" "test.out/touched1"
+                                "test.out/touched2" "test.out/touched3"))
+                (sort (glob "test.out/touched*"))))
+  )
 
 (test* "copy-file (normal)" #t
        (and (copy-file "test.out/test5.o" "test.out/test.copy")
@@ -440,7 +468,27 @@
             (file-equal? "test.out/test1.o" "test.out/test.copy")
             (file-equal? "test.out/test5.o" "test.out/test.copy~")))
 
+(let ()
+  (define (check-copy-file-append . args)
+    (let1 preamble "********************************\n"
+      (sys-unlink "test.out/test.copy")
+      (sys-unlink "test.out/test.copy2")
+      (with-output-to-file "test.out/test.copy"
+        (^[] (display preamble)))
+      (apply copy-file "test.out/test1.o" "test.out/test.copy"
+             :if-exists :append args)
+      (with-output-to-file "test.out/test.copy2"
+        (^[] (display preamble)
+          (call-with-input-file "test.out/test1.o"
+            (^[in] (copy-port in (current-output-port))))))
+      (file-equal? "test.out/test.copy" "test.out/test.copy2")))
+  (test* "copy-file (:if-exists :append)" #t
+         (check-copy-file-append))
+  (test* "copy-file (:if-exists :append :safe #t)" #t
+         (check-copy-file-append :safe #t)))
+
 (sys-unlink "test.out/test.copy")
+(sys-unlink "test.out/test.copy2")
 (sys-unlink "test.out/test.copy~")
 (sys-unlink "test.out/test.copy.orig")
 
@@ -543,7 +591,7 @@
   (define (listdir d)
     (map (cut regexp-replace #/^test2?\.out[\/\\]/ <> "")
          (directory-fold d cons '())))
-    
+
   (test* "copy-directory*" (listdir "test.out")
          (begin
            (copy-directory* "test.out" "test2.out")
@@ -561,7 +609,7 @@
    [gauche.sys.symlink
     (let ()
       (define (tester desc fn expect . opts)
-        (test* #`"dangling symlink - ,desc" expect
+        (test* #"dangling symlink - ~desc" expect
                (begin
                  (sys-symlink "no such file" "test.out")
                  (apply fn "test.out" "test2.out" opts)
@@ -578,6 +626,40 @@
    [else])
   )
 
+;;
+;; temporary file/dir
+;;
+
+(test* "call-with-temporary-file" '(#t #t #f)
+       (unwind-protect
+           (begin
+             (remove-files '("test.o"))
+             (make-directory* "test.o")
+             (receive (exists name)
+                 (call-with-temporary-file 
+                  (^[oport name] (values (file-exists? name) name))
+                  :directory "test.o"
+                  :prefix "ooo")
+               (list exists
+                     (boolean (#/test.o[\/\\]ooo\w+$/ name))
+                     (file-exists? name))))
+         (remove-files '("test.o"))))
+
+(test* "call-with-temporary-directory" '(#t #t #f)
+       (unwind-protect
+           (begin
+             (remove-files '("test.o"))
+             (make-directory* "test.o")
+             (receive (exists name)
+                 (call-with-temporary-directory
+                  (^[name] (values (file-is-directory? name) name))
+                  :directory "test.o"
+                  :prefix "ooo")
+               (list exists
+                     (boolean (#/test.o[\/\\]ooo\w+$/ name))
+                     (file-exists? name))))
+         (remove-files '("test.o"))))
+
 ;;
 ;; file->*
 ;;
@@ -604,22 +686,22 @@
 (define (test-lock-file type)
   (remove-files "test.out" "test.out.2")
 
-  (test* #`"with-lock-file (,type) just lock&release" '(#t #f)
+  (test* #"with-lock-file (~type) just lock&release" '(#t #f)
          (let1 r (with-lock-file "test.out"
-                                 (^() (file-exists? "test.out"))
+                                 (^[] (file-exists? "test.out"))
                                  :type type)
            (list r (file-exists? "test.out"))))
 
   (remove-files "test.out" "test.out.2")
 
-  (test* #`"with-lock-file (,type) lock giveup" `((,<lock-file-failure> #t) #f)
+  (test* #"with-lock-file (~type) lock giveup" `((,<lock-file-failure> #t) #f)
          (let1 r
              (with-lock-file "test.out"
-                             (^()
+                             (^[]
                                (let1 r2
                                    (guard (e (else (class-of e)))
                                      (with-lock-file "test.out"
-                                                     (^() 'boo!)
+                                                     (^[] 'boo!)
                                                      :type type
                                                      :abandon-timeout #f
                                                      :retry-interval 0.1
@@ -630,27 +712,27 @@
 
   (remove-files "test.out" "test.out.2")
 
-  (test* #`"with-lock-file (,type) stealing" 'got
+  (test* #"with-lock-file (~type) stealing" 'got
          (begin
            (case type
              [(file) (touch-file "test.out")]
              [(directory) (make-directory* "test.out")])
            (with-lock-file "test.out"
-                           (^() 'got)
+                           (^[] 'got)
                            :type type
                            :retry-interval 0.1
                            :retry-limit 5
                            :abandon-timeout 1)))
   (remove-files "test.out" "test.out.2")
 
-  (test* #`"with-lock-file (,type) stealing failure"
+  (test* #"with-lock-file (~type) stealing failure"
          (test-error <lock-file-failure>)
          (begin
            (case type
              [(file) (make-directory* "test.out")]
              [(directory) (touch-file "test.out")])
            (with-lock-file "test.out"
-                           (^() 'got)
+                           (^[] 'got)
                            :type type
                            :retry-interval 0.1
                            :retry-limit 0.5
@@ -661,4 +743,57 @@
 (test-lock-file 'file)
 (test-lock-file 'directory)
 
+;;
+;; find-file-in-paths
+;;
+
+(let ()
+  (define (cleanup)
+    (when (file-exists? "test.out")
+      (remove-directory* "test.out")))
+  (define paths '("test.out/bin"
+                  "test.out/usr/bin"
+                  "test.out/usr/sbin"))
+  (cleanup)
+  (dolist [p paths] (make-directory* p))
+  (print (directory-list "."))
+  (touch-files '("test.out/bin/foo"
+                 "test.out/bin/bar"
+                 "test.out/bin/ban.com"
+                 "test.out/usr/bin/foo"
+                 "test.out/usr/bin/bar"
+                 "test.out/usr/bin/baz"
+                 "test.out/usr/bin/ban.exe"
+                 "test.out/usr/sbin/foo"
+                 "test.out/usr/sbin/bar"
+                 "test.out/usr/sbin/ban"))
+
+  (unwind-protect
+      (begin
+        (test* "find-file-in-paths" (n "test.out/bin/foo")
+               (find-file-in-paths "foo"
+                                   :paths paths
+                                   :pred file-exists?))
+        (test* "find-file-in-paths" (n "test.out/usr/bin/foo")
+               (find-file-in-paths "foo"
+                                   :paths (cdr paths)
+                                   :pred file-exists?))
+        (test* "find-file-in-paths" (n "test.out/usr/sbin/ban")
+               (find-file-in-paths "ban"
+                                   :paths paths
+                                   :pred file-exists?))
+        (test* "find-file-in-paths" (n "test.out/usr/bin/ban.exe")
+               (find-file-in-paths "ban"
+                                   :paths paths
+                                   :pred file-exists?
+                                   :extensions '("exe")))
+        (test* "find-file-in-paths" (n "test.out/bin/ban.com")
+               (find-file-in-paths "ban"
+                                   :paths paths
+                                   :pred file-exists?
+                                   :extensions '("exe" "com")))
+        )
+    (cleanup))
+  )
+
 (test-end)
