@@ -2,6 +2,9 @@
 ;;  gauche.generator isn't precompiled (yet), but it depends on gauche.sequence
 ;;  so we test it here.
 
+;;  NB: Some generator interface depends on gauche.uvector, which is tested
+;;  after this.  So they're tested in gauche.uvector.
+
 (use gauche.test)
 (use gauche.sequence)
 (use srfi-1)
@@ -12,7 +15,7 @@
 (use gauche.generator)
 (test-module 'gauche.generator)
 
-(test-section "genreator constructors")
+(test-section "generator constructors")
 
 ;; first, let's test the stuff used by the following tests.
 (test* "giota + generator->list" '(0 1 2 3 4)
@@ -47,6 +50,13 @@
 (test* "grange + generator->list" '(1.0 1.5 2.0 2.5)
        (generator->list (grange 1.0 3 1/2)))
 
+(test* "grange + generator->list" '()
+       (generator->list (grange 5 0)))
+(test* "grange + generator->list" '(5 4 3 2 1)
+       (generator->list (grange 5 0 -1)))
+(test* "grange + generator->list" '(5.0 4.0 3.0 2.0 1.0)
+       (generator->list (grange 5 0 -1.0)))
+
 ;; converters
 (let-syntax ((t (syntax-rules ()
                   [(t dir fn cv data)
@@ -138,20 +148,51 @@
                   (^[yield] (let loop ([i 0]) (yield i) (loop (+ i 1))))))
   )
 
-;; test x->generator <collection>.  this needs to be tested after
-;; the test of 'generate'.
-(test* "x->generator <collection>" '((0 . a) (1 . b) (2 . c))
+(test* "x->generator <hash-table>" '((0 . a) (1 . b) (2 . c))
        (sort-by (generator->list (x->generator (hash-table 'eqv?
                                                            '(0 . a)
                                                            '(1 . b)
                                                            '(2 . c))))
                 car))
+(test* "x->generator <tree-map>" '((0 . a) (1 . b) (2 . c))
+       (generator->list (x->generator (alist->tree-map 
+                                       '((0 . a)
+                                         (1 . b)
+                                         (2 . c))
+                                       = <))))
+(test* "x->generator <tree-map>" '((2 . c) (1 . b) (0 . a))
+       (generator->list (x->generator (alist->tree-map 
+                                       '((0 . a)
+                                         (1 . b)
+                                         (2 . c))
+                                       = >))))
+
+;; test x->generator <collection>.  this needs to be tested after
+;; the test of 'generate'.
+(define-class <mycoll> (<collection>)
+  ((elts :init-value '() :init-keyword :elts)) ;list
+  )
+(define-method call-with-iterator ((coll <mycoll>) proc)
+  (define z (~ coll 'elts))
+  (proc (cut eq? z '())
+        (^[] (pop! z))))
+(test* "x->generator <collection>" '((0 . a) (1 . b) (2 . c))
+       (sort-by (generator->list (x->generator (make <mycoll>
+                                                 :elts '((0 . a)
+                                                         (1 . b)
+                                                         (2 . c)))))
+                car))
 
 (test* "gappend" '(0 1 2 3 a b c d A B C D)
        (generator->list (gappend (giota 4)
                                  (x->generator '(a b c d))
                                  (x->generator '(A B C D)))))
 
+(test* "gflatten" '(0 1 2 3 a b c d A B C D)
+       (generator->list (gflatten (x->generator (list (iota 4)
+                                                      '(a b c d)
+                                                      '(A B C D))))))
+
 (test* "gconcatenate" '(0 1 2 3 0 1 2 3 0 1)
        (generator->list (gconcatenate (gunfold (^v #f)
                                                (^_ (giota 4))
@@ -245,7 +286,11 @@
                 (cut take* <> 3)
                 '(1 2 3 4 5 6) '(1 2))
 
-(test-list-like (cut gtake <> 3 #t 'a)
+(test-list-like (cut gtake* <> 3 #t 'a)
+                (cut take* <> 3 #t 'a)
+                '(1 2 3 4 5 6) '(1 2))
+
+(test-list-like (cut gtake <> 3 'a)
                 (cut take* <> 3 #t 'a)
                 '(1 2 3 4 5 6) '(1 2))
 
@@ -303,6 +348,13 @@
                          '(3 1 4 1 5 9 2 6 5 3 5 8 9 7 9 3 2 3 8))
                         (^[s] `(,(reverse s))))))
 
+(test* "gdelete-neighbor-dups"
+       "Misisipi"
+       (list->string (generator->list (gdelete-neighbor-dups "Mississippi"))))
+(test* "gdelete-neighbor-dups"
+       '()
+       (generator->list (gdelete-neighbor-dups '())))
+
 (test* "grxmatch (string)" '("ab" "cde" "fgh" "jkl")
        (map rxmatch-substring
             (generator->list (grxmatch #/\w{2,}/ " ab x y.cde\nfgh/j/jkl "))))
@@ -356,6 +408,37 @@
           $ grxmatch #/\w+/
           $ gappend "    " (make-string 1999 #\a)))
 
+(test* "ginterval (once, closed)"
+       '(10 11 12 13 14 15 16 17 18 19 20)
+       (generator->list (ginterval (pa$ = 10) (pa$ = 20) (giota 100))))
+(test* "ginterval (once, open-closed)"
+       '(11 12 13 14 15 16 17 18 19 20)
+       (generator->list (ginterval (pa$ = 10) (pa$ = 20) (giota 100)
+                                   :open 'start)))
+(test* "ginterval (once, closed-open)"
+       '(10 11 12 13 14 15 16 17 18 19)
+       (generator->list (ginterval (pa$ = 10) (pa$ = 20) (giota 100)
+                                   :open 'end)))
+(test* "ginterval (once, open)"
+       '(11 12 13 14 15 16 17 18 19)
+       (generator->list (ginterval (pa$ = 10) (pa$ = 20) (giota 100)
+                                   :open 'both)))
+(test* "ginterval (once, open)"
+       '(11 12 13 14 15 16 17 18 19)
+       (generator->list (ginterval (pa$ = 10) (pa$ = 20) (giota 100)
+                                   :open #t)))
+
+(test* "ginterval (repeat, closed)"
+       '(0 1 2 3 6 7 8 9 12 13 14 15 18 19)
+       (generator->list (ginterval (^x (zero? (modulo x 3)))
+                                   (^x (zero? (modulo x 3)))
+                                   (giota 20)
+                                   :repeat #t)))
+(test* "ginterval (repeat, open)"
+       '(1 2 7 8 13 14 19)
+       (generator->list (ginterval (^x (zero? (modulo x 3)))
+                                   (^x (zero? (modulo x 3)))
+                                   (giota 20)
+                                   :open #t :repeat #t)))
 
 (test-end)
-
