@@ -2,7 +2,7 @@
  * mt-random.c - implements MT19937 random number generation algorithm
  *
  * Mersenne Twister algorithm invented by Makoto Matsumoto & Takuji Nishimura.
- *   http://www.math.keio.ac.jp/~matumoto/emt.html
+ *   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
  * This code is based on mt18837ar.c (2002/1/16).   It is released under
  * BSD license.
  * I modified the code in the following parts.
@@ -10,7 +10,6 @@
  *     the allocated memory for random number generator object.
  *   - changed the names of the functions
  *   - added stuff to make it as a Gauche extension module.
- * $Id: mt-random.c,v 1.20 2008-02-03 13:07:43 shirok Exp $
  *
  * The original copyright notice follows.
  */
@@ -18,11 +17,11 @@
    A C-program for MT19937, with initialization improved 2002/1/26.
    Coded by Takuji Nishimura and Makoto Matsumoto.
 
-   Before using, initialize the state by using init_genrand(seed)  
+   Before using, initialize the state by using init_genrand(seed)
    or init_by_array(init_key, key_length).
 
    Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
-   All rights reserved.                          
+   All rights reserved.
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
@@ -35,8 +34,8 @@
         notice, this list of conditions and the following disclaimer in the
         documentation and/or other materials provided with the distribution.
 
-     3. The names of its contributors may not be used to endorse or promote 
-        products derived from this software without specific prior written 
+     3. The names of its contributors may not be used to endorse or promote
+        products derived from this software without specific prior written
         permission.
 
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
@@ -53,14 +52,14 @@
 
 
    Any feedback is very welcome.
-   http://www.math.keio.ac.jp/matumoto/emt.html
-   email: matumoto@math.keio.ac.jp
+   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
+   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
 */
 
 #include <math.h>
 #include "mt-random.h"
 
-/* Period parameters */  
+/* Period parameters */
 #define M 397
 #define MATRIX_A 0x9908b0dfUL   /* constant vector a */
 #define UPPER_MASK 0x80000000UL /* most significant w-r bits */
@@ -72,8 +71,8 @@ void Scm_MTInitByUI(ScmMersenneTwister *mt, unsigned long s)
     int mti;
     mt->mt[0]= s & 0xffffffffUL;
     for (mti=1; mti<N; mti++) {
-        mt->mt[mti] = 
-	    (1812433253UL * (mt->mt[mti-1] ^ (mt->mt[mti-1] >> 30)) + mti); 
+        mt->mt[mti] =
+            (1812433253UL * (mt->mt[mti-1] ^ (mt->mt[mti-1] >> 30)) + mti);
         /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
         /* In the previous versions, MSBs of the seed affect   */
         /* only MSBs of the array mt[].                        */
@@ -95,7 +94,7 @@ void Scm_MTSetSeed(ScmMersenneTwister *mt, ScmObj seed)
         }
         Scm_MTInitByUI(mt, s);
     } else if (SCM_U32VECTORP(seed)) {
-        Scm_MTInitByArray(mt, (ScmInt32*)SCM_U32VECTOR_ELEMENTS(seed),
+        Scm_MTInitByArray(mt, (int32_t*)SCM_U32VECTOR_ELEMENTS(seed),
                           SCM_U32VECTOR_SIZE(seed));
     } else {
         Scm_TypeError("random seed", "an exact integer or u32vector", seed);
@@ -107,7 +106,7 @@ void Scm_MTSetSeed(ScmMersenneTwister *mt, ScmObj seed)
 /* init_key is the array for initializing keys */
 /* key_length is its length */
 void Scm_MTInitByArray(ScmMersenneTwister *mt,
-                       ScmInt32 init_key[],
+                       int32_t init_key[],
                        unsigned long key_length)
 {
     int i, j, k;
@@ -130,7 +129,7 @@ void Scm_MTInitByArray(ScmMersenneTwister *mt,
         if (i>=N) { mt->mt[0] = mt->mt[N-1]; i=1; }
     }
 
-    mt->mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
+    mt->mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */
 }
 
 /* generates a random number on [0,0xffffffff]-interval */
@@ -160,7 +159,7 @@ unsigned long Scm_MTGenrandU32(ScmMersenneTwister *mt)
 
         mti = 0;
     }
-  
+
     y = mt->mt[mti++];
 
     /* Tempering */
@@ -195,7 +194,7 @@ double Scm_MTGenrandF64(ScmMersenneTwister *mt, int exclude0)
         r = (a*67108864.0+b)*(1.0/9007199254740992.0);
     } while (exclude0 && r == 0.0); /*if we get 0.0, try another one. */;
     return r;
-} 
+}
 
 /*
  * Generic integer routine for [0, n-1], 0 < n <= 2^32
@@ -291,11 +290,11 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_MersenneTwisterClass,
                          NULL, NULL, NULL, mt_allocate,
                          NULL);
 
-static ScmObj mt_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj mt_allocate(ScmClass *klass SCM_UNUSED, ScmObj initargs)
 {
     ScmObj seed = Scm_GetKeyword(key_seed, initargs, SCM_FALSE);
     ScmMersenneTwister *mt;
-    
+
     mt = SCM_NEW(ScmMersenneTwister);
     SCM_SET_CLASS(mt, &Scm_MersenneTwisterClass);
     mt->mti = N+1;
@@ -303,15 +302,11 @@ static ScmObj mt_allocate(ScmClass *klass, ScmObj initargs)
     return SCM_OBJ(mt);
 }
 
-extern void Scm_Init_mt_lib(ScmModule*);
-
-SCM_EXTENSION_ENTRY void Scm_Init_mt_random(void)
+void Scm_Init_mt_random(void)
 {
     ScmModule *mod = SCM_FIND_MODULE("math.mt-random", SCM_FIND_MODULE_CREATE);
-    SCM_INIT_EXTENSION(mtrandom);
     Scm_InitStaticClass(&Scm_MersenneTwisterClass, "<mersenne-twister>",
                         mod, NULL, 0);
     key_seed = SCM_MAKE_KEYWORD("seed");
-    Scm_Init_mt_lib(mod);
 }
 
