@@ -1,12 +1,12 @@
 /*
  * addr.c - socket address
  *
- *  Copyright (c) 2001-2011  Shiro Kawai  <shiro@acm.org>
+ *  Copyright (c) 2001-2019  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -64,7 +64,8 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_SockAddrClass, sockaddr_print,
                          NULL, NULL, sockaddr_allocate,
                          NULL);
 
-void sockaddr_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+void sockaddr_print(ScmObj obj, ScmPort *port, 
+                    ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(port, "#<sockaddr %S %S>",
                Scm_SockAddrFamily(SCM_SOCKADDR(obj)),
@@ -90,7 +91,8 @@ ScmObj Scm_SockAddrFamily(ScmSockAddr *addr)
 }
 
 /* Fallback of allocation method */
-static ScmObj sockaddr_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj sockaddr_allocate(ScmClass *klass SCM_UNUSED, 
+                                ScmObj initargs SCM_UNUSED)
 {
     Scm_Error("you can't directly instantiate the abstract class <sockaddr>");
     return SCM_UNDEFINED;       /* dummy */
@@ -99,9 +101,8 @@ static ScmObj sockaddr_allocate(ScmClass *klass, ScmObj initargs)
 /* creates sockaddr from struct sockaddr. */
 ScmObj Scm_MakeSockAddr(ScmClass *klass, struct sockaddr *saddr, int len)
 {
-    ScmSockAddr *addr;
-    addr = SCM_NEW_ATOMIC2(ScmSockAddr*,
-                           sizeof(ScmSockAddr) - sizeof(struct sockaddr) + len);
+    ScmSockAddr *addr = SCM_NEW_ATOMIC2(
+        ScmSockAddr*, sizeof(ScmSockAddr) - sizeof(struct sockaddr) + len);
     if (klass == NULL) {
         switch (saddr->sa_family) {
         case AF_UNIX:
@@ -130,15 +131,15 @@ ScmObj Scm_MakeSockAddr(ScmClass *klass, struct sockaddr *saddr, int len)
  * Unix domain socket
  */
 
-static ScmObj sockaddr_un_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj sockaddr_un_allocate(ScmClass *klass SCM_UNUSED,
+                                   ScmObj initargs)
 {
     ScmObj path = Scm_GetKeyword(key_path, initargs, SCM_FALSE);
-    ScmSockAddrUn *addr;
-    
     if (!SCM_FALSEP(path) && !SCM_STRINGP(path)) {
         Scm_Error(":path parameter must be a string, but got %S", path);
     }
-    addr = SCM_NEW_ATOMIC(ScmSockAddrUn);
+
+    ScmSockAddrUn *addr = SCM_NEW_ATOMIC(ScmSockAddrUn);
     SCM_SET_CLASS(addr, SCM_CLASS_SOCKADDR_UN);
     memset(&addr->addr, 0, sizeof(struct sockaddr_un));
 #ifdef HAVE_STRUCT_SOCKADDR_UN_SUN_LEN
@@ -146,10 +147,10 @@ static ScmObj sockaddr_un_allocate(ScmClass *klass, ScmObj initargs)
 #endif
     addr->addr.sun_family = AF_UNIX;
     if (SCM_STRINGP(path)) {
-        u_int size;
+        ScmSmallInt size;
         const char *cpath = Scm_GetStringContent(SCM_STRING(path), &size,
                                                  NULL, NULL);
-        if (size >= sizeof(addr->addr.sun_path)-1) {
+        if ((size_t)size >= sizeof(addr->addr.sun_path)-1) {
             Scm_Error("path too long: %S", path);
         }
         memcpy(addr->addr.sun_path, cpath, size);
@@ -164,7 +165,7 @@ static int sockaddr_un_compare(ScmObj x, ScmObj y, int equalp)
     ScmSockAddrUn *xx = (ScmSockAddrUn*)x;
     ScmSockAddrUn *yy = (ScmSockAddrUn*)y;
     if (!equalp) Scm_Error("object %S and %S can't be ordered", x, y);
-    
+
     if (xx->addrlen == yy->addrlen
         && memcmp(xx->addr.sun_path, yy->addr.sun_path,
                   sizeof(xx->addr.sun_path)) == 0) {
@@ -182,17 +183,16 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_SockAddrUnClass, sockaddr_print,
  * Inet domain socket
  */
 
-static ScmObj sockaddr_in_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj sockaddr_in_allocate(ScmClass *klass SCM_UNUSED, ScmObj initargs)
 {
     ScmObj host = Scm_GetKeyword(key_host, initargs, key_any);
     ScmObj port = Scm_GetKeyword(key_port, initargs, SCM_MAKE_INT(0));
-    ScmSockAddrIn *addr;
 
     if (!SCM_INTP(port)) {
         Scm_Error(":port parameter must be a small exact integer, but got %S",
                   port);
     }
-    addr = SCM_NEW_ATOMIC(ScmSockAddrIn);
+    ScmSockAddrIn *addr = SCM_NEW_ATOMIC(ScmSockAddrIn);
     SCM_SET_CLASS(addr, SCM_CLASS_SOCKADDR_IN);
     memset(&addr->addr, 0, sizeof(struct sockaddr_in));
 #ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
@@ -206,11 +206,11 @@ static ScmObj sockaddr_in_allocate(ScmClass *klass, ScmObj initargs)
         if (inet_pton(AF_INET, hname, &addr->addr.sin_addr) <= 0) {
             /* Now, we need to look up the host name.
                Call MT-safe Scm_GetHostByName */
-            ScmObj ap, hent = Scm_GetHostByName(hname);
+            ScmObj hent = Scm_GetHostByName(hname);
             if (!SCM_SYS_HOSTENT_P(hent)) {
                 Scm_Error("unknown host: %S", host);
             }
-            ap = SCM_SYS_HOSTENT(hent)->addresses;
+            ScmObj ap = SCM_SYS_HOSTENT(hent)->addresses;
             if (SCM_NULLP(ap) || !SCM_STRINGP(SCM_CAR(ap))) {
                 Scm_Error("host have unknown address type: %S", host);
             }
@@ -231,13 +231,11 @@ static ScmObj sockaddr_in_allocate(ScmClass *klass, ScmObj initargs)
         if (ov) Scm_Error("host address is out of range: %S", host);
         addr->addr.sin_addr.s_addr = htonl(a);
     } else if (SCM_U8VECTORP(host)) {
-        const unsigned char *p;
-        unsigned long a;
         if (SCM_U8VECTOR_SIZE(host) < 4) {
             Scm_Error("host address is too short: %S", host);
         }
-        p = SCM_U8VECTOR_ELEMENTS(host);
-        a = (p[0]<<24)+(p[1]<<16)+(p[2]<<8)+p[3];
+        const unsigned char *p = SCM_U8VECTOR_ELEMENTS(host);
+        unsigned long a = (p[0]<<24)+(p[1]<<16)+(p[2]<<8)+p[3];
         addr->addr.sin_addr.s_addr = htonl(a);
     } else {
         Scm_Error("bad :host parameter: %S", host);
@@ -272,17 +270,16 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_SockAddrInClass, sockaddr_print,
 
 #ifdef HAVE_IPV6
 
-static ScmObj sockaddr_in6_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj sockaddr_in6_allocate(ScmClass *klass SCM_UNUSED, ScmObj initargs)
 {
     ScmObj host = Scm_GetKeyword(key_host, initargs, key_any);
     ScmObj port = Scm_GetKeyword(key_port, initargs, SCM_MAKE_INT(0));
-    ScmSockAddrIn6 *addr;
 
     if (!SCM_INTP(port)) {
         Scm_Error(":port parameter must be a small exact integer, but got %S",
                   port);
     }
-    addr = SCM_NEW_ATOMIC(ScmSockAddrIn6);
+    ScmSockAddrIn6 *addr = SCM_NEW_ATOMIC(ScmSockAddrIn6);
     SCM_SET_CLASS(addr, SCM_CLASS_SOCKADDR_IN6);
     memset(&addr->addr, 0, sizeof(struct sockaddr_in6));
 #ifdef HAVE_STRUCT_SOCKADDR_IN6_SIN6_LEN
@@ -293,12 +290,15 @@ static ScmObj sockaddr_in6_allocate(ScmClass *klass, ScmObj initargs)
     if (SCM_STRINGP(host)) {
         const char *hname = Scm_GetStringConst(SCM_STRING(host));
         struct addrinfo hints, *res;
-        int r;
         memset(&hints, 0, sizeof(hints));
         hints.ai_family = AF_INET6;
         hints.ai_socktype = SOCK_STREAM;
-        r = getaddrinfo(hname, NULL, &hints, &res);
-        if (r) Scm_Error("getaddrinfo: %s", gai_strerror(r));
+        int r = getaddrinfo(hname, NULL, &hints, &res);
+#if !defined(GAUCHE_WINDOWS)
+        if (r) Scm_Error("getaddrinfo failed: %s", gai_strerror(r));
+#else  /*GAUCHE_WINDOWS*/
+        if (r) Scm_SysError("getaddrinfo failed");
+#endif /*GAUCHE_WINDOWS*/
         addr->addr.sin6_addr = ((struct sockaddr_in6*)res->ai_addr)->sin6_addr;
         freeaddrinfo(res);
     } else if (host == key_any) {
@@ -307,20 +307,17 @@ static ScmObj sockaddr_in6_allocate(ScmClass *klass, ScmObj initargs)
         addr->addr.sin6_addr = in6addr_loopback;
     } else if (SCM_INTEGERP(host)) {
         /* NB: Can we have more efficient way? */
-        int i;
-        for (i=15; i>=0; i--) {
+        for (int i=15; i>=0; i--) {
             ScmObj u8 = Scm_LogAnd(host, SCM_MAKE_INT(0xff));
             addr->addr.sin6_addr.s6_addr[i] = SCM_INT_VALUE(u8);
             host = Scm_Ash(host, -8);
         }
     } else if (SCM_U8VECTORP(host)) {
-        const unsigned char *p;
-        int i;
         if (SCM_U8VECTOR_SIZE(host) < 16) {
             Scm_Error("host address is too short: %S", host);
         }
-        p = SCM_U8VECTOR_ELEMENTS(host);
-        for (i=0; i<16; i++) {
+        const unsigned char *p = SCM_U8VECTOR_ELEMENTS(host);
+        for (int i=0; i<16; i++) {
             addr->addr.sin6_addr.s6_addr[i] = p[i];
         }
     } else {
@@ -335,7 +332,7 @@ static int sockaddr_in6_compare(ScmObj x, ScmObj y, int equalp)
     ScmSockAddrIn6 *xx = (ScmSockAddrIn6*)x;
     ScmSockAddrIn6 *yy = (ScmSockAddrIn6*)y;
     if (!equalp) Scm_Error("object %S and %S can't be ordered", x, y);
-    
+
     if (xx->addrlen == yy->addrlen
         && xx->addr.sin6_family == yy->addr.sin6_family
         && xx->addr.sin6_port == yy->addr.sin6_port
@@ -357,8 +354,6 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_SockAddrIn6Class, sockaddr_print,
  * Parse internet addresses
  */
 
-#define S6_ADDR32(in6, offset) (*(uint32_t*)&(((in6).s6_addr)[(offset)*4]))
-
 ScmObj Scm_InetStringToAddress(const char *s,
                                int *proto,     /*out*/
                                ScmUVector *buf /*out*/)
@@ -373,7 +368,8 @@ ScmObj Scm_InetStringToAddress(const char *s,
             if (Scm_UVectorSizeInBytes(buf) < 4) {
                 Scm_Error("uniform vector buffer isn't big enough to hold IPv4 address: %S", buf);
             }
-            SCM_U32VECTOR_ELEMENTS(buf)[0] = in4.s_addr;
+            ScmU32Vector *u32 = (ScmU32Vector*)buf;
+            SCM_U32VECTOR_ELEMENTS(u32)[0] = in4.s_addr;
             return SCM_TRUE;
         } else {
             return Scm_MakeIntegerU(ntohl(in4.s_addr));
@@ -387,17 +383,16 @@ ScmObj Scm_InetStringToAddress(const char *s,
             if (Scm_UVectorSizeInBytes(buf) < 16) {
                 Scm_Error("uniform vector buffer isn't big enough to hold IPv6 address: %S", buf);
             }
-            SCM_U32VECTOR_ELEMENTS(buf)[0] = S6_ADDR32(in6, 0);
-            SCM_U32VECTOR_ELEMENTS(buf)[1] = S6_ADDR32(in6, 1);
-            SCM_U32VECTOR_ELEMENTS(buf)[2] = S6_ADDR32(in6, 2);
-            SCM_U32VECTOR_ELEMENTS(buf)[3] = S6_ADDR32(in6, 3);
+            ScmU8Vector *u8 = (ScmU8Vector*)buf;
+            for (int i=0; i<16; i++) {
+                SCM_U8VECTOR_ELEMENTS(u8)[i] = in6.s6_addr[i];
+            }
             return SCM_TRUE;
         } else {
             ScmObj s = SCM_MAKE_INT(0);
-            int i;
-            for (i=0; i<4; i++) {
-                s = Scm_Add(Scm_Ash(s, 32),
-                            Scm_MakeIntegerU(ntohl(S6_ADDR32(in6, i))));
+            for (int i=0; i<16; i++) {
+                s = Scm_Add(Scm_Ash(s, 8),
+                            SCM_MAKE_INT(in6.s6_addr[i]));
             }
             return s;
         }
@@ -434,22 +429,20 @@ ScmObj Scm_InetAddressToString(ScmObj addr,  /* integer or uvector */
         char buf[INET6_ADDRSTRLEN];
         struct in6_addr in6;
         if (SCM_INTEGERP(addr)) {
-            u_long a;
-            int i;
-            ScmObj mask = Scm_MakeIntegerU(0xffffffffUL);
-            for (i=0; i<4; i++) {
-                a = Scm_GetIntegerU(Scm_LogAnd(addr, mask));
-                S6_ADDR32(in6, 3-i) = htonl(a);
-                addr = Scm_Ash(addr, -32);
+            ScmObj mask = SCM_MAKE_INT(0xff);
+            for (int i=0; i<16; i++) {
+                u_long a = Scm_GetIntegerU(Scm_LogAnd(addr, mask));
+                in6.s6_addr[15-i] = a;
+                addr = Scm_Ash(addr, -8);
             }
         } else if (SCM_UVECTORP(addr)) {
             if (Scm_UVectorSizeInBytes(SCM_UVECTOR(addr)) < 16) {
                 Scm_Error("uvector too short for IPv6 address: %S", addr);
             }
-            S6_ADDR32(in6, 0) = SCM_U32VECTOR_ELEMENTS(addr)[0];
-            S6_ADDR32(in6, 1) = SCM_U32VECTOR_ELEMENTS(addr)[1];
-            S6_ADDR32(in6, 2) = SCM_U32VECTOR_ELEMENTS(addr)[2];
-            S6_ADDR32(in6, 3) = SCM_U32VECTOR_ELEMENTS(addr)[3];
+            ScmU8Vector *u8 = SCM_U8VECTOR(addr);
+            for (int i=0; i<16; i++) {
+                in6.s6_addr[i] = SCM_U8VECTOR_ELEMENTS(u8)[i];
+            }
         } else {
             Scm_TypeError("address", "integer or uvector", addr);
         }
@@ -485,4 +478,3 @@ void Scm_Init_NetAddr(ScmModule *mod)
     Scm_InitStaticClass(&Scm_SockAddrIn6Class, "<sockaddr-in6>", mod, NULL, 0);
 #endif /* HAVE_IPV6 */
 }
-
