@@ -1,7 +1,7 @@
 /*
  * net.c - network interface
  *
- *   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -39,17 +39,27 @@
  * Socket
  */
 
+ScmClass *socket_cpl[] = {
+    SCM_CLASS_STATIC_PTR(Scm_ConnectionClass),
+    SCM_CLASS_STATIC_PTR(Scm_TopClass),
+    NULL
+};
+
 static void socket_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 
-SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SocketClass, socket_print);
+SCM_DEFINE_BUILTIN_CLASS(Scm_SocketClass, socket_print, NULL, NULL, NULL,
+                         socket_cpl);
 
-static void socket_finalize(ScmObj obj, void *data)
+static void socket_finalize(ScmObj obj, void *data SCM_UNUSED)
 {
     ScmSocket *sock = (ScmSocket*)obj;
     /* NB: at this point, sock->inPort and sock->outPort may already
        be GC-ed and finalized, so we don't flush them here.
        Clearing them would help them to be collected earlier, though. */
     if (!(SOCKET_CLOSED(sock->fd))) {
+#if defined(GAUCHE_WINDOWS)
+        if (sock->cfd >= 0) { close(sock->cfd); sock->cfd = -1; }
+#endif /*GAUCHE_WINDOWS*/
         closeSocket(sock->fd);
         sock->fd = INVALID_SOCKET;
         sock->status = SCM_SOCKET_STATUS_CLOSED;
@@ -58,7 +68,8 @@ static void socket_finalize(ScmObj obj, void *data)
     }
 }
 
-static void socket_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static void socket_print(ScmObj obj, ScmPort *port, 
+                         ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmSocket *sock = SCM_SOCKET(obj);
     Scm_Printf(port, "#<socket");
@@ -98,7 +109,7 @@ ScmSocket *make_socket(Socket fd, int type)
     s->name = NULL;
     s->type = type;
 #if defined(GAUCHE_WINDOWS)
-    s->infd = s->outfd = -1;
+    s->cfd = -1;
 #endif /*GAUCHE_WINDOWS*/
     Scm_RegisterFinalizer(SCM_OBJ(s), socket_finalize, NULL);
     return s;
@@ -153,8 +164,7 @@ ScmObj Scm_SocketClose(ScmSocket *s)
         s->outPort = NULL;
     }
 #if defined(GAUCHE_WINDOWS)
-    if (s->infd >= 0)  { close(s->infd);  s->infd = -1 ; }
-    if (s->outfd >= 0) { close(s->outfd); s->outfd = -1 ; }
+    if (s->cfd >= 0) { close(s->cfd); s->cfd = -1; }
 #endif /*GAUCHE_WINDOWS*/
     closeSocket(s->fd);
     s->fd = INVALID_SOCKET;
@@ -180,10 +190,13 @@ ScmObj Scm_SocketInputPort(ScmSocket *sock, int buffering)
 #ifndef GAUCHE_WINDOWS
         infd = sock->fd;
 #else  /*GAUCHE_WINDOWS*/
-        /* infd will be closed when this socket is closed. */
-        sock->infd = infd = _open_osfhandle(sock->fd, O_RDONLY);
+        /* cfd will be closed when this socket is closed. */
+        if (sock->cfd < 0) {
+            sock->cfd = _open_osfhandle(sock->fd, 0);
+        }
+        infd = sock->cfd;
 #endif /*GAUCHE_WINDOWS*/
-        if (infd == INVALID_SOCKET) sockport_err(sock, "input");
+        if (SOCKET_INVALID(infd)) sockport_err(sock, "input");
 
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
@@ -207,10 +220,13 @@ ScmObj Scm_SocketOutputPort(ScmSocket *sock, int buffering)
 #ifndef GAUCHE_WINDOWS
         outfd = sock->fd;
 #else  /*GAUCHE_WINDOWS*/
-        /* outfd will be closed when this socket is closed. */
-        sock->outfd = outfd = _open_osfhandle(sock->fd, 0);
+        /* cfd will be closed when this socket is closed. */
+        if (sock->cfd < 0) {
+            sock->cfd = _open_osfhandle(sock->fd, 0);
+        }
+        outfd = sock->cfd;
 #endif /*GAUCHE_WINDOWS*/
-        if (outfd == INVALID_SOCKET) sockport_err(sock, "output");
+        if (SOCKET_INVALID(outfd)) sockport_err(sock, "output");
 
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
@@ -276,7 +292,12 @@ ScmObj Scm_SocketAccept(ScmSocket *sock)
     ScmClass *addrClass = Scm_ClassOf(SCM_OBJ(sock->address));
 
     CLOSE_CHECK(sock->fd, "accept from", sock);
+#if defined(GAUCHE_WINDOWS)
+    SCM_SYSCALL3(newfd, accept(sock->fd, (struct sockaddr*)&addrbuf, &addrlen),
+                 SOCKET_INVALID(newfd));
+#else  /* !GAUCHE_WINDOWS */
     SCM_SYSCALL(newfd, accept(sock->fd, (struct sockaddr*)&addrbuf, &addrlen));
+#endif /* !GAUCHE_WINDOWS */
     if (SOCKET_INVALID(newfd)) {
         if (errno == EAGAIN) {
             return SCM_FALSE;
@@ -334,7 +355,7 @@ ScmObj Scm_SocketGetPeerName(ScmSocket *sock)
     return SCM_OBJ(Scm_MakeSockAddr(NULL, (struct sockaddr*)&addrbuf, addrlen));
 }
 
-static const char *get_message_body(ScmObj msg, u_int *size)
+static const char *get_message_body(ScmObj msg, ScmSmallInt *size)
 {
     if (SCM_UVECTORP(msg)) {
         *size = Scm_UVectorSizeInBytes(SCM_UVECTOR(msg));
@@ -351,7 +372,7 @@ static const char *get_message_body(ScmObj msg, u_int *size)
 ScmObj Scm_SocketSend(ScmSocket *sock, ScmObj msg, int flags)
 {
     int r;
-    u_int size;
+    ScmSmallInt size;
     CLOSE_CHECK(sock->fd, "send to", sock);
     const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, send(sock->fd, cmsg, size, flags));
@@ -363,7 +384,7 @@ ScmObj Scm_SocketSendTo(ScmSocket *sock, ScmObj msg, ScmSockAddr *to,
                         int flags)
 {
     int r;
-    u_int size;
+    ScmSmallInt size;
     CLOSE_CHECK(sock->fd, "send to", sock);
     const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, sendto(sock->fd, cmsg, size, flags,
@@ -376,13 +397,16 @@ ScmObj Scm_SocketSendMsg(ScmSocket *sock, ScmObj msg, int flags)
 {
 #if !GAUCHE_WINDOWS
     int r;
-    u_int size;
+    ScmSmallInt size;
     CLOSE_CHECK(sock->fd, "send to", sock);
     const char *cmsg = get_message_body(msg, &size);
     SCM_SYSCALL(r, sendmsg(sock->fd, (struct msghdr*)cmsg, flags));
     if (r < 0) Scm_SysError("sendmsg(2) failed");
     return SCM_MAKE_INT(r);
 #else  /*GAUCHE_WINDOWS*/
+    (void)sock;  /* suppress unused var warning */
+    (void)msg;   /* suppress unused var warning */
+    (void)flags; /* suppress unused var warning */
     Scm_Error("sendmsg is not implemented on this platform.");
     return SCM_UNDEFINED;       /* dummy */
 #endif /*GAUCHE_WINDOWS*/
@@ -485,7 +509,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
 {
 #if !GAUCHE_WINDOWS
     struct msghdr *msg;
-    int bufsiz = 0;
+    ScmSize bufsiz = 0;
     char *bufptr = 0;
 
     if (buf != NULL) {
@@ -493,7 +517,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         bufptr = (char*)SCM_UVECTOR_ELEMENTS(buf);
     }
 
-    if (bufsiz >= sizeof(struct msghdr)) {
+    if (bufsiz >= (ScmSize)sizeof(struct msghdr)) {
         msg = (struct msghdr*)bufptr;
         bufptr += sizeof(struct msghdr); bufsiz -= sizeof(struct msghdr);
     } else {
@@ -517,9 +541,9 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         } else {
             msg->msg_iov = SCM_NEW_ARRAY(struct iovec, msg->msg_iovlen);
         }
-        for (int i=0; i < msg->msg_iovlen; i++) {
+        for (ScmSize i=0; i < (ScmSize)msg->msg_iovlen; i++) {
             ScmObj elt = SCM_VECTOR_ELEMENT(iov, i);
-            u_int iovlen;
+            ScmSmallInt iovlen;
             msg->msg_iov[i].iov_base = (char*)get_message_body(elt, &iovlen);
             msg->msg_iov[i].iov_len  = iovlen;
         }
@@ -533,7 +557,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         int ctrllen = 0;
 
         SCM_FOR_EACH(cp, control) {
-            u_int clen;
+            ScmSmallInt clen;
             ScmObj c = SCM_CAR(cp);
             if (!(Scm_Length(c) == 3
                   && SCM_INTP(SCM_CAR(c))
@@ -553,7 +577,7 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
         }
         struct cmsghdr *cmsg = CMSG_FIRSTHDR(msg);
         SCM_FOR_EACH(cp, control) {
-            u_int clen;
+            ScmSmallInt clen;
             ScmObj c = SCM_CAR(cp);
             const char *cdata = get_message_body(SCM_CAR(SCM_CDDR(c)), &clen);
             cmsg->cmsg_level = SCM_INT_VALUE(SCM_CAR(c));
@@ -571,6 +595,11 @@ ScmObj Scm_SocketBuildMsg(ScmSockAddr *name, ScmVector *iov,
     if (buf != NULL) return SCM_OBJ(buf);
     else return Scm_MakeUVector(SCM_CLASS_U8VECTOR, sizeof(struct msghdr), msg);
 #else  /*GAUCHE_WINDOWS*/
+    (void)name;    /* suppress unused var warning */
+    (void)iov;     /* suppress unused var warning */
+    (void)control; /* suppress unused var warning */
+    (void)flags;   /* suppress unused var warning */
+    (void)buf;     /* suppress unused var warning */
     Scm_Error("buildmsg is not implemented on this platform.");
     return SCM_UNDEFINED;
 #endif /*GAUCHE_WINDOWS*/
@@ -587,7 +616,7 @@ ScmObj Scm_SocketSetOpt(ScmSocket *s, int level, int option, ScmObj value)
     int r = 0;
     CLOSE_CHECK(s->fd, "set a socket option of", s);
     if (SCM_STRINGP(value)) {
-        u_int size;
+        ScmSmallInt size;
         const char *cvalue = Scm_GetStringContent(SCM_STRING(value), &size,
                                                   NULL, NULL);
         SCM_SYSCALL(r, setsockopt(s->fd, level, option, cvalue, size));
@@ -626,35 +655,112 @@ ScmObj Scm_SocketGetOpt(ScmSocket *s, int level, int option, int rsize)
     }
 }
 
+#if HAVE_STRUCT_IFREQ
+static void ioctl_by_ifr_name(int fd, struct ifreq *ifr, ScmObj data,
+                              unsigned long req, const char *req_name)
+{
+    if (!SCM_STRINGP(data)) {
+        Scm_Error("string expected for %s ioctl argument, but got %s",
+                  req_name, data);
+    }
+    strncpy(ifr->ifr_name, Scm_GetStringConst(SCM_STRING(data)), IFNAMSIZ-1);
+    int r;
+    SCM_SYSCALL(r, ioctl(fd, req, ifr));
+    if (r < 0) Scm_SysError("ioctl(%s) failed", req_name);
+}
+#endif
+
 /* Low-level ioctl. */
-ScmObj Scm_SocketIoctl(ScmSocket *s, int request, ScmObj data)
+ScmObj Scm_SocketIoctl(ScmSocket *s, u_long request, ScmObj data)
 {
-    int r = 0;
 #if HAVE_STRUCT_IFREQ
     struct ifreq ifreq_pkt;
 
     CLOSE_CHECK(s->fd, "ioctl on", s);
     memset(&ifreq_pkt, 0, sizeof(ifreq_pkt));
     switch (request) {
+#if defined(SIOCGIFNAME)
+    case SIOCGIFNAME:
+        if (!SCM_UINTEGERP(data))
+            Scm_TypeError("SIOCGIFNAME ioctl argument" , "unsigned integer", data);
+#if   HAVE_STRUCT_IFREQ_IFR_IFINDEX
+        ifreq_pkt.ifr_ifindex = Scm_GetIntegerU(data);
+#elif HAVE_STRUCT_IFREQ_IFR_INDEX
+        ifreq_pkt.ifr_index = Scm_GetIntegerU(data);
+#endif /*HAVE_STRUCT_IFREQ_IFR_INDEX*/
+        {
+            int r;
+            SCM_SYSCALL(r, ioctl(s->fd, SIOCGIFNAME, &ifreq_pkt));
+            if (r < 0) Scm_SysError("ioctl(SIOCGIFNAME) failed");
+        }
+        return Scm_MakeString(ifreq_pkt.ifr_name, -1, -1, SCM_STRING_COPYING);
+#endif /*SIOCGIFNAME*/
 #if defined(SIOCGIFINDEX)
     case SIOCGIFINDEX:
-        if (!SCM_STRINGP(data)) {
-            Scm_TypeError("SIOCGIFINDEX ioctl argument", "string", data);
-        }
-        strncpy(ifreq_pkt.ifr_name, Scm_GetStringConst(SCM_STRING(data)),
-                IFNAMSIZ-1);
-        SCM_SYSCALL(r, ioctl(s->fd, SIOCGIFINDEX, &ifreq_pkt));
-        if (r < 0) Scm_SysError("ioctl(SIOCGIFINDEX) failed");
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFINDEX, "SIOCGIFINDEX");
 #if HAVE_STRUCT_IFREQ_IFR_IFINDEX
         return Scm_MakeInteger(ifreq_pkt.ifr_ifindex);
 #elif HAVE_STRUCT_IFREQ_IFR_INDEX
-       return Scm_MakeInteger(ifreq_pkt.ifr_index);
+        return Scm_MakeInteger(ifreq_pkt.ifr_index);
 #endif /*HAVE_STRUCT_IFREQ_IFR_INDEX*/
 #endif /*SIOCGIFINDEX*/
+#if defined(SIOCGIFFLAGS)
+    case SIOCGIFFLAGS:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFFLAGS, "SIOCGIFFLAGS");
+        return Scm_MakeInteger(ifreq_pkt.ifr_flags);
+#endif /*SIOCGIFFLAGS*/
+#if defined(SIOCGIFADDR)
+    case SIOCGIFADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFADDR, "SIOCGIFADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_addr,
+                                sizeof(ifreq_pkt.ifr_addr));
+#endif /*SIOCGIFADDR*/
+#if defined(SIOCGIFDSTADDR)
+    case SIOCGIFDSTADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFDSTADDR, "SIOCGIFDSTADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_dstaddr,
+                                sizeof(ifreq_pkt.ifr_dstaddr));
+#endif  /*SIOCGIFDSTADDR*/
+#if defined(SIOCGIFBRDADDR)
+    case SIOCGIFBRDADDR:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFBRDADDR, "SIOCGIFBRDADDR");
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_broadaddr,
+                                sizeof(ifreq_pkt.ifr_broadaddr));
+#endif  /*SIOCGIFBRDADDR*/
+#if defined(SIOCGIFNETMASK)
+    case SIOCGIFNETMASK:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFNETMASK, "SIOCGIFNETMASK");
+#if defined(HAVE_STRUCT_IFREQ_IFR_NETMASK)
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_netmask,
+                                sizeof(ifreq_pkt.ifr_netmask));
+#else
+        return Scm_MakeSockAddr(NULL,
+                                &ifreq_pkt.ifr_addr,
+                                sizeof(ifreq_pkt.ifr_addr));
+#endif /*HAVE_STRUCT_IFREQ_IFR_NETMASK*/
+#endif  /*SIOCGIFNETMASK*/
+#if defined(SIOCGIFMTU)
+    case SIOCGIFMTU:
+        ioctl_by_ifr_name(s->fd, &ifreq_pkt, data,
+                          SIOCGIFMTU, "SIOCGIFMTU");
+        return Scm_MakeInteger(ifreq_pkt.ifr_mtu);
+#endif  /*SIOCGIFMTU*/
     default:
         Scm_Error("unsupported ioctl operation: %d", request);
     }
 #else  /*!HAVE_STRUCT_IFREQ*/
+    (void)s;    /* suppress unused var warning */
+    (void)data; /* suppress unused var warning */
     Scm_Error("unsupported ioctl operation: %d", request);
 #endif /*!HAVE_STRUCT_IFREQ*/
     return SCM_UNDEFINED;       /* dummy */
@@ -662,8 +768,9 @@ ScmObj Scm_SocketIoctl(ScmSocket *s, int request, ScmObj data)
 
 /*==================================================================
  * Windows/MinGW compatibility layer
+ * Older MinGW lacks inet_pton/ntop.  
  */
-#if defined(GAUCHE_WINDOWS)
+#if defined(GAUCHE_WINDOWS) && !defined(InetNtopA)
 
 int inet_pton(int af, const char *src, void *dst)
 {
@@ -730,7 +837,7 @@ const char *inet_ntop(int af, const void *src, char *dst, socklen_t size)
     return dst;
 #undef ADDR_MAXLEN
 }
-#endif /*GAUCHE_WINDOWS*/
+#endif /*defined(GAUCHE_WINDOWS) && !defined(InetNtopA)*/
 
 /*==================================================================
  * Initialization
@@ -748,9 +855,6 @@ SCM_EXTENSION_ENTRY void Scm_Init_gauche__net(void)
 {
     SCM_INIT_EXTENSION(gauche__net);
     ScmModule *mod = SCM_FIND_MODULE("gauche.net", SCM_FIND_MODULE_CREATE);
-#ifdef HAVE_IPV6
-    Scm_AddFeature("gauche.net.ipv6", NULL);
-#endif
     Scm_InitStaticClass(&Scm_SocketClass, "<socket>", mod, NULL, 0);
     Scm_Init_NetAddr(mod);
     Scm_Init_NetDB(mod);
