@@ -1,12 +1,12 @@
 /*
  * netdb.c - obtain information about network
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -37,12 +37,10 @@
 #define DATA_BUFSIZ  980
 
 static struct netdb_data_rec {
-    int dummy;                  /* dummy var to make sure this structure is
-                                   placed in data area. */
     ScmInternalMutex hostent_mutex;
     ScmInternalMutex protoent_mutex;
     ScmInternalMutex servent_mutex;
-} netdb_data = { 1 };
+} netdb_data;
 
 #define WITH_GLOBAL_LOCK(mutex, body)           \
     SCM_UNWIND_PROTECT {                        \
@@ -64,14 +62,13 @@ static ScmSysHostent *make_hostent(struct hostent *he)
 {
     ScmSysHostent *entry = SCM_NEW(ScmSysHostent);
     ScmObj h = SCM_NIL, t = SCM_NIL;
-    char **p;
-    
+
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_HOSTENT);
     entry->name = SCM_MAKE_STR_COPYING(he->h_name);
     entry->aliases = Scm_CStringArrayToList((const char**)he->h_aliases, -1,
                                             SCM_STRING_COPYING);
     if (he->h_addrtype == AF_INET) {
-        for (p = he->h_addr_list; *p; p++) {
+        for (char **p = he->h_addr_list; *p; p++) {
             char buf[50];
             struct in_addr *addr = (struct in_addr*)*p;
             unsigned long addrval = ntohl(addr->s_addr);
@@ -97,7 +94,7 @@ ScmObj Scm_GetHostByName(const char *name)
         int herr = 0, bufsiz = DATA_BUFSIZ;
         struct hostent he;
         char staticbuf[DATA_BUFSIZ], *buf = staticbuf;
-        
+
         for (;;) {
 #if GETHOSTBYNAME_R_NUMARGS == 5
             if (gethostbyname_r(name, &he, buf, bufsiz, &herr) != NULL) break;
@@ -140,7 +137,7 @@ ScmObj Scm_GetHostByAddr(const char *addr, int type)
     if (inet_pton(AF_INET, addr, &iaddr) <= 0) {
         Scm_Error("bad inet address format: %s", addr);
     }
-    
+
 #if defined(GETHOSTBYADDR_R_NUMARGS)
     {
         ScmObj entry = SCM_FALSE;
@@ -175,7 +172,7 @@ ScmObj Scm_GetHostByAddr(const char *addr, int type)
     {
         volatile ScmObj entry = SCM_FALSE;
         struct hostent *he;
-        
+
         WITH_GLOBAL_LOCK(netdb_data.hostent_mutex,
                          do {
                              he = gethostbyaddr((void*)&iaddr,
@@ -221,7 +218,7 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SysProtoentClass, NULL);
 static ScmSysProtoent *make_protoent(struct protoent *pe)
 {
     ScmSysProtoent *entry = SCM_NEW(ScmSysProtoent);
-    
+
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_PROTOENT);
     entry->name = SCM_MAKE_STR_COPYING(pe->p_name);
     entry->aliases = Scm_CStringArrayToList((const char **)pe->p_aliases, -1,
@@ -238,7 +235,7 @@ ScmObj Scm_GetProtoByName(const char *name)
         int bufsiz = DATA_BUFSIZ;
         struct protoent pe;
         char staticbuf[DATA_BUFSIZ], *buf = staticbuf;
-        
+
         for (;;) {
 #if GETPROTOBYNAME_R_NUMARGS == 4
             if (getprotobyname_r(name, &pe, buf, bufsiz) != NULL) break;
@@ -278,7 +275,7 @@ ScmObj Scm_GetProtoByNumber(int number)
         int bufsiz = DATA_BUFSIZ;
         struct protoent pe;
         char staticbuf[DATA_BUFSIZ], *buf = staticbuf;
-        
+
         for (;;) {
 #if GETPROTOBYNUMBER_R_NUMARGS == 4
             if (getprotobynumber_r(number, &pe, buf, bufsiz) != NULL) break;
@@ -341,7 +338,7 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SysServentClass, NULL);
 static ScmSysServent *make_servent(struct servent *se)
 {
     ScmSysServent *entry = SCM_NEW(ScmSysServent);
-    
+
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_SERVENT);
     entry->name = SCM_MAKE_STR_COPYING(se->s_name);
     entry->aliases = Scm_CStringArrayToList((const char **)se->s_aliases, -1,
@@ -359,7 +356,7 @@ ScmObj Scm_GetServByName(const char *name, const char *proto)
         int bufsiz = DATA_BUFSIZ;
         struct servent se;
         char staticbuf[DATA_BUFSIZ], *buf = staticbuf;
-        
+
         for (;;) {
 #if GETSERVBYNAME_R_NUMARGS == 5
             if (getservbyname_r(name, proto, &se, buf, bufsiz) != NULL) break;
@@ -399,7 +396,7 @@ ScmObj Scm_GetServByPort(int port, const char *proto)
         int bufsiz = DATA_BUFSIZ;
         struct servent se;
         char staticbuf[DATA_BUFSIZ], *buf = staticbuf;
-        
+
         for (;;) {
 #if GETSERVBYPORT_R_NUMARGS == 5
             if (getservbyport_r(htons(port), proto, &se, buf, bufsiz) != NULL)
@@ -467,10 +464,9 @@ static ScmClassStaticSlotSpec servent_slots[] = {
 
 #ifdef HAVE_IPV6
 
-ScmObj addrinfo_allocate(ScmClass *klass, ScmObj initargs)
+ScmObj addrinfo_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmSysAddrinfo *info = SCM_ALLOCATE(ScmSysAddrinfo, klass);
-    SCM_SET_CLASS(info, klass);
+    ScmSysAddrinfo *info = SCM_NEW_INSTANCE(ScmSysAddrinfo, klass);
     info->canonname = NULL;
     info->addr = NULL;
     return SCM_OBJ(info);
@@ -500,13 +496,22 @@ ScmObj Scm_GetAddrinfo(const char *nodename,
                        struct addrinfo *hints)
 {
     ScmObj h = SCM_NIL, t = SCM_NIL;
-    struct addrinfo *res, *res0;
-    int r;
+    struct addrinfo *res0;
 
-    r = getaddrinfo(nodename, servname, hints, &res0);
-    if (r) Scm_Error("getaddrinfo failed: %s", gai_strerror(r));
+    int r = getaddrinfo(nodename, servname, hints, &res0);
+#if !defined(GAUCHE_WINDOWS)
+    if (r) {
+        if (r == EAI_SYSTEM) {
+            Scm_SysError("getaddrinfo failed: %s", gai_strerror(r));
+        } else {
+            Scm_Error("getaddrinfo failed: %s", gai_strerror(r));
+        }
+    }
+#else  /*GAUCHE_WINDOWS*/
+    if (r) Scm_SysError("getaddrinfo failed");
+#endif /*GAUCHE_WINDOWS*/
 
-    for (res = res0; res != NULL; res = res->ai_next) {
+    for (struct addrinfo *res = res0; res != NULL; res = res->ai_next) {
         SCM_APPEND1(h, t, SCM_OBJ(make_addrinfo(res)));
     }
     freeaddrinfo(res0);
@@ -516,11 +521,14 @@ ScmObj Scm_GetAddrinfo(const char *nodename,
 ScmObj Scm_GetNameinfo(ScmSockAddr *addr, int flags)
 {
     char host[NI_MAXHOST], serv[NI_MAXSERV];
-    int r;
 
-    r = getnameinfo(&addr->addr, addr->addrlen,
+    int r = getnameinfo(&addr->addr, addr->addrlen,
                     host, sizeof(host), serv, sizeof(serv), flags);
+#if !defined(GAUCHE_WINDOWS)
     if (r) Scm_Error("getnameinfo failed: %s", gai_strerror(r));
+#else  /*GAUCHE_WINDOWS*/
+    if (r) Scm_SysError("getnameinfo failed");
+#endif /*GAUCHE_WINDOWS*/
     return Scm_Values2(SCM_MAKE_STR_COPYING(host), SCM_MAKE_STR_COPYING(serv));
 }
 
