@@ -33,6 +33,9 @@
 
 (use gauche.test)
 (use srfi-1)
+(use srfi-13)
+(use gauche.generator)
+(use gauche.parameter)
 
 (test-start "parser.peg")
 
@@ -44,34 +47,33 @@
 ;;; helper macros
 ;;;
 (define (%compare test-type accessors)
-  (lambda (e o)
-    (and (test-type o)
-         (every equal? e (map (cut <> o) accessors)))))
+  (^[e o] (and (test-type o)
+               (every equal? e (map (cut <> o) accessors)))))
 
 (define-syntax test-succ
   (syntax-rules ()
-    ((_ label expect parse input)
-     (test* #`",label (success)" expect
-            (peg-parse-string parse input)))))
+    [(_ label expect parse input)
+     (test* #"~label (success)" expect
+            (peg-parse-string parse input))]))
 
 (define-syntax test-fail
   (syntax-rules ()
-    ((_ label expect parse input)
-     (test* #`",label (failure)" expect
-            (guard (e ((<parse-error> e)
-                       (list (ref e 'position) (ref e 'objects)))
-                      (else e))
+    [(_ label expect parse input)
+     (test* #"~label (failure)" expect
+            (guard (e [(<parse-error> e)
+                       (list (ref e 'position) (ref e 'objects))]
+                      [else e])
               (peg-parse-string parse input)
-              (error "test-fail failed"))))))
+              (error "test-fail failed")))]))
 
 ;;;============================================================
 ;;; Ropes
 ;;;
 (receive (make-rope rope->string)
     (with-module parser.peg (values make-rope rope->string))
-  (let* ((rope1 (make-rope "abc"))
-         (rope2 (make-rope `("012" ,rope1 #\X)))
-         (rope3 (make-rope 'foo)))
+  (let* ([rope1 (make-rope "abc")]
+         [rope2 (make-rope `("012" ,rope1 #\X))]
+         [rope3 (make-rope 'foo)])
     (test* "rope->string" "abc"        (rope->string rope1))
     (test* "rope->string" "012abcX"    (rope->string rope2))
     (test* "rope->string" (test-error) (rope->string rope3))))
@@ -118,41 +120,6 @@
 (test-succ "$string-ci" "aBC" ($string-ci "abc") "aBCdef")
 (test-fail "$string-ci" '(0 "abc") ($string-ci "abc") "012")
 
-(define-syntax test-char
-  (syntax-rules ()
-    ((_ parse expin unexpin message)
-     (begin
-       (test-succ (symbol->string 'parse)
-                  (string-ref expin 0)
-                  parse expin)
-       (test-fail (symbol->string 'parse)
-                  '(0 message)
-                  parse unexpin)))))
-
-(test-char anychar  "a" "" "character")
-(test-char upper    "A" "a" "upper case letter")
-(test-char lower    "a" "A" "lower case letter")
-(test-char letter   "a" "." "letter")
-(test-char alphanum "0" "." "letter or digit")
-(test-char digit    "0" "." "digit")
-(test-char hexdigit "f" "." "hexadecimal digit")
-(test-char newline  "\n" "." "newline")
-(test-char tab      "\t" "." "tab")
-(test-char space    "\t" "." "space")
-
-(test-succ "spaces" " " spaces " ")
-(test-succ "spaces" " \n\t" spaces " \n\tj")
-
-(test-succ "$satisfy" #\0
-           ($satisfy char-numeric? "numeric")
-           "0")
-(test-fail "$satisfy" '(0 "numeric")
-           ($satisfy char-numeric? "numeric")
-           "a")
-
-(test-succ "eof" #t eof "")
-(test-fail "eof" '(0 "end of input") eof "a")
-
 ;;;============================================================
 ;;; Combinators
 ;;;
@@ -240,6 +207,51 @@
                 ($string (rope-finalize v)))
            "foobar")
 
+;; $let, $let*
+(test-succ "$let" '("foo" "baz")
+           ($let ([foo ($string "foo")]
+                  [    ($string "bar")]
+                  [baz ($string "baz")])
+                 ($return (list foo baz)))
+           "foobarbaz$$$")
+
+(test-succ "$let" '(#\0 "zyx" #\0)
+           ($let* ([opener ($one-of #[\d])]
+                   [meat   ($->string ($many ($one-of #[a-z])))]
+                   [closer ($char opener)])
+                  ($return (list opener meat closer)))
+           "0zyx0")
+
+;; $parameterize
+(define param1 (make-parameter "zzz"))
+(define match-param1 ($let ([word ($->string ($many ($one-of #[a-z])))])
+                       (if (string-contains word (param1))
+                         ($fail #"word can't have '~(param1)'")
+                         ($return word))))
+(test-succ "$parameterize" "lineament"
+           ($parameterize ((param1 "cicle"))
+              match-param1)
+           "lineament")
+(test-fail "$parameterize" '(9 "word can't have 'line'")
+           ($parameterize ((param1 "line"))
+              match-param1)
+           "lineament")
+
+;; $lift and $lift*
+(test-succ "$lift" '(#\a . #\b)
+           ($lift cons ($any) ($any))
+           "abc")
+(test-fail "$lift" '(1 #\z)
+           ($lift cons ($any) ($char #\z))
+           "abc")
+
+(test-succ "$lift*" "abc"
+           ($lift* list->string ($any) ($any) ($any))
+           "abc")
+(test-fail "$lift" '(2 #\z)
+           ($lift* list->string ($any) ($any) ($char #\z))
+           "abc")
+
 ;; $fold-parsers and $fold-parsers-right
 (test-succ "$fold-parsers" '()                  ; base case
            ($fold-parsers cons '() '())
@@ -292,9 +304,9 @@
 
 ;; $repeat
 (test-succ "$repeat" '(#\a #\b #\a)
-           ($repeat anychar 3) "abab")
+           ($repeat ($any) 3) "abab")
 (test-fail "$repeat" '(2 "character")
-           ($repeat anychar 3) "ab")
+           ($repeat ($any) 3) "ab")
 
 ;; $optional
 (test-succ "$optional" #\a
@@ -328,7 +340,7 @@
            "a,b,")
 
 ;; $sep-end-by
-(let ((p ($sep-end-by ($seq ($one-of #[a-z]) ($one-of #[a-z])) ($string ","))))
+(let1 p ($sep-end-by ($seq ($one-of #[a-z]) ($one-of #[a-z])) ($string ","))
   (define (succ in exp) (test-succ "$sep-end-by" exp p in))
   (define (fail in exp) (test-fail "$sep-end-by" exp p in))
 
@@ -393,7 +405,7 @@
 
 ;; $many-till
 (test-succ "$many-till" '(#\a #\b)
-           ($many-till alphanum digit)
+           ($many-till ($one-of #[[:alnum:]]) ($one-of #[\d]))
            "ab78")
 
 ;; $many-chars
@@ -414,36 +426,36 @@
            "cbad")
 
 ;; $chain-left
-(let ((integer
-       ($do (v ($many digit 1))
-            ($return (string->number (apply string v)))))
-      (op
-       ($or ($do (($char #\*)) ($return *))
-            ($do (($char #\+)) ($return +)))))
+(let ([integer
+       ($do [v ($many ($one-of #[\d])  1)]
+            ($return (string->number (apply string v))))]
+      [op
+       ($or ($do [($char #\*)] ($return *))
+            ($do [($char #\+)] ($return +)))])
   (test-succ "$chain-left" 9
              ($chain-left integer op)
              "1+2*3")
   (test-succ "$chain-left" 3
              ($chain-left integer op)
              "1+2*")
-  (test-fail "$chain-left" '(0 "digit")
+  (test-fail "$chain-left" '(0 #[0-9])
              ($chain-left integer op)
              "abc"))
 
 ;; $chain-right
-(let ((integer
-       ($do (v ($many digit 1))
-            ($return (string->number (apply string v)))))
-      (op
-       ($or ($do (($char #\*)) ($return *))
-            ($do (($char #\+)) ($return +)))))
+(let ([integer
+       ($do [v ($many ($one-of #[\d])  1)]
+            ($return (string->number (apply string v))))]
+      [op
+       ($or ($do [($char #\*)] ($return *))
+            ($do [($char #\+)] ($return +)))])
   (test-succ "$chain-right" 7
              ($chain-right integer op)
              "1+2*3")
   (test-succ "$chain-right" 3
              ($chain-right integer op)
              "1+2*")
-  (test-fail "$chain-right" '(0 "digit")
+  (test-fail "$chain-right" '(0 #[0-9])
              ($chain-right integer op)
              "abc"))
 
@@ -479,6 +491,21 @@
              "abc+efg")
   )
 
+;;;============================================================
+;;; Error handling
+;;;
+
+(test-section "error handling")
+
+;; compound error
+
+(let* ([p1 ($or ($char #\a) ($char #\b))]
+       [p2 ($seq ($any) ($or ($char #\c) p1))])
+  (test* "nested compound error"
+         (test-error <parse-error> "expecting one of (#\\c #\\a #\\b) at 1")
+         (peg-parse-string p2 "dd")))
+
+
 ;;;============================================================
 ;;; Token Parsers
 ;;;
@@ -532,17 +559,17 @@
        (point    ($seq ($char #\.) ($many ($one-of #[\d]) 1)))
        (exponent ($seq ($one-of #[eE]) digits))
        (number?  ($seq digits ($optional point) ($optional exponent)))
-       (p        ($do [number?] [eof] ($return #t))))
+       (p        ($do [number?] [($eos)] ($return #t))))
   (define (%test str . fails?)
     (if (null? fails?)
-      (test-succ #`"number(1) \",str\"" #t p str)
-      (test-fail #`"number(1) \",str\"" (car fails?) p str)))
+      (test-succ #"number(1) \"~str\"" #t p str)
+      (test-fail #"number(1) \"~str\"" (car fails?) p str)))
   (%test "27652") (%test "-27652") (%test "+27652")
   (%test "" '(0 #[0-9])) (%test "-" '(1 #[0-9])) (%test "+" '(1 #[0-9]))
   (%test "+27.46") (%test "0.46")
   (%test ".46" '(0 #[0-9]))
   (%test "0..3" '(2 #[0-9]))
-  (%test "0.4.2" '(3 "end of input"))
+  (%test "0.4.2" '(3 "end of stream"))
   (%test "3e5")
   (%test "3.0e5")
   (%test "3.e5" '(2 #[0-9]))
@@ -560,7 +587,7 @@
   (define double-dquote ($do [($string "\"\"")] ($return #\")))
   (define quoted-body ($many ($or double-dquote ($one-of #[^\"]))))
   (define quoted ($between dquote quoted-body dquote))
-  (define unquoted ($many-till anychar ($or comma newline)))
+  (define unquoted ($many-till ($any) ($or comma ($char #\newline))))
   (define field ($or quoted unquoted))
   (define record ($sep-by ($->rope field) comma))
   (test-succ "CSV" '("a" "b" "c")
@@ -598,39 +625,107 @@
              records "\"a  \"\" \n\" , b  , c"))
 
 ;; Poor-man's XML parser
-(letrec ((open-tag ($->rope ($between ($char #\<) alphanum ($char #\>))))
-         (close-tag (lambda (tagname)
-                      ($seq ($string "</") ($string tagname) ($char #\>))))
-         (text ($->rope ($many ($none-of #[<]))))
-         (body ($do [t text]
+(letrec ([open-tag ($->rope ($between ($char #\<)
+                                      ($one-of #[[:alnum:]])
+                                      ($char #\>)))]
+         [close-tag (^[tagname]
+                      ($seq ($string "</") ($string tagname) ($char #\>)))]
+         [text ($->rope ($many ($none-of #[<])))]
+         [body ($do [t text]
                     [r ($many ($do [e element]
                                    [t text]
                                    ($return (list e t))))]
-                    ($return `(,t ,@(apply append r)))))
-         (element ($do* [tagname ($try open-tag)]
-                        [body body]
-                        [ (close-tag (rope-finalize tagname)) ]
-                        ($return (cons tagname body)))))
+                    ($return `(,t ,@(apply append r))))]
+         [element ($do [tagname ($try open-tag)]
+                       [body body]
+                       [ (close-tag (rope-finalize tagname)) ]
+                       ($return (cons tagname body)))])
   (test-succ "tag element" '("a" "")
              element "<a></a>")
   (test-succ "tag element" '("a" "foo" ("b" "bar") "baz")
              element "<a>foo<b>bar</b>baz</a>"))
 
 ;; Calculator
-(letrec ((integer ($do (v ($many digit 1))
-                       ($return (string->number (apply string v)))))
-         (mulop ($or ($seq ($char #\*) ($return *))
-                     ($seq ($char #\/) ($return /))))
-         (addop ($or ($seq ($char #\+) ($return +))
-                     ($seq ($char #\-) ($return -))))
-         (factor ($lazy ($or ($between ($char #\() expr ($char #\)))
-                             integer)))
-         (term ($chain-left factor mulop))
-         (expr ($chain-left term addop)))
+(letrec ([integer ($do (v ($many ($one-of #[\d]) 1))
+                       ($return (string->number (apply string v))))]
+         [mulop ($or ($seq ($char #\*) ($return *))
+                     ($seq ($char #\/) ($return /)))]
+         [addop ($or ($seq ($char #\+) ($return +))
+                     ($seq ($char #\-) ($return -)))]
+         [factor ($lazy ($or ($between ($char #\() expr ($char #\)))
+                             integer))]
+         [term ($chain-left factor mulop)]
+         [expr ($chain-left term addop)])
   (test-succ "calculator" 11 expr "1+2*3+4")
   (test-succ "calculator" 36 expr "2/2+5*(3+4)")
   (test-succ "calculator" -1 expr "1-2"))
 
+;;;
+;;; deprecated
+;;;
+
+(test-section "deprecated")
+;; kludge - we can't say (use parser.peg.deprecated) until peg/deprecated.scm
+;; is installed in place.  
+(or (load "peg/deprecated" :error-if-not-found #f)
+    (load "parser/peg/deprecated" :error-if-not-found #f))
+(import parser.peg.deprecated)
+(test-module 'parser.peg.deprecated)
+
+(define-syntax test-char
+  (syntax-rules ()
+    ((_ parse expin unexpin message)
+     (begin
+       (test-succ (symbol->string 'parse)
+                  (string-ref expin 0)
+                  parse expin)
+       (test-fail (symbol->string 'parse)
+                  '(0 message)
+                  parse unexpin)))))
+
+(test-char anychar  "a" "" "character")
+(test-char upper    "A" "a" "upper case letter")
+(test-char lower    "a" "A" "lower case letter")
+(test-char letter   "a" "." "letter")
+(test-char alphanum "0" "." "letter or digit")
+(test-char digit    "0" "." "digit")
+(test-char hexdigit "f" "." "hexadecimal digit")
+(test-char newline  "\n" "." "newline")
+(test-char tab      "\t" "." "tab")
+(test-char space    "\t" "." "space")
+
+(test-succ "spaces" " " spaces " ")
+(test-succ "spaces" " \n\t" spaces " \n\tj")
+
+(test-succ "$satisfy" #\0
+           ($satisfy char-numeric? "numeric")
+           "0")
+(test-fail "$satisfy" '(0 "numeric")
+           ($satisfy char-numeric? "numeric")
+           "a")
+
+(test-succ "$match" #\a
+           ($match (#\c x #\r) x)
+           "car foo")
+(test-succ "$match" #\d
+           ($match (#\c x #\r) x)
+           "cdr foo")
+(test-succ "$match" '(#\a (#\space #\f #\o #\o))
+           ($match (#\c x #\r . t) (list x t))
+           "car foo")
+(test-succ "$match (no result arg)" '(#\c #\u #\r)
+           ($match (#\c x #\r))
+           "cur foo")
+(test-fail "$match" '(0 "(#\\c x #\\r)")
+           ($match (#\c x #\r) x)
+           "zxy foo")
+
+(test-succ "$match1" #\a ($match1 #\a) "abc")
+(test-succ "$match1" #\0 ($match1 (and (? #[0-9] d)) d) "0ab")
+(test-fail "$match1" '(0 "#\\a") ($match1 #\a 'yo) "0ab")
+
+(test-succ "eof" (eof-object) eof "")
+(test-fail "eof" '(0 "end of input") eof "a")
 
 ;;;============================================================
 ;;; rfc.json
@@ -661,7 +756,7 @@
 
 (let ()
   (define (t str)
-    (test* #`"parse error ,str" (test-error <json-parse-error>)
+    (test* #"parse error ~str" (test-error <json-parse-error>)
            (parse-json-string str)))
   (t "{\"x\": 100")
   (t "{x : 100}}")
@@ -731,6 +826,21 @@
    }
 ]"))
 
+(test* "Parsing sequence of json objects"
+       '((("a" . 1)("b" . 2)) (("c" . 3) ("d" . 4)))
+       (with-input-from-string "{\"a\":1, \"b\":2}{\"c\":3, \"d\":4}"
+         parse-json*))
+
+(test* "Customizing constructors"
+       '(object ("x" array 1 2 3) ("y" array #f #t null))
+       (parameterize ([json-array-handler (^[elts] (cons 'array elts))]
+                      [json-object-handler (^[pairs] (cons 'object pairs))]
+                      [json-special-handler (^y (case y
+                                                  [(false) #f]
+                                                  [(true) #t]
+                                                  [(null) 'null]))])
+         (parse-json-string "{\"x\":[1,2,3],\"y\":[false,true,null]}")))
+
 (let ()
   (define (test-writer name obj)
     (test* name obj
@@ -745,6 +855,7 @@
                    ("Url"    . "http://www.example.com/image/481989943")
                    ("Height" . 125)
                    ("Width"  . "100"))
+                  ("Description" . "Foo\nbackslash \\and tab\t and \u00a1")
                   ("IDs" . #(116 943 234 38793))
                   ("Misc" . ()))))
 
@@ -767,13 +878,40 @@
                    ("Country"   . "US"))))
   )
 
+(cond-expand
+ [gauche.ces.utf8
+  (let1 data `(("[\"\\u03bb\"]" #("\x3bb;"))
+               ("[\"\\ud800\"]" ,(test-error <json-parse-error>))
+               ("[\"\\ud867\\ude3d\\u03bb\"]" #("\x29e3d;\x3bb;"))
+               ("[\"\\ude3d\\ud867\"]" ,(test-error <json-parse-error>))
+               ("[\"\\uf020\\u03bb\"]"  #("\xf020;\x3bb;")))
+    (dolist [d data]
+      (test* (format "unicode escape reading (~s)" (car d))
+             (cadr d)
+             (parse-json-string (car d)))
+      (when (vector? (cadr data))
+        (test* (format "unicode escape writing (~s)" (cadr d))
+               (car d)
+               (construct-json-string (cadr d))))))]
+ [else])
+
 (let ()
   (define (t obj)
-    (test* #`"writer error ,obj" (test-error <json-construct-error>) 
+    (test* #"writer error ~obj" (test-error <json-construct-error>)
            (construct-json-string obj)))
   (t "a")
   (t '#(1 2 x))
-  (t '(("a" . 2) (3 . "b")))
   (t '(("a" . 2) 9)))
 
+(test* "generalized array" "[1,2,3]"
+       (construct-json-string '#u8(1 2 3)))
+(test* "generalized object" (test-one-of "{\"a\":1,\"b\":2}"
+                                         "{\"b\":2,\"a\":1}")
+       (construct-json-string (hash-table 'eq? '(a . 1) '(b . 2))))
+
+;; https://sourceforge.net/p/gauche/mailman/message/36786284/
+(test* "read/write invariance"
+       '#(48.529166)
+       (parse-json-string (construct-json-string '#(48.529166))))
+       
 (test-end)
