@@ -1,12 +1,12 @@
 /*
  * spvec.c - Sparse vector
  *
- *   Copyright (c) 2009-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2009-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -33,16 +33,7 @@
 
 #include "spvec.h"
 
-#if SIZEOF_LONG == 4
 #define INDEX_CHECK(n)  /*empty*/
-#else   /* SIZEOF_LONG > 4 */
-#define INDEX_CHECK(n)                                                  \
-    do {                                                                \
-        if ((n) >= (1UL << SPARSE_VECTOR_MAX_INDEX_BITS)) {             \
-            Scm_Error("sparse vector index out of range: %lu\n", (n));  \
-        }                                                               \
-    } while (0)
-#endif  /* SIZEOF_LONG > 4 */
 
 /*===================================================================
  * Generic stuff
@@ -59,28 +50,36 @@ ScmClass *spvec_cpl[] = {
     NULL
 };
 
+ScmClass *spmat_cpl[] = {
+    SCM_CLASS_STATIC_PTR(Scm_SparseMatrixBaseClass),
+    SCM_CLASS_STATIC_PTR(Scm_CollectionClass),
+    SCM_CLASS_STATIC_PTR(Scm_TopClass),
+    NULL
+};
+
 SCM_DEFINE_BUILTIN_CLASS(Scm_SparseVectorBaseClass,
                          NULL, NULL, NULL, NULL,
                          spvec_cpl+1);
 
+SCM_DEFINE_BUILTIN_CLASS(Scm_SparseMatrixBaseClass,
+                         NULL, NULL, NULL, NULL,
+                         spmat_cpl+1);
+
 ScmObj SparseVectorRef(SparseVector *sv, u_long index, ScmObj fallback)
 {
-    ScmObj v;
-    Leaf *leaf;
     INDEX_CHECK(index);
-    leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
+    Leaf *leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
     if (leaf == NULL) return fallback;
-    v = sv->desc->ref(leaf, index);
+    ScmObj v = sv->desc->ref(leaf, index);
     if (SCM_UNBOUNDP(v)) return fallback;
     else return v;
 }
 
 void SparseVectorSet(SparseVector *sv, u_long index, ScmObj value)
 {
-    Leaf *leaf;
     INDEX_CHECK(index);
-    leaf = CompactTrieAdd(&sv->trie, index >> sv->desc->shift,
-                          sv->desc->allocate, sv);
+    Leaf *leaf = CompactTrieAdd(&sv->trie, index >> sv->desc->shift,
+                                sv->desc->allocate, sv);
     /* set returns TRUE if this is a new entry */
     if (sv->desc->set(leaf, index, value)) sv->numEntries++;
 }
@@ -88,12 +87,10 @@ void SparseVectorSet(SparseVector *sv, u_long index, ScmObj value)
 /* returns value of the deleted entry, or SCM_UNBOUND if there's no entry */
 ScmObj SparseVectorDelete(SparseVector *sv, u_long index)
 {
-    ScmObj r;
-    Leaf *leaf;
     INDEX_CHECK(index);
-    leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
+    Leaf *leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
     if (leaf == NULL) return SCM_UNBOUND;
-    r = sv->desc->delete(leaf, index);
+    ScmObj r = sv->desc->delete(leaf, index);
     if (!SCM_UNBOUNDP(r)) sv->numEntries--;
     return r;
 }
@@ -108,6 +105,7 @@ ScmObj SparseVectorCopy(const SparseVector *src)
 {
     SparseVector *dst =
         (SparseVector*)MakeSparseVector(Scm_ClassOf(SCM_OBJ(src)),
+                                        src->defaultValue,
                                         src->flags);
     CompactTrieCopy(&dst->trie, &src->trie, src->desc->copy, src->desc);
     dst->numEntries = src->numEntries;
@@ -129,7 +127,7 @@ ScmObj SparseVectorIterNext(SparseVectorIter *iter)
         if (iter->leaf) {
             ScmObj r = iterproc(iter->leaf, &iter->leafIndex);
             if (!SCM_UNBOUNDP(r)) {
-                u_long ind = ((LEAF_KEY(iter->leaf) << iter->sv->desc->shift)
+                u_long ind = ((leaf_key(iter->leaf) << iter->sv->desc->shift)
                               + iter->leafIndex);
                 return Scm_Cons(Scm_MakeIntegerU(ind), r);
             }
@@ -144,11 +142,17 @@ ScmObj SparseVectorIterNext(SparseVectorIter *iter)
 /* TODO: Allow clamp arg */
 ScmObj SparseVectorInc(SparseVector *sv, u_long index,
                        ScmObj delta,    /* number */
-                       ScmObj fallback) /* number */
+                       ScmObj fallback) /* number or unbound */
 {
-    Leaf *leaf;
+    if (!SCM_NUMBERP(fallback)) {
+        if (SCM_NUMBERP(sv->defaultValue)) {
+            fallback = sv->defaultValue;
+        } else {
+            fallback = SCM_MAKE_INT(0);
+        }
+    }
     INDEX_CHECK(index);
-    leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
+    Leaf *leaf = CompactTrieGet(&sv->trie, index >> sv->desc->shift);
     if (leaf == NULL) {
         ScmObj v = Scm_Add(fallback, delta);
         SparseVectorSet(sv, index, v);
@@ -162,12 +166,10 @@ ScmObj SparseVectorInc(SparseVector *sv, u_long index,
     }
 }
 
-#if SCM_DEBUG_HELPER
 void SparseVectorDump(SparseVector *sv)
 {
     CompactTrieDump(SCM_CUROUT, &sv->trie, sv->desc->dump, sv->desc);
 }
-#endif /*SCM_DEBUG_HELPER*/
 
 /*===================================================================
  * Individual vector types
@@ -175,7 +177,7 @@ void SparseVectorDump(SparseVector *sv)
 
 /*-------------------------------------------------------------------
  * General vector
-v */
+ */
 
 typedef struct GLeafRec {
     Leaf hdr;
@@ -187,7 +189,7 @@ static ScmObj g_ref(Leaf *leaf, u_long index)
     return ((GLeaf*)leaf)->val[index&1];
 }
 
-static Leaf *g_allocate(void *data)
+static Leaf *g_allocate(void *data SCM_UNUSED)
 {
     GLeaf *z = SCM_NEW(GLeaf);
     z->val[0] = z->val[1] = SCM_UNBOUND;
@@ -208,13 +210,13 @@ static ScmObj g_delete(Leaf *leaf, u_long index)
     return v;
 }
 
-static void g_clear(Leaf *leaf, void *data)
+static void g_clear(Leaf *leaf, void *data SCM_UNUSED)
 {
     GLeaf *z = (GLeaf*)leaf;
     z->val[0] = z->val[1] = NULL;
 }
 
-static Leaf *g_copy(Leaf *leaf, void *data)
+static Leaf *g_copy(Leaf *leaf, void *data SCM_UNUSED)
 {
     GLeaf *dst = SCM_NEW(GLeaf);
     *dst = *(GLeaf*)leaf;
@@ -236,20 +238,15 @@ static ScmObj g_iter(Leaf *leaf, int *index)
     return SCM_UNBOUND;
 }
 
-#if SCM_DEBUG_HELPER
-static void g_dump(ScmPort *out, Leaf *leaf, int indent, void *data)
+static void g_dump(ScmPort *out, Leaf *leaf, int indent, void *data SCM_UNUSED)
 {
-    int i;
     GLeaf *z = (GLeaf*)leaf;
-    for (i=0; i<2; i++) {
+    for (int i=0; i<2; i++) {
         if (!SCM_UNBOUNDP(z->val[i])) {
             Scm_Printf(out, "\n  %*s%2d: %25.1S", indent, "", i, z->val[i]);
         }
     }
 }
-#else
-#define g_dump NULL
-#endif /*SCM_DEBUG_HELPER*/
 
 static SparseVectorDescriptor g_desc = {
     g_ref, g_set, g_allocate, g_delete, g_clear, g_copy, g_iter, g_dump, 1
@@ -258,6 +255,9 @@ static SparseVectorDescriptor g_desc = {
 SCM_DEFINE_BUILTIN_CLASS(Scm_SparseVectorClass, NULL, NULL, NULL, NULL,
                          spvec_cpl);
 
+SCM_DEFINE_BUILTIN_CLASS(Scm_SparseMatrixClass, NULL, NULL, NULL, NULL,
+                         spmat_cpl);
+
 /*-------------------------------------------------------------------
  * Uniform sparse vector common stuff
  */
@@ -266,14 +266,14 @@ typedef struct ULeafRec {
     Leaf hdr;                   /* LEAF_DATA has bitmap of used entries */
     union {
         ScmWord dummy[2];       /* placeholder */
-        signed char    s8[2*SIZEOF_LONG];
-        unsigned char  u8[2*SIZEOF_LONG];
-        signed short   s16[SIZEOF_LONG];
-        unsigned short u16[SIZEOF_LONG];
-        ScmInt32       s32[SIZEOF_LONG/2];
-        ScmUInt32      u32[SIZEOF_LONG/2];
-        ScmInt64       s64[SIZEOF_LONG/4];
-        ScmUInt64      u64[SIZEOF_LONG/4];
+        int8_t         s8[2*SIZEOF_LONG];
+        uint8_t        u8[2*SIZEOF_LONG];
+        int16_t        s16[SIZEOF_LONG];
+        uint16_t       u16[SIZEOF_LONG];
+        int32_t        s32[SIZEOF_LONG/2];
+        uint32_t       u32[SIZEOF_LONG/2];
+        int64_t        s64[SIZEOF_LONG/4];
+        uint64_t       u64[SIZEOF_LONG/4];
         ScmHalfFloat   f16[SIZEOF_LONG];
         float          f32[SIZEOF_LONG/2];
         double         f64[SIZEOF_LONG/4];
@@ -299,23 +299,23 @@ typedef struct ULeafRec {
 #define MASK32 ((1UL<<SHIFT32)-1)
 #define MASK64 ((1UL<<SHIFT64)-1)
 
-#define U_HAS_ENTRY(leaf, ind, mask) LEAF_DATA_BIT_TEST(leaf, (ind)&(mask))
-#define U_SET_ENTRY(leaf, ind, mask) LEAF_DATA_BIT_SET(leaf, (ind)&(mask))
-#define U_RESET_ENTRY(leaf, ind, mask) LEAF_DATA_BIT_RESET(leaf, (ind)&(mask))
+#define U_HAS_ENTRY(leaf, ind, mask) leaf_data_bit_test(leaf, (ind)&(mask))
+#define U_SET_ENTRY(leaf, ind, mask) leaf_data_bit_set(leaf, (ind)&(mask))
+#define U_RESET_ENTRY(leaf, ind, mask) leaf_data_bit_reset(leaf, (ind)&(mask))
 
-static Leaf *u_allocate(void *data)
+static Leaf *u_allocate(void *data SCM_UNUSED)
 {
     ULeaf *z = SCM_NEW_ATOMIC(ULeaf);
     z->dummy[0] = z->dummy[1] = 0;
     return (Leaf*)z;
 }
 
-static void u_clear(Leaf *leaf, void *data)
+static void u_clear(Leaf *leaf SCM_UNUSED, void *data SCM_UNUSED)
 {
     /* nothing to do */
 }
 
-static Leaf *u_copy(Leaf *leaf, void *data)
+static Leaf *u_copy(Leaf *leaf, void *data SCM_UNUSED)
 {
     ULeaf *z = SCM_NEW_ATOMIC(ULeaf);
     memcpy(z, leaf, sizeof(ULeaf));
@@ -360,7 +360,7 @@ U_REF(f64, MASK64, Scm_VMReturnFlonum)
         ULEAF(leaf)->elt[index&mask] = getter(val, SCM_CLAMP_ERROR, NULL); \
         U_SET_CHECK(mask);                                              \
     } while(0)
-    
+
 #define U_SET_CHECK(mask)                       \
     do {                                        \
         int z = U_HAS_ENTRY(leaf, index, mask); \
@@ -544,7 +544,9 @@ U_ITER(f64, MASK64)
         NULL, shift,                                                    \
     };                                                                  \
     SCM_DEFINE_BUILTIN_CLASS(SCM_CPP_CAT3(Scm_Sparse,TAG,VectorClass),  \
-                             NULL, NULL, NULL, NULL, spvec_cpl)
+                             NULL, NULL, NULL, NULL, spvec_cpl);        \
+    SCM_DEFINE_BUILTIN_CLASS(SCM_CPP_CAT3(Scm_Sparse,TAG,MatrixClass),  \
+                             NULL, NULL, NULL, NULL, spmat_cpl)
 
 U_DECL(s8, S8, SHIFT8);
 U_DECL(u8, U8, SHIFT8);
@@ -562,10 +564,9 @@ U_DECL(f64, F64, SHIFT64);
  * Generic constructor
  */
 
-ScmObj MakeSparseVector(ScmClass *klass, u_long flags)
+ScmObj MakeSparseVector(ScmClass *klass, ScmObj defaultValue, u_long flags)
 {
     SparseVectorDescriptor *desc = NULL;
-    SparseVector *v = SCM_NEW(SparseVector);
 
     if (SCM_EQ(SCM_CLASS_SPARSE_VECTOR, klass))         desc = &g_desc;
     else if (SCM_EQ(SCM_CLASS_SPARSE_S8VECTOR, klass))  desc = &s8_desc;
@@ -579,16 +580,30 @@ ScmObj MakeSparseVector(ScmClass *klass, u_long flags)
     else if (SCM_EQ(SCM_CLASS_SPARSE_F16VECTOR, klass)) desc = &f16_desc;
     else if (SCM_EQ(SCM_CLASS_SPARSE_F32VECTOR, klass)) desc = &f32_desc;
     else if (SCM_EQ(SCM_CLASS_SPARSE_F64VECTOR, klass)) desc = &f64_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_MATRIX, klass))    desc = &g_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_S8MATRIX, klass))  desc = &s8_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_U8MATRIX, klass))  desc = &u8_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_S16MATRIX, klass)) desc = &s16_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_U16MATRIX, klass)) desc = &u16_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_S32MATRIX, klass)) desc = &s32_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_U32MATRIX, klass)) desc = &u32_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_S64MATRIX, klass)) desc = &s64_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_U64MATRIX, klass)) desc = &u64_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_F16MATRIX, klass)) desc = &f16_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_F32MATRIX, klass)) desc = &f32_desc;
+    else if (SCM_EQ(SCM_CLASS_SPARSE_F64MATRIX, klass)) desc = &f64_desc;
     else {
         Scm_TypeError("class", "subclass of <sparse-vector-base>",
                       SCM_OBJ(klass));
     }
-    
+
+    SparseVector *v = SCM_NEW(SparseVector);
     SCM_SET_CLASS(v, klass);
     CompactTrieInit(&v->trie);
     v->numEntries = 0;
     v->desc = desc;
     v->flags = flags;
+    v->defaultValue = defaultValue;
     return SCM_OBJ(v);
 }
 
@@ -614,5 +629,18 @@ void Scm_Init_spvec(ScmModule *mod)
     INITC(Scm_SparseF16VectorClass, "<sparse-f16vector>");
     INITC(Scm_SparseF32VectorClass, "<sparse-f32vector>");
     INITC(Scm_SparseF64VectorClass, "<sparse-f64vector>");
-}
 
+    INITC(Scm_SparseMatrixBaseClass, "<sparse-matrix-base>");
+    INITC(Scm_SparseMatrixClass, "<sparse-matrix>");
+    INITC(Scm_SparseS8MatrixClass, "<sparse-s8matrix>");
+    INITC(Scm_SparseU8MatrixClass, "<sparse-u8matrix>");
+    INITC(Scm_SparseS16MatrixClass, "<sparse-s16matrix>");
+    INITC(Scm_SparseU16MatrixClass, "<sparse-u16matrix>");
+    INITC(Scm_SparseS32MatrixClass, "<sparse-s32matrix>");
+    INITC(Scm_SparseU32MatrixClass, "<sparse-u32matrix>");
+    INITC(Scm_SparseS64MatrixClass, "<sparse-s64matrix>");
+    INITC(Scm_SparseU64MatrixClass, "<sparse-u64matrix>");
+    INITC(Scm_SparseF16MatrixClass, "<sparse-f16matrix>");
+    INITC(Scm_SparseF32MatrixClass, "<sparse-f32matrix>");
+    INITC(Scm_SparseF64MatrixClass, "<sparse-f64matrix>");
+}
