@@ -1,23 +1,23 @@
 ;;;
 ;;; syslog - syslog interface
-;;;  
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;;   
+;;;
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;;   
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
-;;;  
+;;;
 ;;;   2. Redistributions in binary form must reproduce the above copyright
 ;;;      notice, this list of conditions and the following disclaimer in the
 ;;;      documentation and/or other materials provided with the distribution.
-;;;  
+;;;
 ;;;   3. Neither the name of the authors nor the names of its contributors
 ;;;      may be used to endorse or promote products derived from this
 ;;;      software without specific prior written permission.
-;;;  
+;;;
 ;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 ;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 ;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -29,7 +29,7 @@
 ;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-;;;  
+;;;
 
 #!no-fold-case
 
@@ -40,10 +40,9 @@
 
 (inline-stub
  (declcode
-  "#include <gauche/extend.h>"
-  "#ifdef HAVE_SYSLOG_H"
-  "#include <syslog.h>"
-  "#endif")
+  (.include <gauche/extend.h>)
+  (.if (defined HAVE_SYSLOG_H)
+       (.include <syslog.h>)))
 
  ;; On the platforms which don't have syslog facility, should we raise
  ;; a "not implemented" error, or simply discard the requests?  I tend to
@@ -51,38 +50,39 @@
 
  (define-cproc sys-openlog
    (ident::<const-cstring> option::<fixnum> facility::<fixnum>) ::<void>
-   (.if "defined(HAVE_SYSLOG)" (openlog ident option facility)))
+   (.if (defined HAVE_SYSLOG)
+        (openlog ident option facility)
+        (begin (cast void ident)       ; suppress unused var warning
+               (cast void option)      ; suppress unused var warning
+               (cast void facility)))) ; suppress unused var warning
 
  (define-cproc sys-syslog (prio::<fixnum> message::<const-cstring>) ::<void>
-   (.if "defined(HAVE_SYSLOG)" (syslog prio "%s" message)))
+   (.if (defined HAVE_SYSLOG)
+        (syslog prio "%s" message)
+        (begin (cast void prio)       ; suppress unused var warning
+               (cast void message)))) ; suppress unused var warning
 
  (define-cproc sys-closelog () ::<void>
-   (.if "defined(HAVE_SYSLOG)" (closelog)))
-
- (when "defined(HAVE_SYSLOG)"
-   (initcode "Scm_AddFeature(\"gauche.sys.syslog\", NULL);"))
+   (.if (defined HAVE_SYSLOG) (closelog)))
 
  (define-cproc sys-logmask (prio::<fixnum>) ::<fixnum>
-   (.if "defined(HAVE_SETLOGMASK)" (result (LOG_MASK prio)) (result 0)))
+   (.if (defined HAVE_SETLOGMASK)
+        (return (LOG_MASK prio))
+        (begin (cast void prio) ; suppress unused var warning
+               (return 0))))
 
  (define-cproc sys-setlogmask (mask::<fixnum>) ::<fixnum>
-   (.if "defined(HAVE_SETLOGMASK)" (result (setlogmask mask)) (result 0)))
-
- (when "defined(HAVE_SETLOGMASK)"
-   (initcode "Scm_AddFeature(\"gauche.sys.setlogmask\", NULL);"))
+   (.if (defined HAVE_SETLOGMASK)
+        (return (setlogmask mask))
+        (begin (cast void mask) ; suppress unused var warning
+               (return 0))))
 
  ;; We need to define at least these three logging option constants,
  ;; for they are referenced in gauche.logger.  The actual value doesn't matter.
  (declcode
-  "#ifndef LOG_PID"
-  "#define LOG_PID 0"
-  "#endif"
-  "#ifndef LOG_INFO"
-  "#define LOG_INFO 1"
-  "#endif"
-  "#ifndef LOG_USER"
-  "#define LOG_USER 2"
-  "#endif")
+  (.unless (defined LOG_PID) (.define LOG_PID 0))
+  (.unless (defined LOG_INFO) (.define LOG_INFO 1))
+  (.unless (defined LOG_USER) (.define LOG_USER 2)))
 
  ;; openlog options
  (define-enum-conditionally LOG_CONS)
