@@ -1,12 +1,12 @@
 /*
  * termios.c - termios interface
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -50,7 +50,8 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_SysTermiosClass,
                          termios_allocate,
                          NULL);
 
-static ScmObj termios_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj termios_allocate(ScmClass *klass SCM_UNUSED,
+                               ScmObj initargs SCM_UNUSED)
 {
     ScmSysTermios *t = SCM_NEW(ScmSysTermios);
     SCM_SET_CLASS(t, SCM_CLASS_SYS_TERMIOS);
@@ -143,25 +144,22 @@ ScmObj Scm_ForkptyAndExec(ScmString *file, ScmObj args, ScmObj iomap,
                           ScmObj slaveterm, ScmSysSigset *mask)
 {
     int argc = Scm_Length(args);
-    char **argv;
-    const char *program;
-    int *fds;
-    int master;
-    pid_t pid;
     struct termios *term = NULL;
 
     if (argc < 1) {
         Scm_Error("argument list must have at least one element: %S", args);
     }
-    argv = Scm_ListToCStringArray(args, TRUE, NULL);
-    program = Scm_GetStringConst(file);
+    char **argv = Scm_ListToCStringArray(args, TRUE, NULL);
+    const char *program = Scm_GetStringConst(file);
 
     if (SCM_SYS_TERMIOS_P(slaveterm)) {
         term = &SCM_SYS_TERMIOS(slaveterm)->term;
     }
 
-    fds = Scm_SysPrepareFdMap(iomap);
+    int *fds = Scm_SysPrepareFdMap(iomap);
 
+    int master;
+    pid_t pid;
     if ((pid = forkpty(&master, NULL, term, NULL)) < 0) {
         Scm_SysError("forkpty failed");
     }
@@ -185,14 +183,11 @@ ScmObj Scm_ForkptyAndExec(ScmString *file, ScmObj args, ScmObj iomap,
  * Initializaion
  */
 
-extern void Scm_Init_termiolib(ScmModule *mod);
-
-SCM_EXTENSION_ENTRY void Scm_Init_gauche__termios(void)
+void Scm_Init_termios(void)
 {
-    ScmModule *mod;
     SCM_INIT_EXTENSION(gauche__termios);
-    mod = SCM_FIND_MODULE("gauche.termios", SCM_FIND_MODULE_CREATE);
-    Scm_Init_termiolib(mod);
+    ScmModule *mod = SCM_FIND_MODULE("gauche.termios", SCM_FIND_MODULE_CREATE);
+    (void)mod; /* suppress unused var warning */
 
 #if !defined(GAUCHE_WINDOWS)
     Scm_InitStaticClass(&Scm_SysTermiosClass, "<sys-termios>", mod,
@@ -349,7 +344,7 @@ SCM_EXTENSION_ENTRY void Scm_Init_gauche__termios(void)
 
     /* c_cc size */
     DEFSYM(NCCS);
-    
+
     /* disable character */
     DEFSYM(_POSIX_VDISABLE);
 
@@ -392,7 +387,7 @@ SCM_EXTENSION_ENTRY void Scm_Init_gauche__termios(void)
 #ifdef VSWTC
     DEFSYM(VSWTC);
 #endif
-    
+
     /* extra baudrates.   <= B38400 is defined in termiolib.stub */
 #ifdef B57600
     DEFSYM(B57600);
