@@ -1,23 +1,23 @@
 ;;;
 ;;; threads.scm - thread related procedures.  to be autoloaded
-;;;  
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;;   
+;;;
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;;   
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
-;;;  
+;;;
 ;;;   2. Redistributions in binary form must reproduce the above copyright
 ;;;      notice, this list of conditions and the following disclaimer in the
 ;;;      documentation and/or other materials provided with the distribution.
-;;;  
+;;;
 ;;;   3. Neither the name of the authors nor the names of its contributors
 ;;;      may be used to endorse or promote products derived from this
 ;;;      software without specific prior written permission.
-;;;  
+;;;
 ;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 ;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 ;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -29,11 +29,12 @@
 ;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-;;;  
+;;;
 
 (define-module gauche.threads
   (use gauche.record)
   (export gauche-thread-type
+          current-thread                ;re-exporting the builtin
 
           thread? make-thread thread-name thread-specific-set! thread-specific
           thread-state thread-start! thread-yield! thread-sleep!
@@ -42,6 +43,7 @@
           mutex? make-mutex mutex-name mutex-state
           mutex-specific-set! mutex-specific
           with-locking-mutex mutex-lock! mutex-unlock!
+          mutex-locker mutex-unlocker
 
           condition-variable? make-condition-variable condition-variable-name
           condition-variable-specific condition-variable-specific-set!
@@ -55,25 +57,27 @@
 (select-module gauche.threads)
 
 (inline-stub
- "#include \"threads.h\""
-
  (declcode
-  "extern void Scm_Init_mutex(ScmModule*);"
-  "extern void Scm_Init_threads(ScmModule*);")
+  (.include "threads.h"))
+
+ (declare-cfn Scm_Init_mutex (mod::ScmModule*) ::void)
+ (declare-cfn Scm_Init_threads (mod::ScmModule*) ::void)
 
  (initcode
-  "Scm_Init_threads(Scm_CurrentModule());"
-  "Scm_Init_mutex(Scm_CurrentModule());"))
+  (Scm_Init_threads (Scm_CurrentModule))
+  (Scm_Init_mutex (Scm_CurrentModule))))
 
 ;;===============================================================
 ;; System query
 ;;
 
+;; TODO: for the consistency of ./configure --enable-threads,
+;; it should return 'pthreads' instead of 'pthread' on pthreads platform.
 (inline-stub
  (define-cproc gauche-thread-type ()
-   (.if "defined(GAUCHE_USE_PTHREADS)"
-        (result 'pthread)
-        (result 'none))))
+   (.cond [(defined GAUCHE_USE_PTHREADS) (return 'pthread)]
+          [(defined GAUCHE_USE_WTHREADS) (return 'win32)]
+          [else (return 'none)])))
 
 ;;===============================================================
 ;; Thread
@@ -82,32 +86,36 @@
 (define (thread? obj) (is-a? obj <thread>))
 
 (define (thread-name thread)
-  (check-arg thread? thread)
+  (assume-type thread <thread>)
   (slot-ref thread 'name))
 
 (define (thread-specific-set! thread value)
-  (check-arg thread? thread)
+  (assume-type thread <thread>)
   (slot-set! thread 'specific value))
 
 (define thread-specific
   (getter-with-setter
-   (lambda (thread)
-     (check-arg thread? thread)
+   (^[thread]
+     (assume-type thread <thread>)
      (slot-ref thread 'specific))
    thread-specific-set!))
 
+(define (make-thread thunk :optional (name #f))
+  (rlet1 t (%make-thread thunk name)
+    ((with-module gauche.internal %vm-custom-error-reporter-set!) t (^e #f))))
+
 (inline-stub
  (define-cproc thread-state (vm::<thread>)
    (case (-> vm state)
-     [(SCM_VM_NEW)       (result 'new)]
-     [(SCM_VM_RUNNABLE)  (result 'runnable)]
-     [(SCM_VM_STOPPED)   (result 'stopped)]
-     [(SCM_VM_TERMINATED)(result 'terminated)]
+     [(SCM_VM_NEW)       (return 'new)]
+     [(SCM_VM_RUNNABLE)  (return 'runnable)]
+     [(SCM_VM_STOPPED)   (return 'stopped)]
+     [(SCM_VM_TERMINATED)(return 'terminated)]
      [else (Scm_Error "[internal] thread state has invalid value: %d"
-                      (-> vm state))]))
+                      (-> vm state))
+           (return SCM_UNDEFINED)]))   ;dummy
 
- (define-cproc make-thread (thunk::<procedure> :optional (name #f))
-   Scm_MakeThread)
+ (define-cproc %make-thread (thunk::<procedure> name) Scm_MakeThread)
 
  (define-cproc thread-start! (vm::<thread>) Scm_ThreadStart)
 
@@ -125,7 +133,7 @@
    Scm_ThreadStop)
 
  (define-cproc thread-cont! (target::<thread>) Scm_ThreadCont)
- ) 
+ )
 
 ;;===============================================================
 ;; Mutex
@@ -134,29 +142,29 @@
 (define (mutex? obj)  (is-a? obj <mutex>))
 
 (define (mutex-name mutex)
-  (check-arg mutex? mutex)
+  (assume-type mutex <mutex>)
   (slot-ref mutex 'name))
 
 (define (mutex-state mutex)
-  (check-arg mutex? mutex)
+  (assume-type mutex <mutex>)
   (slot-ref mutex 'state))
 
 (define (mutex-specific-set! mutex value)
-  (check-arg mutex? mutex)
+  (assume-type mutex <mutex>)
   (slot-set! mutex 'specific value))
 
 (define mutex-specific
   (getter-with-setter
-   (lambda (mutex)
-     (check-arg mutex? mutex)
+   (^[mutex]
+     (assume-type mutex <mutex>)
      (slot-ref mutex 'specific))
    mutex-specific-set!))
 
-(define (with-locking-mutex mutex thunk)
+(define-inline (with-locking-mutex mutex thunk)
   (dynamic-wind
-   (lambda () (mutex-lock! mutex))
-   thunk
-   (lambda () (mutex-unlock! mutex))))
+      (mutex-locker mutex)
+      thunk
+      (mutex-unlocker mutex)))
 
 (inline-stub
  (define-cproc make-mutex (:optional (name #f)) Scm_MakeMutex)
@@ -173,13 +181,16 @@
      (cond [(SCM_VMP thread) (set! owner (SCM_VM thread))]
            [(SCM_UNBOUNDP thread) (set! owner (Scm_VM))]
            [(not (SCM_FALSEP thread)) (SCM_TYPE_ERROR thread "thread or #f")])
-     (result (Scm_MutexLock mutex timeout owner))))
- 
+     (return (Scm_MutexLock mutex timeout owner))))
+
  (define-cproc mutex-unlock! (mutex::<mutex> :optional (cv #f) (timeout #f))
    (let* ([cond::ScmConditionVariable* NULL])
      (cond [(SCM_CONDITION_VARIABLE_P cv) (set! cond (SCM_CONDITION_VARIABLE cv))]
            [(not (SCM_FALSEP cv)) (SCM_TYPE_ERROR cv "condition variale or #f")])
-     (result (Scm_MutexUnlock mutex cond timeout))))
+     (return (Scm_MutexUnlock mutex cond timeout))))
+
+ (define-cproc mutex-locker (mutex::<mutex>) Scm_MutexLocker)
+ (define-cproc mutex-unlocker (mutex::<mutex>) Scm_MutexUnlocker)
  )
 
 ;;===============================================================
@@ -189,17 +200,17 @@
 (define (condition-variable? obj) (is-a? obj <condition-variable>))
 
 (define (condition-variable-name cv)
-  (check-arg condition-variable? cv)
+  (assume-type cv <condition-variable>)
   (slot-ref cv 'name))
 
 (define (condition-variable-specific-set! cv value)
-  (check-arg condition-variable? cv)
+  (assume-type cv <condition-variable>)
   (slot-set! cv 'specific value))
 
 (define condition-variable-specific
   (getter-with-setter
    (lambda (cv)
-     (check-arg condition-variable? cv)
+     (assume-type cv <condition-variable>)
      (slot-ref cv 'specific))
    condition-variable-specific-set!))
 
@@ -212,8 +223,8 @@
 
   (define-cproc condition-variable-broadcast! (cv::<condition-variable>)
     Scm_ConditionVariableBroadcast)
-  )  
-     
+  )
+
 ;;===============================================================
 ;; Exceptions
 ;;
@@ -231,7 +242,7 @@
   (is-a? obj <uncaught-exception>))
 
 (define (uncaught-exception-reason exc)
-  (check-arg uncaught-exception? exc)
+  (assume-type exc <uncaught-exception>)
   (slot-ref exc 'reason))
 
 ;;===============================================================
@@ -246,20 +257,22 @@
   (define m (make-mutex))
   (let-syntax ([with-lock
                 (syntax-rules ()
-                  [(_ timeout timeout-val . form)
+                  [(_ timeout timeout-val timeout-vals . form)
                    (unwind-protect
                        (if (mutex-lock! m timeout)
                          (begin . form)
-                         timeout-val)
+                         (if (null? timeout-vals)
+                           timeout-val
+                           (apply values (cons timeout-val timeout-vals))))
                      (when (eq? (mutex-state m) (current-thread))
                        (mutex-unlock! m)))])])
     ;; TODO: we may expand special cases like vals is 1 to 3 elements long,
     ;; avoiding creation of lists every time updater is called.
     (%make-atom
-     (lambda (proc timeout timeout-val)
-       (with-lock timeout timeout-val (apply proc vals)))
-     (lambda (proc timeout timeout-val)
-       (with-lock timeout timeout-val
+     (^[proc timeout timeout-val timeout-vals]
+       (with-lock timeout timeout-val timeout-vals (apply proc vals)))
+     (^[proc timeout timeout-val timeout-vals]
+       (with-lock timeout timeout-val timeout-vals
                   (call-with-values (cut apply proc vals)
                     (^ vs
                       (unless (= (length vs) (length vals))
@@ -269,14 +282,14 @@
                       (set! vals vs)
                       (apply values vals))))))))
 
-(define (atomic atom proc :optional (timeout #f) (timeout-val #f))
+(define (atomic atom proc :optional (timeout #f) (timeout-val #f) :rest vals)
   (unless (atom? atom) (error "atom required, but got:" atom))
-  ((atom-applier atom) proc timeout timeout-val))
+  ((atom-applier atom) proc timeout timeout-val vals))
 
-(define (atomic-update! atom proc :optional (timeout #f) (timeout-val #f))
+(define (atomic-update! atom proc :optional (timeout #f) (timeout-val #f) :rest vals)
   (unless (atom? atom) (error "atom required, but got:" atom))
-  ((atom-updater atom) proc timeout timeout-val))
+  ((atom-updater atom) proc timeout timeout-val vals))
 
 (define (atom-ref atom :optional (index 0) (timeout #f) (timeout-val #f))
   (unless (atom? atom) (error "atom required, but got:" atom))
-  ((atom-applier atom) (^ xs (list-ref xs index)) timeout timeout-val))
+  ((atom-applier atom) (^ xs (list-ref xs index)) timeout timeout-val '()))
