@@ -1,12 +1,12 @@
 /*
  * uvectorP.h - internal macros for the uniform vector module
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -53,7 +53,7 @@ static void range_error(const char *type, ScmObj obj)
 static inline long range_s8hi(long val, int clamp)
 {
     if (!(clamp & SCM_CLAMP_HI)) range_error("s8", Scm_MakeInteger(val));
-    return 127; 
+    return 127;
 }
 
 static inline long range_s8lo(long val, int clamp)
@@ -171,31 +171,31 @@ static inline long clamp_u32(u_long val, int clamp)
 }
 #endif
 
-static inline ScmInt64 range_s64hi(ScmInt64 val, int clamp)
+static inline int64_t range_s64hi(int64_t val, int clamp)
 {
     if (!(clamp & SCM_CLAMP_HI)) range_error("s64", Scm_MakeInteger64(val));
-    SCM_SET_INT64_MAX(val);
+    val = INT64_MAX;
     return val;
 }
 
-static inline ScmInt64 range_s64lo(ScmInt64 val, int clamp)
+static inline int64_t range_s64lo(int64_t val, int clamp)
 {
     if (!(clamp & SCM_CLAMP_LO)) range_error("s64", Scm_MakeInteger64(val));
-    SCM_SET_INT64_MIN(val);
+    val = INT64_MIN;
     return val;
 }
 
-static inline ScmUInt64 range_u64hi(ScmUInt64 val, int clamp)
+static inline uint64_t range_u64hi(uint64_t val, int clamp)
 {
     if (!(clamp & SCM_CLAMP_HI)) range_error("u64", Scm_MakeIntegerU64(val));
-    SCM_SET_UINT64_MAX(val);
+    val = UINT64_MAX;
     return val;
 }
 
-static inline ScmUInt64 range_u64lo(ScmUInt64 val, int clamp)
+static inline uint64_t range_u64lo(uint64_t val, int clamp)
 {
     if (!(clamp & SCM_CLAMP_LO)) range_error("u64", Scm_MakeIntegerU64(val));
-    SCM_SET_INT64_ZERO(val);
+    val = 0;
     return val;
 }
 
