@@ -1,12 +1,12 @@
 /*
  * vport.c - 'virtual port'
  *
- *   Copyright (c) 2004-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2004-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -39,6 +39,9 @@
 
 #undef close    /* windows black magic */
 
+static ScmObj key_bufsize = SCM_FALSE; /* :buffer-size */
+static ScmObj key_name = SCM_FALSE;    /* :name */
+
 /*================================================================
  * <virtual-port>
  */
@@ -90,26 +93,23 @@ static int vport_getb(ScmPort *p)
     if (SCM_FALSEP(data->getb_proc)) {
         /* If the port doesn't have get-byte method, use get-char
            if possible. */
-        ScmObj ch;
-        ScmChar c;
         char buf[SCM_CHAR_MAX_BYTES];
-        int nb, i;
 
         if (SCM_FALSEP(data->getc_proc)) return EOF;
-        ch = Scm_ApplyRec(data->getc_proc, SCM_NIL);
+        ScmObj ch = Scm_ApplyRec(data->getc_proc, SCM_NIL);
         if (!SCM_CHARP(ch)) return EOF;
 
-        c = SCM_CHAR_VALUE(ch);
-        nb = SCM_CHAR_NBYTES(c);
+        ScmChar c = SCM_CHAR_VALUE(ch);
+        int nb = SCM_CHAR_NBYTES(c);
         SCM_CHAR_PUT(buf, c);
-        
-        for (i=1; i<nb; i++) {
+
+        for (int i=1; i<nb; i++) {
             /* pushback for later use.  this isn't very efficient;
                if efficiency becomes a problem, we need another API
                to pushback multiple bytes. */
-            Scm_UngetbUnsafe(buf[i], p); 
+            Scm_UngetbUnsafe(buf[i], p);
         }
-        return buf[0];
+        return (unsigned char)buf[0];
     } else {
         ScmObj b = Scm_ApplyRec(data->getb_proc, SCM_NIL);
         if (!SCM_INTP(b)) return EOF;
@@ -127,17 +127,14 @@ static int vport_getc(ScmPort *p)
 
     if (SCM_FALSEP(data->getc_proc)) {
         /* If the port doesn't have get-char method, try get-byte */
-        ScmObj b;
-        int n, i;
-        ScmChar ch;
         char buf[SCM_CHAR_MAX_BYTES];
 
         if (SCM_FALSEP(data->getb_proc)) return EOF;
-        b = Scm_ApplyRec(data->getb_proc, SCM_NIL);
+        ScmObj b = Scm_ApplyRec(data->getb_proc, SCM_NIL);
         if (!SCM_INTP(b)) return EOF;
         buf[0] = (char)SCM_INT_VALUE(b);
-        n = SCM_CHAR_NFOLLOWS(p->scratch[0]);
-        for (i=0; i<n; i++) {
+        int n = SCM_CHAR_NFOLLOWS(p->scratch[0]);
+        for (int i=0; i<n; i++) {
             b = Scm_ApplyRec(data->getb_proc, SCM_NIL);
             if (!SCM_INTP(b)) {
                 /* TODO: should raise an exception? */
@@ -145,6 +142,7 @@ static int vport_getc(ScmPort *p)
             }
             buf[i+1] = (char)SCM_INT_VALUE(b);
         }
+        ScmChar ch;
         SCM_CHAR_GET(buf, ch);
         return ch;
     } else {
@@ -157,19 +155,19 @@ static int vport_getc(ScmPort *p)
 /*------------------------------------------------------------
  * Vport Gets
  */
-static int vport_getz(char *buf, int buflen, ScmPort *p)
+static ScmSize vport_getz(char *buf, ScmSize buflen, ScmPort *p)
 {
     vport *data = (vport*)p->src.vt.data;
     SCM_ASSERT(data != NULL);
 
     if (!SCM_FALSEP(data->gets_proc)) {
-        u_int size;
-        const char *start;
+        ScmSmallInt size;
         ScmObj s = Scm_ApplyRec(data->gets_proc,
                                 SCM_LIST1(SCM_MAKE_INT(buflen)));
         if (!SCM_STRINGP(s)) return EOF;
-        start = Scm_GetStringContent(SCM_STRING(s), &size, NULL, NULL);
-        if ((int)size > buflen) {
+        const char *start = Scm_GetStringContent(SCM_STRING(s), &size,
+                                                 NULL, NULL);
+        if (size > buflen) {
             /* NB: should raise an exception? */
             memcpy(buf, start, buflen);
             return buflen;
@@ -178,9 +176,9 @@ static int vport_getz(char *buf, int buflen, ScmPort *p)
             return size;
         }
     } else {
-        int byte, i;
-        for (i=0; i<buflen; i++) {
-            byte = vport_getb(p);
+        ScmSize i = 0;
+        for (; i<buflen; i++) {
+            int byte = vport_getb(p);
             if (byte == EOF) break;
             buf[i] = byte;
         }
@@ -245,9 +243,9 @@ static void vport_putc(ScmChar c, ScmPort *p)
                           "cannot perform output to the port %S", p);
         } else {
             unsigned char buf[SCM_CHAR_MAX_BYTES];
-            int i, n=SCM_CHAR_NBYTES(c);
+            int n = SCM_CHAR_NBYTES(c);
             SCM_CHAR_PUT(buf, c);
-            for (i=0; i<n; i++) {
+            for (int i=0; i<n; i++) {
                 Scm_ApplyRec(data->putb_proc, SCM_LIST1(SCM_MAKE_INT(buf[i])));
             }
         }
@@ -259,7 +257,7 @@ static void vport_putc(ScmChar c, ScmPort *p)
 /*------------------------------------------------------------
  * Vport putz
  */
-static void vport_putz(const char *buf, int size, ScmPort *p)
+static void vport_putz(const char *buf, ScmSize size, ScmPort *p)
 {
     vport *data = (vport*)p->src.vt.data;
     SCM_ASSERT(data != NULL);
@@ -269,8 +267,7 @@ static void vport_putz(const char *buf, int size, ScmPort *p)
                      SCM_LIST1(Scm_MakeString(buf, size, -1,
                                               SCM_STRING_COPYING)));
     } else if (!SCM_FALSEP(data->putb_proc)) {
-        int i;
-        for (i=0; i<size; i++) {
+        for (ScmSize i=0; i<size; i++) {
             unsigned char b = buf[i];
             Scm_ApplyRec(data->putb_proc, SCM_LIST1(SCM_MAKE_INT(b)));
         }
@@ -291,16 +288,15 @@ static void vport_puts(ScmString *s, ScmPort *p)
 
     if (!SCM_FALSEP(data->puts_proc)) {
         Scm_ApplyRec(data->puts_proc, SCM_LIST1(SCM_OBJ(s)));
-    } else if (SCM_STRING_BODY_INCOMPLETE_P(b) 
+    } else if (SCM_STRING_BODY_INCOMPLETE_P(b)
                || (SCM_FALSEP(data->putc_proc)
                    && !SCM_FALSEP(data->putb_proc))) {
         /* we perform binary output */
         vport_putz(SCM_STRING_BODY_START(b), SCM_STRING_BODY_SIZE(b), p);
     } else if (!SCM_FALSEP(data->putc_proc)) {
-        ScmChar c;
-        int i;
         const char *cp = SCM_STRING_BODY_START(b);
-        for (i=0; i < (int)SCM_STRING_BODY_LENGTH(b); i++) {
+        for (int i=0; i < (int)SCM_STRING_BODY_LENGTH(b); i++) {
+            ScmChar c;
             SCM_CHAR_GET(cp, c);
             cp += SCM_CHAR_NFOLLOWS(*cp)+1;
             Scm_ApplyRec(data->putc_proc, SCM_LIST1(SCM_MAKE_CHAR(c)));
@@ -359,10 +355,7 @@ static off_t vport_seek(ScmPort *p, off_t off, int whence)
 
 static ScmObj vport_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmObj port;
     vport *data = SCM_NEW(vport);
-    ScmPortVTable vtab;
-    int dir = 0;
 
     data->getb_proc = SCM_FALSE;
     data->getc_proc = SCM_FALSE;
@@ -375,6 +368,7 @@ static ScmObj vport_allocate(ScmClass *klass, ScmObj initargs)
     data->close_proc = SCM_FALSE;
     data->seek_proc = SCM_FALSE;
 
+    ScmPortVTable vtab;
     vtab.Getb = vport_getb;
     vtab.Getc = vport_getc;
     vtab.Getz = vport_getz;
@@ -387,22 +381,25 @@ static ScmObj vport_allocate(ScmClass *klass, ScmObj initargs)
     vtab.Close = vport_close;
     vtab.Seek  = vport_seek;
 
+    int dir = 0;
     if (Scm_SubtypeP(klass, SCM_CLASS_VIRTUAL_INPUT_PORT)) {
         dir = SCM_PORT_INPUT;
     } else if (Scm_SubtypeP(klass, SCM_CLASS_VIRTUAL_OUTPUT_PORT)) {
         dir = SCM_PORT_OUTPUT;
     } else {
-        Scm_Panic("vport_allocate: implementaion error (class wiring screwed?)");
+        Scm_Panic("vport_allocate: implementation error (class wiring screwed?)");
     }
-    port = Scm_MakeVirtualPort(klass, dir, &vtab);
+    ScmObj port = Scm_MakeVirtualPort(klass, dir, &vtab);
     SCM_PORT(port)->src.vt.data = data;
+    SCM_PORT(port)->name = Scm_GetKeyword(key_name, initargs, SCM_FALSE);
     return port;
 }
 
-static void vport_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static void vport_print(ScmObj obj, ScmPort *port, 
+                        ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(port, "#<%A%s %A %p>",
-               Scm__InternalClassName(Scm_ClassOf(obj)),
+               Scm_ShortClassName(Scm_ClassOf(obj)),
                SCM_PORT_CLOSED_P(obj)? "(closed)" : "",
                Scm_PortName(SCM_PORT(obj)),
                obj);
@@ -489,7 +486,6 @@ SCM_DEFINE_BASE_CLASS(Scm_BufferedOutputPortClass, ScmPort,
                       vport_print, NULL, NULL,
                       bport_allocate, vport_cpa);
 
-static ScmObj key_bufsize = SCM_FALSE; /* :buffer-size */
 
 /*
  * Scheme handlers.  They are visible from Scheme as instance slots.
@@ -507,17 +503,16 @@ typedef struct bport_rec {
 /*------------------------------------------------------------
  * Bport fill
  */
-static int bport_fill(ScmPort *p, int cnt)
+static ScmSize bport_fill(ScmPort *p, ScmSize cnt)
 {
     bport *data = (bport*)p->src.buf.data;
-    ScmObj vec, r;
     SCM_ASSERT(data != NULL);
     if (SCM_FALSEP(data->fill_proc)) {
         return 0;               /* indicates EOF */
     }
-    vec = Scm_MakeU8VectorFromArrayShared(cnt,
-                                          (unsigned char*)p->src.buf.buffer);
-    r = Scm_ApplyRec(data->fill_proc, SCM_LIST1(vec));
+    ScmObj vec =
+        Scm_MakeU8VectorFromArrayShared(cnt, (unsigned char*)p->src.buf.buffer);
+    ScmObj r = Scm_ApplyRec(data->fill_proc, SCM_LIST1(vec));
     if (SCM_INTP(r)) return SCM_INT_VALUE(r);
     else if (SCM_EOFP(r)) return 0;
     else return -1;
@@ -526,17 +521,17 @@ static int bport_fill(ScmPort *p, int cnt)
 /*------------------------------------------------------------
  * Bport flush
  */
-static int bport_flush(ScmPort *p, int cnt, int forcep)
+static ScmSize bport_flush(ScmPort *p, ScmSize cnt, int forcep)
 {
     bport *data = (bport*)p->src.buf.data;
-    ScmObj vec, r;
     SCM_ASSERT(data != NULL);
     if (SCM_FALSEP(data->flush_proc)) {
         return cnt;             /* blackhole */
     }
-    vec = Scm_MakeU8VectorFromArrayShared(cnt,
-                                          (unsigned char*)p->src.buf.buffer);
-    r = Scm_ApplyRec(data->flush_proc, SCM_LIST2(vec, SCM_MAKE_BOOL(forcep)));
+    ScmObj vec = 
+        Scm_MakeU8VectorFromArrayShared(cnt, (unsigned char*)p->src.buf.buffer);
+    ScmObj r = Scm_ApplyRec(data->flush_proc,
+                            SCM_LIST2(vec, SCM_MAKE_BOOL(forcep)));
     if (SCM_INTP(r)) return SCM_INT_VALUE(r);
     else if (SCM_EOFP(r)) return 0;
     else return -1;
@@ -612,10 +607,7 @@ static off_t bport_seek(ScmPort *p, off_t off, int whence)
 
 static ScmObj bport_allocate(ScmClass *klass, ScmObj initargs)
 {
-    ScmObj port;
     bport *data = SCM_NEW(bport);
-    ScmPortBuffer buf;
-    int dir = 0;
     int bufsize = Scm_GetInteger(Scm_GetKeyword(key_bufsize, initargs,
                                                 SCM_MAKE_INT(0)));
 
@@ -626,6 +618,7 @@ static ScmObj bport_allocate(ScmClass *klass, ScmObj initargs)
     data->filenum_proc = SCM_FALSE;
     data->seek_proc  = SCM_FALSE;
 
+    ScmPortBuffer buf;
     if (bufsize > 0) {
         buf.buffer = SCM_NEW_ATOMIC2(char*, bufsize);
         buf.size = bufsize;
@@ -633,7 +626,7 @@ static ScmObj bport_allocate(ScmClass *klass, ScmObj initargs)
         buf.buffer = NULL;
         buf.size = 0;
     }
-    
+
     buf.current = NULL;
     buf.end     = NULL;
     buf.mode    = SCM_PORT_BUFFER_FULL;
@@ -645,14 +638,16 @@ static ScmObj bport_allocate(ScmClass *klass, ScmObj initargs)
     buf.seeker  = bport_seek;
     buf.data    = data;
 
+    int dir = 0;
     if (Scm_SubtypeP(klass, SCM_CLASS_BUFFERED_INPUT_PORT)) {
         dir = SCM_PORT_INPUT;
     } else if (Scm_SubtypeP(klass, SCM_CLASS_BUFFERED_OUTPUT_PORT)) {
         dir = SCM_PORT_OUTPUT;
     } else {
-        Scm_Panic("bport_allocate: implementaion error (class wiring screwed?)");
+        Scm_Panic("bport_allocate: implementation error (class wiring screwed?)");
     }
-    port = Scm_MakeBufferedPort(klass, SCM_FALSE, dir, TRUE, &buf);
+    ScmObj port = Scm_MakeBufferedPort(klass, SCM_FALSE, dir, TRUE, &buf);
+    SCM_PORT(port)->name = Scm_GetKeyword(key_name, initargs, SCM_FALSE);
     return port;
 }
 
@@ -704,14 +699,11 @@ static ScmClassStaticSlotSpec boport_slots[] = {
  * Initialization
  */
 
-void Scm_Init_vportlib(ScmModule*);
-
 SCM_EXTENSION_ENTRY void Scm_Init_gauche__vport(void)
 {
-    ScmModule *mod;
     SCM_INIT_EXTENSION(gauche__vport);
-    mod = SCM_FIND_MODULE("gauche.vport", SCM_FIND_MODULE_CREATE);
-    
+    ScmModule *mod = SCM_FIND_MODULE("gauche.vport", SCM_FIND_MODULE_CREATE);
+
     Scm_InitStaticClass(&Scm_VirtualInputPortClass,
                         "<virtual-input-port>", mod, viport_slots, 0);
     Scm_InitStaticClass(&Scm_VirtualOutputPortClass,
@@ -722,7 +714,5 @@ SCM_EXTENSION_ENTRY void Scm_Init_gauche__vport(void)
                         "<buffered-output-port>", mod, boport_slots, 0);
 
     key_bufsize = SCM_MAKE_KEYWORD("buffer-size");
-
-    Scm_Init_vportlib(mod);
+    key_name = SCM_MAKE_KEYWORD("name");
 }
-
