@@ -1,7 +1,7 @@
 ;;;
 ;;; file/util.scm - filesystem utility functions
 ;;;
-;;;   Copyright (c) 2000-2012  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -40,19 +40,18 @@
 
 (define-module file.util
   (use srfi-1)
-  (use srfi-11)
   (use srfi-13)
-  (use util.list)
   (use util.match)
   (use gauche.parameter)
   (export current-directory directory-list directory-list2 directory-fold
           home-directory temporary-directory
           make-directory* create-directory* remove-directory* delete-directory*
           copy-directory*
+          call-with-temporary-file call-with-temporary-directory
           create-directory-tree check-directory-tree
           build-path resolve-path expand-path simplify-path decompose-path
           absolute-path? relative-path? find-file-in-paths
-	  path-separator
+          path-separator
           path-extension path-sans-extension path-swap-extension
           file-is-readable? file-is-writable? file-is-executable?
           file-is-symlink?
@@ -64,9 +63,11 @@
           file-ctime=? file-ctime<? file-ctime<=? file-ctime>? file-ctime>=?
           touch-file touch-files copy-file move-file
           ;copy-files move-files
+          remove-file delete-file
           remove-files delete-files
           null-device console-device
           file->string file->string-list file->list file->sexp-list
+          string->file string-list->file list->file sexp-list->file
           <lock-file-failure> with-lock-file
           ))
 (select-module file.util)
@@ -112,6 +113,21 @@
 (define temporary-directory
   (make-parameter (sys-tmpdir)))
 
+(define (call-with-temporary-file proc
+                                  :key (directory (temporary-directory))
+                                       (prefix "gtemp"))
+  (receive (oport name) (sys-mkstemp (build-path directory prefix))  
+    (unwind-protect (proc oport name)
+      (close-output-port oport)
+      (sys-unlink name))))
+
+(define (call-with-temporary-directory proc 
+                                       :key (directory (temporary-directory))
+                                            (prefix "gtemp"))
+  (let1 dir (sys-mkdtemp (build-path directory prefix))
+    (unwind-protect (proc dir)
+      (remove-directory* dir))))
+
 (define (null-device)
   (cond-expand
    [gauche.os.windows "NUL"]
@@ -125,7 +141,7 @@
 ;; utility for directory-list and directory-list2
 (define (%directory-filter dir pred filter-add-path?)
   (if filter-add-path?
-    (filter (lambda (e) (pred (build-path dir e))) (sys-readdir dir))
+    (filter (^e (pred (build-path dir e))) (sys-readdir dir))
     (filter pred (sys-readdir dir))))
 
 (define (%directory-filter-compose opts)
@@ -133,7 +149,7 @@
                       (filter #f))
     (apply every-pred
            (cond-list
-            [children? (lambda (e) (not (member (sys-basename e) '("." ".."))))]
+            [children? (^e (not (member (sys-basename e) '("." ".."))))]
             [filter]))))
 
 ;; directory-list DIR &keyword ADD-PATH? FILTER-ADD-PATH? CHILDREN? FILTER
@@ -159,7 +175,7 @@
          [entries (sort (%directory-filter dir filters filter-add-path?))])
     (if add-path?
       (partition selector (map (cut build-path dir <>) entries))
-      (partition (lambda (e) (selector (build-path dir e))) entries))))
+      (partition (^e (selector (build-path dir e))) entries))))
 
 ;; directory-fold DIR PROC KNIL &keyword LISTER FOLDER FOLLOW-LINK?
 (define (directory-fold dir proc knil
@@ -470,13 +486,20 @@
                                                         [gauche.os.windows #\;]
                                                         [else #\:]))]
                                               [else '()]))
-                                 (pred file-is-executable?))
+                                 (pred file-is-executable?)
+                                 (extensions '()))
+  (define names
+    (if (null? extensions)
+      `(,name)
+      (cons name
+            (map (^e (string-append name "." e)) extensions))))
+  (define (try n) (and (pred n) n))
   (if (absolute-path? name)
-    (and (pred name) name)
+    (any try names)
     (let loop ((paths paths))
       (and (not (null? paths))
-           (let1 p (build-path (car paths) name)
-             (if (pred p) p (loop (cdr paths))))))))
+           (or (any try (map (cute build-path (car paths) <>) names))
+               (loop (cdr paths)))))))
 
 ;;;=============================================================
 ;;; File attributes
@@ -529,7 +552,7 @@
  [gauche.os.windows
   (define (%stat-compare s1 s2 f1 f2)
     (let ([p1 (sys-normalize-pathname f1 :absolute #t :canonicalize #t)]
-	  [p2 (sys-normalize-pathname f2 :absolute #t :canonicalize #t)])
+          [p2 (sys-normalize-pathname f2 :absolute #t :canonicalize #t)])
       (equal? p1 p2)))
   (define (file-eq? f1 f2)  (%stat-compare #f #f f1 f2))
   (define (file-eqv? f1 f2) (%stat-compare #f #f f1 f2))
@@ -637,7 +660,7 @@
   (for-each (cut touch-file <> :time time :type type :create create) pathnames))
 
 ;; copy-file
-;;  if-exists     - :error :supersede :backup #f
+;;  if-exists     - :error :supersede :backup :append #f
 ;;  follow-link?  - follow symlinks (default #t)
 ;;  backup-suffix
 ;;  safe
@@ -674,23 +697,22 @@
       (and (not (file-is-symlink? path)) (sys-chmod path mode)))
     (define (open-destination)
       (if safe
-        (cond
-         [(and (eq? if-exists :error) (file-exists? dst))
-          (error "destination file exists" dst)]
-         [(and (not if-exists) (file-exists? dst)) #f]
-         [else
-          (set!-values (outport tmpfile) (sys-mkstemp dst)) #t])
-        (cond
-         [(eq? if-exists :error)
-          (set! outport (open-output-file dst :if-exists :error)) #t]
-         [(not if-exists)
-          (set! outport (open-output-file dst :if-exists #f)) outport]
-         [(eq? if-exists :backup)
-          (when (file-exists? dst) (sys-rename dst backfile))
-          (set! outport (open-output-file dst)) #t]
-         [else
-          (set! outport (open-output-file dst :if-exists :supersede)) #t])
-        ))
+        (cond [(and (eq? if-exists :error) (file-exists? dst))
+               (error "destination file exists" dst)]
+              [(and (not if-exists) (file-exists? dst)) #f]
+              [(and (eq? if-exists :append) (file-exists? dst))
+               (set!-values (outport tmpfile) (sys-mkstemp dst))
+               (call-with-input-file dst (cut copy-port <> outport))
+               #t]
+              [else
+               (set!-values (outport tmpfile) (sys-mkstemp dst)) #t])
+        (ecase if-exists
+          [(#f) (set! outport (open-output-file dst :if-exists #f)) outport]
+          [(:backup)
+           (when (file-exists? dst) (sys-rename dst backfile))
+           (set! outport (open-output-file dst)) #t]
+          [(:supersede :append :error)
+           (set! outport (open-output-file dst :if-exists if-exists)) #t])))
     (define (get-times stat)
       (map (cut slot-ref stat <>) '(atime mtime)))
     (define (do-symlink)
@@ -701,8 +723,8 @@
       (when keeptime (set! times (get-times (sys-lstat src))))
       (if (file-exists? dst)
         (case if-exists
-          [(:error) (error <system-error> :errno EEXIST
-                           "destination file exists" dst)]
+          [(:error :append) (error <system-error> :errno EEXIST
+                                   "destination file exists" dst)]
           [(#f) #f]
           [(:backup) (sys-rename dst backfile) (doit)]
           [else      (sys-unlink dst) (doit)])
@@ -720,8 +742,8 @@
           (commit))))
 
     ;; body of copy-file
-    (unless (memq if-exists '(#f :error :supersede :backup))
-      (error "argument for :if-exists must be either :error, :supersede, :backup or #f, but got" if-exists))
+    (unless (memq if-exists '(#f :error :supersede :backup :append))
+      (error "argument for :if-exists must be either :error, :supersede, :backup, :append or #f, but got" if-exists))
     (when (and (file-exists? src) (file-exists? dst)
                ((if follow-link? file-eqv? file-eq?) src dst))
       (errorf "source ~s and destination ~s are the same file" src dst))
@@ -785,7 +807,15 @@
 ; (define copy-files (%multifile-cpmv copy-file))
 ; (define move-files (%multifile-cpmv move-file))
 
-;; remove files
+;; removing.
+;; Unlike sys-unlink, we throw an error if FILE doesn't exist,
+;; to align with R7RS.  (Technically we should throw a file-error
+;; object.  We'll fix that later.)
+(define (remove-file file)
+  (or (sys-unlink file)
+      (error <system-error> :errno ENOENT "File does not exist:" file)))
+(define delete-file remove-file)
+
 (define (remove-files . paths)
   (dolist (p paths)
     (cond [(list? p) (apply remove-files p)] ; for the convenience
@@ -814,6 +844,19 @@
 (define (file->sexp-list file . opts)
   (apply call-with-input-file file (%maybe (cut port->list read <>)) opts))
 
+;; inverse of file->*
+(define (string->file file str . opts)
+  (apply call-with-output-file file (cut display str <>) opts))
+
+(define (list->file writer file lis . opts)
+  (apply call-with-output-file file (^p (dolist [s lis] (writer s p))) opts))
+
+(define string-list->file
+  (cute list->file (^[s p] (display s p) (newline p)) <> <> <...>))
+
+(define sexp-list->file
+  (cute list->file (^[s p] (write s p) (newline p)) <> <> <...>))
+
 ;;;=============================================================
 ;;; File locking
 ;;;
@@ -847,7 +890,7 @@
                              (retry-interval 1)
                              (retry-limit 10)
                              (abandon-timeout 600)
-                             (secondary-lock-name #`",|lock-name|.2")
+                             (secondary-lock-name #"~|lock-name|.2")
                              (retry2-interval 1)
                              (retry2-limit 10)
                              (perms (if (eq? type 'directory)
@@ -858,7 +901,7 @@
     (case type
       [(file)
        (with-output-to-file path
-         (^() (print (sys-getpid)) (sys-chmod path perms))
+         (^() (print (sys-getpid)) (sys-chmod path perms) #t)
          :if-exists #f)]
       [(directory)
        (guard (e [(and (<system-error> e) (eqv? (~ e'errno) EEXIST)) #f])
