@@ -1,23 +1,23 @@
 ;;;
 ;;; srfi-13.scm - string library
-;;;  
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;;   
+;;;
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;;   
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
-;;;  
+;;;
 ;;;   2. Redistributions in binary form must reproduce the above copyright
 ;;;      notice, this list of conditions and the following disclaimer in the
 ;;;      documentation and/or other materials provided with the distribution.
-;;;  
+;;;
 ;;;   3. Neither the name of the authors nor the names of its contributors
 ;;;      may be used to endorse or promote products derived from this
 ;;;      software without specific prior written permission.
-;;;  
+;;;
 ;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 ;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 ;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -29,13 +29,7 @@
 ;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-;;;  
-
-;; Natively implemented functions:
-;;   string? make-string string string->list list->string
-;;   string-join string-length string-ref string-copy
-;;   string-set! string-fill!
-;;   string-append string-contains
+;;;
 
 (define-module srfi-13
   (export string-null? string-every string-any
@@ -75,12 +69,21 @@
           make-kmp-restart-vector
           kmp-step
           string-kmp-partial-search
+
+          ;; Gauche supports the following functions natively, but
+          ;; we re-export them so that they will be available by
+          ;; importing srfi-13 into vanilla environment.
+          string? make-string string string->list list->string
+          string-join string-length string-ref string-copy
+          string-set! string-fill!
+          string-append
           ))
 (select-module srfi-13)
 
 ;; Utility macro to handle char/char-set/pred match.
 ;; To avoid closure creation in inner loop, this returns a static procedure
 ;; that takes char/char-set/pred and char.
+;; NB: This is not exported, but shared with srfi-152.
 (define-inline (%get-char-pred char/char-set/pred)
   (cond [(char? char/char-set/pred) char=?]
         [(char-set? char/char-set/pred) char-set-contains?]
@@ -90,12 +93,15 @@
 
 (define (%char-pred/pred c/s/p x) (c/s/p x))
 
+(define %maybe-substring (with-module gauche.internal %maybe-substring))
+(define %hash-string (with-module gauche.internal %hash-string))
+(define %string-replace-body! (with-module gauche.internal %string-replace-body!))
 ;;;
 ;;; Predicates
 ;;;
 
 (define (string-null? str)
-  (check-arg string? str)
+  (assume-type str <string>)
   (equal? "" str))
 
 (define (string-every c/s/p s . args)
@@ -120,7 +126,7 @@
 ;;;
 
 (define (string-tabulate proc len)
-  (check-arg (lambda (l) (and (integer? l) (>= l 0))) len)
+  (check-arg (^l (and (integer? l) (>= l 0))) len)
   (let ((sink (open-output-string)))
     (do ((i 0 (+ i 1)))
         ((>= i len) (get-output-string sink))
@@ -136,8 +142,8 @@
 (define substring/shared string-copy)  ; same in Gauche
 
 (define (string-copy! target tstart s . args)
-  (check-arg string? target)
-  (check-arg (lambda (x) (and (integer? x) (>= x 0))) tstart)
+  (assume-type target <string>)
+  (check-arg (^x (and (integer? x) (>= x 0))) tstart)
   (let* ((str (apply %maybe-substring s args))
          (slen (string-length str))
          (tlen (string-length target)))
@@ -153,7 +159,7 @@
                                                        tlen))))))
 
 (define (string-pad s len :optional (char #\space) start end)
-  (check-arg char? char)
+  (assume-type char <char>)
   (let* ((str (%maybe-substring s start end))
          (slen (string-length str)))
     (cond ((< slen len)
@@ -163,7 +169,7 @@
           (else str))))
 
 (define (string-pad-right s len :optional (char #\space) start end)
-  (check-arg char? char)
+  (assume-type char <char>)
   (let* ((str (%maybe-substring s start end))
          (slen (string-length str)))
     (cond ((< slen len)
@@ -173,31 +179,31 @@
           (else str))))
 
 (define (string-take s nchars)
-  (check-arg string? s)
+  (assume-type s <string>)
   (when (or (< nchars 0) (> nchars (string-length s)))
     (error "argument out of range:" nchars))
   (%maybe-substring s 0 nchars))
 
 (define (string-drop s nchars)
-  (check-arg string? s)
+  (assume-type s <string>)
   (when (or (< nchars 0) (> nchars (string-length s)))
     (error "argument out of range:" nchars))
   (%maybe-substring s nchars))
 
 (define (string-take-right s nchars)
-  (check-arg string? s)
+  (assume-type s <string>)
   (when (or (< nchars 0) (> nchars (string-length s)))
     (error "argument out of range:" nchars))
   (%maybe-substring s (- (string-length s) nchars)))
 
 (define (string-drop-right s nchars)
-  (check-arg string? s)
+  (assume-type s <string>)
   (when (or (< nchars 0) (> nchars (string-length s)))
     (error "argument out of range:" nchars))
   (%maybe-substring s 0 (- (string-length s) nchars)))
 
 (define (string-trim s :optional (c/s/p #[\s]) start end)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (sp (make-string-pointer (%maybe-substring s start end))))
     (let loop ((ch (string-pointer-next! sp)))
@@ -208,7 +214,7 @@
     ))
 
 (define (string-trim-right s :optional (c/s/p #[\s]) start end)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (sp (make-string-pointer (%maybe-substring s start end) -1)))
     (let loop ((ch (string-pointer-prev! sp)))
@@ -219,7 +225,7 @@
     ))
 
 (define (string-trim-both s :optional (c/s/p #[\s]) start end)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (sp (make-string-pointer (%maybe-substring s start end))))
     (let loop ((ch (string-pointer-next! sp)))
@@ -259,8 +265,8 @@
 
 (define (string-compare s1 s2 proc< proc= proc>
                         :optional (start1 0) end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-compare-int (- start1 1) str1 str2
@@ -269,8 +275,8 @@
 
 (define (string-compare-ci s1 s2 proc< proc= proc>
                            :optional (start1 0) end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-compare-int (- start1 1) str1 str2
@@ -278,87 +284,87 @@
                          proc< proc= proc>)))
 
 (define (string= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #f) (lambda (_) #t) (lambda (_) #f)
+         s1 s2 (^_ #f) (^_ #t) (^_ #f)
          args))
 
 (define (string<> s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #t) (lambda (_) #f) (lambda (_) #t)
+         s1 s2 (^_ #t) (^_ #f) (^_ #t)
          args))
 
 (define (string< s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #t) (lambda (_) #f) (lambda (_) #f)
+         s1 s2 (^_ #t) (^_ #f) (^_ #f)
          args))
 
 (define (string<= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #t) (lambda (_) #t) (lambda (_) #f)
+         s1 s2 (^_ #t) (^_ #t) (^_ #f)
          args))
 
 (define (string> s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #f) (lambda (_) #f) (lambda (_) #t)
+         s1 s2 (^_ #f) (^_ #f) (^_ #t)
          args))
 
 (define (string>= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare
-         s1 s2 (lambda (_) #f) (lambda (_) #t) (lambda (_) #t)
+         s1 s2 (^_ #f) (^_ #t) (^_ #t)
          args))
 
 (define (string-ci= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #f) (lambda (_) #t) (lambda (_) #f)
+         s1 s2 (^_ #f) (^_ #t) (^_ #f)
          args))
 
 (define (string-ci<> s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #t) (lambda (_) #f) (lambda (_) #t)
+         s1 s2 (^_ #t) (^_ #f) (^_ #t)
          args))
 
 (define (string-ci< s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #t) (lambda (_) #f) (lambda (_) #f)
+         s1 s2 (^_ #t) (^_ #f) (^_ #f)
          args))
 
 (define (string-ci<= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #t) (lambda (_) #t) (lambda (_) #f)
+         s1 s2 (^_ #t) (^_ #t) (^_ #f)
          args))
 
 (define (string-ci> s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #f) (lambda (_) #f) (lambda (_) #t)
+         s1 s2 (^_ #f) (^_ #f) (^_ #t)
          args))
 
 (define (string-ci>= s1 s2 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (apply string-compare-ci
-         s1 s2 (lambda (_) #f) (lambda (_) #t) (lambda (_) #t)
+         s1 s2 (^_ #f) (^_ #t) (^_ #t)
          args))
 
 ;;;
@@ -366,11 +372,11 @@
 ;;;
 
 (define (string-hash s :optional bound start end)
-  (check-arg string? s)
+  (assume-type s <string>)
   (%hash-string (%maybe-substring s start end) bound))
 
 (define (string-hash-ci s :optional bound start end)
-  (check-arg string? s)
+  (assume-type s <string>)
   ;; oops! optimization required!
   (%hash-string (string-upcase (%maybe-substring s start end)) bound))
 
@@ -391,29 +397,29 @@
       )))
 
 (define (string-prefix-length s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-prefix-int str1 str2 char=? (lambda (cnt flag) cnt))))
 
 (define (string-prefix-length-ci s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-prefix-int str1 str2 char-ci=? (lambda (cnt flag) cnt))))
 
 (define (string-prefix? s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-prefix-int str1 str2 char=? (lambda (cnt flag) flag))))
 
 (define (string-prefix-ci? s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-prefix-int str1 str2 char-ci=? (lambda (cnt flag) flag))))
@@ -434,29 +440,29 @@
       )))
 
 (define (string-suffix-length s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-suffix-int str1 str2 char=? (lambda (cnt flag) cnt))))
 
 (define (string-suffix-length-ci s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-suffix-int str1 str2 char-ci=? (lambda (cnt flag) cnt))))
 
 (define (string-suffix? s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-suffix-int str1 str2 char=? (lambda (cnt flag) flag))))
 
 (define (string-suffix-ci? s1 s2 :optional start1 end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let ((str1 (%maybe-substring s1 start1 end1))
         (str2 (%maybe-substring s2 start2 end2)))
     (%string-suffix-int str1 str2 char-ci=? (lambda (cnt flag) flag))))
@@ -466,7 +472,7 @@
 ;;;
 
 (define (string-index s c/s/p . args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (offset (if (pair? args) (car args) 0))
         (sp (apply make-string-pointer s 0 args)))
@@ -476,7 +482,7 @@
             (else (loop (string-pointer-next! sp)))))))
 
 (define (string-index-right s c/s/p . args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (offset (if (pair? args) (car args) 0))
         (sp (apply make-string-pointer s -1 args)))
@@ -486,7 +492,7 @@
             (else (loop (string-pointer-prev! sp)))))))
 
 (define (string-skip s c/s/p . args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (offset (if (pair? args) (car args) 0))
         (sp (apply make-string-pointer s 0 args)))
@@ -496,7 +502,7 @@
             (else (+ offset (- (string-pointer-index sp) 1)))))))
 
 (define (string-skip-right s c/s/p . args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (offset (if (pair? args) (car args) 0))
         (sp (apply make-string-pointer s -1 args)))
@@ -506,7 +512,7 @@
             (else (+ offset (string-pointer-index sp)))))))
 
 (define (string-count s c/s/p . args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((pred (%get-char-pred c/s/p))
         (sp (apply make-string-pointer s 0 args)))
     (let loop ((ch (string-pointer-next! sp))
@@ -516,8 +522,8 @@
             (else      (loop (string-pointer-next! sp) count))))))
 
 (define (string-contains s1 s2 :optional (start1 0) end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let* ((str1 (%maybe-substring s1 start1 end1))
          (str2 (%maybe-substring s2 start2 end2))
          (res  (string-scan str1 str2)))
@@ -525,8 +531,8 @@
 
 ;; not tuned (maybe to be moved to native)
 (define (string-contains-ci s1 s2 :optional (start1 0) end1 start2 end2)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (let* ((str1 (string-upcase (%maybe-substring s1 start1 end1)))
          (str2 (string-upcase (%maybe-substring s2 start2 end2)))
          (res  (string-scan str1 str2)))
@@ -643,8 +649,7 @@
 ;;;
 
 (define (string-map proc s . args)
-  (check-arg procedure? proc)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((src  (apply make-string-pointer s 0 args))
         (dest (open-output-string)))
     (let loop ((ch (string-pointer-next! src)))
@@ -655,14 +660,12 @@
     ))
 
 (define (string-map! proc s . args)
-  (check-arg procedure? proc)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((mapped (apply string-map proc s args)))
     (string-copy! s (get-optional args 0) mapped)))
 
 (define (string-fold kons knil s . args)
-  (check-arg procedure? kons)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((src (apply make-string-pointer s 0 args)))
     (let loop ((ch (string-pointer-next! src))
                (r  knil))
@@ -672,8 +675,7 @@
     ))
 
 (define (string-fold-right kons knil s . args)
-  (check-arg procedure? kons)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((src (apply make-string-pointer s -1 args)))
     (let loop ((ch (string-pointer-prev! src))
                (r  knil))
@@ -683,10 +685,7 @@
     ))
 
 (define (string-unfold p f g seed
-                       :optional (base "") (make-final (lambda (_) "")))
-  (check-arg procedure? p)
-  (check-arg procedure? f)
-  (check-arg procedure? g)
+                       :optional (base "") (make-final (^_ "")))
   (let ((dest (open-output-string)))
     (display base dest)
     (let loop ((seed seed))
@@ -697,10 +696,7 @@
                (loop (g seed)))))))
 
 (define (string-unfold-right p f g seed
-                             :optional (base "") (make-final (lambda (_) "")))
-  (check-arg procedure? p)
-  (check-arg procedure? f)
-  (check-arg procedure? g)
+                             :optional (base "") (make-final (^_ "")))
   (let ((dest (open-output-string)))
     (let loop ((seed seed))
       (if (p seed)
@@ -711,8 +707,7 @@
                (loop (g seed)))))))
 
 (define (string-for-each proc s . args)
-  (check-arg procedure? proc)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((src (apply make-string-pointer s 0 args)))
     (let loop ((ch (string-pointer-next! src)))
       (unless (eof-object? ch)
@@ -721,8 +716,7 @@
   )
 
 (define (string-for-each-index proc s :optional (start 0) (end (string-length s)))
-  (check-arg procedure? proc)
-  (check-arg string? s)
+  (assume-type s <string>)
   (do ((i start (+ i 1)))
       ((>= i end))
     (proc i)))
@@ -732,43 +726,39 @@
 ;;;
 
 (define (xsubstring s from :optional to start end)
-  (check-arg string? s)
-  (check-arg (lambda (x) (and (integer? x) (exact? x))) from)
-  (let* ((str (%maybe-substring s start end))
-         (len (string-length str))
-         (from-rank (quotient from len))
-         (from-mod  (modulo from len))
-         (dest (open-output-string)))
-    (cond ((undefined? to)
-           (set! from from-mod)
-           (set! to (+ from len)))
-          ((not (and (integer? to) (exact? to)))
-           (error "argument out of domain:" to))
-          ((= len 0)
-           (error "zero length source string is not allowed"))
-          ((< to from)
-           (errorf "argument out of range (from, to): (~s, ~s)" from to))
-          (else
-           (set! from from-mod)
-           (set! to (- to (* from-rank len)))))
-    (let ((sp (make-string-pointer str from)))
-      (let loop ((count from)
-                 (ch (string-pointer-next! sp)))
-        (cond ((>= count to) (get-output-string dest))
-              ((eof-object? ch)
-               (string-pointer-set! sp 0)
-               (write-char (string-pointer-next! sp) dest)
-               (loop (+ count 1) (string-pointer-next! sp)))
-              (else
-               (write-char ch dest)
-               (loop (+ count 1) (string-pointer-next! sp)))))
-      )
-    ))
+  (assume-type s <string>)
+  (check-arg (^x (and (integer? x) (exact? x))) from)
+  (let* ([str (%maybe-substring s start end)]
+         [len (string-length str)]
+         [dest (open-output-string)])
+    ;; Fill TO if it's not provided; also check the validity.
+    (cond [(undefined? to) (set! to (+ from len))]
+          [(not (and (integer? to) (exact? to)))
+           (error "argument out of domain:" to)]
+          [(= len 0)
+           (error "zero length source string is not allowed")]
+          [(< to from)
+           (errorf "argument out of range (from, to): (~s, ~s)" from to)])
+    ;; Adjust the range
+    (let* ([from. (modulo from len)]
+           [to. (- to (- from from.))])
+      (let1 sp (make-string-pointer str from.)
+        (let loop ([count from.]
+                   [ch (string-pointer-next! sp)])
+          (cond [(>= count to.) (get-output-string dest)]
+                [(eof-object? ch)
+                 (string-pointer-set! sp 0)
+                 (write-char (string-pointer-next! sp) dest)
+                 (loop (+ count 1) (string-pointer-next! sp))]
+                [else
+                 (write-char ch dest)
+                 (loop (+ count 1) (string-pointer-next! sp))]))
+        ))))
 
 (define (string-xcopy! target tstart s sfrom . args)
-  (check-arg string? target)
-  (check-arg (lambda (x) (and (integer? x) (exact? x))) tstart)
-  (let ((result (apply xsubstring s sfrom args)))
+  (assume-type target <string>)
+  (check-arg (^x (and (integer? x) (exact? x))) tstart)
+  (let1 result (apply xsubstring s sfrom args)
     (string-copy! target tstart result)))
 
 ;;;
@@ -776,66 +766,70 @@
 ;;;
 
 (define (string-replace s1 s2 start1 end1 . args)
-  (check-arg string? s1)
-  (check-arg string? s2)
+  (assume-type s1 <string>)
+  (assume-type s2 <string>)
   (string-append (substring s1 0 start1)
                  (apply %maybe-substring s2 args)
                  (substring s1 end1 (string-length s1))))
 
 (define (string-tokenize s :optional (token-set #[\S]) start end)
-  (check-arg string? s)
-  (letrec ((src (open-input-string (%maybe-substring s start end)))
-           (in-word (lambda (ch dst r)
-                      (cond ((eof-object? ch)
-                             (reverse! (cons (get-output-string dst) r)))
-                            ((char-set-contains? token-set ch)
+  (assume-type s <string>)
+  (letrec ([src (open-input-string (%maybe-substring s start end))]
+           [in-word (^[ch dst r]
+                      (cond [(eof-object? ch)
+                             (reverse! (cons (get-output-string dst) r))]
+                            [(char-set-contains? token-set ch)
                              (write-char ch dst)
-                             (in-word (read-char src) dst r))
-                            (else
+                             (in-word (read-char src) dst r)]
+                            [else
                              (out-word (read-char src)
-                                       (cons (get-output-string dst) r))))))
-           (out-word (lambda (ch r)
-                       (cond ((eof-object? ch) (reverse! r))
-                             ((char-set-contains? token-set ch)
+                                       (cons (get-output-string dst) r))]))]
+           [out-word (^[ch r]
+                       (cond [(eof-object? ch) (reverse! r)]
+                             [(char-set-contains? token-set ch)
                               (let ((dst (open-output-string)))
                                 (write-char ch dst)
-                                (in-word (read-char src) dst r)))
-                             (else
-                              (out-word (read-char src) r))))))
+                                (in-word (read-char src) dst r))]
+                             [else
+                              (out-word (read-char src) r)]))])
     (out-word (read-char src) '())))
 
 ;;;
 ;;; Filter
 ;;;
 
-(define (string-filter s c/s/p . args)
-  (check-arg string? s)
-  (let ((src (open-input-string (apply %maybe-substring s args)))
-        (dest (open-output-string))
-        (pred (%get-char-pred c/s/p)))
-    (let loop ((ch (read-char src)))
-      (cond ((eof-object? ch) (get-output-string dest))
-            ((pred c/s/p ch) (write-char ch dest) (loop (read-char src)))
-            (else (loop (read-char src)))))
-    ))
+;; The argument order of string-filter and string-delete is changed
+;; by post-finalization amendment.  In order to avoid breaking existing
+;; code, we accept old argument oder as well.
 
-(define (string-delete s c/s/p . args)
-  (check-arg string? s)
-  (let ((src (open-input-string (apply %maybe-substring s args)))
-        (dest (open-output-string))
-        (pred (%get-char-pred c/s/p)))
-    (let loop ((ch (read-char src)))
-      (cond ((eof-object? ch) (get-output-string dest))
-            ((pred c/s/p ch) (loop (read-char src)))
-            (else (write-char ch dest) (loop (read-char src)))))
-    ))
+(define (string-filter c/s/p s . args)
+  (if (string? c/s/p)
+    (apply string-filter s c/s/p args) ;; for the backward compatibility
+    (let ([src (open-input-string (apply %maybe-substring s args))]
+          [dest (open-output-string)]
+          [pred (%get-char-pred c/s/p)])
+      (let loop ([ch (read-char src)])
+        (cond [(eof-object? ch) (get-output-string dest)]
+              [(pred c/s/p ch) (write-char ch dest) (loop (read-char src))]
+              [else (loop (read-char src))])))))
+
+(define (string-delete c/s/p s . args)
+  (if (string? c/s/p)
+    (apply string-delete s c/s/p args) ;; for the backward compatibility
+    (let ([src (open-input-string (apply %maybe-substring s args))]
+          [dest (open-output-string)]
+          [pred (%get-char-pred c/s/p)])
+      (let loop ([ch (read-char src)])
+        (cond [(eof-object? ch) (get-output-string dest)]
+              [(pred c/s/p ch) (loop (read-char src))]
+              [else (write-char ch dest) (loop (read-char src))])))))
 
 ;;; Low-level procedures.  These are included for completeness, but
 ;;; I'm not using these in other SRFI-13 routines, since it is more
 ;;; efficient to let %maybe-substring handle argument checking as well.
 
 (define (string-parse-start+end proc s args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((slen (string-length s)))
     (let-optionals* args (start end)
       (if (undefined? start)
@@ -853,7 +847,7 @@
       )))
 
 (define (string-parse-final-start+end proc s args)
-  (check-arg string? s)
+  (assume-type s <string>)
   (let ((slen (string-length s)))
     (let-optionals* args (start end)
       (if (undefined? start)
@@ -934,11 +928,11 @@
   (let ((patlen (vector-length rv)))
     (let lp ((si start)
              (vi i))
-      (cond ((= vi patlen) (- si)) 
+      (cond ((= vi patlen) (- si))
             ((= si end) vi)
             (else (lp (+ si 1)
                       (kmp-step pat rv (string-ref s si)
                                 vi c= p-start)))))))
-    
 
-          
+
+
