@@ -7,8 +7,6 @@
 ;;
 
 (define-module util.match
-  (use srfi-1)
-  (use srfi-2)
   (export match
           match-lambda
           match-lambda*
@@ -18,6 +16,7 @@
           match-let1
           match-define
           match:$-ref
+          |setter of match:$-ref|
           match:every
           match:error))
 (select-module util.match)
@@ -154,6 +153,12 @@
 ;;; procedure in gauche.internal.  If you're reading this code out of
 ;;; curiousity, DO NOT USE IT in your code.  The identifier stuff may be
 ;;; changed greatly when I implement proper low-level hygienic stuff.
+;;;
+;;; The plan is that we revise the code to use er-macro in a couple of
+;;; releases (we need to release er-macro support first, for util.match
+;;; is used to build HEAD version of source so in this module we can only
+;;; depend on the features available in the current release of Gauche.
+;;;
 
 (define (%match-id sym)
   ((with-module gauche.internal make-identifier) sym (current-module) '()))
@@ -166,23 +171,21 @@
 (define match-define.  (%match-id 'match-define))
 (define match:length>=?. (%match-id 'match:length>=?))
 
-;; In the original code, length>= is introduced as a local binding
-;; so that optimizer can inline the call.   In Gauche we can use
-;; define-inline, so we don't need to bother introducing the local
-;; bindings.
-(define-inline (match:length>=? n) (lambda (l) (>= (length l) n)))
+(define lambda.        ((with-module gauche.internal make-identifier) 'lambda
+                        (find-module 'null) '()))
+
+(define-inline (match:length>=? n) (cut length>=? <> n))
 
 ;;; [SK] End of black magic
 
-
 (define match:error
   (lambda (val . args)
     (errorf "~a: no matching clause for ~a"
             (string-join (map x->string args) " ")
             val)))
 
-(define match:every every) ; alias srfi-1's every, in case the user of
-                           ; util.match isn't using srfi-1.
+(define match:every every) ; alias gauche's every, in case the user of
+                           ; util.match isn't inheriting gauche.
 
 (define match:syntax-err
   (lambda (obj msg) (error msg obj)))
@@ -204,16 +207,15 @@
           procedure?
           vector?)))
 
-(define match:$-ref
-  (getter-with-setter
-   (lambda (class fnum obj)
-     (let ((slot (list-ref (class-slots class) fnum #f)))
-       (and slot
-            (slot-ref obj (slot-definition-name slot)))))
-   (lambda (class fnum obj val)
-     (let ((slot (list-ref (class-slots class) fnum #f)))
-       (and slot
-            (slot-set! obj (slot-definition-name slot) val))))))
+;; These two methods are used for positional match of objects using $.
+;; The default is to take the order of class-slots, which is suffice for
+;; most cases.  A metaclass can override these if desired.
+(define-method match:$-ref ((class <class>) fnum obj)
+  (and-let1 slot (list-ref (class-slots class) fnum #f)
+    (slot-ref obj (slot-definition-name slot))))
+(define-method (setter match:$-ref) ((class <class>) fnum obj val)
+  (and-let1 slot (list-ref (class-slots class) fnum #f)
+    (slot-set! obj (slot-definition-name slot) val)))
 
 (define (genmatch x clauses match-expr)
   (let* ((eb-errf (error-maker match-expr))
@@ -227,15 +229,15 @@
                               (fail (and (pair? (cdr c))
                                          (pair? (cadr c))
                                          (equal? (caadr c) '=>)
-                                         (symid? (cadadr c))
                                          (pair? (cdadr c))
+                                         (symid? (cadadr c))
                                          (null? (cddadr c))
                                          (pair? (cddr c))
                                          (cadadr c)))
                               (bv2 (if fail (cons fail bv) bv))
                               (body (if fail (cddr c) (cdr c))))
                          (set! blist
-                               (cons `(,code (lambda ,bv2 ,@body))
+                               (cons `(,code (,lambda. ,bv2 ,@body))
                                      (append bindings blist)))
                          (list p code bv (and fail (gensym)) #f)))
                      clauses))
@@ -258,7 +260,7 @@
     (unreachable plist match-expr)
     `(letrec (,@(map (lambda (v) `(,v #f)) bv)
               (,x ,exp)
-              (,code (lambda ,gs
+              (,code (,lambda. ,gs
                        ,@(map (lambda (v g) `(set! ,v ,g)) bv gs)
                        ,@body))
               ,@bindings
@@ -278,13 +280,18 @@
          (gs (map (lambda (_) (gensym)) bv)))
     (unreachable plist match-expr)
     `(begin ,@(map (lambda (v) `(define ,v #f)) bv)
-            `(let ((,x ,exp)
-                   (,code (lambda ,gs
-                            ,@(map (lambda (v g) `(set! ,v ,g)) bv gs)
-                            (cond (#f #f))))
-                   ,@bindings
-                   ,@(car eb-errf))
-               ,m))))
+            (let ((,x ,exp)
+                  (,code (,lambda. ,gs
+                           ,@(map (lambda (v g) `(set! ,v ,g)) bv gs)
+                           (cond (#f #f))))
+                  ,@bindings
+                  ,@(car eb-errf))
+              ,m))))
+
+(define (symbolize x)
+  (cond [(symbol? x) x]
+        [(identifier? x) (unwrap-syntax x)]
+        [else x]))
 
 (define (pattern-var? x)
   (and-let* ([x (cond [(symbol? x) x]
@@ -311,7 +318,7 @@
    ((eq? match:error-control 'match)
     (let ((errf (gensym))
           (arg (gensym)))
-      (cons `((,errf (lambda (,arg)
+      (cons `((,errf (,lambda. (,arg)
                        (,match:error. ,arg ',match-expr))))
             (lambda (x) `(,errf ,x)))))
    (else (match:syntax-err
@@ -327,7 +334,11 @@
 
 (define (validate-pattern pattern)
   (define (simple? x)
-    (or (string? x) (boolean? x) (char? x) (number? x) (null? x) (keyword? x)))
+    (or (string? x) (boolean? x) (char? x) (number? x) (null? x)
+        ;; This last term is to support both disjoint-keyword and
+        ;; keyword-is-symbol runtime.  After we fully migrated to
+        ;; keyword-is-symbol, just remove this term.
+        (and (not (symbol? :x)) (keyword? x))))
   (define (ordinary p)
     (let ((cons-ordinary (lambda (x y) (cons (ordinary x) (ordinary y)))))
       (cond
@@ -335,7 +346,7 @@
        ((equal? p '_) '_)
        ((pattern-var? p) p)
        ((pair? p)
-        (case (car p)
+        (case (symbolize (car p))
           ((quasiquote)
            (if (and (pair? (cdr p)) (null? (cddr p)))
              (quasi (cadr p))
@@ -408,7 +419,7 @@
        ((simple? p) p)
        ((symid? p) `',p)
        ((pair? p)
-        (cond 
+        (cond
          ((eq? (car p) 'unquote)
           (if (and (pair? (cdr p)) (null? (cddr p)))
             (ordinary (cadr p))
@@ -429,9 +440,10 @@
         (let* ((pl (vector->list p))
                (rpl (reverse pl)))
           (list->vector
-           (if (dot-dot-k? (car rpl))
+           (if (and (not (null? rpl))
+                    (dot-dot-k? (car rpl)))
              (reverse (cons (car rpl) (map quasi (cdr rpl))))
-             (map ordinary pl)))))
+             (map quasi pl)))))
        (else
         (match:syntax-err pattern "syntax error in pattern")))))
   (define (ordlist p)
@@ -452,9 +464,9 @@
       (when (memq p a)
         (match:syntax-err pattern "duplicate variable in pattern"))
       (k p (cons p a)))
-     ((and (pair? p) (eq? 'quote (car p)))
+     ((and (pair? p) (equal? 'quote (car p)))
       (k p a))
-     ((and (pair? p) (eq? '? (car p)))
+     ((and (pair? p) (equal? '? (car p)))
       (cond
        ((not (null? (cddr p)))
         (bound `(and (? ,(cadr p)) ,@(cddr p)) a k))
@@ -614,11 +626,11 @@
                         (if fail-sym
                           (let ((ap `(,code ,fail-sym ,@(map val bv))))
                             `(call-with-current-continuation
-                              (lambda (,fail-sym)
+                              (,lambda. (,fail-sym)
                                 (let ((,fail-sym
-                                       (lambda ()
+                                       (,lambda. ()
                                          (call-with-values
-                                             (lambda () ,(fail sf))
+                                             (,lambda. () ,(fail sf))
                                            ,fail-sym))))
                                   ,ap))))
                           `(,code ,@(map val bv)))))))
@@ -629,6 +641,15 @@
                  (ks success))
         (cond
          ((eq? '_ p) (ks sf))
+         ;; The check of (symbol? :x) is to support both disjoint-keyword and
+         ;; keyword-is-symbol runtime.  Remove this clause once we fully
+         ;; migrated to keyword-is-symbol.
+         ((and (not (symbol? :x)) (keyword? p))
+          (warn "Unquoted keyword `~s' in match pattern: ~s.  \
+                 This would likely break in future versions of Gauche.  \
+                 See the ``Keyword and symbol integration'' section \
+                 of the manual for the details.\n" p x)
+          (emit `(equal? ,e ,p) sf kf ks))
          ((symid? p) (set! v (cons (cons p e) v))
           (ks sf))
          ((null? p) (emit `(null? ,e) sf kf ks))
@@ -637,15 +658,14 @@
          ((boolean? p) (emit `(equal? ,e ,p) sf kf ks))
          ((char? p) (emit `(equal? ,e ,p) sf kf ks))
          ((number? p) (emit `(equal? ,e ,p) sf kf ks))
-         ((keyword? p) (emit `(equal? ,e ,p) sf kf ks))
-         ((and (pair? p) (eq? 'quote (car p)))
+         ((and (pair? p) (equal? 'quote (car p)))
           (emit `(equal? ,e ,p) sf kf ks))
          ((and (pair? p) (eq? '? (car p)))
           (let ((tst `(,(cadr p) ,e)))
             (emit tst sf kf ks)))
          ((and (pair? p) (eq? '= (car p)))
           (if (and (pair? (cadr p))
-                   (eq? (caadr p) 'quote))
+                   (equal? (caadr p) 'quote))
             (next (caddr p) `(ref ,(cadr p) ,e) sf kf ks)
             (next (caddr p) `(,(cadr p) ,e) sf kf ks)))
          ((and (pair? p) (eq? 'and (car p)))
@@ -717,7 +737,7 @@
                                                         (eq? eta (cadr ptst))
                                                         (null? (cddr ptst)))
                                                  (car ptst)
-                                                 `(lambda (,eta) ,ptst))))
+                                                 `(,lambda. (,eta) ,ptst))))
                                      (assm `(match:every ,tst ,e)
                                            (kf sf)
                                            (ks sf))))
@@ -858,7 +878,7 @@
                          ((char? p) `((char? ,e)))
                          ((number? p) `((number? ,e)))
                          ((and (pair? p)
-                               (eq? 'quote (car p)))
+                               (equal? 'quote (car p)))
                           `((symbol? ,e)))
                          (else '()))))
                      ((eq? (car tst) 'null?)
@@ -893,18 +913,18 @@
       `(if (and ,tst ,(cadr s))
          ,(caddr s)
          ,f)))
-   (;; (call/cc (lambda (X) 
-    ;;            (let ((Y (lambda () (call-with-values (lambda () f) X)))) 
+   (;; (call/cc (lambda (X)
+    ;;            (let ((Y (lambda () (call-with-values (lambda () f) X))))
     ;;              BODY)))
     ;; =>
-    ;; (call/cc (lambda (X) 
-    ;;            (let ((Y (lambda () (call-with-values (lambda () f) X)))) 
+    ;; (call/cc (lambda (X)
+    ;;            (let ((Y (lambda () (call-with-values (lambda () f) X))))
     ;;              (assm tst (Y) BODY))))
     (and (pair? s)
          (equal? (car s) 'call-with-current-continuation)
          (pair? (cdr s))
          (pair? (cadr s))
-         (equal? (caadr s) 'lambda)
+         (equal? (caadr s) lambda.)
          (pair? (cdadr s))
          (pair? (cadadr s))
          (null? (cdr (cadadr s)))
@@ -916,15 +936,15 @@
          (pair? (caadar (cddadr s)))
          (pair? (cdr (caadar (cddadr s))))
          (pair? (cadr (caadar (cddadr s))))
-         (equal? (caadr (caadar (cddadr s))) 'lambda)
+         (equal? (caadr (caadar (cddadr s))) lambda.)
          (pair? (cdadr (caadar (cddadr s))))
          (null? (cadadr (caadar (cddadr s))))
          (pair? (cddadr (caadar (cddadr s))))
          (pair? (car (cddadr (caadar (cddadr s)))))
-         (equal? (caar (cddadr (caadar (cddadr s)))) 'call-with-values) 
+         (equal? (caar (cddadr (caadar (cddadr s)))) 'call-with-values)
          (pair? (cdar (cddadr (caadar (cddadr s)))))
          (pair? (cadar (cddadr (caadar (cddadr s)))))
-         (equal? (caadar (cddadr (caadar (cddadr s)))) 'lambda)
+         (equal? (caadar (cddadr (caadar (cddadr s)))) lambda.)
          (pair? (cdadar (cddadr (caadar (cddadr s)))))
          (null? (car (cdadar (cddadr (caadar (cddadr s))))))
          (pair? (cdr (cdadar (cddadr (caadar (cddadr s))))))
@@ -944,7 +964,7 @@
           (s2 (caddar (cddadr s))))
       `(call-with-current-continuation
         (lambda (,k)
-          (let ((,fail (lambda () (call-with-values (lambda () ,f) ,k))))
+          (let ((,fail (,lambda. () (call-with-values (,lambda. () ,f) ,k))))
             ,(assm tst `(,fail) s2))))))
    ((and #f
          (pair? s)
@@ -966,7 +986,7 @@
          (equal? (caddar (cdaadr s)) f))
     (let ((fail (caaadr s))
           (s2 (caddr s)))
-      `(let ((,fail (lambda () ,f)))
+      `(let ((,fail (,lambda. () ,f)))
          ,(assm tst `(,fail) s2))))
    (else `(if ,tst ,s ,f))))
 
@@ -991,8 +1011,10 @@
 (define (in e l)
   (or (member e l)
       (and (eq? (car e) 'list?)
-           (or (member `(null? ,(cadr e)) l)
-               (member `(pair? ,(cadr e)) l)))
+           ;; NB: Original Wright's code allows `(pair? ,(cadr e)) here as well,
+           ;; but they are not the same condition - pair? allows improper list
+           ;; but list? doesn't.  Ref: https://github.com/shirok/Gauche/issues/47
+           (member `(null? ,(cadr e)) l))
       (and (eq? (car e) 'not)
            (let* ((srch (cadr e))
                   (const-class (equal-test? srch)))
@@ -1031,7 +1053,7 @@
               (else #f))))))
 
 (define (equal-test? tst)
-  (and (eq? (car tst) 'equal?)
+  (and (equal? (car tst) 'equal?)
        (let ((p (caddr tst)))
          (cond
           ((string? p) 'string?)
@@ -1041,7 +1063,7 @@
           ((and (pair? p)
                 (pair? (cdr p))
                 (null? (cddr p))
-                (eq? 'quote (car p))
+                (equal? 'quote (car p))
                 (symid? (cadr p)))
            'symbol?)
           (else #f)))))
@@ -1116,7 +1138,7 @@
                    #f))
              args))
       ((lambda ()
-         (let ((e (gensym))) `(lambda (,e) (,match. ,e ,@args)))))
+         (let ((e (gensym))) `(,lambda. (,e) (,match. ,e ,@args)))))
       ((lambda ()
          (match:syntax-err
            `(match-lambda ,@args)
@@ -1131,7 +1153,7 @@
                    #f))
              args))
       ((lambda ()
-         (let ((e (gensym))) `(lambda ,e (,match. ,e ,@args)))))
+         (let ((e (gensym))) `(,lambda. ,e (,match. ,e ,@args)))))
       ((lambda ()
          (match:syntax-err
            `(match-lambda* ,@args)
