@@ -1,12 +1,12 @@
 /*
  * bignum.c - multiple precision exact integer arithmetic
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -80,8 +80,9 @@ char *alloca ();
 #include <limits.h>
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/arith.h"
+#include "gauche/priv/arith.h"
 #include "gauche/bits.h"
+#include "gauche/bits_inline.h"
 #include "gauche/bignum.h"
 
 #undef min
@@ -89,10 +90,10 @@ char *alloca ();
 #undef max
 #define max(x, y)   (((x) > (y))? (x) : (y))
 
-static ScmBignum *bignum_rshift(ScmBignum *br, ScmBignum *bx, int amount);
-static ScmBignum *bignum_lshift(ScmBignum *br, ScmBignum *bx, int amount);
-static int bignum_safe_size_for_add(ScmBignum *x, ScmBignum *y);
-static ScmBignum *bignum_add_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by);
+static ScmBignum *bignum_rshift(ScmBignum *br, const ScmBignum *bx, int amount);
+static ScmBignum *bignum_lshift(ScmBignum *br, const ScmBignum *bx, int amount);
+static int bignum_safe_size_for_add(const ScmBignum *x, const ScmBignum *y);
+static ScmBignum *bignum_add_int(ScmBignum *br, const ScmBignum *bx, const ScmBignum *by);
 static ScmBignum *bignum_2scmpl(ScmBignum *br);
 
 /*---------------------------------------------------------------------
@@ -102,8 +103,7 @@ static ScmBignum *bignum_2scmpl(ScmBignum *br);
  */
 static ScmBignum *bignum_clear(ScmBignum *b)
 {
-    u_int i;
-    for (i=0; i<b->size; i++) b->values[i] = 0;
+    for (u_int i=0; i<b->size; i++) b->values[i] = 0;
     return b;
 }
 
@@ -111,10 +111,9 @@ static ScmBignum *bignum_clear(ScmBignum *b)
 
 static ScmBignum *make_bignum(int size)
 {
-    ScmBignum *b;
     if (size < 0) Scm_Error("invalid bignum size (internal error): %d", size);
     if (size > (int)SCM_BIGNUM_MAX_DIGITS) Scm_Error("too large bignum");
-    b = SCM_NEW_ATOMIC2(ScmBignum*, BIGNUM_SIZE(size));
+    ScmBignum *b = SCM_NEW_ATOMIC2(ScmBignum*, BIGNUM_SIZE(size));
     SCM_SET_CLASS(b, SCM_CLASS_INTEGER);
     b->size = size;
     b->sign = 1;
@@ -164,16 +163,15 @@ ScmObj Scm_MakeBignumFromUI(u_long val)
 
 /* If sign > 0 or sign < 0, values[] has absolute value.
    If sign == 0, values[] has 2's complement signed representation */
-ScmObj Scm_MakeBignumFromUIArray(int sign, u_long *values, int size)
+ScmObj Scm_MakeBignumFromUIArray(int sign, const u_long *values, int size)
 {
     ScmBignum *b = make_bignum(size);
-    int i;
     if (sign != 0) {
         b->sign = (sign > 0)? 1 : -1;
-        for (i=0; i<size; i++) b->values[i] = values[i];
+        for (int i=0; i<size; i++) b->values[i] = values[i];
     } else {
         int nonzerop = FALSE;
-        for (i=0; i<size; i++) {
+        for (int i=0; i<size; i++) {
             if ((b->values[i] = values[i]) != 0) nonzerop = TRUE;
         }
         if (nonzerop) {
@@ -188,18 +186,21 @@ ScmObj Scm_MakeBignumFromUIArray(int sign, u_long *values, int size)
 
 ScmObj Scm_MakeBignumFromDouble(double val)
 {
-    int exponent, sign;
-    ScmObj mantissa, b;
-
-    if (val >= LONG_MIN && val <= LONG_MAX) {
+    if (LONG_MIN <= val
+#if SIZEOF_LONG == 4
+        && val <= LONG_MAX
+#else
+        && val <= nextafter((double)LONG_MAX, 0.0)
+#endif
+        )
         return Scm_MakeBignumFromSI((long)val);
-    }
 
-    mantissa = Scm_DecodeFlonum(val, &exponent, &sign);
+    int exponent, sign;
+    ScmObj mantissa = Scm_DecodeFlonum(val, &exponent, &sign);
     if (!SCM_NUMBERP(mantissa)) {
         Scm_Error("can't convert %lf to an integer", val);
     }
-    b = Scm_Ash(mantissa, exponent);
+    ScmObj b = Scm_Ash(mantissa, exponent);
     if (sign < 0) b = Scm_Negate(b);
     /* always returns bignum */
     if (SCM_INTP(b)) {
@@ -209,12 +210,11 @@ ScmObj Scm_MakeBignumFromDouble(double val)
     }
 }
 
-ScmObj Scm_BignumCopy(ScmBignum *b)
+ScmObj Scm_BignumCopy(const ScmBignum *b)
 {
-    u_int i;
     ScmBignum *c = make_bignum(b->size);
     c->sign = b->sign;
-    for (i=0; i<b->size; i++) c->values[i] = b->values[i];
+    for (u_int i=0; i<b->size; i++) c->values[i] = b->values[i];
     return SCM_OBJ(c);
 }
 
@@ -222,6 +222,7 @@ ScmObj Scm_BignumCopy(ScmBignum *b)
  * Conversion
  */
 
+/* Modifies B and return it. */
 ScmObj Scm_NormalizeBignum(ScmBignum *b)
 {
     int size = b->size;
@@ -246,7 +247,7 @@ ScmObj Scm_NormalizeBignum(ScmBignum *b)
 }
 
 /* b must be normalized.  */
-long Scm_BignumToSI(ScmBignum *b, int clamp, int *oor)
+long Scm_BignumToSI(const ScmBignum *b, int clamp, int *oor)
 {
     if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign >= 0) {
@@ -274,7 +275,7 @@ long Scm_BignumToSI(ScmBignum *b, int clamp, int *oor)
 }
 
 /* b must be normalized. */
-u_long Scm_BignumToUI(ScmBignum *b, int clamp, int *oor)
+u_long Scm_BignumToUI(const ScmBignum *b, int clamp, int *oor)
 {
     if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign >= 0) {
@@ -299,32 +300,8 @@ u_long Scm_BignumToUI(ScmBignum *b, int clamp, int *oor)
 
 #if SIZEOF_LONG == 4
 /* we need special routines for int64 */
-ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamp, int *oor)
+int64_t Scm_BignumToSI64(const ScmBignum *b, int clamp, int *oor)
 {
-#if SCM_EMULATE_INT64
-    ScmInt64 r = {0, 0};
-    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
-    if (b->sign > 0) {
-        if (b->size > 2 || b->values[1] > LONG_MAX) {
-            if (!(clamp & SCM_CLAMP_HI)) goto err;
-            SCM_SET_INT64_MAX(r);
-        } else {
-            r.lo = b->values[0];
-            if (b->size == 2) r.hi = b->values[1];
-        }
-    } else if (b->sign < 0) {
-        if (b->size > 2 || b->values[1] > (u_long)LONG_MAX + 1) {
-            if (!(clamp&SCM_CLAMP_LO)) goto err;
-            SCM_SET_INT64_MIN(r);
-        } else {
-            b = SCM_BIGNUM(Scm_BignumComplement(b));
-            r.lo = b->values[0];
-            if (b->size == 2) r.hi = b->values[1];
-            else              r.hi = -1;
-        }
-    }
-    return r;
-#else  /*!SCM_EMULATE_INT64*/
     int64_t r = 0;
     if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
@@ -347,7 +324,6 @@ ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamp, int *oor)
         }
     }
     return r;
-#endif /*!SCM_EMULATE_INT64*/
   err:
     if (clamp == SCM_CLAMP_NONE && oor != NULL) {
         *oor = TRUE;
@@ -357,24 +333,8 @@ ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamp, int *oor)
     return r;
 }
 
-ScmUInt64 Scm_BignumToUI64(ScmBignum *b, int clamp, int *oor)
+uint64_t Scm_BignumToUI64(const ScmBignum *b, int clamp, int *oor)
 {
-#if SCM_EMULATE_INT64
-    ScmInt64 r = {0, 0};
-    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
-    if (b->sign > 0) {
-        if (b->size > 2) {
-            if (!(clamp&SCM_CLAMP_HI)) goto err;
-            SCM_SET_UINT64_MAX(r);
-        } else {
-            r.lo = b->values[0];
-            if (b->size == 2) r.hi = b->values[1];
-        }
-    } else if (b->sign < 0) {
-        if (!(clamp&SCM_CLAMP_LO)) goto err;
-    }
-    return r;
-#else  /*!SCM_EMULATE_INT64*/
     uint64_t r = 0;
     if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
@@ -390,7 +350,6 @@ ScmUInt64 Scm_BignumToUI64(ScmBignum *b, int clamp, int *oor)
         if (!(clamp&SCM_CLAMP_LO)) goto err;
     }
     return r;
-#endif /*!SCM_EMULATE_INT64*/
   err:
     if (clamp == SCM_CLAMP_NONE && oor != NULL) {
         *oor = TRUE;
@@ -401,7 +360,6 @@ ScmUInt64 Scm_BignumToUI64(ScmBignum *b, int clamp, int *oor)
 }
 #endif /* SIZEOF_LONG == 4 */
 
-
 extern double Scm__EncodeDouble(u_long, u_long, int, int);
 
 /* Converts a bignum b to a double.  b must be normalized.
@@ -410,17 +368,16 @@ extern double Scm__EncodeDouble(u_long, u_long, int, int);
    extract bits directly from the bignum values array.
    (We use ScmBits API by casting b->values to ScmBits*).
  */
-double Scm_BignumToDouble(ScmBignum *b)
+double Scm_BignumToDouble(const ScmBignum *b)
 {
     ScmBits *bits = (ScmBits*)b->values;
     ScmBits dst[2];
-    int maxbit, exponent;
 
     /* first, filter out a special case. */
     if (b->size == 0) return 0.0;
 
-    maxbit = Scm_BitsHighest1(bits, 0, b->size*WORD_BITS);
-    exponent = maxbit+1023;
+    int maxbit = Scm_BitsHighest1(bits, 0, b->size*WORD_BITS);
+    int exponent = maxbit+1023;
 #if SIZEOF_LONG >= 8
     SCM_ASSERT(maxbit >= 54);   /* because b is normalized */
     dst[0] = 0;
@@ -440,7 +397,7 @@ double Scm_BignumToDouble(ScmBignum *b)
         }
     }
     if (exponent > 2046) return Scm__EncodeDouble(0, 0, 2047, (b->sign < 0));
-    else return Scm__EncodeDouble(0, dst[0], exponent, (b->sign < 0));
+    else return Scm__EncodeDouble(dst[0], 0, exponent, (b->sign < 0));
 #else  /*SIZEOF_LONG == 4 */
     dst[0] = dst[1] = 0;
     if (maxbit < 53) {
@@ -457,7 +414,7 @@ double Scm_BignumToDouble(ScmBignum *b)
             if (dst[1] >= (1UL<<(52-32))) {
                 /* Overflow.  We mask the hidden bit, then shift. */
                 dst[1] &= ~(1UL<<(52-32));
-                dst[0] = (dst[0] >> 1) | (dst[1]&1 << 31);
+                dst[0] = (dst[0] >> 1) | ((u_long)(dst[1]&1) << 31);
                 dst[1] >>= 1;
                 exponent++;
             }
@@ -469,28 +426,26 @@ double Scm_BignumToDouble(ScmBignum *b)
 }
 
 /* return -b, normalized */
-ScmObj Scm_BignumNegate(ScmBignum *b)
+ScmObj Scm_BignumNegate(const ScmBignum *b)
 {
     ScmObj c = Scm_BignumCopy(b);
     SCM_BIGNUM_SIGN(c) = -SCM_BIGNUM_SIGN(c);
     return Scm_NormalizeBignum(SCM_BIGNUM(c));
 }
-    
+
 /*-----------------------------------------------------------------------
  * Compare
  */
 
 /* bx and by must be normalized */
-int Scm_BignumCmp(ScmBignum *bx, ScmBignum *by)
+int Scm_BignumCmp(const ScmBignum *bx, const ScmBignum *by)
 {
-    int i;
-    
     if (bx->sign < by->sign) return -1;
     if (bx->sign > by->sign) return 1;
     if (bx->size < by->size) return (bx->sign > 0) ? -1 : 1;
     if (bx->size > by->size) return (bx->sign > 0) ? 1 : -1;
 
-    for (i=(int)bx->size-1; i>=0; i--) {
+    for (int i=(int)bx->size-1; i>=0; i--) {
         if (bx->values[i] < by->values[i]) return (bx->sign > 0) ? -1 : 1;
         if (bx->values[i] > by->values[i]) return (bx->sign > 0) ? 1 : -1;
     }
@@ -498,13 +453,11 @@ int Scm_BignumCmp(ScmBignum *bx, ScmBignum *by)
 }
 
 /* compare absolute values.  assume bx and by are nomalized. */
-int Scm_BignumAbsCmp(ScmBignum *bx, ScmBignum *by)
+int Scm_BignumAbsCmp(const ScmBignum *bx, const ScmBignum *by)
 {
-    int i;
-    
     if (bx->size < by->size) return -1;
     if (bx->size > by->size) return 1;
-    for (i=(int)bx->size-1; i>=0; i--) {
+    for (int i=(int)bx->size-1; i>=0; i--) {
         if (bx->values[i] < by->values[i]) return -1;
         if (bx->values[i] > by->values[i]) return 1;
     }
@@ -518,13 +471,12 @@ int Scm_BignumAbsCmp(ScmBignum *bx, ScmBignum *by)
    Screen out the obvious case without actually calculating bx+off.
    Experimentary, the following set of conditions avoid 93% of cases from
    doing actual bignum addition. */
-int Scm_BignumCmp3U(ScmBignum *bx, ScmBignum *off, ScmBignum *by)
+int Scm_BignumCmp3U(const ScmBignum *bx,
+                    const ScmBignum *off,
+                    const ScmBignum *by)
 {
     u_int xsize = SCM_BIGNUM_SIZE(bx), ysize = SCM_BIGNUM_SIZE(by);
     u_int osize = SCM_BIGNUM_SIZE(off);
-    u_int tsize;
-    int i;
-    ScmBignum *br;
 
     if (xsize > ysize) return 1;
     if (xsize < ysize) {
@@ -552,12 +504,13 @@ int Scm_BignumCmp3U(ScmBignum *bx, ScmBignum *off, ScmBignum *by)
         }
         /* fallthrough */
     }
-    tsize = bignum_safe_size_for_add(bx, off);
+    u_int tsize = bignum_safe_size_for_add(bx, off);
+    ScmBignum *br;
     ALLOC_TEMP_BIGNUM(br, tsize);
     bignum_add_int(br, bx, off);
-    
+
     if (br->size < by->size) return -1;
-    for (i=(int)br->size-1; i>=0; i--) {
+    for (int i=(int)br->size-1; i>=0; i--) {
         if (i >= (int)by->size) {
             if (br->values[i]) return 1;
             continue;
@@ -571,7 +524,7 @@ int Scm_BignumCmp3U(ScmBignum *bx, ScmBignum *off, ScmBignum *by)
 /*-----------------------------------------------------------------------
  * Add & subtract
  */
-static int bignum_safe_size_for_add(ScmBignum *x, ScmBignum *y)
+static int bignum_safe_size_for_add(const ScmBignum *x, const ScmBignum *y)
 {
     int xsize = SCM_BIGNUM_SIZE(x);
     int ysize = SCM_BIGNUM_SIZE(y);
@@ -590,16 +543,15 @@ static int bignum_safe_size_for_add(ScmBignum *x, ScmBignum *y)
 static ScmBignum *bignum_2scmpl(ScmBignum *br)
 {
     int rsize = SCM_BIGNUM_SIZE(br);
-    int i;
-    u_long c;
-    for (i=0, c=1; i<rsize; i++) {
+    u_long c = 1;
+    for (int i=0; i<rsize; i++) {
         unsigned long x = ~br->values[i];
         UADD(br->values[i], c, x, 0);
     }
     return br;
 }
 
-ScmObj Scm_BignumComplement(ScmBignum *bx)
+ScmObj Scm_BignumComplement(const ScmBignum *bx)
 {
     ScmBignum *r = SCM_BIGNUM(Scm_BignumCopy(bx));
     return SCM_OBJ(bignum_2scmpl(r));
@@ -607,31 +559,31 @@ ScmObj Scm_BignumComplement(ScmBignum *bx)
 
 /* br = abs(bx) + abs(by), assuming br has enough size. br and bx can be
    the same object. */
-static ScmBignum *bignum_add_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by)
+static ScmBignum *bignum_add_int(ScmBignum *br,
+                                 const ScmBignum *bx, const ScmBignum *by)
 {
     int rsize = SCM_BIGNUM_SIZE(br);
     int xsize = SCM_BIGNUM_SIZE(bx);
     int ysize = SCM_BIGNUM_SIZE(by);
-    int i;
-    u_long c, x, y;
+    u_long c = 0;
 
-    for (i=0, c=0; i<rsize; i++, xsize--, ysize--) {
+    for (int i=0; i<rsize; i++, xsize--, ysize--) {
         if (xsize <= 0) {
             if (ysize <= 0) {
                 UADD(br->values[i], c, 0, 0);
                 continue;
             }
-            y = by->values[i];
+            u_long y = by->values[i];
             UADD(br->values[i], c, 0, y);
             continue;
         }
         if (ysize <= 0) {
-            x = bx->values[i];
+            u_long x = bx->values[i];
             UADD(br->values[i], c, x, 0);
             continue;
         }
-        x = bx->values[i];
-        y = by->values[i];
+        u_long x = bx->values[i];
+        u_long y = by->values[i];
         UADD(br->values[i], c, x, y);
     }
     return br;
@@ -639,31 +591,31 @@ static ScmBignum *bignum_add_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by)
 
 /* br = abs(bx) - abs(by), assuming br has enough size.  br and bx can be
    the same object. */
-static ScmBignum *bignum_sub_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by)
+static ScmBignum *bignum_sub_int(ScmBignum *br,
+                                 const ScmBignum *bx, const ScmBignum *by)
 {
     int rsize = SCM_BIGNUM_SIZE(br);
     int xsize = SCM_BIGNUM_SIZE(bx);
     int ysize = SCM_BIGNUM_SIZE(by);
-    int i;
-    u_long c, x, y;
+    u_long c = 0;
 
-    for (i=0, c=0; i<rsize; i++, xsize--, ysize--) {
+    for (int i=0; i<rsize; i++, xsize--, ysize--) {
         if (xsize <= 0) {
             if (ysize <= 0) {
                 USUB(br->values[i], c, 0, 0);
                 continue;
             }
-            y = by->values[i];
+            u_long y = by->values[i];
             USUB(br->values[i], c, 0, y);
             continue;
         }
         if (ysize <= 0) {
-            x = bx->values[i];
+            u_long x = bx->values[i];
             USUB(br->values[i], c, x, 0);
             continue;
         }
-        x = bx->values[i];
-        y = by->values[i];
+        u_long x = bx->values[i];
+        u_long y = by->values[i];
         USUB(br->values[i], c, x, y);
     }
     if (c != 0) {
@@ -674,7 +626,7 @@ static ScmBignum *bignum_sub_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by)
 }
 
 /* returns bx + by, not normalized */
-static ScmBignum *bignum_add(ScmBignum *bx, ScmBignum *by)
+static ScmBignum *bignum_add(const ScmBignum *bx, const ScmBignum *by)
 {
     int rsize = bignum_safe_size_for_add(bx, by);
     ScmBignum *br = make_bignum(rsize);
@@ -688,7 +640,7 @@ static ScmBignum *bignum_add(ScmBignum *bx, ScmBignum *by)
 }
 
 /* returns bx - by, not normalized */
-static ScmBignum *bignum_sub(ScmBignum *bx, ScmBignum *by)
+static ScmBignum *bignum_sub(const ScmBignum *bx, const ScmBignum *by)
 {
     int rsize = bignum_safe_size_for_add(bx, by);
     ScmBignum *br = make_bignum(rsize);
@@ -702,25 +654,21 @@ static ScmBignum *bignum_sub(ScmBignum *bx, ScmBignum *by)
 }
 
 /* returns bx + y, not nomalized */
-static ScmBignum *bignum_add_si(ScmBignum *bx, long y)
+static ScmBignum *bignum_add_si(const ScmBignum *bx, long y)
 {
-    long c;
-    u_int i, rsize = bx->size+1;
+    long c = 0;
+    u_int rsize = bx->size+1;
     u_long yabs = ((y < 0)? -y : y);
     int ysign = ((y < 0)? -1 : 1);
-    ScmBignum *br;
-
-    if (y == 0) return bx;
-    
-    br = make_bignum(rsize);
+    ScmBignum *br = make_bignum(rsize);
     br->sign = bx->sign;
     if (SCM_BIGNUM_SIGN(bx) == ysign) {
-        for (c=0, i=0; i<bx->size; i++) {
+        for (u_int i=0; i<bx->size; i++) {
             UADD(br->values[i], c, bx->values[i], yabs);
             yabs = 0;
         }
     } else {
-        for (c=0, i=0; i<bx->size; i++) {
+        for (u_int i=0; i<bx->size; i++) {
             USUB(br->values[i], c, bx->values[i], yabs);
             yabs = 0;
         }
@@ -729,24 +677,26 @@ static ScmBignum *bignum_add_si(ScmBignum *bx, long y)
     return br;
 }
 
-ScmObj Scm_BignumAdd(ScmBignum *bx, ScmBignum *by)
+ScmObj Scm_BignumAdd(const ScmBignum *bx, const ScmBignum *by)
 {
     return Scm_NormalizeBignum(bignum_add(bx, by));
 }
 
-ScmObj Scm_BignumSub(ScmBignum *bx, ScmBignum *by)
+ScmObj Scm_BignumSub(const ScmBignum *bx, const ScmBignum *by)
 {
     return Scm_NormalizeBignum(bignum_sub(bx, by));
 }
 
-ScmObj Scm_BignumAddSI(ScmBignum *bx, long y)
+ScmObj Scm_BignumAddSI(const ScmBignum *bx, long y)
 {
-    return Scm_NormalizeBignum(bignum_add_si(bx, y));
+    if (y == 0) return SCM_OBJ(bx);
+    else return Scm_NormalizeBignum(bignum_add_si(bx, y));
 }
 
-ScmObj Scm_BignumSubSI(ScmBignum *bx, long y)
+ScmObj Scm_BignumSubSI(const ScmBignum *bx, long y)
 {
-    return Scm_NormalizeBignum(bignum_add_si(bx, -y));
+    if (y == 0) return SCM_OBJ(bx);
+    else return Scm_NormalizeBignum(bignum_add_si(bx, -y));
 }
 
 /*-----------------------------------------------------------------------
@@ -755,12 +705,12 @@ ScmObj Scm_BignumSubSI(ScmBignum *bx, long y)
 
 /* br = bx >> amount.  amount >= 0.  no normalization.  assumes br
    has enough size to hold the result.  br and bx can be the same object. */
-static ScmBignum *bignum_rshift(ScmBignum *br, ScmBignum *bx, int amount)
+static ScmBignum *bignum_rshift(ScmBignum *br, const ScmBignum *bx, int amount)
 {
     u_int nwords = amount / WORD_BITS;
     u_int nbits = amount % WORD_BITS;
     int i;
-    
+
     if (bx->size <= nwords) {
         br->size = 0; br->values[0] = 0;
     } else if (nbits == 0) {
@@ -784,26 +734,25 @@ static ScmBignum *bignum_rshift(ScmBignum *br, ScmBignum *bx, int amount)
 
 /* br = bx << amount, amount > 0.   no normalization.   assumes br
    has enough size.  br and bx can be the same object. */
-static ScmBignum *bignum_lshift(ScmBignum *br, ScmBignum *bx, int amount)
+static ScmBignum *bignum_lshift(ScmBignum *br, const ScmBignum *bx, int amount)
 {
-    int nwords, nbits, i;
-    u_long x;
-    
-    nwords = amount / WORD_BITS;
-    nbits = amount % WORD_BITS;
+    int nwords = amount / WORD_BITS;
+    int nbits = amount % WORD_BITS;
+
     if (nbits == 0) {
         /* short path */
-        for (i = (int)bx->size-1; i>=0; i--) {
+        for (int i = (int)bx->size-1; i>=0; i--) {
             if ((int)br->size > i+nwords) br->values[i+nwords] = bx->values[i];
         }
-        for (i = nwords-1; i>=0; i--) br->values[i] = 0;
+        for (int i = nwords-1; i>=0; i--) br->values[i] = 0;
     } else {
         if (br->size > bx->size + nwords) {
             br->values[bx->size+nwords] =
                 bx->values[bx->size-1]>>(WORD_BITS-nbits);
         }
+        int i;
         for (i = (int)bx->size-1; i > 0; i--) {
-            x = (bx->values[i]<<nbits)|(bx->values[i-1]>>(WORD_BITS-nbits));
+            u_long x = (bx->values[i]<<nbits)|(bx->values[i-1]>>(WORD_BITS-nbits));
             if ((int)br->size > i+nwords) br->values[i+nwords] = x;
         }
         br->values[nwords] = bx->values[0]<<nbits;
@@ -820,18 +769,16 @@ static ScmBignum *bignum_lshift(ScmBignum *br, ScmBignum *bx, int amount)
  */
 
 /* br += bx * (y << off*WORD_BITS).   br must have enough size. */
-static ScmBignum *bignum_mul_word(ScmBignum *br, ScmBignum *bx,
+static ScmBignum *bignum_mul_word(ScmBignum *br, const ScmBignum *bx,
                                   u_long y, int off)
 {
-    u_long hi, lo, x, r0, r1, c;
-    u_int i,j;
-    
-    for (i=0; i<bx->size; i++) {
-        x = bx->values[i];
+    for (int i=0; i<bx->size; i++) {
+        u_long hi, lo, r1;
+        u_long x = bx->values[i];
         UMUL(hi, lo, x, y);
-        c = 0;
 
-        r0 = br->values[i+off];
+        u_long c = 0;
+        u_long r0 = br->values[i+off];
         UADD(r1, c, r0, lo);
         br->values[i+off] = r1;
 
@@ -839,7 +786,7 @@ static ScmBignum *bignum_mul_word(ScmBignum *br, ScmBignum *bx,
         UADD(r1, c, r0, hi);
         br->values[i+off+1] = r1;
 
-        for (j=i+off+2; c && j<br->size; j++) {
+        for (int j=i+off+2; c && j<br->size; j++) {
             r0 = br->values[j];
             UADD(r1, c, r0, 0);
             br->values[j] = r1;
@@ -849,53 +796,45 @@ static ScmBignum *bignum_mul_word(ScmBignum *br, ScmBignum *bx,
 }
 
 /* returns bx * by.  not normalized */
-static ScmBignum *bignum_mul(ScmBignum *bx, ScmBignum *by)
+static ScmBignum *bignum_mul(const ScmBignum *bx, const ScmBignum *by)
 {
-    u_int i;
     ScmBignum *br = make_bignum(bx->size + by->size);
-    for (i=0; i<by->size; i++) {
+    for (u_int i=0; i<by->size; i++) {
         bignum_mul_word(br, bx, by->values[i], i);
     }
     br->sign = bx->sign * by->sign;
     return br;
 }
 
-static ScmBignum *bignum_mul_si(ScmBignum *bx, long y)
+/* return bx * y,  y != 0 and y != 1 */
+static ScmBignum *bignum_mul_si(const ScmBignum *bx, long y)
 {
-    ScmBignum *br;
-    u_long yabs;
-    
-    if (y == 1) return bx;
-    if (y == 0) {
-        br = make_bignum(1);
-        br->sign = 1;
-        br->values[0] = 0;
-        return br;
-    }
     if (y == -1) {
-        br = SCM_BIGNUM(Scm_BignumCopy(bx));
+        ScmBignum *br = SCM_BIGNUM(Scm_BignumCopy(bx));
         br->sign = -br->sign;
         return br;
     }
     /* TODO: optimize for 2^n case !*/
+    ScmBignum *br;
     br = make_bignum(bx->size + 1); /* TODO: more accurate estimation */
-    yabs = (y<0)? -y:y;
+    u_long yabs = (y<0)? -y:y;
     br->sign = bx->sign;
     bignum_mul_word(br, bx, yabs, 0);
     if (y<0) br->sign = -br->sign;
     return br;
 }
 
-ScmObj Scm_BignumMul(ScmBignum *bx, ScmBignum *by)
+ScmObj Scm_BignumMul(const ScmBignum *bx, const ScmBignum *by)
 {
     ScmBignum *br = bignum_mul(bx, by);
     return Scm_NormalizeBignum(br);
 }
 
-ScmObj Scm_BignumMulSI(ScmBignum *bx, long y)
+ScmObj Scm_BignumMulSI(const ScmBignum *bx, long y)
 {
-    ScmBignum *br = bignum_mul_si(bx, y);
-    return Scm_NormalizeBignum(br);
+    if (y == 1) return SCM_OBJ(bx);
+    else if (y == 0) return SCM_MAKE_INT(0);
+    else return Scm_NormalizeBignum(bignum_mul_si(bx, y));
 }
 
 /*-----------------------------------------------------------------------
@@ -914,17 +853,17 @@ static inline int div_normalization_factor(u_long w)
     return 0;                   /* dummy */
 }
 
-/* General case of division.  We use each half word as a digit. 
+/* General case of division.  We use each half word as a digit.
    Assumes digitsof(dividend) >= digitsof(divisor) > 1.
    Assumes enough digits are allocated to quotient.
    Remainder is returned (not normalized) */
-static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
+static ScmBignum *bignum_gdiv(const ScmBignum *dividend,
+                              const ScmBignum *divisor,
                               ScmBignum *quotient)
 {
-    ScmBignum *u, *v;
     int d = div_normalization_factor(divisor->values[divisor->size-1]);
-    int j, k, n, m;
-    u_long vn_1, vn_2, vv, uj, uj2, cy;
+    int n, m;
+    u_long vv, uj, uj2, cy;
 
 #define DIGIT(num, n) (((n)%2)? HI((num)->values[(n)/2]) : LO((num)->values[(n)/2]))
 #define DIGIT2(num, n) \
@@ -942,7 +881,8 @@ static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
      (num->values[(n)/2] = (v)))
 
     /* normalize */
-    u = make_bignum(dividend->size + 1); /*will be returned as a remainder */
+    ScmBignum *u, *v;
+    u = make_bignum(dividend->size + 1); /* will be returned as a remainder */
     ALLOC_TEMP_BIGNUM(v, divisor->size);
     if (d >= HALF_BITS) {
         d -= HALF_BITS;
@@ -954,8 +894,8 @@ static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
     }
     bignum_lshift(u, dividend, d);
     bignum_lshift(v, divisor, d);
-    vn_1 = DIGIT(v, n-1);
-    vn_2 = DIGIT(v, n-2);
+    u_long vn_1 = DIGIT(v, n-1);
+    u_long vn_2 = DIGIT(v, n-2);
 #undef DIV_DEBUG
 #ifdef DIV_DEBUG
     Scm_Printf(SCM_CUROUT, "shift=%d, n=%d, m=%d\n", d, n, m);
@@ -964,7 +904,7 @@ static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
     Scm_Printf(SCM_CUROUT, "\nvn_1=%08lx, vn_2=%08lx\n", vn_1, vn_2);
 #endif
 
-    for (j = m; j >= 0; j--) {
+    for (int j = m; j >= 0; j--) {
         u_long uu = (DIGIT(u, j+n) << HALF_BITS) + DIGIT(u, j+n-1);
         u_long qq = uu/vn_1;
         u_long rr = uu%vn_1;
@@ -981,7 +921,7 @@ static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
                    uu, qq, rr);
 #endif
         cy = 0;
-        for (k = 0; k < n; k++) {
+        for (int k = 0; k < n; k++) {
             vv = qq * DIGIT(v, k);
             uj = DIGIT2(u, j+k);
             uj2 = uj - vv - cy;
@@ -996,7 +936,7 @@ static ScmBignum *bignum_gdiv(ScmBignum *dividend, ScmBignum *divisor,
         if (cy) {
             qq--;
             cy = 0;
-            for (k = 0; k < n; k++) {
+            for (int k = 0; k < n; k++) {
                 vv = DIGIT(v, k);
                 uj = DIGIT(u, j+k) + vv + cy;
                 cy = (uj >= HALF_WORD)? 1 : 0;
@@ -1039,7 +979,7 @@ static u_long bignum_sdiv(ScmBignum *dividend, u_long divisor)
 }
 
 /* assuming dividend is normalized. */
-ScmObj Scm_BignumDivSI(ScmBignum *dividend, long divisor, long *remainder)
+ScmObj Scm_BignumDivSI(const ScmBignum *dividend, long divisor, long *remainder)
 {
     u_long dd = (divisor < 0)? -divisor : divisor;
     u_long rr;
@@ -1063,32 +1003,70 @@ ScmObj Scm_BignumDivSI(ScmBignum *dividend, long divisor, long *remainder)
     return Scm_NormalizeBignum(q);
 }
 
+/* If we only need rem(bignum, fixnum), we don't need to copy bignum
+   to keep quotient. */
+long Scm_BignumRemSI(const ScmBignum *dividend, long divisor)
+{
+#if (SIZEOF_LONG == 4) && HAVE_UINT64_T
+    /* ILP32 with 64bit integer - easy one. */
+    uint64_t dd = (divisor < 0)? -divisor : divisor;
+    int sign = SCM_BIGNUM_SIGN(dividend);
+    int k = SCM_BIGNUM_SIZE(dividend) - 1;
+    uint64_t m = 0;
+    for (;k >= 0; k--) {
+        uint64_t x = (uint64_t)dividend->values[k] + (m << WORD_BITS);
+        m = x % dd;
+    }
+    return (long)m * sign;
+#else /* SIZEOF_LONG >= 4 || !HAVE_UINT64_T*/
+    /* Here we need double-machine-word `rem` machine-word. */
+    u_long dd = (divisor < 0)? -divisor : divisor;
+    int sign = SCM_BIGNUM_SIGN(dividend);
+    int k = SCM_BIGNUM_SIZE(dividend) - 1;
+    int shift = WORD_BITS - Scm__HighestBitNumber(dd) - 1;
+    u_long m = 0;
+    for (;k >= 0; k--) {
+        u_long x = dividend->values[k];
+        int total_shift = 0;
+        /* we calculate [m;x] modulo dd.  since always m < dd it's safe to
+           shift m first. */
+        while (total_shift < WORD_BITS) {
+            int s = ((total_shift + shift < WORD_BITS)?
+                     shift : WORD_BITS-total_shift);
+            m = (m << s) | (x >> (WORD_BITS - s));
+            x <<= s;
+            m %= dd;
+            total_shift += s;
+        }
+    }
+    return (long)m * sign;
+#endif
+}
+
 /* assuming dividend and divisor is normalized.  returns quotient and
    remainder */
-ScmObj Scm_BignumDivRem(ScmBignum *dividend, ScmBignum *divisor)
+ScmObj Scm_BignumDivRem(const ScmBignum *dividend, const ScmBignum *divisor)
 {
-    ScmBignum *q, *r;
-
     /* special case */
     if (Scm_BignumAbsCmp(dividend, divisor) < 0) {
         return Scm_Cons(SCM_MAKE_INT(0), SCM_OBJ(dividend));
     }
 
-    q = make_bignum(dividend->size - divisor->size + 1);
-    r = bignum_gdiv(dividend, divisor, q);
+    ScmBignum *q = make_bignum(dividend->size - divisor->size + 1);
+    ScmBignum *r = bignum_gdiv(dividend, divisor, q);
     q->sign = dividend->sign * divisor->sign;
     r->sign = dividend->sign;
-    
+
     return Scm_Cons(Scm_NormalizeBignum(q), Scm_NormalizeBignum(r));
 }
 
 /*-----------------------------------------------------------------------
- * Logical (bitwise) opertaions
+ * Logical (bitwise) operations
  */
 
-ScmObj Scm_BignumAsh(ScmBignum *x, int cnt)
+ScmObj Scm_BignumAsh(const ScmBignum *x, int cnt)
 {
-    if (cnt == 0) return Scm_NormalizeBignum(x);
+    if (cnt == 0) return SCM_OBJ(x);
     if (cnt > 0) {
         int rsize = SCM_BIGNUM_SIZE(x) + (cnt+WORD_BITS-1)/WORD_BITS;
         ScmBignum *r = make_bignum(rsize);
@@ -1119,7 +1097,8 @@ ScmObj Scm_BignumAsh(ScmBignum *x, int cnt)
 /* internal routine for logand.  z = x & y.  assumes z has enough size.
  * assumes x and y are in 2's complement form (sign is ignored).
  */
-static ScmBignum *bignum_and(ScmBignum *z, ScmBignum *x, ScmBignum *y,
+static ScmBignum *bignum_and(ScmBignum *z,
+                             const ScmBignum *x, const ScmBignum *y,
                              int commsize, int xsize, int ysize)
 {
     int i;
@@ -1134,32 +1113,35 @@ static ScmBignum *bignum_and(ScmBignum *z, ScmBignum *x, ScmBignum *y,
     return z;
 }
 
-ScmObj Scm_BignumLogAnd(ScmBignum *x, ScmBignum *y)
+ScmObj Scm_BignumLogAnd(const ScmBignum *x, const ScmBignum *y)
 {
     int xsize = SCM_BIGNUM_SIZE(x), xsign = SCM_BIGNUM_SIGN(x);
     int ysize = SCM_BIGNUM_SIZE(y), ysign = SCM_BIGNUM_SIGN(y);
     int zsize, minsize = min(xsize, ysize);
-    ScmBignum *xx, *yy, *z;
 
     if (xsign > 0) {
         if (ysign > 0) {
-            z = bignum_and(make_bignum(minsize), x, y, minsize, 0, 0);
+            ScmBignum *z = bignum_and(make_bignum(minsize), x, y, minsize,
+                                      0, 0);
             return Scm_NormalizeBignum(z);
         } else {
-            yy = SCM_BIGNUM(Scm_BignumComplement(y));
-            z = bignum_and(make_bignum(xsize), x, yy, minsize, xsize, 0);
+            ScmBignum *yy = SCM_BIGNUM(Scm_BignumComplement(y));
+            ScmBignum *z = bignum_and(make_bignum(xsize), x, yy, minsize,
+                                      xsize, 0);
             return Scm_NormalizeBignum(z);
         }
     } else {
         if (ysign > 0) {
-            xx = SCM_BIGNUM(Scm_BignumComplement(x));
-            z = bignum_and(make_bignum(ysize), xx, y, minsize, 0, ysize);
+            ScmBignum *xx = SCM_BIGNUM(Scm_BignumComplement(x));
+            ScmBignum *z = bignum_and(make_bignum(ysize), xx, y, minsize,
+                                      0, ysize);
             return Scm_NormalizeBignum(z);
         } else {
-            xx = SCM_BIGNUM(Scm_BignumComplement(x));
-            yy = SCM_BIGNUM(Scm_BignumComplement(y));
+            ScmBignum *xx = SCM_BIGNUM(Scm_BignumComplement(x));
+            ScmBignum *yy = SCM_BIGNUM(Scm_BignumComplement(y));
             zsize = max(xsize, ysize);
-            z = bignum_and(make_bignum(zsize), xx, yy, minsize, xsize, ysize);
+            ScmBignum *z = bignum_and(make_bignum(zsize), xx, yy, minsize,
+                                      xsize, ysize);
             SCM_BIGNUM_SIGN(z) = -1;
             bignum_2scmpl(z);
             return Scm_NormalizeBignum(z);
@@ -1170,7 +1152,8 @@ ScmObj Scm_BignumLogAnd(ScmBignum *x, ScmBignum *y)
 /* internal routine for logior.  z = x | y.  assumes z has enough size.
  * assumes x and y are in 2's complement form (sign is ignored).
  */
-static ScmBignum *bignum_ior(ScmBignum *z, ScmBignum *x, ScmBignum *y,
+static ScmBignum *bignum_ior(ScmBignum *z,
+                             const ScmBignum *x, const ScmBignum *y,
                              int commsize, int xsize, int ysize)
 {
     int i;
@@ -1185,36 +1168,39 @@ static ScmBignum *bignum_ior(ScmBignum *z, ScmBignum *x, ScmBignum *y,
     return z;
 }
 
-ScmObj Scm_BignumLogIor(ScmBignum *x, ScmBignum *y)
+ScmObj Scm_BignumLogIor(const ScmBignum *x, const ScmBignum *y)
 {
     int xsize = SCM_BIGNUM_SIZE(x), xsign = SCM_BIGNUM_SIGN(x);
     int ysize = SCM_BIGNUM_SIZE(y), ysign = SCM_BIGNUM_SIGN(y);
-    int zsize, minsize = min(xsize, ysize);
-    ScmBignum *xx, *yy, *z;
+    int minsize = min(xsize, ysize);
 
     if (xsign >= 0) {
         if (ysign >= 0) {
-            zsize = max(xsize, ysize);
-            z = bignum_ior(make_bignum(zsize), x, y, minsize, xsize, ysize);
+            int zsize = max(xsize, ysize);
+            ScmBignum *z = bignum_ior(make_bignum(zsize), x, y, minsize,
+                                      xsize, ysize);
             return Scm_NormalizeBignum(z);
         } else {
-            yy = SCM_BIGNUM(Scm_BignumComplement(y));
-            z = bignum_ior(make_bignum(ysize), x, yy, minsize, 0, ysize);
+            ScmBignum *yy = SCM_BIGNUM(Scm_BignumComplement(y));
+            ScmBignum *z = bignum_ior(make_bignum(ysize), x, yy, minsize,
+                                      0, ysize);
             SCM_BIGNUM_SIGN(z) = -1;
             bignum_2scmpl(z);
             return Scm_NormalizeBignum(z);
         }
     } else {
         if (ysign >= 0) {
-            xx = SCM_BIGNUM(Scm_BignumComplement(x));
-            z = bignum_ior(make_bignum(xsize), xx, y, minsize, xsize, 0);
+            ScmBignum *xx = SCM_BIGNUM(Scm_BignumComplement(x));
+            ScmBignum *z = bignum_ior(make_bignum(xsize), xx, y, minsize,
+                                      xsize, 0);
             SCM_BIGNUM_SIGN(z) = -1;
             bignum_2scmpl(z);
             return Scm_NormalizeBignum(z);
         } else {
-            xx = SCM_BIGNUM(Scm_BignumComplement(x));
-            yy = SCM_BIGNUM(Scm_BignumComplement(y));
-            z = bignum_ior(make_bignum(minsize), xx, yy, minsize, 0, 0);
+            ScmBignum *xx = SCM_BIGNUM(Scm_BignumComplement(x));
+            ScmBignum *yy = SCM_BIGNUM(Scm_BignumComplement(y));
+            ScmBignum *z = bignum_ior(make_bignum(minsize), xx, yy, minsize,
+                                      0, 0);
             SCM_BIGNUM_SIGN(z) = -1;
             bignum_2scmpl(z);
             return Scm_NormalizeBignum(z);
@@ -1222,7 +1208,7 @@ ScmObj Scm_BignumLogIor(ScmBignum *x, ScmBignum *y)
     }
 }
 
-ScmObj Scm_BignumLogXor(ScmBignum *x, ScmBignum *y)
+ScmObj Scm_BignumLogXor(const ScmBignum *x, const ScmBignum *y)
 {
     /* TODO: more efficient implementation */
     ScmObj xandy = Scm_BignumLogAnd(x, y);
@@ -1230,13 +1216,12 @@ ScmObj Scm_BignumLogXor(ScmBignum *x, ScmBignum *y)
     return Scm_LogAnd(xory, Scm_LogNot(xandy));
 }
 
-int Scm_BignumLogCount(ScmBignum *b)
+int Scm_BignumLogCount(const ScmBignum *b)
 {
-    ScmBits *bits;
-    ScmBignum *z = (SCM_BIGNUM_SIGN(b)>0)? b : SCM_BIGNUM(Scm_BignumComplement(b));
+    const ScmBignum *z = (SCM_BIGNUM_SIGN(b)>0)? b : SCM_BIGNUM(Scm_BignumComplement(b));
     int size = SCM_BIGNUM_SIZE(z) * SCM_WORD_BITS;
 
-    bits = (ScmBits*)z->values;
+    ScmBits *bits = (ScmBits*)z->values;
     if (b->sign > 0) {
         return Scm_BitsCount1(bits, 0, size);
     } else {
@@ -1249,19 +1234,17 @@ int Scm_BignumLogCount(ScmBignum *b)
  * Printing
  */
 
-ScmObj Scm_BignumToString(ScmBignum *b, int radix, int use_upper)
+ScmObj Scm_BignumToString(const ScmBignum *b, int radix, int use_upper)
 {
     static const char ltab[] = "0123456789abcdefghijklmnopqrstuvwxyz";
     static const char utab[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
     const char *tab = use_upper? utab : ltab;
     ScmObj h = SCM_NIL, t = SCM_NIL;
-    ScmBignum *q;
-    long rem;
-    if (radix < 2 || radix > 36)
+    if (radix < SCM_RADIX_MIN || radix > SCM_RADIX_MAX)
         Scm_Error("radix out of range: %d", radix);
-    q = SCM_BIGNUM(Scm_BignumCopy(b));
+    ScmBignum *q = SCM_BIGNUM(Scm_BignumCopy(b));
     for (;q->size > 0;) {
-        rem = bignum_sdiv(q, radix);
+        long rem = bignum_sdiv(q, radix);
         SCM_ASSERT(rem >= 0 && rem < radix);
         SCM_APPEND1(h, t, SCM_MAKE_CHAR(tab[rem]));
         for (; q->values[q->size-1] == 0 && q->size > 0; q->size--)
@@ -1271,19 +1254,16 @@ ScmObj Scm_BignumToString(ScmBignum *b, int radix, int use_upper)
     return Scm_ListToString(Scm_ReverseX(h));
 }
 
-#if SCM_DEBUG_HELPER
-int Scm_DumpBignum(ScmBignum *b, ScmPort *out)
+int Scm_DumpBignum(const ScmBignum *b, ScmPort *out)
 {
-    int i;
     Scm_Printf(out, "#<bignum ");
     if (b->sign < 0) SCM_PUTC('-', out);
-    for (i=(int)b->size-1; i>=0; i--) {
-        Scm_Printf(out, "%08x ", b->values[i]);
+    for (int i=(int)b->size-1; i>=0; i--) {
+        Scm_Printf(out, "%08lx ", b->values[i]);
     }
     SCM_PUTC('>', out);
     return 0;
 }
-#endif
 
 /*-----------------------------------------------------------------------
  * Denormalized bignum API
@@ -1307,12 +1287,12 @@ ScmBignum *Scm_MakeBignumWithSize(int size, u_long init)
 ScmBignum *Scm_BignumAccMultAddUI(ScmBignum *acc, u_long coef, u_long c)
 {
     ScmBignum *r;
-    u_int rsize = acc->size + 1, i;
+    u_int rsize = acc->size + 1;
     ALLOC_TEMP_BIGNUM(r, rsize);
     r->values[0] = c;
     bignum_mul_word(r, acc, coef, 0);
     if (r->values[rsize-1] == 0) {
-        for (i=0; i<acc->size; i++) {
+        for (u_int i=0; i<acc->size; i++) {
             acc->values[i] = r->values[i];
         }
         return acc;
@@ -1320,10 +1300,9 @@ ScmBignum *Scm_BignumAccMultAddUI(ScmBignum *acc, u_long coef, u_long c)
         ScmBignum *rr;
         rr = make_bignum(rsize + 3); /* 3 is arbitrary size increment */
         rr->sign = acc->sign;
-        for (i=0; i<rsize; i++) {
+        for (u_int i=0; i<rsize; i++) {
             rr->values[i] = r->values[i];
         }
         return rr;
     }
 }
-
