@@ -1,12 +1,12 @@
 /*
  * code.c - compiled code builder/handler
  *
- *   Copyright (c) 2005-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2005-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -36,7 +36,9 @@
 #include "gauche/class.h"
 #include "gauche/code.h"
 #include "gauche/vminsn.h"
-#include "gauche/builtin-syms.h"
+#include "gauche/priv/codeP.h"
+#include "gauche/priv/identifierP.h"
+#include "gauche/priv/builtin-syms.h"
 
 /*===============================================================
  * NVM related stuff
@@ -65,14 +67,6 @@ ScmObj Scm_CompiledCodeFullName(ScmCompiledCode *cc)
     }
 }
 
-static void compiled_code_print(ScmObj obj, ScmPort *out, ScmWriteContext *c)
-{
-    Scm_Printf(out, "#<compiled-code %S@%p>",
-               Scm_CompiledCodeFullName(SCM_COMPILED_CODE(obj)), obj);
-}
-
-SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_CompiledCodeClass, compiled_code_print);
-
 static ScmCompiledCode *make_compiled_code(void)
 {
     ScmCompiledCode *cc = SCM_NEW(ScmCompiledCode);
@@ -80,14 +74,27 @@ static ScmCompiledCode *make_compiled_code(void)
     cc->code = NULL;
     cc->constants = NULL;
     cc->maxstack = -1;
-    cc->info = SCM_NIL;
-    cc->argInfo = SCM_FALSE;
+    cc->debugInfo = SCM_NIL;
+    cc->signatureInfo = SCM_FALSE;
     cc->name = SCM_FALSE;
     cc->parent = SCM_FALSE;
     cc->builder = NULL;
     return cc;
 }
 
+/* Copy the source compiled-code into the destination.  This is used in
+   the external optimizer to 'edit' code vector---such routines can
+   create a new compiled-code, then copy it to the original so that
+   the identity of compiled-code is kept. */
+void Scm_CompiledCodeCopyX(ScmCompiledCode *dest,
+                           const ScmCompiledCode *src)
+{
+    SCM_ASSERT(dest->builder == NULL);
+    SCM_ASSERT(src->builder == NULL);
+
+    memcpy(dest, src, sizeof(ScmCompiledCode));
+}
+
 /*----------------------------------------------------------------------
  * An API to execute statically compiled toplevel code.  *PROVISIONAL*
  */
@@ -99,20 +106,21 @@ void Scm_VMExecuteToplevels(ScmCompiledCode *cs[])
     Scm_ApplyRec(proc, SCM_NIL);
 }
 
-static ScmObj execute_toplevels_cc(ScmObj result, void **data)
+static ScmObj execute_toplevels_cc(ScmObj result SCM_UNUSED, void **data)
 {
     ScmCompiledCode **cs = (ScmCompiledCode **)data[0];
-    ScmVM *vm;
     if (cs[0] == NULL) return SCM_UNDEFINED;
     data[0] = cs+1;
-    vm = Scm_VM();
+    ScmVM *vm = Scm_VM();
     Scm_VMPushCC(execute_toplevels_cc, data, 1);
     vm->base = cs[0];
     vm->pc = vm->base->code;
     return SCM_UNDEFINED;
 }
 
-static ScmObj execute_toplevels(ScmObj *args, int nargs, void *cv)
+static ScmObj execute_toplevels(ScmObj *args SCM_UNUSED,
+                                int nargs SCM_UNUSED,
+                                void *cv)
 {
     Scm_VMPushCC(execute_toplevels_cc, &cv, 1);
     return SCM_UNDEFINED;
@@ -121,31 +129,25 @@ static ScmObj execute_toplevels(ScmObj *args, int nargs, void *cv)
 /*----------------------------------------------------------------------
  * Disassembler
  */
+static ScmObj check_lifted_closure(ScmWord *p, ScmObj lifted);
+static void print_header(const char *prefix, ScmObj name, ScmCompiledCode *cc);
+
 void Scm_CompiledCodeDump(ScmCompiledCode *cc)
 {
-    int i;
-    ScmWord *p;
-    ScmObj closures = SCM_NIL, cp;
-    int clonum = 0;
+    ScmObj closures = SCM_NIL, lifted = SCM_NIL, shown_lifted = SCM_NIL;
+    int clonum = 0, more = FALSE;
 
-    Scm_Printf(SCM_CUROUT, "main_code (name=%S, code=%p, size=%d, const=%d, stack=%d):\n",
-               cc->name, cc->code, cc->codeSize, cc->constantSize,
-               cc->maxstack);
+    print_header("main_code", SCM_MAKE_STR(""), cc);
     do {
-      loop:
-        p = cc->code;
-        Scm_Printf(SCM_CUROUT, "args: %S\n", cc->argInfo);
-        for (i=0; i < cc->codeSize; i++) {
+        ScmWord *p = cc->code;
+        Scm_Printf(SCM_CUROUT, "signatureInfo: %S\n", cc->signatureInfo);
+        for (int i=0; i < cc->codeSize; i++) {
             ScmWord insn = p[i];
-            ScmObj info, s;
             ScmPort *out = SCM_PORT(Scm_MakeOutputStringPort(TRUE));
-            u_int code;
-            const char *insn_name;
+            ScmObj info = Scm_Assq(SCM_MAKE_INT(i), cc->debugInfo);
+            u_int code = SCM_VM_INSN_CODE(insn);
+            const char *insn_name = Scm_VMInsnName(code);
 
-            info = Scm_Assq(SCM_MAKE_INT(i), cc->info);
-            code = SCM_VM_INSN_CODE(insn);
-            insn_name = Scm_VMInsnName(code);
-            
             switch (Scm_VMInsnNumParams(code)) {
             case 0:
                 Scm_Printf(out, "  %4d %s ", i, insn_name);
@@ -165,6 +167,8 @@ void Scm_CompiledCodeDump(ScmCompiledCode *cc)
                 i++;
                 break;
             case SCM_VM_OPERAND_OBJ:
+                /* Check if we're referring to a lifted closure. */
+                lifted = check_lifted_closure(p+i, lifted);
                 Scm_Printf(out, "%S", p[i+1]);
                 i++;
                 break;
@@ -179,8 +183,9 @@ void Scm_CompiledCodeDump(ScmCompiledCode *cc)
                 clonum++;
                 i++;
                 break;
-            case SCM_VM_OPERAND_CODES:
+            case SCM_VM_OPERAND_CODES: {
                 Scm_Printf(out, "(");
+                ScmObj cp;
                 SCM_FOR_EACH(cp, SCM_OBJ(p[i+1])) {
                     if (SCM_COMPILED_CODE_P(SCM_CAR(cp))) {
                         closures = Scm_Acons(SCM_CAR(cp),
@@ -193,12 +198,13 @@ void Scm_CompiledCodeDump(ScmCompiledCode *cc)
                 Scm_Printf(out, ")");
                 i++;
                 break;
+            }
             default:
                 /*nothing*/;
             }
 
             /* Show info */
-            s = Scm_GetOutputStringUnsafe(out, 0);
+            ScmObj s = Scm_GetOutputStringUnsafe(out, 0);
             if (!SCM_PAIRP(info)) {
                 Scm_Puts(SCM_STRING(s), SCM_CUROUT);
                 Scm_Putc('\n', SCM_CUROUT);
@@ -220,15 +226,78 @@ void Scm_CompiledCodeDump(ScmCompiledCode *cc)
                 }
             }
         }
+        more = FALSE;
         if (!SCM_NULLP(closures)) {
             cc = SCM_COMPILED_CODE(SCM_CAAR(closures));
-            Scm_Printf(SCM_CUROUT, "internal_closure_%S (name=%S, code=%p, size=%d, const=%d stack=%d):\n",
-                       SCM_CDAR(closures), cc->name, cc->code,
-                       cc->codeSize, cc->constantSize, cc->maxstack);
+            print_header("closure:", SCM_CDAR(closures), cc);
             closures = SCM_CDR(closures);
-            goto loop;
+            more = TRUE;
+        } else if (!SCM_NULLP(lifted)) {
+            while (!SCM_NULLP(lifted)) {
+                if (SCM_FALSEP(Scm_Memq(SCM_CAAR(lifted), shown_lifted))) {
+                    cc = SCM_COMPILED_CODE(SCM_CAAR(lifted));
+                    print_header("lifted:", SCM_CDAR(lifted), cc);
+                    shown_lifted = Scm_Cons(SCM_CAAR(lifted), shown_lifted);
+                    lifted = SCM_CDR(lifted);
+                    more = TRUE;
+                    break;
+                } else {
+                    lifted = SCM_CDR(lifted);
+                    continue;
+                }
+            }
         }
-    } while (0);
+    } while (more);
+}
+
+static void print_header(const char *prefix, ScmObj name, ScmCompiledCode *cc)
+{
+    Scm_Printf(SCM_CUROUT, "=== %s%A (name=%S, cc=%p, codevec=%p, size=%d, const=%d stack=%d):\n",
+               prefix, name, cc->name, cc, cc->code,
+               cc->codeSize, cc->constantSize, cc->maxstack);
+}
+
+/* The compiler may have lifted an internal closure to a global procedure.
+   We can tell so if the opcode is GREF_x, and the operand is an identifier,
+   whose name is an uninterned symbol and it is globally bound to a procedure.
+
+   If we indeed have a lifted closure, we chain the closure's code and
+   the identifier into the lifted list, returns the updated list.
+   Otherwise, we return lifted list as is.
+ */
+static ScmObj check_lifted_closure(ScmWord *p, ScmObj lifted)
+{
+    ScmWord code = SCM_VM_INSN_CODE(p[0]);
+    static ScmWord gref_insns[] = {
+        SCM_VM_GREF,
+        SCM_VM_GREF_PUSH,
+        SCM_VM_GREF_CALL,
+        SCM_VM_GREF_TAIL_CALL,
+        SCM_VM_PUSH_GREF,
+        SCM_VM_PUSH_GREF_CALL,
+        SCM_VM_PUSH_GREF_TAIL_CALL
+    };
+
+    if (!SCM_IDENTIFIERP(p[1])) return lifted;
+    ScmIdentifier *id = Scm_OutermostIdentifier(SCM_IDENTIFIER(p[1]));
+    if (SCM_SYMBOL_INTERNED(id->name)) return lifted;
+
+    for (u_int i=0; i < sizeof(gref_insns)/sizeof(ScmWord); i++) {
+        if (code == gref_insns[i]) {
+            ScmObj g = Scm_GlobalVariableRef(id->module, SCM_SYMBOL(id->name),
+                                             SCM_BINDING_STAY_IN_MODULE);
+            if (SCM_CLOSUREP(g)) {
+                if (SCM_FALSEP(Scm_Assq(SCM_CLOSURE(g)->code, lifted))) {
+                    return Scm_Acons(SCM_CLOSURE(g)->code,
+                                     SCM_OBJ(id->name),
+                                     lifted);
+                } else {
+                    return lifted;
+                }
+            }
+        }
+    }
+    return lifted;
 }
 
 /*------------------------------------------------------------------
@@ -252,7 +321,12 @@ typedef struct cc_builder_rec {
     int numChunks;
     ScmObj constants;           /* list of constants */
     int currentIndex;
-    ScmWord currentInsn;        /* buffer for instruction combining. */
+    ScmWord currentInsn;        /* buffer for instruction combining.
+                                   this can be a special value either
+                                   CC_BUILDER_BUFFER_EMPTY or
+                                   CC_BUILDER_BUFFER_TRANS.  see below. */
+    int    prevOpcode;          /* previous saved insn opcode */
+    int    currentOpcode;       /* saved insn opcode */
     int    currentArg0;         /* ditto */
     int    currentArg1;         /* ditto */
     ScmObj currentOperand;      /* ditto */
@@ -262,7 +336,7 @@ typedef struct cc_builder_rec {
     ScmObj labelDefs;           /* alist of (name . offset) */
     ScmObj labelRefs;           /* alist of (name . offset-to-fill) */
     int labelCount;             /* counter to generate unique labels */
-    ScmObj info;                /* alist of (offset (source-info obj)) */
+    ScmObj debugInfo;           /* alist of (offset (source-info obj)) */
 } cc_builder;
 
 /* Indicates that there's no pending instruction. */
@@ -288,18 +362,18 @@ typedef struct cc_builder_rec {
 
 static cc_builder *make_cc_builder(void)
 {
-    cc_builder *b;
-    b = SCM_NEW(cc_builder);
+    cc_builder *b = SCM_NEW(cc_builder);
     b->chunks = NULL;
     b->numChunks = 0;
     b->constants = SCM_NIL;
     b->currentIndex = 0;
     b->currentInsn = CC_BUILDER_BUFFER_EMPTY;
+    b->currentOpcode = b->prevOpcode = -1;
     b->currentOperand = b->currentInfo = SCM_FALSE;
     b->currentState = -1;
     b->labelDefs = b->labelRefs = SCM_NIL;
     b->labelCount = 0;
-    b->info = SCM_NIL;
+    b->debugInfo = SCM_NIL;
     return b;
 }
 
@@ -326,10 +400,10 @@ static void cc_builder_add_constant(cc_builder *b, ScmObj obj)
 static void cc_builder_add_info(cc_builder *b)
 {
     if (SCM_FALSEP(b->currentInfo)) return;
-    b->info = Scm_Acons(SCM_MAKE_INT(b->currentIndex),
-                        SCM_LIST1(Scm_Cons(SCM_SYM_SOURCE_INFO,
-                                           b->currentInfo)),
-                        b->info);
+    b->debugInfo = Scm_Acons(SCM_MAKE_INT(b->currentIndex),
+                             SCM_LIST1(Scm_Cons(SCM_SYM_SOURCE_INFO,
+                                                b->currentInfo)),
+                             b->debugInfo);
     b->currentInfo = SCM_FALSE;
 }
 
@@ -350,8 +424,6 @@ static void finish_transition(cc_builder *b);
 /* Flush the currentInsn buffer. */
 static void cc_builder_flush(cc_builder *b)
 {
-    u_int code;
-    
     if ((b)->currentInsn == CC_BUILDER_BUFFER_EMPTY) return;
     if ((b)->currentInsn == CC_BUILDER_BUFFER_TRANS) {
         finish_transition(b);
@@ -359,7 +431,7 @@ static void cc_builder_flush(cc_builder *b)
     cc_builder_add_info(b);
     cc_builder_add_word(b, b->currentInsn);
 
-    code = SCM_VM_INSN_CODE(b->currentInsn);
+    u_int code = SCM_VM_INSN_CODE(b->currentInsn);
     switch (Scm_VMInsnOperandType(code)) {
     case SCM_VM_OPERAND_ADDR:
         /* Addr should be a label.  We just push the label reference
@@ -397,6 +469,7 @@ static void cc_builder_flush(cc_builder *b)
     }
     b->currentInsn = CC_BUILDER_BUFFER_EMPTY;
     b->currentState = -1;
+    b->currentOpcode = -1;
     return;
   badoperand:
     b->currentInsn = CC_BUILDER_BUFFER_EMPTY;
@@ -415,11 +488,9 @@ static void cc_builder_flush(cc_builder *b)
 static void cc_builder_jumpopt(ScmCompiledCode *cc)
 {
     ScmWord *cp = cc->code;
-    u_int code, i;
-    ScmWord *target;
 
-    for (i=0; i<(u_int)cc->codeSize; i++) {
-        code = SCM_VM_INSN_CODE(*cp); cp++;
+    for (u_int i=0; i<(u_int)cc->codeSize; i++) {
+        u_int code = SCM_VM_INSN_CODE(*cp); cp++;
         switch (Scm_VMInsnOperandType(code)) {
         case SCM_VM_OPERAND_OBJ:;
         case SCM_VM_OPERAND_CODE:;
@@ -429,8 +500,8 @@ static void cc_builder_jumpopt(ScmCompiledCode *cc)
         case SCM_VM_OPERAND_OBJ_ADDR:
             i++; cp++;
             /*FALLTHROUGH*/
-        case SCM_VM_OPERAND_ADDR:
-            target = (ScmWord*)*cp;
+        case SCM_VM_OPERAND_ADDR: {
+            ScmWord *target = (ScmWord*)*cp;
             while (SCM_VM_INSN_CODE(*target) == SCM_VM_JUMP
                    || (code == SCM_VM_BF
                        && SCM_VM_INSN_CODE(*target) == SCM_VM_BF)) {
@@ -441,6 +512,7 @@ static void cc_builder_jumpopt(ScmCompiledCode *cc)
             }
             i++; cp++;
             break;
+        }
         default:
             break;
         }
@@ -465,10 +537,9 @@ ScmObj Scm_MakeCompiledCodeBuilder(int reqargs, int optargs,
 /* Returns a label identifier (integer) unique to this code block */
 ScmObj Scm_CompiledCodeNewLabel(ScmCompiledCode *cc)
 {
-    ScmObj label;
     cc_builder *b;
     CC_BUILDER_GET(b, cc);
-    label = SCM_MAKE_INT(b->labelCount);
+    ScmObj label = SCM_MAKE_INT(b->labelCount);
     b->labelCount++;
     return label;
 }
@@ -477,7 +548,7 @@ ScmObj Scm_CompiledCodeNewLabel(ScmCompiledCode *cc)
 void Scm_CompiledCodeSetLabel(ScmCompiledCode *cc, ScmObj label)
 {
     cc_builder *b;
-    
+
     CC_BUILDER_GET(b, cc);
 
     /* Flush buffered insn first. */
@@ -488,22 +559,31 @@ void Scm_CompiledCodeSetLabel(ScmCompiledCode *cc, ScmObj label)
                              b->labelDefs);
 }
 
+/* Push arbitrary debug-info into the builder.
+   <debug-info> : (<insn-offset> <item> ...)
+   <insn-offset> : <integer> or 'definition
+   Currently supported item:
+   <item> : (<source-info> . source)
+*/
+void Scm_CompiledCodePushInfo(ScmCompiledCode *cc, ScmObj info)
+{
+    cc_builder *b;
+    CC_BUILDER_GET(b, cc);
+    b->debugInfo = Scm_Cons(info, b->debugInfo);
+}
+
 /* Pack the code accumulated in the builder into a code vector.
    Perform label resolution and jump optimization. */
 void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
 {
-    ScmObj cp;
     cc_builder *b;
-    cc_builder_chunk *bc, *bcprev;
-    int i, j, numConstants;
-
     CC_BUILDER_GET(b, cc);
     cc_builder_flush(b);
     cc->code = SCM_NEW_ATOMIC2(ScmWord *, b->currentIndex * sizeof(ScmWord));
     cc->codeSize = b->currentIndex;
 
     /* reverse chunks, leaving the first chunk in bcprev. */
-    bcprev = NULL;
+    cc_builder_chunk *bc, *bcprev = NULL;
     for (bc = b->chunks; bc;) {
         cc_builder_chunk *next = bc->prev;
         bc->prev = bcprev;
@@ -513,7 +593,7 @@ void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
 
     /* pack words */
     bc = bcprev;
-    for (i=0, j=0; i<b->currentIndex; i++, j++) {
+    for (int i=0, j=0; i<b->currentIndex; i++, j++) {
         if (j >= CC_BUILDER_CHUNK_SIZE) {
             bc = bc->prev;
             j = 0;
@@ -522,17 +602,18 @@ void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
     }
 
     /* pack constants */
-    numConstants = Scm_Length(b->constants);
+    int numConstants = Scm_Length(b->constants);
     if (numConstants > 0) {
-        ScmObj cp;
         cc->constants = SCM_NEW_ARRAY(ScmObj, numConstants);
-        for (i=0, cp=b->constants; i<numConstants; i++, cp=SCM_CDR(cp)) {
+        ScmObj cp = b->constants;
+        for (int i=0; i<numConstants; i++, cp=SCM_CDR(cp)) {
             cc->constants[i] = SCM_CAR(cp);
         }
     }
     cc->constantSize = numConstants;
 
     /* resolve labels */
+    ScmObj cp;
     SCM_FOR_EACH(cp, b->labelRefs) {
         int destAddr = cc_builder_label_def(b, SCM_CAAR(cp));
         int operandAddr;
@@ -549,11 +630,11 @@ void Scm_CompiledCodeFinishBuilder(ScmCompiledCode *cc, int maxstack)
     cc_builder_jumpopt(cc);
 
     /* record debug info */
-    cc->info = b->info;
+    cc->debugInfo = b->debugInfo;
 
     /* set max stack depth */
     cc->maxstack = maxstack;
-    
+
     /* make sure this code is 'fixed'---no more building */
     cc->builder = NULL;
 }
@@ -592,6 +673,8 @@ static inline void save_params(cc_builder *b, int code,
                                int arg0, int arg1, ScmObj operand,
                                ScmObj info)
 {
+    b->prevOpcode = b->currentOpcode;
+    b->currentOpcode = code;
     switch (Scm_VMInsnNumParams(code)) {
     case 2: b->currentArg1 = arg1;
         /* FALLTHROUGH */
@@ -607,9 +690,37 @@ static inline void save_params(cc_builder *b, int code,
     }
 }
 
+static int vm_insn_flags(u_int code);
+
 /* Fill the current insn word */
 static inline void fill_current_insn(cc_builder *b, int code)
 {
+    /* A special handling of fold-lref insn.
+       Fold-lref insn is a combined insn LREF-XXXX(depth,offset).  What's
+       special about it is that we 'fold' specialized LREF insn
+       (e.g. LREF10) into generic LREF.
+     */
+#define SET_LREF_ARGS(dep, off) \
+    b->currentArg0 = (dep); b->currentArg1 = (off); break
+    if (vm_insn_flags(code) & SCM_VM_INSN_FOLD_LREF) {
+        switch (b->prevOpcode) {
+        case SCM_VM_LREF0:  SET_LREF_ARGS(0, 0);
+        case SCM_VM_LREF1:  SET_LREF_ARGS(0, 1);
+        case SCM_VM_LREF2:  SET_LREF_ARGS(0, 2);
+        case SCM_VM_LREF3:  SET_LREF_ARGS(0, 3);
+        case SCM_VM_LREF10: SET_LREF_ARGS(1, 0);
+        case SCM_VM_LREF11: SET_LREF_ARGS(1, 1);
+        case SCM_VM_LREF12: SET_LREF_ARGS(1, 2);
+        case SCM_VM_LREF20: SET_LREF_ARGS(2, 0);
+        case SCM_VM_LREF21: SET_LREF_ARGS(2, 1);
+        case SCM_VM_LREF30: SET_LREF_ARGS(3, 0);
+        case SCM_VM_LREF: /* args are already set */ break;
+        default: Scm_Error("[internal] Compiler internal error: FOLD_LREF insn needs to be combined with LREF*, but prevOpcode = %d", b->prevOpcode);
+        }
+    }
+#undef SET_LREF_ARGS
+    
+    /* Compose insn word */
     switch (Scm_VMInsnNumParams(code)) {
     case 0: b->currentInsn = SCM_VM_INSN(code); break;
     case 1: b->currentInsn = SCM_VM_INSN1(code, b->currentArg0); break;
@@ -626,13 +737,13 @@ static inline void fill_current_insn(cc_builder *b, int code)
 static void finish_transition(cc_builder *b)
 {
     int i = b->currentState;
-    SCM_ASSERT(i >= 0 && i < sizeof(stn)/sizeof(struct stn_arc[1]));
+    SCM_ASSERT(i >= 0 && i < (int)(sizeof(stn)/sizeof(struct stn_arc[1])));
     for (;; i++) {
         if (stn[i].input < 0) {
             fill_current_insn(b, stn[i].next);
             break;
         }
-        SCM_ASSERT(i < sizeof(stn)/sizeof(struct stn_arc[1]));
+        SCM_ASSERT(i < (int)(sizeof(stn)/sizeof(struct stn_arc[1])));
     }
 }
 
@@ -717,7 +828,7 @@ void Scm_CompiledCodeEmit(ScmCompiledCode *cc,
 }
 
 /*----------------------------------------------------------------
- * CompiledCode - Scheme interface
+ * CompiledCode - introspection
  */
 
 /* Converts the code vector into a list.
@@ -726,16 +837,15 @@ void Scm_CompiledCodeEmit(ScmCompiledCode *cc,
    Addr operand -> integer offset from the beginning of the code */
 ScmObj Scm_CompiledCodeToList(ScmCompiledCode *cc)
 {
-    u_int i, off;
     ScmObj h = SCM_NIL, t = SCM_NIL;
-    
-    for (i=0; i<(u_int)cc->codeSize; i++) {
+
+    for (u_int i=0; i<(u_int)cc->codeSize; i++) {
         ScmWord insn = cc->code[i];
         u_int code = SCM_VM_INSN_CODE(insn);
         const char *name = Scm_VMInsnName(code);
-        
+
         switch (Scm_VMInsnNumParams(code)) {
-        case 0: 
+        case 0:
             SCM_APPEND1(h, t, SCM_LIST1(SCM_INTERN(name)));
             break;
         case 1:
@@ -755,91 +865,35 @@ ScmObj Scm_CompiledCodeToList(ScmCompiledCode *cc)
         case SCM_VM_OPERAND_CODES:;
             SCM_APPEND1(h, t, SCM_OBJ(cc->code[++i]));
             break;
-        case SCM_VM_OPERAND_ADDR:
-            off = (u_int)((ScmWord*)cc->code[++i] - cc->code);
+        case SCM_VM_OPERAND_ADDR: {
+            u_int off = (u_int)((ScmWord*)cc->code[++i] - cc->code);
             SCM_APPEND1(h, t, SCM_MAKE_INT(off));
             break;
-        case SCM_VM_OPERAND_OBJ_ADDR:
-            off = (u_int)((ScmWord*)cc->code[i+2] - cc->code);
+        }
+        case SCM_VM_OPERAND_OBJ_ADDR: {
+            u_int off = (u_int)((ScmWord*)cc->code[i+2] - cc->code);
             SCM_APPEND(h, t, SCM_LIST2(SCM_OBJ(cc->code[i+1]),
                                        SCM_MAKE_INT(off)));
             i += 2;
             break;
         }
+        }
     }
     return h;
 }
 
-static ScmObj code_size_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->codeSize);
-}
-
-static ScmObj code_maxstack_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->maxstack);
-}
-
-static ScmObj code_info_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->info;
-}
-
-static ScmObj code_arginfo_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->argInfo;
-}
-
-static ScmObj code_reqargs_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->requiredArgs);
-}
-
-static ScmObj code_optargs_get(ScmObj cc)
-{
-    return SCM_MAKE_INT(SCM_COMPILED_CODE(cc)->optionalArgs);
-}
-
-static ScmObj code_name_get(ScmObj cc)
-{
-    return SCM_COMPILED_CODE(cc)->name;
-}
-
-static ScmObj code_parent_get(ScmObj cc)
-{
-    return SCM_OBJ(SCM_COMPILED_CODE(cc)->parent);
-}
-
-static ScmObj code_iform_get(ScmObj cc)
-{
-    return SCM_OBJ(SCM_COMPILED_CODE(cc)->intermediateForm);
-}
-
-static ScmClassStaticSlotSpec code_slots[] = {
-    SCM_CLASS_SLOT_SPEC("parent", code_parent_get, NULL),
-    SCM_CLASS_SLOT_SPEC("arg-info", code_arginfo_get, NULL),
-    SCM_CLASS_SLOT_SPEC("info", code_info_get, NULL),
-    SCM_CLASS_SLOT_SPEC("required-args", code_reqargs_get, NULL),
-    SCM_CLASS_SLOT_SPEC("optional-args", code_optargs_get, NULL),
-    SCM_CLASS_SLOT_SPEC("name", code_name_get, NULL),
-    SCM_CLASS_SLOT_SPEC("full-name", Scm_CompiledCodeFullName, NULL),
-    SCM_CLASS_SLOT_SPEC("size", code_size_get, NULL),
-    SCM_CLASS_SLOT_SPEC("max-stack", code_maxstack_get, NULL),
-    SCM_CLASS_SLOT_SPEC("intermediate-form", code_iform_get, NULL),
-    SCM_CLASS_SLOT_SPEC_END()
-};
-
 /*===========================================================
  * VM Instruction introspection
  */
 
-static struct insn_info {
+static struct {
     const char *name;           /* name */
     int nparams;                /* # of parameters */
     int operandType;            /* operand type */
+    int flags;                  /* flags */
 } insn_table[] = {
-#define DEFINSN(sym, nam, np, type) \
-    { nam, np, SCM_CPP_CAT(SCM_VM_OPERAND_, type) },
+#define DEFINSN(sym, nam, np, type, flags)                     \
+    { nam, np, SCM_CPP_CAT(SCM_VM_OPERAND_, type), flags },
 #include "vminsn.c"
 #undef DEFINSN
 };
@@ -863,6 +917,12 @@ int Scm_VMInsnNumParams(u_int code)
     return insn_table[code].nparams;
 }
 
+int vm_insn_flags(u_int code)   /* private for the time being */
+{
+    CHECK_CODE(code);
+    return insn_table[code].flags;
+}
+
 int Scm_VMInsnOperandType(u_int code)
 {
     CHECK_CODE(code);
@@ -871,17 +931,12 @@ int Scm_VMInsnOperandType(u_int code)
 
 int Scm_VMInsnNameToCode(ScmObj name)
 {
-    const char *n;
-    struct insn_info *info;
-    int i;
-    
     if (SCM_SYMBOLP(name))  name = SCM_OBJ(SCM_SYMBOL_NAME(name));
     else if (!SCM_STRINGP(name)) {
         Scm_Error("vm-insn-name->code: requires a symbol or a string, but got %S", name);
     }
-    n = Scm_GetStringConst(SCM_STRING(name));
-    info = insn_table;
-    for (i=0; i<SCM_VM_NUM_INSNS; i++) {
+    const char *n = Scm_GetStringConst(SCM_STRING(name));
+    for (int i=0; i<SCM_VM_NUM_INSNS; i++) {
         if (strcmp(insn_table[i].name, n) == 0) {
             return i;
         }
@@ -893,11 +948,11 @@ int Scm_VMInsnNameToCode(ScmObj name)
 /* (kind of) inversion of VMInsnInspect. */
 ScmWord Scm_VMInsnBuild(ScmObj obj)
 {
-    int len = Scm_Length(obj), code, arg0, arg1;
-    
+    int len = Scm_Length(obj);
+
     if (len < 1 || len > 3 || !SCM_SYMBOLP(SCM_CAR(obj))) goto badspec;
-    code = Scm_VMInsnNameToCode(SCM_CAR(obj));
-    
+    int code = Scm_VMInsnNameToCode(SCM_CAR(obj));
+
     switch (Scm_VMInsnNumParams(code)) {
     case 0:
         if (len != 1) {
@@ -905,36 +960,30 @@ ScmWord Scm_VMInsnBuild(ScmObj obj)
                       SCM_CAR(obj), obj);
         }
         return SCM_VM_INSN(code);
-    case 1:
+    case 1: {
         if (len != 2) {
             Scm_Error("VM instruction %S takes one parameter, but got %S",
                       SCM_CAR(obj), obj);
         }
         if (!SCM_INTP(SCM_CADR(obj))) goto badspec;
-        arg0 = SCM_INT_VALUE(SCM_CADR(obj));
+        int arg0 = SCM_INT_VALUE(SCM_CADR(obj));
         return SCM_VM_INSN1(code, arg0);
-    case 2:
+    }
+    case 2: {
         if (len != 3) {
             Scm_Error("VM instruction %S takes two parameters, but got %S",
                       SCM_CAR(obj), obj);
         }
         if (!SCM_INTP(SCM_CADR(obj))) goto badspec;
         if (!SCM_INTP(SCM_CAR(SCM_CDDR(obj)))) goto badspec;
-        arg0 = SCM_INT_VALUE(SCM_CADR(obj));
-        arg1 = SCM_INT_VALUE(SCM_CAR(SCM_CDDR(obj)));
+        int arg0 = SCM_INT_VALUE(SCM_CADR(obj));
+        int arg1 = SCM_INT_VALUE(SCM_CAR(SCM_CDDR(obj)));
         return SCM_VM_INSN2(code, arg0, arg1);
     }
+    }
     /*FALLTHROUGH*/
-  badspec:    
+  badspec:
     Scm_Error("Bad VM insn spec: %S", obj);
     return 0;       /* dummy */
 }
 
-/*===========================================================
- * Initialization
- */
-void Scm__InitCode(void)
-{
-    Scm_InitStaticClass(SCM_CLASS_COMPILED_CODE, "<compiled-code>",
-                        Scm_GaucheModule(), code_slots, 0);
-}
