@@ -1,12 +1,12 @@
 /*
  * compaux.c - C API bridge for the compiler
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -41,7 +41,8 @@
 #include "gauche/vminsn.h"
 #include "gauche/class.h"
 #include "gauche/code.h"
-#include "gauche/builtin-syms.h"
+#include "gauche/priv/identifierP.h"
+#include "gauche/priv/builtin-syms.h"
 
 /*
  * Syntax
@@ -91,7 +92,8 @@ void Scm_CompileFinish(ScmCompiledCode *cc)
  * Syntactic closure object
  */
 
-static void synclo_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static void synclo_print(ScmObj obj, ScmPort *port,
+                         ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(port, "#<syntactic-closure %S>",
                SCM_SYNTACTIC_CLOSURE(obj)->expr);
@@ -135,7 +137,26 @@ static ScmClassStaticSlotSpec synclo_slots[] = {
  * Identifier object
  */
 
-static void identifier_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+/* 
+ * About identifier's ENV slot.
+ * We close identifier's binding environment (list of frames), so that later 
+ * we can look up its bindings hygienically.  We truncate the frames up to
+ * where the binding occur, for the efficient lookup and comparison.
+ * Notably, the identifiers that are unbound or refer to toplevel variable 
+ * has () in its env.
+ * A caveat--we can't truncate frames at the time of construction, since
+ * the entire frame structure may not be fixed while we're processing internal
+ * defines.  The identifier may refer to another identifier that will be 
+ * inserted later.  Thus, we delay the truncation operation until it is
+ * needed.
+ * The ENV slot itself now contains (<flag> . <frames>), where <flag> is #f
+ * if truncation hasn't be done, #t otherwise.  ENV should be treated as
+ * an opaque data for the others; you should always get it with 
+ * Scm_IdentifierEnv(), and never directly access ENV slot itself.
+ */
+
+static void identifier_print(ScmObj obj, ScmPort *port,
+                             ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmIdentifier *id = SCM_IDENTIFIER(obj);
     /* We may want to have an external identifier syntax, so that an
@@ -145,47 +166,88 @@ static void identifier_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
        module-based sandbox implementation, so further consideration is
        required.
     */
-    Scm_Printf(port, "#<identifier %S#%S>", id->module->name, id->name);
+    Scm_Printf(port, "#<identifier %S#%S.%x>",
+               id->module->name, id->name, SCM_WORD(id));
 }
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_IdentifierClass, identifier_print);
 
+/* Truncate local frames */
 static ScmObj get_binding_frame(ScmObj var, ScmObj env)
 {
     ScmObj frame, fp;
     SCM_FOR_EACH(frame, env) {
         if (!SCM_PAIRP(SCM_CAR(frame))) continue;
         SCM_FOR_EACH(fp, SCM_CDAR(frame)) {
-            if (SCM_CAAR(fp) == var) return frame;
+            if (SCM_CAAR(fp) == var) {
+                SCM_ASSERT(SCM_LISTP(frame));
+                return frame;
+            }
         }
     }
     return SCM_NIL;
 }
 
-ScmObj Scm_MakeIdentifier(ScmSymbol *name, ScmModule *mod, ScmObj env)
+ScmObj Scm_MakeIdentifier(ScmObj name, ScmModule *mod, ScmObj env)
 {
     ScmIdentifier *id = SCM_NEW(ScmIdentifier);
     SCM_SET_CLASS(id, SCM_CLASS_IDENTIFIER);
     id->name = name;
     id->module = mod? mod : SCM_CURRENT_MODULE();
-    id->env = (env == SCM_NIL)? SCM_NIL : get_binding_frame(SCM_OBJ(name), env);
+    id->frames = Scm_Cons(SCM_FALSE, env); /* see the above comment */
     return SCM_OBJ(id);
 }
 
+ScmObj Scm_IdentifierEnv(ScmIdentifier *id)
+{
+    SCM_ASSERT(SCM_PAIRP(id->frames));
+    if (SCM_FALSEP(SCM_CAR(id->frames))) {
+        /* MT safety: This operation is idempotent, so it's ok if more than
+           one thread execute here. */
+        ScmObj f = get_binding_frame(id->name, SCM_CDR(id->frames));
+        SCM_SET_CDR(id->frames, f);
+        SCM_SET_CAR(id->frames, SCM_TRUE);
+    }
+    return SCM_CDR(id->frames);
+}
+
+ScmIdentifier *Scm_OutermostIdentifier(ScmIdentifier *id)
+{
+    while (SCM_IDENTIFIERP(id->name)) {
+        id = SCM_IDENTIFIER(SCM_IDENTIFIER(id)->name);
+    }
+    return id;
+}
+
+ScmSymbol *Scm_UnwrapIdentifier(ScmIdentifier *id)
+{
+    ScmObj z = Scm_OutermostIdentifier(id)->name;
+    SCM_ASSERT(SCM_SYMBOLP(z));
+    return SCM_SYMBOL(z);
+}
+
+/* returns global binding of the identifier */
+ScmGloc *Scm_IdentifierGlobalBinding(ScmIdentifier *id)
+{
+    ScmIdentifier *z = Scm_OutermostIdentifier(id);
+    return Scm_FindBinding(z->module, SCM_SYMBOL(z->name), 0);
+}
+
 /* returns true if SYM has the same binding with ID in ENV. */
 int Scm_IdentifierBindingEqv(ScmIdentifier *id, ScmSymbol *sym, ScmObj env)
 {
-    ScmObj bf = get_binding_frame(SCM_OBJ(sym), env);
-    return (bf == id->env);
+    ScmObj env1 = Scm_IdentifierEnv(id);
+    ScmObj env2 = get_binding_frame(SCM_OBJ(sym), env);
+    return (env1 == env2);
 }
 
-ScmObj Scm_CopyIdentifier(ScmIdentifier *orig)
+ScmObj Scm_WrapIdentifier(ScmIdentifier *orig)
 {
     ScmIdentifier *id = SCM_NEW(ScmIdentifier);
     SCM_SET_CLASS(id, SCM_CLASS_IDENTIFIER);
-    id->name = orig->name;
+    id->name = SCM_OBJ(orig);
     id->module = orig->module;
-    id->env = orig->env;
+    id->frames = orig->frames;
     return SCM_OBJ(id);
 }
 
@@ -196,10 +258,10 @@ static ScmObj identifier_name_get(ScmObj obj)
 
 static void   identifier_name_set(ScmObj obj, ScmObj val)
 {
-    if (!SCM_SYMBOLP(val)) {
-        Scm_Error("symbol required, but got %S", val);
+    if (!SCM_SYMBOLP(val) && !SCM_IDENTIFIERP(val)) {
+        Scm_Error("symbol or identifier required, but got %S", val);
     }
-    SCM_IDENTIFIER(obj)->name = SCM_SYMBOL(val);
+    SCM_IDENTIFIER(obj)->name = val;
 }
 
 static ScmObj identifier_module_get(ScmObj obj)
@@ -207,31 +269,19 @@ static ScmObj identifier_module_get(ScmObj obj)
     return SCM_OBJ(SCM_IDENTIFIER(obj)->module);
 }
 
-static void   identifier_module_set(ScmObj obj, ScmObj val)
-{
-    if (!SCM_MODULEP(val)) {
-        Scm_Error("module required, but got %S", val);
-    }
-    SCM_IDENTIFIER(obj)->module = SCM_MODULE(val);
-}
-
 static ScmObj identifier_env_get(ScmObj obj)
 {
-    return SCM_IDENTIFIER(obj)->env;
-}
-
-static void   identifier_env_set(ScmObj obj, ScmObj val)
-{
-    if (!SCM_LISTP(val)) {
-        Scm_Error("list required, but got %S", val);
-    }
-    SCM_IDENTIFIER(obj)->env = val;
+    return Scm_IdentifierEnv(SCM_IDENTIFIER(obj));
 }
 
+/* Identifier name can be mutated during macro expansion to avoid
+   conflicts on macro-inserted toplevel identifiers.  See
+   %rename-toplevel-identifier! in compiler pass1.
+   Other than that, identifiers must be treated as immutable objects. */
 static ScmClassStaticSlotSpec identifier_slots[] = {
     SCM_CLASS_SLOT_SPEC("name", identifier_name_get, identifier_name_set),
-    SCM_CLASS_SLOT_SPEC("module", identifier_module_get, identifier_module_set),
-    SCM_CLASS_SLOT_SPEC("env", identifier_env_get, identifier_env_set),
+    SCM_CLASS_SLOT_SPEC("module", identifier_module_get, NULL),
+    SCM_CLASS_SLOT_SPEC("env", identifier_env_get, NULL),
     SCM_CLASS_SLOT_SPEC_END()
 };
 
@@ -239,17 +289,17 @@ static ScmClassStaticSlotSpec identifier_slots[] = {
  * Utility functions
  */
 
-/* Convert all identifiers in form into a symbol. 
+/* Convert all identifiers in form into a symbol.
    This keeps linear history to avoid entering infinite loop if
    the given form is circular; but it doens't recover the shared
-   substricture. */
+   substructure. */
 static ScmObj unwrap_rec(ScmObj form, ScmObj history)
 {
-    ScmObj newh;
-    
     if (!SCM_PTRP(form)) return form;
     if (!SCM_FALSEP(Scm_Memq(form, history))) return form;
-    
+
+    ScmObj newh;
+
     if (SCM_PAIRP(form)) {
         ScmObj ca, cd;
         newh = Scm_Cons(form, history);
@@ -262,17 +312,18 @@ static ScmObj unwrap_rec(ScmObj form, ScmObj history)
         }
     }
     if (SCM_IDENTIFIERP(form)) {
-        return SCM_OBJ(SCM_IDENTIFIER(form)->name);
+        return SCM_OBJ(Scm_UnwrapIdentifier(SCM_IDENTIFIER(form)));
     }
     if (SCM_VECTORP(form)) {
-        int i, j, len = SCM_VECTOR_SIZE(form);
-        ScmObj elt, *pelt = SCM_VECTOR_ELEMENTS(form);
+        int len = SCM_VECTOR_SIZE(form);
+        ScmObj *pelt = SCM_VECTOR_ELEMENTS(form);
         newh = Scm_Cons(form, history);
-        for (i=0; i<len; i++, pelt++) {
-            elt = unwrap_rec(*pelt, newh);
+        for (int i=0; i<len; i++, pelt++) {
+            ScmObj elt = unwrap_rec(*pelt, newh);
             if (elt != *pelt) {
                 ScmObj newvec = Scm_MakeVector(len, SCM_FALSE);
                 pelt = SCM_VECTOR_ELEMENTS(form);
+                int j;
                 for (j=0; j<i; j++, pelt++) {
                     SCM_VECTOR_ELEMENT(newvec, j) = *pelt;
                 }
@@ -326,4 +377,3 @@ void Scm__InitCompaux(void)
 
     Scm_ApplyRec0(SCM_GLOC_GET(init_compiler_gloc));
 }
-
