@@ -1,12 +1,12 @@
 /*
  * core.c - core kernel interface
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -33,13 +33,11 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/arch.h"
-#include "gauche/paths.h"
-#include "gauche/builtin-syms.h"
+#include "gauche/priv/builtin-syms.h"
 
 /* GC_print_static_roots() is declared in private/gc_priv.h.  It is too much
    hassle to include it with other GC internal baggages, so we just declare
-   it.  It's a private funciton of GC, so watch out the changes in GC. */
+   it.  It's a private function of GC, so watch out the changes in GC. */
 extern void GC_print_static_roots(void);
 
 /*
@@ -56,18 +54,18 @@ static void *oom_handler(size_t bytes)
  * Features list used by cond-expand macro
  */
 static struct {
-    ScmObj alist;
+    ScmObj alist;               /* list of (feature) or (feature module) */
+    ScmObj dlist;               /* list of disabled features */
     ScmInternalMutex mutex;
-} cond_features = { SCM_NIL };
+} cond_features = { SCM_NIL, SCM_NIL, SCM_INTERNAL_MUTEX_INITIALIZER };
 
 /*=============================================================
  * Program initialization
  */
 
 extern void Scm__InitModule(void);
-extern void Scm__InitModulePost(void);
+extern void Scm__InitHash(void);
 extern void Scm__InitSymbol(void);
-extern void Scm__InitKeyword(void);
 extern void Scm__InitNumber(void);
 extern void Scm__InitChar(void);
 extern void Scm__InitClass(void);
@@ -77,25 +75,44 @@ extern void Scm__InitWrite(void);
 extern void Scm__InitCompaux(void);
 extern void Scm__InitMacro(void);
 extern void Scm__InitLoad(void);
+extern void Scm__InitParameter(void);
 extern void Scm__InitProc(void);
 extern void Scm__InitRegexp(void);
 extern void Scm__InitRead(void);
 extern void Scm__InitSignal(void);
 extern void Scm__InitSystem(void);
-extern void Scm__InitCode(void);
 extern void Scm__InitVM(void);
-extern void Scm__InitRepl(void);
 extern void Scm__InitAutoloads(void);
 extern void Scm__InitCollection(void);
+extern void Scm__InitComparator(void);
 
-extern void Scm_Init_stdlib(ScmModule *);
-extern void Scm_Init_extlib(ScmModule *);
-extern void Scm_Init_syslib(ScmModule *);
-extern void Scm_Init_intlib(ScmModule *);
-
-extern void Scm_Init_scmlib(void);
+extern void Scm_Init_libalpha(void);
+extern void Scm_Init_libbool(void);
+extern void Scm_Init_libchar(void);
+extern void Scm_Init_libcode(void);
+extern void Scm_Init_libcmp(void);
+extern void Scm_Init_libdict(void);
+extern void Scm_Init_libeval(void);
+extern void Scm_Init_libexc(void);
+extern void Scm_Init_libfmt(void);
+extern void Scm_Init_libio(void);
+extern void Scm_Init_liblazy(void);
+extern void Scm_Init_liblist(void);
+extern void Scm_Init_libmisc(void);
+extern void Scm_Init_libmod(void);
+extern void Scm_Init_libnum(void);
+extern void Scm_Init_libobj(void);
+extern void Scm_Init_libproc(void);
+extern void Scm_Init_librx(void);
+extern void Scm_Init_libsrfis(void);
+extern void Scm_Init_libstr(void);
+extern void Scm_Init_libsym(void);
+extern void Scm_Init_libsys(void);
+extern void Scm_Init_libvec(void);
+extern void Scm_Init_libmacbase(void);
 extern void Scm_Init_compile(void);
-extern void Scm_Init_objlib(void);
+extern void Scm_Init_libmacro(void);
+extern void Scm_Init_libomega(void);
 
 static void finalizable(void);
 static void init_cond_features(void);
@@ -105,11 +122,16 @@ static void init_cond_features(void);
 static int (*ptr_pthread_create)(void) = NULL;
 #endif
 
+/* flag to see if Scheme infrastructure is fully initialized or not */
+static int scheme_initialized = FALSE;
+
 /*
  * Entry point of initlalizing Gauche runtime
  */
 void Scm_Init(const char *signature)
 {
+    if (scheme_initialized) return;
+    
     /* make sure the main program links the same version of libgauche */
     if (strcmp(signature, GAUCHE_SIGNATURE) != 0) {
         Scm_Panic("libgauche ABI version mismatch: libgauche %s, expected %s",
@@ -131,36 +153,54 @@ void Scm_Init(const char *signature)
 
     /* Initialize components.  The order is important, for some components
        rely on the other components to be initialized. */
+    Scm__InitParameter();
     Scm__InitVM();
+    Scm__InitHash();
     Scm__InitSymbol();
     Scm__InitModule();
-    Scm__InitKeyword();
     Scm__InitNumber();
     Scm__InitChar();
     Scm__InitClass();
-    Scm__InitModulePost();
     Scm__InitCollection();
     Scm__InitExceptions();
     Scm__InitProc();
     Scm__InitPort();
     Scm__InitWrite();
-    Scm__InitCode();
     Scm__InitMacro();
     Scm__InitLoad();
     Scm__InitRegexp();
     Scm__InitRead();
     Scm__InitSignal();
     Scm__InitSystem();
-    Scm__InitRepl();
+    Scm__InitComparator();
 
-    Scm_Init_stdlib(Scm_SchemeModule());
-    Scm_Init_extlib(Scm_GaucheModule());
-    Scm_Init_syslib(Scm_GaucheModule());
-    Scm_Init_intlib(Scm_GaucheInternalModule());
-
-    Scm_Init_scmlib();
+    Scm_Init_libalpha();
+    Scm_Init_libbool();
+    Scm_Init_libchar();
+    Scm_Init_libcode();
+    Scm_Init_libcmp();
+    Scm_Init_libdict();
+    Scm_Init_libeval();
+    Scm_Init_libexc();
+    Scm_Init_libfmt();
+    Scm_Init_libio();
+    Scm_Init_liblazy();
+    Scm_Init_liblist();
+    Scm_Init_libmisc();
+    Scm_Init_libmod();
+    Scm_Init_libnum();
+    Scm_Init_libobj();
+    Scm_Init_libproc();
+    Scm_Init_librx();
+    Scm_Init_libsrfis();
+    Scm_Init_libstr();
+    Scm_Init_libsym();
+    Scm_Init_libsys();
+    Scm_Init_libvec();
+    Scm_Init_libmacbase();
     Scm_Init_compile();
-    Scm_Init_objlib();
+    Scm_Init_libmacro();
+    Scm_Init_libomega();
 
     Scm__InitCompaux();
 
@@ -176,6 +216,13 @@ void Scm_Init(const char *signature)
     /* a trick to make sure the gc thread object is linked */
     ptr_pthread_create = (int (*)(void))GC_pthread_create;
 #endif
+
+    scheme_initialized = TRUE;
+}
+
+int Scm_InitializedP()
+{
+    return scheme_initialized;
 }
 
 /*=============================================================
@@ -244,12 +291,14 @@ void Scm_UnregisterFinalizer(ScmObj z)
                                    &ofn, &ocd);
 }
 
-/* GC calls this back when finalizers are queued */
+/* GC calls this back when finalizers are queued. */
 void finalizable(void)
 {
     ScmVM *vm = Scm_VM();
-    vm->finalizerPending = TRUE;
-    vm->attentionRequest = TRUE;
+    if (vm != NULL) {
+        vm->finalizerPending = TRUE;
+        vm->attentionRequest = TRUE;
+    }
 }
 
 /* Called from VM loop.  Queue is not empty. */
@@ -273,7 +322,7 @@ struct cleanup_handler_rec {
 static struct {
     int dirty;                  /* Flag to avoid cleaning up more than once. */
     struct cleanup_handler_rec *handlers;
-} cleanup = { TRUE, NULL }; 
+} cleanup = { TRUE, NULL };
 
 /* Add cleanup handler.  Returns an opaque handle, which can be
    passed to DeleteCleanupHandler. */
@@ -305,14 +354,13 @@ void Scm_DeleteCleanupHandler(void *handle)
     }
 }
 
-/* pthread specific cleanup handler to unlock mutex */
-#ifdef GAUCHE_USE_PTHREADS
+#if defined(GAUCHE_HAS_THREADS)
 void Scm__MutexCleanup(void *mutex_)
 {
     ScmInternalMutex *mutex = mutex_;
     (void)SCM_INTERNAL_MUTEX_UNLOCK(*mutex);
 }
-#endif /* GAUCHE_USE_PTHREADS */
+#endif /* GAUCHE_HAS_THREADS */
 
 
 /* Scm_Cleanup and Scm_Exit
@@ -332,28 +380,26 @@ void Scm_Exit(int code)
 
 void Scm_Cleanup(void)
 {
-    ScmVM *vm = Scm_VM();
-    ScmObj hp;
-    struct cleanup_handler_rec *ch;
-
     if (!cleanup.dirty) return;
     cleanup.dirty = FALSE;
-    
+
     /* Execute pending dynamic handlers.
        NB: we ignore errors here.  It may not be accurate behavior, though.
        (A handler may intentionally raise an error to skip the rest of
        handlers).  We may change to break from SCM_FOR_EACH in case if
        we detect error in Scm_Apply. */
+    ScmVM *vm = Scm_VM();
+    ScmObj hp;
     SCM_FOR_EACH(hp, vm->handlers) {
         vm->handlers = SCM_CDR(hp);
         Scm_Apply(SCM_CDAR(hp), SCM_NIL, NULL);
     }
 
     /* Call the C-registered cleanup handlers. */
-    for (ch = cleanup.handlers; ch; ch = ch->next) {
+    for (struct cleanup_handler_rec *ch = cleanup.handlers; ch; ch = ch->next) {
         ch->handler(ch->data);
     }
-    
+
     /* Flush Scheme ports. */
     Scm_FlushAllPorts(TRUE);
 }
@@ -390,182 +436,181 @@ Scm_GetFeatures()
 }
 
 /*
- * Append FEATURE to the cond-features alist.  Once added, the symbol FEATURE
- * will be recognized by cond-expand.
+ * Enables FEATURE.  Once added, the symbol FEATURE will be recognized
+ * by cond-expand.
+ *
+ *  (cond-expand (FEATURE body ...))
  *
- *  (cond-expand (FEATURE body ...)) 
- *   
  * If loading a module is required in order to make FEATURE available,
  * such module name can be specified in MODULE argument.  It can be NULL
  * if the feature is built-in.
  *
+ * If FEATURE is already in the disabled list, it is not added.
+ *
  * This API is meant to expose configuration information to a Scheme
  * level.
  */
 void
 Scm_AddFeature(const char *feature, const char *module)
 {
-    ScmObj cell;
+    ScmObj f = SCM_INTERN(feature);
+    ScmObj cell = module? SCM_LIST2(f, SCM_INTERN(module)): SCM_LIST1(f);
 
-    if (module) {
-        cell = SCM_LIST2(SCM_INTERN(feature), SCM_INTERN(module));
-    } else {
-        cell = SCM_LIST1(SCM_INTERN(feature));
-    }
-    
     (void)SCM_INTERNAL_MUTEX_LOCK(cond_features.mutex);
-    cond_features.alist = Scm_Cons(cell, cond_features.alist);
+    if (SCM_FALSEP(Scm_Memq(f, cond_features.dlist))) {
+        cond_features.alist = Scm_Cons(cell, cond_features.alist);
+    }
     (void)SCM_INTERNAL_MUTEX_UNLOCK(cond_features.mutex);
 }
 
+/*
+ * Disables FEATURE.  If FEATURE is already enabled, it is removed
+ * from the features alist.
+ * Mainly used by gosh's -F-feature command-line option.
+ */
+void
+Scm_DisableFeature(const char *feature)
+{
+    ScmObj f = SCM_INTERN(feature);
+    (void)SCM_INTERNAL_MUTEX_LOCK(cond_features.mutex);
+    cond_features.dlist = Scm_Cons(f, cond_features.dlist);
+    cond_features.alist = Scm_AssocDelete(f, cond_features.alist, SCM_CMP_EQ);
+    (void)SCM_INTERNAL_MUTEX_UNLOCK(cond_features.mutex);
+}
+
+
 static void
 init_cond_features()
 {
-    int i;
     /* The initial cond-features list. */
     static struct {
         const char *feature;
         const char *module;
     } init_features[] = {
         { "gauche", NULL },
-#ifdef GAUCHE_WINDOWS
+        { "gauche-" GAUCHE_VERSION , NULL },
+
+        /* Platform */
+#if   defined(GAUCHE_WINDOWS)
         { "gauche.os.windows", NULL },
         { "gauche-windows", NULL }, /* for backward compatibility */
-#endif
-#ifdef __CYGWIN__              /* cygwin is different enough to deserve this */
+#elif defined(__CYGWIN__)  /* cygwin is different enough to deserve this */
         { "gauche.os.cygwin", NULL },
 #endif
-#ifdef GAUCHE_USE_PTHREADS
-        { "gauche.sys.pthreads", "gauche.threads" },
+
+        /* R7RS */
+        /* NB: We should probably make checking 'r7rs' trigger loading
+           r7rs module and setting up the environment. */
+        { "r7rs", NULL },
+            
+        /* R7RS Appendix B */
+        { "exact-closed", NULL },
+        // { "exact-complex", NULL }, /* not yet */
+        { "ieee-float", NULL },
+        { "full-unicode", NULL },
+        { "ratios", NULL },
+#if   defined(GAUCHE_WINDOWS)
+        { "windows", NULL },
+#else
+        { "posix", NULL },
 #endif
-        { "srfi-0", NULL },         /* autoloaded */
-        { "srfi-1", "srfi-1" },
-        { "srfi-2", NULL },         /* builtin */
-        { "srfi-4", "gauche.uvector" },
-        { "srfi-5", "srfi-5" },
-        { "srfi-6", NULL },         /* builtin */
-        { "srfi-7", NULL },         /* autoloaded */
-        { "srfi-8", NULL },         /* builtin */
-        { "srfi-9", "gauche.record" },
-        { "srfi-10", NULL },
-        { "srfi-11", "srfi-11" },
-        { "srfi-13", "srfi-13" },
-        { "srfi-14", "srfi-14" },
-        { "srfi-16", NULL },
-        { "srfi-17", NULL },
-        { "srfi-18", "gauche.threads" },
-        { "srfi-19", "srfi-19" },
-        { "srfi-22", NULL },
-        { "srfi-23", NULL },
-        { "srfi-25", "gauche.array" },
-        { "srfi-26", NULL },
-        { "srfi-27", "srfi-27" },
-        { "srfi-28", NULL },
-        { "srfi-29", "srfi-29" },
-        { "srfi-30", NULL },
-        { "srfi-31", NULL },
-        { "srfi-34", NULL },
-        { "srfi-35", NULL },
-        { "srfi-36", NULL },
-        { "srfi-37", "srfi-37" },
-        { "srfi-38", NULL },
-        { "srfi-39", "gauche.parameter" },
-        { "srfi-40", "util.stream" },
-        { "srfi-42", "srfi-42" },
-        { "srfi-43", "srfi-43" },
-        { "srfi-45", NULL },
-        { "srfi-55", NULL },
-        { "srfi-61", NULL },
-        { "srfi-62", NULL },
-        { "srfi-87", NULL },
-        { "srfi-98", "srfi-98" },
-        { "srfi-99", "gauche.record" },
+        /* TODO: OS Flags, CPU arch flags and C memory model flags */
+#if   defined(WORDS_BIGENDIAN)
+        { "big-endian", NULL },
+#else
+        { "little-endian", NULL }, /* NB: r7rs say nothing on mixed endian */
+#endif
+
+        /* Threads */
+#if   defined(GAUCHE_USE_PTHREADS)
+        { "gauche.sys.threads", "gauche.threads" },
+        { "gauche.sys.pthreads", "gauche.threads" },
+#elif defined(GAUCHE_USE_WTHREADS)
+        { "gauche.sys.threads", "gauche.threads" },
+        { "gauche.sys.wthreads", "gauche.threads" },
+#endif
+
+        /* TLS/SSL.  This nees to be in the core in order to switch
+           code _before_ loading rfc.tls */
+#if defined(GAUCHE_USE_AXTLS) || defined(GAUCHE_USE_MBEDTLS)
+        { "gauche.net.tls", "rfc.tls" },
+#endif
+#if defined(GAUCHE_USE_AXTLS)
+        { "gauche.net.tls.axtls", "rfc.tls" },
+#endif
+#if defined(GAUCHE_USE_MBEDTLS)
+        /* NB: Kludge - mbedTLS is implemented in a separate module rfc.tls.mbed,
+           but we don't autoload it with cond-expand gauche.net.tls.mbedtls.
+           If the gauche is compiled with mbedTLS support but the target system
+           lacks mbed DSO, using rfc.tls.mbed causes runtime error.  We don't
+           want that merely by checking feature identifier.  We set autoload
+           in rfc.tls, so the runtime error only occur when the program actually
+           try to use <mbed-tls>. */
+        { "gauche.net.tls.mbedtls", NULL },
+#endif
+
+        /* zlib */
+#if defined(USE_ZLIB)
+        { "gauche.sys.zlib", "rfc.zlib" },
+#endif
+
+        { "regexp-non-greedy", "srfi.115" },
+        { "regexp-look-around", "srfi.115" },
+        { "regexp-backrefs", "srfi.115" },
+        { "regexp-unicode", "srfi.115" },
+
+#include "features.c"
+        
         { NULL, NULL }
     };
 
-    for (i=0; init_features[i].feature; i++) {
+    for (int i=0; init_features[i].feature; i++) {
         Scm_AddFeature(init_features[i].feature, init_features[i].module);
     }
 }
 
-const char *Scm_HostArchitecture(void)
+/* TRANSIENT: These used to be in paths.c but no longer needed.  We keep
+   the entries just for ABI compatibility.  These shouldn't be called.
+   Remove on 1.0 release.
+*/
+const char *Scm_GetLibraryDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
 {
-    return GAUCHE_ARCH;
+    Scm_Abort("Scm_GetLibraryDirectory is called (it shouldn't).");
 }
 
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
-ScmObj Scm_LibraryDirectory(void)
+const char *Scm_GetArchitectureDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
 {
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-        char buf[PATH_MAX];
-        Scm_GetLibraryDirectory(buf, PATH_MAX, Scm_Error);
-        dir = Scm_MakeString(buf, -1, -1,
-                             SCM_STRING_COPYING|SCM_STRING_IMMUTABLE);
-    }
-    return dir;
+    Scm_Abort("Scm_GetArchitectureDirectory is called (it shouldn't).");
 }
 
-ScmObj Scm_ArchitectureDirectory(void)
+const char *Scm_GetSiteLibraryDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
 {
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-        char buf[PATH_MAX];
-        Scm_GetArchitectureDirectory(buf, PATH_MAX, Scm_Error);
-        dir = Scm_MakeString(buf, -1, -1,
-                             SCM_STRING_COPYING|SCM_STRING_IMMUTABLE);
-    }
-    return dir;
+    Scm_Abort("Scm_GetSiteLibraryDirectory is called (it shouldn't).");
 }
 
-ScmObj Scm_SiteLibraryDirectory(void)
+const char *Scm_GetSiteArchitectureDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
 {
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-        char buf[PATH_MAX];
-        Scm_GetSiteLibraryDirectory(buf, PATH_MAX, Scm_Error);
-        dir = Scm_MakeString(buf, -1, -1,
-                             SCM_STRING_COPYING|SCM_STRING_IMMUTABLE);
-    }
-    return dir;
+    Scm_Abort("Scm_GetSiteArchitectureDirectory is called (it shouldn't).");
 }
 
-ScmObj Scm_SiteArchitectureDirectory(void)
+const char *Scm_GetRuntimeDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
 {
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-        char buf[PATH_MAX];
-        Scm_GetSiteArchitectureDirectory(buf, PATH_MAX, Scm_Error);
-        dir = Scm_MakeString(buf, -1, -1,
-                             SCM_STRING_COPYING|SCM_STRING_IMMUTABLE);
-    }
-    return dir;
+    Scm_Abort("Scm_GetRuntimeDirectory is called (it shouldn't).");
 }
 
-/* Temporary hack - this only works on Windows native and Darwin.
-   Returns a pathname that can be replaced for '@' prefix in various
-   configuration paths.  Note that replacement is already done in
-   the returned values of other Scm_*Directory functions above.
-   This is needed to augument the returned values of gauche.config
-   module.  Some refactorization is needed in future.
-   
-   Aside from the current purpose, it will be useful to have a procedure
-   to obtain the directory of the running executable.  This function
-   can eventually evlove to it.
+/*=============================================================
+ * Command line arguments
  */
-ScmObj Scm__RuntimeDirectory(void)
+
+ScmObj Scm_InitCommandLine(int argc, const char *argv[])
 {
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-        char buf[PATH_MAX];
-        Scm_GetRuntimeDirectory(buf, PATH_MAX, Scm_Error);
-        dir = Scm_MakeString(buf, -1, -1,
-                             SCM_STRING_COPYING|SCM_STRING_IMMUTABLE);
-    }
-    return dir;
+    static ScmObj command_line_proc = SCM_UNDEFINED;
+    ScmObj args = Scm_CStringArrayToList(argv, argc, SCM_STRING_IMMUTABLE);
+    SCM_BIND_PROC(command_line_proc, "command-line", Scm_GaucheModule());
+    Scm_ApplyRec1(command_line_proc, args);
+    SCM_DEFINE(Scm_UserModule(), "*program-name*", SCM_CAR(args));
+    SCM_DEFINE(Scm_UserModule(), "*argv*", SCM_CDR(args));
+    return args;
 }
 
 /*=============================================================
@@ -589,24 +634,10 @@ int main(void)
  */
 
 void Scm_SimpleMain(int argc, const char *argv[],
-                    const char *script, u_long flags)
+                    const char *script, u_long flags SCM_UNUSED)
 {
-    ScmModule *user = Scm_UserModule();
-    ScmObj mainproc, args;
-    ScmLoadPacket lpak;
-
     SCM_ASSERT(argc > 0);
-    if (Scm_Load("gauche-init.scm", 0, NULL)) {
-        Scm_Printf(SCM_CURERR, "%s: Couldn't load gauche-init.scm: %A(%A).\n",
-                   argv[0], 
-                   Scm_ConditionMessage(lpak.exception),
-                   Scm_ConditionTypeName(lpak.exception));
-        Scm_Exit(1);
-    }
-
-    args = Scm_CStringArrayToList(argv, argc, SCM_STRING_IMMUTABLE);
-    SCM_DEFINE(user, "*program-name*", SCM_CAR(args));
-    SCM_DEFINE(user, "*argv*", SCM_CDR(args));
+    ScmObj args = Scm_InitCommandLine(argc, argv);
 
     if (script) {
         ScmObj s = SCM_MAKE_STR(script);
@@ -614,13 +645,18 @@ void Scm_SimpleMain(int argc, const char *argv[],
         Scm_LoadFromPort(SCM_PORT(p), SCM_LOAD_PROPAGATE_ERROR, NULL);
     }
 
-    mainproc = Scm_GlobalVariableRef(user, SCM_SYMBOL(SCM_INTERN("main")), 0);
+    ScmModule *user = Scm_UserModule();
+    ScmObj mainproc = Scm_GlobalVariableRef(user, SCM_SYMBOL(SCM_INTERN("main")), 0);
     if (SCM_PROCEDUREP(mainproc)) {
-        ScmObj r = Scm_ApplyRec(mainproc, SCM_LIST1(args));
-        if (SCM_INTP(r)) Scm_Exit(SCM_INT_VALUE(r));
-        else             Scm_Exit(70);
+        static ScmObj run_main_proc = SCM_UNDEFINED;
+        SCM_BIND_PROC(run_main_proc, "run-main", Scm_GaucheInternalModule());
+        SCM_ASSERT(SCM_PROCEDUREP(run_main_proc));
+
+        ScmEvalPacket epak;
+        int r = Scm_Apply(run_main_proc, SCM_LIST2(mainproc, args), &epak);
+        SCM_ASSERT(r == 1 && SCM_INTP(epak.results[0]));
+        Scm_Exit(SCM_INT_VALUE(epak.results[0]));
     } else {
         Scm_Exit(70);
     }
 }
-
