@@ -1,12 +1,12 @@
 /*
  * error.c - error handling
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -36,12 +36,13 @@
 #include "gauche/class.h"
 #include "gauche/exception.h"
 #include "gauche/vm.h"
-#include "gauche/builtin-syms.h"
+#include "gauche/priv/builtin-syms.h"
 
 #include <errno.h>
 #include <string.h>
 #include <ctype.h>
 
+static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs);
 static void   message_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static ScmObj message_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs);
@@ -75,43 +76,96 @@ static ScmClass *condition_cpl[] = {
 };
 
 SCM_DEFINE_BASE_CLASS(Scm_ConditionClass, ScmInstance,
-                      NULL, NULL, NULL,
-                      Scm_ObjectAllocate, SCM_CLASS_DEFAULT_CPL);
+                      NULL, NULL, NULL, 
+                      condition_allocate, SCM_CLASS_DEFAULT_CPL);
 SCM_DEFINE_BASE_CLASS(Scm_MessageConditionClass, ScmMessageCondition,
                       message_print, NULL, NULL,
                       message_allocate, condition_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_SeriousConditionClass, ScmSeriousCondition,
                       NULL, NULL, NULL,
-                      Scm_ObjectAllocate, condition_cpl);
+                      condition_allocate, condition_cpl);
 
-static void message_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
+{
+    return SCM_OBJ(SCM_NEW_INSTANCE(ScmCondition, klass));
+}
+
+static void message_print(ScmObj obj, ScmPort *port, 
+                          ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmClass *k = Scm_ClassOf(obj);
     Scm_Printf(port, "#<%A \"%30.1A\">",
-               Scm__InternalClassName(k),
+               Scm_ShortClassName(k),
                SCM_ERROR_MESSAGE(obj));
 }
 
-static ScmObj message_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj message_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmError *e = SCM_ALLOCATE(ScmError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmError *e = SCM_NEW_INSTANCE(ScmError, klass);
     e->message = SCM_FALSE;     /* would be set by initialize */
     return SCM_OBJ(e);
 }
 
+/* See comment on gauche/exception.h about hack in 'message' slot.
+   TRANSIENT: Remove this hack on 1.0 release. */
 static ScmObj message_get(ScmMessageCondition *obj)
 {
-    return SCM_MESSAGE_CONDITION(obj)->message;
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist)) return SCM_CAR(msglist);
+    else return msglist;
 }
 
 static void message_set(ScmMessageCondition *obj, ScmObj val)
 {
-    obj->message = val;
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist)) SCM_SET_CAR(msglist, val);
+    else SCM_MESSAGE_CONDITION(obj)->message = SCM_LIST2(val, val);
+}
+
+static ScmObj message_prefix_get(ScmMessageCondition *obj)
+{
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist) && SCM_PAIRP(SCM_CDR(msglist))) {
+        return SCM_CADR(msglist);
+    } else {
+        return msglist;
+    }
+}
+
+static void message_prefix_set(ScmMessageCondition *obj, ScmObj val)
+{
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist) && SCM_PAIRP(SCM_CDR(msglist))) {
+        SCM_SET_CAR(SCM_CDR(msglist), val);
+    } else {
+        obj->message = SCM_LIST2(msglist, val);
+    }
+}
+
+static ScmObj message_args_get(ScmMessageCondition *obj)
+{
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist) && SCM_PAIRP(SCM_CDR(msglist))) {
+        return SCM_CDDR(msglist);
+    } else {
+        return SCM_NIL;
+    }
+}
+
+static void message_args_set(ScmMessageCondition *obj, ScmObj val)
+{
+    ScmObj msglist = obj->message;
+    if (SCM_PAIRP(msglist) && SCM_PAIRP(SCM_CDR(msglist))) {
+        SCM_SET_CDR(SCM_CDR(msglist), val);
+    } else {
+        obj->message = Scm_Cons(msglist, Scm_Cons(msglist, val));
+    }
 }
 
 static ScmClassStaticSlotSpec message_slots[] = {
     SCM_CLASS_SLOT_SPEC("message", message_get, message_set),
+    SCM_CLASS_SLOT_SPEC("message-prefix",   message_prefix_get, message_prefix_set),
+    SCM_CLASS_SLOT_SPEC("message-args", message_args_get, message_args_set),
     SCM_CLASS_SLOT_SPEC_END()
 };
 
@@ -130,10 +184,10 @@ static ScmClass *porterror_cpl[] = {
     ERROR_CPL,
     NULL
 };
-    
+
 
 SCM_DEFINE_BASE_CLASS(Scm_ErrorClass, ScmError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       message_allocate, error_cpl+1);
 SCM_DEFINE_BASE_CLASS(Scm_SystemErrorClass, ScmSystemError,
                       message_print, NULL, NULL,
@@ -145,56 +199,53 @@ SCM_DEFINE_BASE_CLASS(Scm_ReadErrorClass, ScmReadError,
                       message_print, NULL, NULL,
                       readerror_allocate, error_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_IOErrorClass, ScmIOError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       message_allocate, error_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_PortErrorClass, ScmPortError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       porterror_allocate, porterror_cpl+1);
 SCM_DEFINE_BASE_CLASS(Scm_IOReadErrorClass, ScmIOReadError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       porterror_allocate, porterror_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_IOWriteErrorClass, ScmIOWriteError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       porterror_allocate, porterror_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_IOClosedErrorClass, ScmIOClosedError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       porterror_allocate, porterror_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_IOUnitErrorClass, ScmIOUnitError,
-                      message_print, NULL, NULL, 
+                      message_print, NULL, NULL,
                       porterror_allocate, porterror_cpl);
 
-static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmSystemError *e = SCM_ALLOCATE(ScmSystemError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmSystemError *e = SCM_NEW_INSTANCE(ScmSystemError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->error_number = 0;           /* set by initialize */
     return SCM_OBJ(e);
 }
 
-static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmUnhandledSignalError *e = SCM_ALLOCATE(ScmUnhandledSignalError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmUnhandledSignalError *e = SCM_NEW_INSTANCE(ScmUnhandledSignalError,
+                                                  klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->signal = 0;                 /* set by initialize */
     return SCM_OBJ(e);
 }
 
-static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmReadError *e = SCM_ALLOCATE(ScmReadError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmReadError *e = SCM_NEW_INSTANCE(ScmReadError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->port = NULL;                /* set by initialize */
     e->line = -1;                  /* set by initialize */
     return SCM_OBJ(e);
 }
 
-static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmPortError *e = SCM_ALLOCATE(ScmPortError, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmPortError *e = SCM_NEW_INSTANCE(ScmPortError, klass);
     e->common.message = SCM_FALSE; /* set by initialize */
     e->port = NULL;                /* set by initialize */
     return SCM_OBJ(e);
@@ -258,12 +309,13 @@ static void readerror_line_set(ScmReadError *obj, ScmObj val)
     obj->line = SCM_INT_VALUE(val);
 }
 
-static ScmObj readerror_dummy_get(ScmReadError *obj)
+static ScmObj readerror_dummy_get(ScmReadError *obj SCM_UNUSED)
 {
     return SCM_FALSE;
 }
 
-static void readerror_dummy_set(ScmReadError *obj, ScmObj val)
+static void readerror_dummy_set(ScmReadError *obj SCM_UNUSED,
+                                ScmObj val SCM_UNUSED)
 {
     /* nothing */
 }
@@ -316,24 +368,36 @@ static ScmClass *compound_cpl[] = {
     NULL
 };
 
+static void compound_print(ScmObj obj, ScmPort *port, 
+                           ScmWriteContext *ctx SCM_UNUSED)
+{
+    ScmClass *k = Scm_ClassOf(obj);
+    Scm_Printf(port, "#<%A", Scm_ShortClassName(k));
+    ScmCompoundCondition *c = SCM_COMPOUND_CONDITION(obj);
+    ScmObj cp;
+    SCM_FOR_EACH(cp, c->conditions) {
+        Scm_Printf(port, " %A", SCM_CAR(cp));
+    }
+    Scm_Printf(port, ">");
+}
+
 SCM_DEFINE_BASE_CLASS(Scm_CompoundConditionClass, ScmCompoundCondition,
-                      NULL, NULL, NULL, 
+                      compound_print, NULL, NULL,
                       compound_allocate, compound_cpl+2);
 SCM_DEFINE_BASE_CLASS(Scm_SeriousCompoundConditionClass, ScmCompoundCondition,
-                      NULL, NULL, NULL, 
+                      compound_print, NULL, NULL,
                       compound_allocate, compound_cpl);
 
-static ScmObj compound_allocate(ScmClass *klass, ScmObj initargs)
+static ScmObj compound_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmCompoundCondition *e = SCM_ALLOCATE(ScmCompoundCondition, klass);
-    SCM_SET_CLASS(e, klass);
+    ScmCompoundCondition *e = SCM_NEW_INSTANCE(ScmCompoundCondition, klass);
     e->conditions = SCM_NIL;
     return SCM_OBJ(e);
 }
 
 ScmObj Scm_MakeCompoundCondition(ScmObj conditions)
 {
-    ScmObj h = SCM_NIL, t = SCM_NIL, cp, cond;
+    ScmObj h = SCM_NIL, t = SCM_NIL;
     int serious = FALSE;
     int nconds = Scm_Length(conditions);
 
@@ -353,6 +417,7 @@ ScmObj Scm_MakeCompoundCondition(ScmObj conditions)
     }
 
     /* collect conditions and creates compound one */
+    ScmObj cp;
     SCM_FOR_EACH(cp, conditions) {
         ScmObj c = SCM_CAR(cp);
         if (!SCM_CONDITIONP(c)) {
@@ -361,7 +426,7 @@ ScmObj Scm_MakeCompoundCondition(ScmObj conditions)
         if (SCM_SERIOUS_CONDITION_P(c)) {
             serious = TRUE;
         }
-        
+
         if (SCM_COMPOUND_CONDITION_P(c)) {
             ScmCompoundCondition *cc = SCM_COMPOUND_CONDITION(c);
             SCM_APPEND(h, t, cc->conditions);
@@ -369,14 +434,34 @@ ScmObj Scm_MakeCompoundCondition(ScmObj conditions)
             SCM_APPEND1(h, t, c);
         }
     }
-    cond = compound_allocate((serious?
-                              SCM_CLASS_COMPOUND_CONDITION :
-                              SCM_CLASS_SERIOUS_COMPOUND_CONDITION),
-                             SCM_NIL);
+    ScmObj cond = compound_allocate((serious?
+                                     SCM_CLASS_COMPOUND_CONDITION :
+                                     SCM_CLASS_SERIOUS_COMPOUND_CONDITION),
+                                    SCM_NIL);
     SCM_COMPOUND_CONDITION(cond)->conditions = h;
     return cond;
 }
 
+/* Extract condition of type TYPE from a compound condition CONDITION.
+   The returned condition may be of subtype of TYPE.
+   CONDITION may be simple condition; in that case, CONDITION itself
+   is returned iff condition is of a subtype of TYPE.
+   This isn't the same as `extract-condition' of SRFI-35, which always
+   returns a condition of the specified type; it also has different
+   rule to refer to the slots.
+ */
+ScmObj Scm_ExtractSimpleCondition(ScmObj condition, ScmClass *type)
+{
+    ScmObj cs = (SCM_COMPOUND_CONDITION_P(condition)
+                 ? SCM_COMPOUND_CONDITION(condition)->conditions
+                 : SCM_LIST1(condition));
+    ScmObj cp;
+    SCM_FOR_EACH(cp, cs) {
+        if (SCM_ISA(SCM_CAR(cp), type)) return SCM_CAR(cp);
+    }
+    return SCM_FALSE;
+}
+
 static ScmObj conditions_get(ScmCompoundCondition *obj)
 {
     return obj->conditions;
@@ -418,7 +503,7 @@ ScmObj Scm_MakeThreadException(ScmClass *klass, ScmVM *thread)
 ScmObj Scm_MakeError(ScmObj message)
 {
     ScmError *e = SCM_ERROR(message_allocate(SCM_CLASS_ERROR, SCM_NIL));
-    e->message = message;
+    e->message = SCM_LIST2(message, message);
     return SCM_OBJ(e);
 }
 
@@ -426,7 +511,7 @@ ScmObj Scm_MakeSystemError(ScmObj message, int en)
 {
     ScmSystemError *e =
         SCM_SYSTEM_ERROR(syserror_allocate(SCM_CLASS_SYSTEM_ERROR, SCM_NIL));
-    e->common.message = message;
+    e->common.message = SCM_LIST2(message, message);
     e->error_number = en;
     return SCM_OBJ(e);
 }
@@ -435,7 +520,7 @@ ScmObj Scm_MakeReadError(ScmObj message, ScmPort *port, int line)
 {
     ScmReadError *e =
         SCM_READ_ERROR(readerror_allocate(SCM_CLASS_READ_ERROR, SCM_NIL));
-    e->common.message = message;
+    e->common.message = SCM_LIST2(message, message);
     e->port = port;
     e->line = line;
     return SCM_OBJ(e);
@@ -443,11 +528,11 @@ ScmObj Scm_MakeReadError(ScmObj message, ScmPort *port, int line)
 
 int Scm_ConditionHasType(ScmObj c, ScmObj k)
 {
-    ScmObj cp;
-    
     if (!SCM_CONDITIONP(c)) return FALSE;
     if (!SCM_CLASSP(k)) return FALSE;
     if (!SCM_COMPOUND_CONDITION_P(c)) return SCM_ISA(c, SCM_CLASS(k));
+
+    ScmObj cp;
     SCM_FOR_EACH(cp, SCM_COMPOUND_CONDITION(c)->conditions) {
         if (SCM_ISA(SCM_CAR(cp), SCM_CLASS(k))) return TRUE;
     }
@@ -457,12 +542,12 @@ int Scm_ConditionHasType(ScmObj c, ScmObj k)
 ScmObj Scm_ConditionMessage(ScmObj c)
 {
     if (SCM_MESSAGE_CONDITION_P(c)) {
-        return SCM_MESSAGE_CONDITION(c)->message;
+        return message_get(SCM_MESSAGE_CONDITION(c));
     } else if (SCM_COMPOUND_CONDITION_P(c)) {
         ScmObj cp;
         SCM_FOR_EACH(cp, SCM_COMPOUND_CONDITION(c)->conditions) {
             if (SCM_MESSAGE_CONDITION_P(SCM_CAR(cp))) {
-                return SCM_MESSAGE_CONDITION(SCM_CAR(cp))->message;
+                return message_get(SCM_MESSAGE_CONDITION(SCM_CAR(cp)));
             }
         }
     }
@@ -473,7 +558,7 @@ ScmObj Scm_ConditionMessage(ScmObj c)
    suitable for the error message.  Because of personal preference
    and backward compatibility, I upcase the class name of the condition
    sans brackets.  If it is a composite condition, the component's typenames
-   are joind with commas.
+   are joind with commas, excluding mixin conditions.
 */
 ScmObj Scm_ConditionTypeName(ScmObj c)
 {
@@ -482,23 +567,32 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
 
     /* just a safety net */
     if (!SCM_CONDITIONP(c)) return SCM_MAKE_STR("(not a condition)");
-    
+
     if (!SCM_COMPOUND_CONDITION_P(c)) {
-        sname = Scm__InternalClassName(Scm_ClassOf(c));
+        sname = Scm_ShortClassName(Scm_ClassOf(c));
     } else {
         ScmObj h = SCM_NIL, t = SCM_NIL, cp;
         SCM_FOR_EACH(cp, SCM_COMPOUND_CONDITION(c)->conditions) {
             ScmObj cc = SCM_CAR(cp);
-            SCM_APPEND1(h, t, Scm__InternalClassName(Scm_ClassOf(cc)));
+            if (SCM_MIXIN_CONDITION_P(cc)) continue;
+            SCM_APPEND1(h, t, Scm_ShortClassName(Scm_ClassOf(cc)));
         }
         if (SCM_NULLP(h)) {
             /* not usual, but tolerate */
-            sname = Scm__InternalClassName(Scm_ClassOf(c));
+            sname = Scm_ShortClassName(Scm_ClassOf(c));
         } else {
             sname = Scm_StringJoin(h, &cond_name_delim, SCM_STRING_JOIN_INFIX);
         }
     }
-    return sname;
+
+    ScmDString ds;
+    Scm_DStringInit(&ds);
+    ScmObj p = Scm_MakeInputStringPort(SCM_STRING(sname), TRUE);
+    int ch;
+    while ((ch = Scm_Getc(SCM_PORT(p))) != EOF) {
+        Scm_DStringPutc(&ds, Scm_CharUpcase(ch));
+    }
+    return Scm_DStringGet(&ds, 0);
 }
 
 /*================================================================
@@ -519,33 +613,64 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
  *   seem to use over 2^31 for the time being).
  */
 
+/* Double fault check
+ * In order to aviod infinite loop when error throwing routine
+ * throws an error, we use vm flag SCM_ERROR_BEING_HANDLED to
+ * check that.  Ideally a common single API should handle it,   
+ * but for the time being, we add the check at the beginning
+ * of Scm_*Error APIs.
+ * The SCM_ERROR_BEING_HANDLED flag is cleared in Scm_VMThrowException().
+ */
+#define SCM_ERROR_DOUBLE_FAULT_CHECK(vm, msg)                           \
+    do {                                                                \
+        if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_HANDLED)) {  \
+            ScmDString ds;                                              \
+            Scm_DStringInit(&ds);                                       \
+            Scm_DStringPutz(&ds, "Error occurred in error handler (", -1); \
+            Scm_DStringPutz(&ds, msg, -1);                              \
+            Scm_DStringPutz(&ds, ")", -1);                              \
+            ScmObj e = Scm_MakeError(Scm_DStringGet(&ds, 0));           \
+            Scm_VMThrowException(vm, e, SCM_RAISE_NON_CONTINUABLE);     \
+        }                                                               \
+        SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);           \
+    } while (0)
+
+/* Common part to format error message passed as varargs.  
+ * ostr must be declared as ScmObj, and msg must be the last
+ * arg of function before '...'.
+ */
+#define SCM_ERROR_MESSAGE_FORMAT(ostr, msg)             \
+    do {                                                \
+        ostr = Scm_MakeOutputStringPort(TRUE);          \
+        va_list args__;                                 \
+        va_start(args__, msg);                          \
+        Scm_Vprintf(SCM_PORT(ostr), msg, args__, TRUE); \
+        va_end(args__);                                 \
+    } while (0)
+
+/* Like SCM_ERROR_MESSAGE_FORMAT, but we also append system error message.
+ * Need to pass errno, which should be fetched by get_errno() *before*
+ * calling Scm_VM().
+ */
+#define SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en)      \
+    do {                                                \
+        SCM_ERROR_MESSAGE_FORMAT(ostr, msg);            \
+        ScmObj syserr = get_syserrmsg(en);              \
+        SCM_PUTZ(": ", -1, ostr);                       \
+        SCM_PUTS(syserr, ostr);                         \
+    } while (0)
+
 /*
  * C-like interface
  */
-
 void Scm_Error(const char *msg, ...)
 {
-    ScmObj e;
     ScmVM *vm = Scm_VM();
-    va_list args;
-
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_HANDLED)) {
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-        Scm_VMThrowException(vm, e);
-    }
-    SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);
-    
-    SCM_UNWIND_PROTECT {
-        va_start(args, msg);
-        e = Scm_MakeError(Scm_Vsprintf(msg, args, TRUE));
-        va_end(args);
-    }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation? */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-    }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm, msg);
+    ScmObj ostr;
+    SCM_ERROR_MESSAGE_FORMAT(ostr, msg);
+    ScmObj e = Scm_MakeError(Scm_GetOutputString(SCM_PORT(ostr), TRUE));
+    Scm_VMThrowException(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -560,17 +685,23 @@ static ScmObj get_syserrmsg(int en)
     syserr = SCM_MAKE_STR_COPYING(strerror(en));
 #else  /*GAUCHE_WINDOWS*/
     if (en < 0) {
-      LPTSTR msgbuf;
+      ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
+      LPTSTR msgbuf = NULL;
       const char *xmsgbuf;
-      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
-                    NULL,
-                    -en,
-                    0,
-                    (LPTSTR)&msgbuf,
-                    0, NULL);
-      xmsgbuf = SCM_WCS2MBS(msgbuf);
-      syserr = SCM_MAKE_STR_COPYING(xmsgbuf);
+      if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
+                        FORMAT_MESSAGE_FROM_SYSTEM |
+                        FORMAT_MESSAGE_IGNORE_INSERTS,
+                        NULL,
+                        -en,
+                        MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT),
+                        (LPTSTR)&msgbuf,
+                        0, NULL)) {
+        xmsgbuf = SCM_WCS2MBS(msgbuf);
+        SCM_PUTZ(xmsgbuf, -1, ostr);
+      }
       LocalFree(msgbuf);
+      Scm_Printf(SCM_PORT(ostr), "(error code = %d)", -en);
+      syserr = Scm_GetOutputString(SCM_PORT(ostr), 0);
     } else {
       syserr = SCM_MAKE_STR_COPYING(strerror(en));
     }
@@ -583,46 +714,35 @@ static int get_errno(void)
 #if !defined(GAUCHE_WINDOWS)
     return errno;
 #else  /*GAUCHE_WINDOWS*/
+    int en;
+    
     if (errno == 0) {
-      return -(int)GetLastError();
+        en = -(int)GetLastError();
     } else {
-      return errno;             /* NB: MSDN says we should use _get_errno,
+        en = errno;             /* NB: MSDN says we should use _get_errno,
                                    but MinGW doesn't seem to have it yet.*/
     }
-#endif /*GAUCHE_WINDOWS*/
-}
 
-void Scm_SysError(const char *msg, ...)
-{
-    ScmObj e;
-    va_list args;
-    ScmVM *vm = Scm_VM();
-    int en = get_errno();
-    ScmObj syserr = get_syserrmsg(en);
-
-#if defined(GAUCHE_WINDOWS)
     /* Reset the error code, so that we can find which is the actual
        error code in the next occasion. */
     errno = 0;                  /* NB: MSDN says we should use _set_errno,
                                    but MinGW doesn't seem to have it yet. */
     SetLastError(0);
+
+    return en;
 #endif /*GAUCHE_WINDOWS*/
-    
-    SCM_UNWIND_PROTECT {
-        ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
-        va_start(args, msg);
-        Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
-        va_end(args);
-        SCM_PUTZ(": ", -1, ostr);
-        SCM_PUTS(syserr, ostr);
-        e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr), 0), en);
-    }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
-    }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+}
+
+void Scm_SysError(const char *msg, ...)
+{
+    int en = get_errno();       /* must take this before Scm_VM() */
+    ScmVM *vm = Scm_VM();
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm, msg);
+
+    ScmObj ostr;
+    SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en);
+    ScmObj e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr), TRUE), en);
+    Scm_VMThrowException(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -633,68 +753,55 @@ void Scm_SysError(const char *msg, ...)
  */
 void Scm_TypeError(const char *what, const char *expected, ScmObj got)
 {
-    Scm_Error("%s expected for %s, but got %S", expected, what, got);
+    Scm_Error("%s is supposed to be of type %s, but got %S",
+              what, expected, got);
 }
 
 
 /*
- * A convenience function to raise port-relates errors.
+ * A convenience function to raise port-related errors.
  * It creates either one of <port-error>, <io-read-error>,
  * <io-write-error>, <io-closed-error>, or <io-unit-error>,
- * depending on the 'reason' argument being 
+ * depending on the 'reason' argument being
  * SCM_PORT_ERROR_{OTHER,INPUT,OUTPUT,CLOSED,UNIT}, respectively.
  * If errno isn't zero, it also creates a <system-error> and throws
  * a compound condition of both.
  */
 void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
 {
-    ScmObj e, smsg, pe;
-    ScmClass *peclass;
+    int en = get_errno();       /* must take this before Scm_VM() */
     ScmVM *vm = Scm_VM();
-    va_list args;
-    int en = get_errno();
+    SCM_ERROR_DOUBLE_FAULT_CHECK(vm, msg);
 
-    SCM_UNWIND_PROTECT {
-        ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
-        va_start(args, msg);
-        Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
-        va_end(args);
-        if (en != 0) {
-            ScmObj syserr = get_syserrmsg(en);
-            SCM_PUTZ(": ", -1, ostr);
-            SCM_PUTS(syserr, ostr);
-        }
-        smsg = Scm_GetOutputString(SCM_PORT(ostr), 0);
+    ScmObj ostr;
+    if (en != 0) SCM_SYSERROR_MESSAGE_FORMAT(ostr, msg, en);
+    else         SCM_ERROR_MESSAGE_FORMAT(ostr, msg);
+    ScmObj smsg = Scm_GetOutputString(SCM_PORT(ostr), TRUE);
 
-        switch (reason) {
-        case SCM_PORT_ERROR_INPUT:
-            peclass = SCM_CLASS_IO_READ_ERROR; break;
-        case SCM_PORT_ERROR_OUTPUT:
-            peclass = SCM_CLASS_IO_WRITE_ERROR; break;
-        case SCM_PORT_ERROR_CLOSED:
-            peclass = SCM_CLASS_IO_CLOSED_ERROR; break;
-        case SCM_PORT_ERROR_UNIT:
-            peclass = SCM_CLASS_IO_UNIT_ERROR; break;
-        default:
-            peclass = SCM_CLASS_PORT_ERROR; break;
-        }
-        pe = porterror_allocate(peclass, SCM_NIL);
-        SCM_ERROR(pe)->message = smsg;
-        SCM_PORT_ERROR(pe)->port = port;
-        
-        if (en != 0) {
-            e = Scm_MakeCompoundCondition(SCM_LIST2(Scm_MakeSystemError(smsg, en),
-                                                    pe));
-        } else {
-            e = pe;
-        }
+    ScmClass *peclass;
+    switch (reason) {
+    case SCM_PORT_ERROR_INPUT:
+        peclass = SCM_CLASS_IO_READ_ERROR; break;
+    case SCM_PORT_ERROR_OUTPUT:
+        peclass = SCM_CLASS_IO_WRITE_ERROR; break;
+    case SCM_PORT_ERROR_CLOSED:
+        peclass = SCM_CLASS_IO_CLOSED_ERROR; break;
+    case SCM_PORT_ERROR_UNIT:
+        peclass = SCM_CLASS_IO_UNIT_ERROR; break;
+    default:
+        peclass = SCM_CLASS_PORT_ERROR; break;
     }
-    SCM_WHEN_ERROR {
-        /* TODO: should check continuation */
-        e = Scm_MakeError(SCM_MAKE_STR("Error occurred in error handler"));
+
+    ScmObj pe = porterror_allocate(peclass, SCM_NIL);
+    SCM_ERROR(pe)->message = SCM_LIST2(smsg, smsg);
+    SCM_PORT_ERROR(pe)->port = port;
+
+    ScmObj e = pe;
+    if (en != 0) {
+        e = Scm_MakeCompoundCondition(SCM_LIST2(Scm_MakeSystemError(smsg, en),
+                                                pe));
     }
-    SCM_END_PROTECT;
-    Scm_VMThrowException(vm, e);
+    Scm_VMThrowException(vm, e, SCM_RAISE_NON_CONTINUABLE);
     Scm_Panic("Scm_Error: Scm_VMThrowException returned.  something wrong.");
 }
 
@@ -705,6 +812,7 @@ void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
 
 void Scm_Warn(const char *msg, ...)
 {
+    if (Scm_GetEnv("GAUCHE_SUPPRESS_WARNING") != NULL) return;
     va_list args;
     va_start(args, msg);
     Scm_Printf(SCM_CURERR, "WARNING: %A\n", Scm_Vsprintf(msg, args, TRUE));
@@ -712,14 +820,10 @@ void Scm_Warn(const char *msg, ...)
     va_end(args);
 }
 
-/* format & warn */
-void Scm_FWarn(ScmString *fmt, ScmObj args)
+/* OBSOLETED: 'warn' is now in Scheme. */
+void Scm_FWarn(ScmString *fmt SCM_UNUSED, ScmObj args SCM_UNUSED)
 {
-    ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
-    Scm_Format(SCM_PORT(ostr), fmt, args, TRUE);
-    Scm_Printf(SCM_CURERR, "WARNING: %A\n",
-               Scm_GetOutputString(SCM_PORT(ostr), 0));
-    Scm_Flush(SCM_CURERR);
+    Scm_Error("Scm_FWarn is obsoleted");
 }
 
 /*
@@ -727,11 +831,12 @@ void Scm_FWarn(ScmString *fmt, ScmObj args)
  */
 
 /* An external API to hide Scm_VMThrowException. */
-ScmObj Scm_Raise(ScmObj condition)
+ScmObj Scm_Raise(ScmObj condition, u_long flags)
 {
-    return Scm_VMThrowException(Scm_VM(), condition);
+    return Scm_VMThrowException(Scm_VM(), condition, flags);
 }
 
+
 /* A convenient API---allows to call user-defined condition easily,
    even the condition type is defined in Scheme.  For example:
 
@@ -751,6 +856,13 @@ ScmObj Scm_Raise(ScmObj condition)
 
    This function isn't very efficient; but sometimes you want the convenience
    more, right?
+
+   The argument list format:
+
+     <condition-type> {<string-slot-name> <value>}* <terminator>
+
+     <terminator> : SCM_RAISE_CONDITION_MESSAGE <fmtstr> <fmtarg> ...
+                  | NULL
 */
 
 ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
@@ -787,168 +899,112 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
 }
 
 /*
- * Show stack trace.
- *   stacklite - return value of Scm_GetStackLite
- *   maxdepth - maximum # of stacks to be shown.
+ * Show stack or call trace.
+ *   stacklite - return value of Scm_VMGetStackLite or Scm_VMGetCallTraceLite
+ *   maxdepth - maximum # of entries to be shown.
  *              0 to use the default.  -1 for unlimited.
  *   skip     - ignore this number of frames.  Useful to call this from
  *              a Scheme error handling routine, in order to skip the
  *              frames of the handler itself.
  *   offset   - add this to the frame number.  Useful to show a middle part
  *              of frames only, by combining the skip parameter.
- *   format   - SCM_STACK_TRACE_FORMAT_* enum value.  EXPERIMENTAL.
+ *   format   - SCM_STACK_TRACE_FORMAT_* enum value.  No longer used.
  */
-
-#define STACK_DEPTH_LIMIT 30
-
-#define FMT_ORIG SCM_STACK_TRACE_FORMAT_ORIGINAL
-#define FMT_CC   SCM_STACK_TRACE_FORMAT_CC
-
-#define SHOW_EXPR(depth, expr) \
-    Scm_Printf(out, "%3d  %66.1S\n", (depth), Scm_UnwrapSyntax(expr));
-
 void Scm_ShowStackTrace(ScmPort *out, ScmObj stacklite,
-                        int maxdepth, int skip, int offset, int format)
+                        int maxdepth, int skip, int offset, 
+                        int format SCM_UNUSED)
 {
-    ScmObj cp;
-    int depth = offset;
-    
-    if (maxdepth == 0) maxdepth = STACK_DEPTH_LIMIT;
-    
-    SCM_FOR_EACH(cp, stacklite) {
-        if (skip-- > 0) continue;
-        if (format == FMT_ORIG) {
-            SHOW_EXPR(depth++, SCM_CAR(cp));
-        }
-        if (SCM_PAIRP(SCM_CAR(cp))) {
-            ScmObj srci = Scm_PairAttrGet(SCM_PAIR(SCM_CAR(cp)),
-                                          SCM_SYM_SOURCE_INFO, SCM_FALSE);
-            if (SCM_PAIRP(srci) && SCM_PAIRP(SCM_CDR(srci))) {
-                switch (format) {
-                case FMT_ORIG:
-                    Scm_Printf(out, "        At line %S of %S\n",
-                               SCM_CADR(srci), SCM_CAR(srci));
-                    break;
-                case FMT_CC:
-                    Scm_Printf(out, "%A:%S:\n",
-                               SCM_CAR(srci), SCM_CADR(srci));
-                    break;
-                }
-            } else {
-                switch (format) {
-                case FMT_ORIG:
-                    Scm_Printf(out, "        [unknown location]\n");
-                    break;
-                case FMT_CC:
-                    Scm_Printf(out, "[unknown location]:\n");
-                    break;
-                }
-            }
-        } else {
-            Scm_Printf(out, "\n");
-        }
-        if (format == FMT_CC) {
-            SHOW_EXPR(depth++, SCM_CAR(cp));
-        }
-
-        if (maxdepth >= 0 && depth >= STACK_DEPTH_LIMIT) {
-            Scm_Printf(out, "... (more stack dump truncated)\n");
-            break;
-        }
-    }
+    static ScmObj show_stack_trace = SCM_UNDEFINED;
+    SCM_BIND_PROC(show_stack_trace,
+                  "%show-stack-trace",
+                  Scm_GaucheInternalModule());
+    Scm_ApplyRec5(show_stack_trace, stacklite, SCM_OBJ(out),
+                  SCM_MAKE_INT(maxdepth), SCM_MAKE_INT(skip),
+                  SCM_MAKE_INT(offset));
 }
 
-#undef SHOW_EXPR
-
+/* Dump stack trace.  Called from the default error reporter.
+   Also intended to be called from the debugger, so we allow vm to be NULL
+   to mean the current VM, and port to be NULL for the current error port. */
+void Scm_DumpStackTrace(ScmVM *vm, ScmPort *port)
+{
+    if (vm == NULL) vm = Scm_VM();
+    if (port == NULL) port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    ScmObj stack = Scm_VMGetStackLite(vm);
+    ScmObj calls = Scm_VMGetCallTraceLite(vm);
+    SCM_PUTZ("Stack Trace:\n", -1, port);
+    SCM_PUTZ("_______________________________________\n", -1, port);
+    Scm_ShowStackTrace(port, stack, 0, 0, 0, 0);
+    if (SCM_PAIRP(calls)) {
+        SCM_PUTZ("Call Trace:\n", -1, port);
+        SCM_PUTZ("_______________________________________\n", -1, port);
+        Scm_ShowStackTrace(port, calls, 0, 0, 0, 0);
+    }
+    SCM_FLUSH(port);
+}
 
 /*
  * Default error reporter
  */
 
 /* The default procedure to display the header of error message.
-   E is a thrown condition, not necessarily an error object. */
+   E is a thrown condition, not necessarily an error object.
+
+   The actual operation is written in Scheme (libexc.scm).  However,
+   we can't use that before the infrastructure is fully booted; so
+   this routine has a fallback which will be used only during initialization.
+*/
 static void Scm_PrintDefaultErrorHeading(ScmObj e, ScmPort *out)
 {
-    ScmObj msg;
-    char *heading, *p;
-
-    if (SCM_CONDITIONP(e)) {
-        heading = Scm_GetString(SCM_STRING(Scm_ConditionTypeName(e)));
-        /* TODO: considring that the class name may contain multibyte
-           characters, we should use string-upcase here. */
-        for (p=heading; *p; p++) {
-            *p = toupper(*p);
-        }
-        msg = Scm_ConditionMessage(e);
-        if (!SCM_FALSEP(msg)) {
-            Scm_Printf(out, "*** %s: %A\n", heading, msg);
-        } else {
-            Scm_Printf(out, "*** %s\n", heading);
-        }
+    if (Scm_InitializedP()) {
+        static ScmObj print_default_error_heading = SCM_UNDEFINED;
+        SCM_BIND_PROC(print_default_error_heading,
+                      "print-default-error-heading",
+                      Scm_GaucheModule());
+        Scm_ApplyRec2(print_default_error_heading, e, SCM_OBJ(out));
     } else {
-        Scm_Printf(out, "*** ERROR: unhandled exception: %S\n", e);
+        /* Error during initialization. */
+        if (SCM_CONDITIONP(e)) {
+            Scm_Printf(out, "*** %A: %A\n",
+                       Scm_ConditionTypeName(e),
+                       Scm_ConditionMessage(e));
+        } else {
+            Scm_Printf(out, "*** ERROR: Unhandled condition: %S\n", e);
+        }
     }
 }
 
-static void report_error_inner(ScmVM *vm, ScmObj e)
-{
-    ScmObj stack = Scm_VMGetStackLite(vm);
-    ScmPort *err = SCM_VM_CURRENT_ERROR_PORT(vm);
-
-    Scm_PrintDefaultErrorHeading(e, err);
-    SCM_PUTZ("Stack Trace:\n", -1, err);
-    SCM_PUTZ("_______________________________________\n", -1, err);
-    Scm_ShowStackTrace(err, stack, 0, 0, 0, FMT_ORIG);
-    /* NB: stderr is autoflushed by default, but in case err is replaced
-       by some other port, we explicitly flush it. */
-    SCM_FLUSH(err);
-}
-
-void Scm_ReportError(ScmObj e)
+/* We treat out == #f or #t just like 'format' - #t for the current output
+   port, and #f for string port.  If it's output port, use it.  For any
+   other objects, we use current error port.  This permissive behavior is
+   intentional - report-error is usually called during error handling,
+   and raising an error there masks the original error.  However, the
+   caller should explicitly pass SCM_OBJ(SCM_CURERR) if that's what it intends,
+   instead of relying on the permissive behavior.
+*/
+ScmObj Scm_ReportError(ScmObj e, ScmObj out)
 {
     ScmVM *vm = Scm_VM();
-
-    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_ERROR_BEING_REPORTED)) {
-        /* An _uncaptured_ error occurred during reporting an error.
-           We can't proceed, for it will cause infinite loop.
-           Note that it is OK for an error to occur inside the error
-           reporter, as far as the error is handled by user-installed
-           handler.   The user-installed handler can even invoke a
-           continuation that is captured outside; the flag is reset
-           in such case. 
-           Be careful that it is possible that stderr is no longer
-           available here (since it may be the very cause of the
-           recursive error).  All we can do is to abort. */
-        Scm_Abort("Unhandled error occurred during reporting an error.  Process aborted.\n");
+    ScmPort *port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    if (SCM_FALSEP(out)) {
+        port = SCM_PORT(Scm_MakeOutputStringPort(TRUE));
+    } else if (SCM_TRUEP(out)) {
+        port = SCM_VM_CURRENT_OUTPUT_PORT(vm);
+    } else if (SCM_OPORTP(out)) {
+        port = SCM_PORT(out);
     }
-
-    SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_REPORTED);
-    SCM_UNWIND_PROTECT {
-        if (SCM_PROCEDUREP(vm->defaultEscapeHandler)) {
-            Scm_ApplyRec(vm->defaultEscapeHandler, SCM_LIST1(e));
-        } else {
-            report_error_inner(vm, e);
-        }
-    }
-    SCM_WHEN_ERROR {
-        /* NB: this is called when a continuation captured outside is
-           invoked inside the error reporter.   It may be invoked by
-           the user's error handler.  */
-        SCM_VM_RUNTIME_FLAG_CLEAR(vm, SCM_ERROR_BEING_REPORTED);
-    }
-    SCM_END_PROTECT;
-    SCM_VM_RUNTIME_FLAG_CLEAR(vm, SCM_ERROR_BEING_REPORTED);
+    Scm_PrintDefaultErrorHeading(e, port);
+    Scm_DumpStackTrace(vm, port);
+    if (SCM_FALSEP(out)) return Scm_GetOutputString(SCM_PORT(port), 0);
+    else return SCM_UNDEFINED;
 }
 
 /*
  * Initialization
  */
-extern void Scm_Init_exclib(ScmModule *module);
-
 void Scm__InitExceptions(void)
 {
     ScmModule *mod = Scm_GaucheModule();
-    ScmClass *cond_meta;
 
     ScmObj mes_ser_supers
         = SCM_LIST2(SCM_OBJ(SCM_CLASS_MESSAGE_CONDITION),
@@ -960,7 +1016,7 @@ void Scm__InitExceptions(void)
     Scm_InitStaticClassWithMeta(SCM_CLASS_CONDITION,
                                 "<condition>",
                                 mod, NULL, SCM_FALSE, NULL, 0);
-    cond_meta = Scm_ClassOf(SCM_OBJ(SCM_CLASS_CONDITION));
+    ScmClass *cond_meta = Scm_ClassOf(SCM_OBJ(SCM_CLASS_CONDITION));
     Scm_InitStaticClassWithMeta(SCM_CLASS_SERIOUS_CONDITION,
                                 "<serious-condition>",
                                 mod, cond_meta, SCM_FALSE, NULL, 0);
@@ -1009,6 +1065,7 @@ void Scm__InitExceptions(void)
                                 mod, cond_meta, SCM_FALSE,
                                 porterror_slots, 0);
 
+
     Scm_InitStaticClassWithMeta(SCM_CLASS_COMPOUND_CONDITION,
                                 "<compound-condition>",
                                 mod, cond_meta, SCM_FALSE,
@@ -1017,7 +1074,4 @@ void Scm__InitExceptions(void)
                                 "<serious-compound-condition>",
                                 mod, cond_meta, com_ser_supers,
                                 compound_slots, 0);
-
-    Scm_Init_exclib(mod);
 }
-
