@@ -1,12 +1,12 @@
 /*
  * gloc.c - gloc implementation
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -38,15 +38,19 @@
  * GLOCs
  */
 
-static void gloc_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static void gloc_print(ScmObj obj, ScmPort *port,
+                       ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmGloc *g = SCM_GLOC(obj);
-    Scm_Printf(port, "#<gloc %S%s%S%s>", g->module->name,
-               (g->exported?"#":"##"),
+    Scm_Printf(port, "#<gloc %S#%S%s>", g->module->name,
                g->name,
                (Scm_GlocConstP(g)
                 ? " const"
-                : (Scm_GlocInlinableP(g) ? " inlinable" : "")));
+                : (Scm_GlocInlinableP(g)
+                   ? " inlinable"
+                   : (SCM_GLOC_PHANTOM_BINDING_P(g)
+                      ? " phantom"
+                      : ""))));
 }
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_GlocClass, gloc_print);
@@ -58,7 +62,6 @@ ScmObj Scm_MakeGloc(ScmSymbol *sym, ScmModule *module)
     g->name = sym;
     g->module = module;
     g->value = SCM_UNBOUND;
-    g->exported = FALSE;
     g->hidden = FALSE;
     g->getter = NULL;
     g->setter = NULL;
@@ -66,7 +69,7 @@ ScmObj Scm_MakeGloc(ScmSymbol *sym, ScmModule *module)
 }
 
 /* special setters for const and inlinable bindings. */
-ScmObj Scm_GlocConstSetter(ScmGloc *gloc, ScmObj val)
+ScmObj Scm_GlocConstSetter(ScmGloc *gloc, ScmObj val SCM_UNUSED)
 {
     Scm_Error("cannot change constant value of %S#%S",
               gloc->module->name, gloc->name);
