@@ -1,12 +1,12 @@
 /*
  * list.c - List related functions
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -81,13 +81,11 @@ ScmObj Scm_Acons(ScmObj caar, ScmObj cdar, ScmObj cdr)
 
 ScmObj Scm_List(ScmObj elt, ...)
 {
-    va_list pvar;
-    ScmObj cdr;
-
     if (elt == NULL) return SCM_NIL;
-        
+
+    va_list pvar;
     va_start(pvar, elt);
-    cdr = Scm_VaList(pvar);
+    ScmObj cdr = Scm_VaList(pvar);
     va_end(pvar);
     return Scm_Cons(elt, cdr);
 }
@@ -95,13 +93,11 @@ ScmObj Scm_List(ScmObj elt, ...)
 
 ScmObj Scm_Conses(ScmObj elt, ...)
 {
-    va_list pvar;
-    ScmObj cdr;
-
     if (elt == NULL) return SCM_NIL;
-    
+
+    va_list pvar;
     va_start(pvar, elt);
-    cdr = Scm_VaCons(pvar);
+    ScmObj cdr = Scm_VaCons(pvar);
     va_end(pvar);
     if (cdr == NULL) return elt;
     else             return Scm_Cons(elt, cdr);
@@ -111,12 +107,12 @@ ScmObj Scm_Conses(ScmObj elt, ...)
 ScmObj Scm_VaList(va_list pvar)
 {
     ScmObj start = SCM_NIL, cp = SCM_NIL, obj;
-    
+
     for (obj = va_arg(pvar, ScmObj);
-	 obj != NULL;
-	 obj = va_arg(pvar, ScmObj))
+         obj != NULL;
+         obj = va_arg(pvar, ScmObj))
     {
-	if (SCM_NULLP(start)) {
+        if (SCM_NULLP(start)) {
             start = SCM_OBJ(SCM_NEW(ScmPair));
             SCM_SET_CAR(start, obj);
             SCM_SET_CDR(start, SCM_NIL);
@@ -128,39 +124,39 @@ ScmObj Scm_VaList(va_list pvar)
             SCM_SET_CAR(item, obj);
             SCM_SET_CDR(item, SCM_NIL);
             cp = item;
-	}
+        }
     }
     return start;
 }
 
 
-ScmObj Scm_VaCons(va_list pvar)
+ScmObj Scm_VaCons(va_list pvar SCM_UNUSED)
 {
     Scm_Panic("Scm_VaCons: not implemented");
     return SCM_UNDEFINED;
 }
 
-ScmObj Scm_ArrayToList(ScmObj *elts, int nelts)
+ScmObj Scm_ArrayToList(ScmObj *elts, ScmSize nelts)
 {
     return Scm_ArrayToListWithTail(elts, nelts, SCM_NIL);
 }
 
-ScmObj Scm_ArrayToListWithTail(ScmObj *elts, int nelts, ScmObj tail)
+ScmObj Scm_ArrayToListWithTail(ScmObj *elts, ScmSize nelts, ScmObj tail)
 {
     ScmObj h = SCM_NIL, t = SCM_NIL;
     if (elts) {
-        int i;
-        for (i=0; i<nelts; i++) SCM_APPEND1(h, t, *elts++);
+        for (ScmSize i=0; i<nelts; i++) SCM_APPEND1(h, t, *elts++);
     }
     if (!SCM_NULLP(tail)) SCM_APPEND(h, t, tail);
     return h;
 }
 
-ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
+ScmObj *Scm_ListToArray(ScmObj list, ScmSize *nelts, ScmObj *store, int alloc)
 {
-    ScmObj *array, lp;
-    int len = Scm_Length(list), i;
+    ScmSize len = Scm_Length(list);
     if (len < 0) Scm_Error("proper list required, but got %S", list);
+
+    ScmObj *array;
     if (store == NULL) {
         array = SCM_NEW_ARRAY(ScmObj, len);
     } else {
@@ -172,7 +168,8 @@ ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
             array = store;
         }
     }
-    for (i=0, lp=list; i<len; i++, lp=SCM_CDR(lp)) {
+    ScmSize i = 0;
+    for (ScmObj lp=list; i<len; i++, lp=SCM_CDR(lp)) {
         array[i] = SCM_CAR(lp);
     }
     *nelts = len;
@@ -181,7 +178,7 @@ ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
 
 /* cXr stuff */
 
-#define	CXR(cname, sname, body)                 \
+#define CXR(cname, sname, body)                 \
 ScmObj cname (ScmObj obj)                       \
 {                                               \
    ScmObj obj2 = obj;                           \
@@ -189,11 +186,11 @@ ScmObj cname (ScmObj obj)                       \
    return obj2;                                 \
 }
 
-#define	A                                                       \
+#define A                                                       \
    if (!SCM_PAIRP(obj2)) Scm_Error("bad object: %S", obj);      \
    obj2 = SCM_CAR(obj2);
 
-#define	D                                                       \
+#define D                                                       \
    if (!SCM_PAIRP(obj2)) Scm_Error("bad object: %S", obj);      \
    obj2 = SCM_CDR(obj2);
 
@@ -213,24 +210,24 @@ CXR(Scm_Cddr, "cddr", D D)
    If the argument is a dotted list, return -1.
    If the argument is a circular list, return -2. */
 
-int Scm_Length(ScmObj obj)
+ScmSize Scm_Length(ScmObj obj)
 {
     ScmObj slow = obj;
-    int len = 0;
+    ScmSize len = 0;
 
     for (;;) {
         if (SCM_NULLP(obj)) break;
         if (!SCM_PAIRP(obj)) return SCM_LIST_DOTTED;
-	
-	obj = SCM_CDR(obj);
-	len++;
+
+        obj = SCM_CDR(obj);
+        len++;
         if (SCM_NULLP(obj)) break;
         if (!SCM_PAIRP(obj)) return SCM_LIST_DOTTED;
 
-	obj = SCM_CDR(obj);
-	slow = SCM_CDR(slow);
-	if (obj == slow) return SCM_LIST_CIRCULAR;
-	len++;
+        obj = SCM_CDR(obj);
+        slow = SCM_CDR(slow);
+        if (obj == slow) return SCM_LIST_CIRCULAR;
+        len++;
     }
     return len;
 }
@@ -242,10 +239,9 @@ int Scm_Length(ScmObj obj)
 
 ScmObj Scm_CopyList(ScmObj list)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
-
     if (!SCM_PAIRP(list)) return list;
-    
+
+    ScmObj start = SCM_NIL, last = SCM_NIL;
     SCM_FOR_EACH(list, list) {
         SCM_APPEND1(start, last, SCM_CAR(list));
     }
@@ -258,19 +254,19 @@ ScmObj Scm_CopyList(ScmObj list)
  *    Note that <len> is C-integer.
  */
 
-ScmObj Scm_MakeList(int len, ScmObj fill)
+ScmObj Scm_MakeList(ScmSmallInt len, ScmObj fill)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
     if (len < 0) {
-        Scm_Error("make-list: negative length given: %d", len);
+        Scm_Error("make-list: negative length given: %ld", len);
     }
+    ScmObj start = SCM_NIL, last = SCM_NIL;
     while (len--) {
         SCM_APPEND1(start, last, fill);
     }
     return start;
 }
-     
-    
+
+
 /* Scm_Append2X(list, obj)
  *    Replace cdr of last pair of LIST for OBJ.
  *    If LIST is not a pair, return OBJ.
@@ -280,7 +276,7 @@ ScmObj Scm_Append2X(ScmObj list, ScmObj obj)
 {
     ScmObj cp;
     SCM_FOR_EACH(cp, list) {
-	if (SCM_NULLP(SCM_CDR(cp))) {
+        if (SCM_NULLP(SCM_CDR(cp))) {
             SCM_SET_CDR(cp, obj);
             return list;
         }
@@ -295,12 +291,14 @@ ScmObj Scm_Append2X(ScmObj list, ScmObj obj)
 
 ScmObj Scm_Append2(ScmObj list, ScmObj obj)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
+    if (SCM_NULLP(list)) return obj;
 
-    if (!SCM_PAIRP(list)) return obj;
-
-    SCM_FOR_EACH(list, list) {
-	SCM_APPEND1(start, last, SCM_CAR(list));
+    ScmObj cp, start = SCM_NIL, last = SCM_NIL;
+    SCM_FOR_EACH(cp, list) {
+        SCM_APPEND1(start, last, SCM_CAR(cp));
+    }
+    if (!SCM_NULLP(cp)) {
+        Scm_Error("proper list required, but got %S", list);
     }
     SCM_SET_CDR(last, obj);
 
@@ -321,6 +319,9 @@ ScmObj Scm_Append(ScmObj args)
             Scm_Error("pair required, but got %S", SCM_CAR(cp));
         } else {
             SCM_APPEND(start, last, Scm_CopyList(SCM_CAR(cp)));
+            if (SCM_PAIRP(last) && !SCM_NULLP(SCM_CDR(last))) {
+                Scm_Error("proper list required, but got %S", SCM_CAR(cp));
+            }
         }
     }
     return start;
@@ -336,17 +337,15 @@ ScmObj Scm_Append(ScmObj args)
 
 ScmObj Scm_Reverse2(ScmObj list, ScmObj tail)
 {
-    ScmObj cp, result;
-    ScmPair *p;
-
     if (!SCM_PAIRP(list)) return tail;
 
-    p = SCM_NEW(ScmPair);
+    ScmPair *p = SCM_NEW(ScmPair);
     SCM_SET_CAR(p, SCM_NIL);
     SCM_SET_CDR(p, tail);
-    result = SCM_OBJ(p);
+    ScmObj result = SCM_OBJ(p);
+    ScmObj cp;
     SCM_FOR_EACH(cp, list) {
-	SCM_SET_CAR(result, SCM_CAR(cp));
+        SCM_SET_CAR(result, SCM_CAR(cp));
         p = SCM_NEW(ScmPair);
         SCM_SET_CAR(p, SCM_NIL);
         SCM_SET_CDR(p, result);
@@ -360,7 +359,7 @@ ScmObj Scm_Reverse(ScmObj list)
     return Scm_Reverse2(list, SCM_NIL);
 }
 
-    
+
 /* Scm_Reverse2X(list, tail)
  *   Return reversed list of LIST.  Pairs in previous LIST is used to
  *   create new list.  TAIL is appended to the result.
@@ -370,8 +369,8 @@ ScmObj Scm_Reverse(ScmObj list)
 
 ScmObj Scm_Reverse2X(ScmObj list, ScmObj tail)
 {
+    if (!SCM_PAIRP(list)) return tail;
     ScmObj first, next, result = tail;
-    if (!SCM_PAIRP(list)) return result;
     for (first = list; SCM_PAIRP(first); first = next) {
         next = SCM_CDR(first);
         SCM_SET_CDR(first, result);
@@ -390,25 +389,24 @@ ScmObj Scm_ReverseX(ScmObj list)
  *    Note that i is C-INTEGER.  If i is out of bound, signal error.
  */
 
-ScmObj Scm_ListTail(ScmObj list, int i, ScmObj fallback)
+ScmObj Scm_ListTail(ScmObj list, ScmSmallInt i, ScmObj fallback)
 {
-    int cnt = i;
     if (i < 0) goto err;
+    ScmSmallInt cnt = i;
     while (cnt-- > 0) {
         if (!SCM_PAIRP(list)) goto err;
         list = SCM_CDR(list);
     }
     return list;
-  err:  
-    if (SCM_UNBOUNDP(fallback)) Scm_Error("argument out of range: %d", i);
+  err:
+    if (SCM_UNBOUNDP(fallback)) Scm_Error("argument out of range: %ld", i);
     return fallback;
 }
 
-ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
+ScmObj Scm_ListRef(ScmObj list, ScmSmallInt i, ScmObj fallback)
 {
-    int k;
     if (i < 0) goto err;
-    for (k=0; k<i; k++) {
+    for (ScmSmallInt k=0; k<i; k++) {
         if (!SCM_PAIRP(list)) goto err;
         list = SCM_CDR(list);
     }
@@ -416,7 +414,7 @@ ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
     return SCM_CAR(list);
   err:
     if (SCM_UNBOUNDP(fallback)) {
-        Scm_Error("argument out of range: %d", i);
+        Scm_Error("argument out of range: %ld", i);
     }
     return fallback;
 }
@@ -428,12 +426,12 @@ ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
 
 ScmObj Scm_LastPair(ScmObj l)
 {
-    ScmObj cp;
-
     if (!SCM_PAIRP(l)) Scm_Error("pair required: %S", l);
+
+    ScmObj cp;
     SCM_FOR_EACH(cp, l) {
-	ScmObj cdr = SCM_CDR(cp);
-	if (!SCM_PAIRP(cdr)) return cp;
+        ScmObj cdr = SCM_CDR(cp);
+        if (!SCM_PAIRP(cdr)) return cp;
     }
     return SCM_UNDEFINED;       /* NOTREACHED */
 }
@@ -470,9 +468,9 @@ ScmObj Scm_Member(ScmObj obj, ScmObj list, int cmpmode)
 /* delete. */
 ScmObj Scm_Delete(ScmObj obj, ScmObj list, int cmpmode)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL, cp, prev = list;
-
     if (SCM_NULLP(list)) return SCM_NIL;
+
+    ScmObj start = SCM_NIL, last = SCM_NIL, cp, prev = list;
     SCM_FOR_EACH(cp, list) {
         if (Scm_EqualM(obj, SCM_CAR(cp), cmpmode)) {
             for (; prev != cp; prev = SCM_CDR(prev))
@@ -512,8 +510,8 @@ ScmObj Scm_DeleteX(ScmObj obj, ScmObj list, int cmpmode)
 
 ScmObj Scm_Assq(ScmObj obj, ScmObj alist)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assq: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
         ScmObj entry = SCM_CAR(cp);
         if (!SCM_PAIRP(entry)) continue;
@@ -524,20 +522,20 @@ ScmObj Scm_Assq(ScmObj obj, ScmObj alist)
 
 ScmObj Scm_Assv(ScmObj obj, ScmObj alist)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assv: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
-	ScmObj entry = SCM_CAR(cp);
-	if (!SCM_PAIRP(entry)) continue;
-	if (Scm_EqvP(obj, SCM_CAR(entry))) return entry;
+        ScmObj entry = SCM_CAR(cp);
+        if (!SCM_PAIRP(entry)) continue;
+        if (Scm_EqvP(obj, SCM_CAR(entry))) return entry;
     }
     return SCM_FALSE;
 }
 
 ScmObj Scm_Assoc(ScmObj obj, ScmObj alist, int cmpmode)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assoc: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
         ScmObj entry = SCM_CAR(cp);
         if (!SCM_PAIRP(entry)) continue;
@@ -549,12 +547,12 @@ ScmObj Scm_Assoc(ScmObj obj, ScmObj alist, int cmpmode)
 /* Assoc-delete */
 ScmObj Scm_AssocDelete(ScmObj elt, ScmObj alist, int cmpmode)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL, cp, p, prev = alist;
     if (!SCM_LISTP(alist)) {
         Scm_Error("assoc-delete: list required, but got %S", alist);
     }
     if (SCM_NULLP(alist)) return SCM_NIL;
-    
+
+    ScmObj start = SCM_NIL, last = SCM_NIL, cp, p, prev = alist;
     SCM_FOR_EACH(cp, alist) {
         p = SCM_CAR(cp);
         if (SCM_PAIRP(p)) {
@@ -573,10 +571,10 @@ ScmObj Scm_AssocDelete(ScmObj elt, ScmObj alist, int cmpmode)
 
 ScmObj Scm_AssocDeleteX(ScmObj elt, ScmObj alist, int cmpmode)
 {
-    ScmObj cp, prev = SCM_NIL;
     if (!SCM_LISTP(alist)) {
         Scm_Error("assoc-delete!: list required, but got %S", alist);
     }
+    ScmObj cp, prev = SCM_NIL;
     SCM_FOR_EACH(cp, alist) {
         ScmObj e = SCM_CAR(cp);
         if (SCM_PAIRP(e)) {
@@ -627,7 +625,7 @@ ScmObj Scm_DeleteDuplicatesX(ScmObj list, int cmpmode)
  *  list.   If there's more than one way to order an element, choose the
  *  first one appears in the given list of lists.
  *  Returns SCM_FALSE if the lists are inconsistent to be ordered
- *  in the way. 
+ *  in the way.
  *
  *  START is an item of the starting point.  It is inserted into the result
  *  first.  SEQUENCES is a list of lists describing the order of preference.
@@ -638,7 +636,7 @@ ScmObj Scm_DeleteDuplicatesX(ScmObj list, int cmpmode)
  *  Since the algorithm is generally useful, I implement the core routine
  *  of the algorithm here.
  *
- *  TODO at 1.0: I noticed START argument acutally isn't used in the
+ *  TODO at 1.0: I noticed START argument actually isn't used in the
  *  algorithm at all.  We can drop it and the caller can just say
  *  Scm_Cons(start, Scm_MonotonicMerge(sequences)).   We can't change it
  *  now because of ABI compatibility, but it will be nice to do so when
@@ -656,28 +654,30 @@ ScmObj Scm_MonotonicMerge(ScmObj start, ScmObj sequences)
 /* WILL RENAME IN 1.0 */
 ScmObj Scm_MonotonicMerge1(ScmObj sequences)
 {
-    ScmObj result = SCM_NIL, next, h;
-    ScmObj *seqv, *sp, *tp;
+    ScmObj result = SCM_NIL;
     int nseqs = Scm_Length(sequences);
-
     if (nseqs < 0) Scm_Error("bad list of sequences: %S", sequences);
-    seqv = SCM_NEW_ARRAY(ScmObj, nseqs);
-    for (sp=seqv; SCM_PAIRP(sequences); sp++, sequences=SCM_CDR(sequences)) {
+    ScmObj *seqv = SCM_NEW_ARRAY(ScmObj, nseqs);
+    for (ScmObj *sp=seqv;
+         SCM_PAIRP(sequences);
+         sp++, sequences=SCM_CDR(sequences)) {
         *sp = SCM_CAR(sequences);
     }
 
     for (;;) {
         /* have we consumed all the inputs? */
+        ScmObj *sp;
         for (sp=seqv; sp<seqv+nseqs; sp++) {
             if (!SCM_NULLP(*sp)) break;
         }
         if (sp == seqv+nseqs) return Scm_ReverseX(result);
 
         /* select candidate */
-        next = SCM_FALSE;
+        ScmObj next = SCM_FALSE;
         for (sp = seqv; sp < seqv+nseqs; sp++) {
             if (!SCM_PAIRP(*sp)) continue;
-            h = SCM_CAR(*sp);
+            ScmObj h = SCM_CAR(*sp);
+            ScmObj *tp;
             for (tp = seqv; tp < seqv+nseqs; tp++) {
                 if (!SCM_PAIRP(*tp)) continue;
                 if (!SCM_FALSEP(Scm_Memq(h, SCM_CDR(*tp)))) {
@@ -715,23 +715,29 @@ ScmObj Scm_PairAttr(ScmPair *pair)
     }
 }
 
-ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr)
+ScmObj Scm_MakeExtendedPair(ScmObj car, ScmObj cdr, ScmObj attrs)
 {
     ScmExtendedPair *xp = SCM_NEW(ScmExtendedPair);
     xp->car = car;
     xp->cdr = cdr;
-    xp->attributes = SCM_NIL;
+    xp->attributes = attrs;
     return SCM_OBJ(xp);
 }
 
+/* The common scenario is to use ExtendedCons in place of Cons,
+   then add attributes later.  So we provide this.  */
+ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr)
+{
+    return Scm_MakeExtendedPair(car, cdr, SCM_NIL);
+}
+
 ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback)
 {
-    ScmObj p;
     if (!SCM_EXTENDED_PAIR_P(pair)) {
         goto fallback;
     }
-    
-    p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
+
+    ScmObj p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
     if (SCM_PAIRP(p)) return SCM_CDR(p);
   fallback:
     if (fallback == SCM_UNBOUND)
@@ -742,13 +748,12 @@ ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback)
 
 ScmObj Scm_PairAttrSet(ScmPair *pair, ScmObj key, ScmObj value)
 {
-    ScmObj p;
     if (!SCM_EXTENDED_PAIR_P(pair)) {
         Scm_Error("Cannot set pair attribute (%S) to non-extended pair: %S",
                   key, SCM_OBJ(pair));
     }
-    
-    p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
+
+    ScmObj p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
     if (SCM_PAIRP(p)) SCM_SET_CDR(p, value);
     else SCM_EXTENDED_PAIR(pair)->attributes
         = Scm_Acons(key, value, SCM_EXTENDED_PAIR(pair)->attributes);
