@@ -1,12 +1,12 @@
 /*
  * signal.c - signal handling
  *
- *   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -57,7 +57,7 @@
  *  the system to handle the signal unless the Scheme program installs
  *  the handler.   Such signals are the ones that can't be caught, or
  *  are ignored by default.  SIGPWR and SIGXCPU are also left to the system
- *  since GC uses it in the Linux/pthread environment.  
+ *  since GC uses it in the Linux/pthread environment.
  *
  *  About the signal behavior on windows, see "Note on windows port" below.
  */
@@ -75,21 +75,21 @@ static int sigprocmask_win(int how, const sigset_t *set, sigset_t *oldset);
 
 /* Master signal handler vector. */
 static struct sigHandlersRec {
-    ScmObj handlers[NSIG];      /* Scheme signal handlers.  This is #<undef> on
+    ScmObj handlers[SCM_NSIG];  /* Scheme signal handlers.  This is #<undef> on
                                    signals to which Gauche does not install
                                    C-level signal handler (sig_handle). */
-    ScmSysSigset *masks[NSIG];  /* Signal masks during executing Scheme
+    ScmSysSigset *masks[SCM_NSIG];/* Signal masks during executing Scheme
                                    handlers.  Can be NULL, which means
                                    the handling signal(s) are blocked. */
     sigset_t masterSigset;      /* The signals Gauche is _allowed_ to handle.
-                                   set by Scm_SetMasterSigset.
+                                   set by Scm_SetMasterSigmask.
                                    For some signals in this set Gauche sets
                                    the default signal handlers; for other
                                    signals in this set Gauche leaves them
                                    for the system to handle.  These can be
                                    overridden by Scm_SetSignalHandler. */
     ScmInternalMutex mutex;
-} sigHandlers = {{NULL}};
+} sigHandlers;
 
 /* Maximum # of the same signals before it is processed by the VM loop.
    If any one of signals exceeds this count, Gauche exits with Scm_Abort.
@@ -97,13 +97,45 @@ static struct sigHandlersRec {
    long-running C-routine and do not returns to VM.
    The actual limit can be changed at runtime by Scm_SetSignalPendingLimit().
    If signalPendingLimit is 0, the number of pending signals is unlimited. */
-#define SIGNAL_PENDING_LIMIT_DEFALT 3
+#define SIGNAL_PENDING_LIMIT_DEFAULT 3
 #define SIGNAL_PENDING_LIMIT_MAX 255
 
-static unsigned int signalPendingLimit = SIGNAL_PENDING_LIMIT_DEFALT;
+static unsigned int signalPendingLimit = SIGNAL_PENDING_LIMIT_DEFAULT;
 
 
-/* Table of signals and its initial behavior. */
+/* Table of signals and its initial behavior.   If Application asks
+   Gauche to handle a specific signal (via Scm_SetMasterSigmask),
+   Gauche installs a signal handler specified in this table.
+   (In general, the signals reserved by GC are excluded from the master
+   signal mask, so we don't need to consider them here.  See main.c.)
+
+   Note on SIGPIPE behavior: The Unix convention is that the default
+   behavior or writing to closed pipe terminates the process.  It is
+   suitable for small commands designed to be piped together, for
+   a failure of any one of the commands in the chunk causes
+   entire chain to be finished promptly.  However it is rather an
+   annoyance if we use an fd for transient communication (e.g.
+   sockets, most notably).  Since signal handlers are process-global,
+   it is quite tricky for general libraries/middlewares to control
+   SIGPIPE behavior transparently.
+
+   As a library, Gauche employs simpler model: Application will specify
+   which signals they want handle and which signals they want Gauche
+   to handle.  Once app delegates handling of a specific signal,
+   Gauche has a right to install its own handler.
+
+   So, here's the deal: If the app wants to handle SIGPIPE by itself,
+   it's fine.  By default SIGPIPE terminates the process.  If the app
+   installs its own SIGPIPE handler that returns, then Gauche sees
+   EPIPE and raises <system-error>.
+
+   If the app let Gauche to handle SIGPIPE, we install our own handler,
+   which does nothing.  Scheme code can handle the situation via EPIPE
+   <system-error>.  To emulate Unix default behavior, however, EPIPE
+   caused by standard output and standard error will terminate the
+   process by default.  It can be configured per-port basis.
+*/
+
 #define SIGDEF_NOHANDLE 0       /* Gauche doesn't install a signal handler,
                                    leaving it to the application. */
 #define SIGDEF_DFL      1       /* Gauche resets the singal handler to
@@ -112,6 +144,8 @@ static unsigned int signalPendingLimit = SIGNAL_PENDING_LIMIT_DEFALT;
                                    that raises an error. */
 #define SIGDEF_EXIT     3       /* Gauche installs a handler that calls
                                    Scm_Exit(). */
+#define SIGDEF_INDIFFERENT 4    /* Gauche installs a handler that does
+                                   nothing. */
 
 #define SIGDEF(x, flag)  { #x, x, flag }
 
@@ -150,7 +184,8 @@ static struct sigdesc {
     SIGDEF(SIGUSR2, SIGDEF_ERROR),    /* User-defined signal 2 (POSIX) */
 #endif
 #ifdef SIGPIPE
-    SIGDEF(SIGPIPE, SIGDEF_ERROR),    /* Broken pipe (POSIX) */
+    SIGDEF(SIGPIPE, SIGDEF_INDIFFERENT), /* Broken pipe (POSIX).
+                                            See above note on SIGPIPE. */
 #endif
 #ifdef SIGALRM
     SIGDEF(SIGALRM, SIGDEF_ERROR),    /* Alarm clock (POSIX) */
@@ -204,7 +239,10 @@ static struct sigdesc {
 #ifdef SIGPWR
     SIGDEF(SIGPWR,  SIGDEF_NOHANDLE), /* Power failure restart (System V) */
 #endif
-    { NULL, -1 }
+#ifdef SIGTHR
+    SIGDEF(SIGTHR,  SIGDEF_NOHANDLE), /* Thread interrupt / AST (FreeBSD, OpenBSD) */
+#endif
+    { NULL, -1, 0 }
 };
 
 /*===============================================================
@@ -269,17 +307,17 @@ static ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs);
 SCM_DEFINE_BUILTIN_CLASS(Scm_SysSigsetClass, sigset_print,
                          NULL, NULL, sigset_allocate, SCM_CLASS_DEFAULT_CPL);
 
-void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
+void sigset_print(ScmObj obj, ScmPort *out, 
+                  ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(out, "#<sys-sigset [");
     display_sigset(&SCM_SYS_SIGSET(obj)->set, out);
     Scm_Printf(out, "]>");
 }
 
-ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs)
+ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
-    ScmSysSigset *s = SCM_ALLOCATE(ScmSysSigset, klass);
-    SCM_SET_CLASS(s, klass);
+    ScmSysSigset *s = SCM_NEW_INSTANCE(ScmSysSigset, klass);
     sigemptyset(&s->set);
     return SCM_OBJ(s);
 }
@@ -296,11 +334,11 @@ ScmSysSigset *make_sigset(void)
 */
 ScmObj Scm_SysSigsetOp(ScmSysSigset *set, ScmObj signals, int delp)
 {
-    ScmObj cp;
-    
     if (!SCM_PAIRP(signals)) {
         Scm_Error("list of signals required, but got %S", signals);
     }
+
+    ScmObj cp;
     SCM_FOR_EACH(cp, signals) {
         ScmObj s = SCM_CAR(cp);
         if (SCM_TRUEP(s)) {
@@ -329,6 +367,43 @@ ScmObj Scm_SysSigsetFill(ScmSysSigset *set, int emptyp)
     return SCM_OBJ(set);
 }
 
+/* Fills sigset_t with as many signals as we can reasonably handle. */
+void Scm_SigFillSetMostly(sigset_t *set) /* out */
+{
+    sigfillset(set);
+    sigdelset(set, SIGABRT);
+    sigdelset(set, SIGILL);
+#ifdef SIGKILL
+    sigdelset(set, SIGKILL);
+#endif
+#ifdef SIGCONT
+    sigdelset(set, SIGCONT);
+#endif
+#ifdef SIGSTOP
+    sigdelset(set, SIGSTOP);
+#endif
+    sigdelset(set, SIGSEGV);
+#ifdef SIGBUS
+    sigdelset(set, SIGBUS);
+#endif
+#ifdef SIGTHR
+    sigdelset(set, SIGTHR);
+#endif
+
+    /* Exclude signals used by GC to stop and restart the world. */
+#ifdef GC_THREADS
+    int sig_suspend = GC_get_suspend_signal();
+    if (sig_suspend >= 0) sigdelset(set, sig_suspend);
+    int sig_restart = GC_get_thr_restart_signal();
+    if (sig_restart >= 0) sigdelset(set, sig_restart);
+#endif /*GC_THREADS*/
+
+    /* Exclude SIGPROF when we're using gperftools profiler. */
+#if defined(HAVE_GPERFTOOLS_PROFILER_H)
+    sigdelset(set, SIGPROF);
+#endif /*WITHGPERFTOOLS*/
+}
+
 /*=============================================================
  * C-level signal handling
  */
@@ -363,8 +438,7 @@ static void sig_handle(int signum)
  */
 void Scm_SignalQueueClear(ScmSignalQueue* q)
 {
-    int i;
-    for (i=0; i<NSIG; i++) q->sigcounts[i] = 0;
+    for (int i=0; i<SCM_NSIG; i++) q->sigcounts[i] = 0;
 }
 
 /*
@@ -372,6 +446,7 @@ void Scm_SignalQueueClear(ScmSignalQueue* q)
  */
 void Scm_SignalQueueInit(ScmSignalQueue* q)
 {
+    q->sigcounts = SCM_NEW_ARRAY(unsigned char, SCM_NSIG);
     Scm_SignalQueueClear(q);
     q->pending = SCM_NIL;
 }
@@ -398,31 +473,38 @@ void Scm_SetSignalPendingLimit(int num)
  */
 void Scm_SigCheck(ScmVM *vm)
 {
-    ScmObj tail, cell, sp;
     ScmSignalQueue *q = &vm->sigq;
-    sigset_t omask;
-    int i;
-    unsigned char sigcounts[NSIG]; /* copy of signal counter */
+    unsigned char sigcounts[SCM_NSIG]; /* copy of signal counter */
 
     /* Copy VM's signal counter to local storage, for we can't call
        storage allocation during blocking signals. */
+    sigset_t omask;
     SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
-    memcpy(sigcounts, vm->sigq.sigcounts, NSIG * sizeof(unsigned char));
+    memcpy(sigcounts, vm->sigq.sigcounts, SCM_NSIG * sizeof(unsigned char));
     Scm_SignalQueueClear(&vm->sigq);
     vm->signalPending = FALSE;
     SIGPROCMASK(SIG_SETMASK, &omask, NULL);
 
-    /* Now, prepare queued signal handlers
-       If an error is thrown in this loop, the queued signals will be
-       lost---it doesn't look like so, but I may overlook something. */
-    tail = q->pending;
+#if defined(GAUCHE_USE_PTHREADS) && defined(GAUCHE_PTHREAD_SIGNAL)
+    /* We may use GAUCHE_PTHREAD_SIGNAL signal to terminate a thread
+       gracefully.  See Scm_ThreadTerminate in ext/threads/threads.c */
+    if (sigcounts[GAUCHE_PTHREAD_SIGNAL] > 0) {
+        /* The thread state will be set to TERMINATED by the cleanup
+           handler so we don't need to change it. */
+        SCM_INTERNAL_THREAD_EXIT();
+        /* NOTREACHED */
+    }
+#endif  /* defined(GAUCHE_USE_PTHREADS) && defined(GAUCHE_PTHREAD_SIGNAL) */
+
+    /* Now, queue the signal handlers */
+    ScmObj tail = q->pending;
     if (!SCM_NULLP(tail)) tail = Scm_LastPair(tail);
-    for (i=0; i<NSIG; i++) {
+    for (int i=0; i<SCM_NSIG; i++) {
         if (sigcounts[i] == 0) continue;
         if (SCM_PROCEDUREP(sigHandlers.handlers[i])) {
-            cell = Scm_Cons(SCM_LIST3(sigHandlers.handlers[i],
-                                      SCM_MAKE_INT(i),
-                                      SCM_OBJ_SAFE(sigHandlers.masks[i])),
+            ScmObj cell = Scm_Cons(SCM_LIST3(sigHandlers.handlers[i],
+                                             SCM_MAKE_INT(i),
+                                             SCM_OBJ_SAFE(sigHandlers.masks[i])),
                             SCM_NIL);
             if (SCM_NULLP(tail)) {
                 q->pending = tail = cell;
@@ -432,19 +514,21 @@ void Scm_SigCheck(ScmVM *vm)
             }
         }
     }
-    
+
     /* Call the queued signal handlers.  If an error is thrown in one
        of those handlers, the rest of handlers remain in the queue. */
     /* TODO: if VM is active, it'd be better to make the active VM to handle
        those handler procs, instead of calling Scm_Eval. */
+    ScmObj sp;
     SCM_FOR_EACH(sp, q->pending) {
-        ScmObj e = SCM_CAR(sp), handler, num, mask;
+        ScmObj e = SCM_CAR(sp);
         q->pending = SCM_CDR(sp);
-        handler = SCM_CAR(e);
-        num = SCM_CADR(e);
-        mask = SCM_CAR(SCM_CDDR(e));
+        ScmObj handler = SCM_CAR(e);
+        ScmObj num = SCM_CADR(e);
+        ScmObj mask = SCM_CAR(SCM_CDDR(e));
         if (SCM_SYS_SIGSET_P(mask)) {
             sigset_t omask;
+            sigemptyset(&omask);
             SCM_UNWIND_PROTECT {
                 SIGPROCMASK(SIG_BLOCK, &SCM_SYS_SIGSET(mask)->set, &omask);
                 Scm_ApplyRec(handler, SCM_LIST1(num));
@@ -469,15 +553,13 @@ void Scm_SigCheck(ScmVM *vm)
  * Default Scheme-level handlers
  */
 /* For most signals, default handler raises an error. */
-static ScmObj default_sighandler(ScmObj *args, int nargs, void *data)
+static ScmObj default_sighandler(ScmObj *args, int nargs, void *data SCM_UNUSED)
 {
-    int signum;
+    SCM_ASSERT(nargs == 1 && SCM_INTP(args[0]));
+    int signum = SCM_INT_VALUE(args[0]);
+
     struct sigdesc *desc;
     const char *name = NULL;
-    
-    SCM_ASSERT(nargs == 1 && SCM_INTP(args[0]));
-    signum = SCM_INT_VALUE(args[0]);
-
     for (desc = sigDesc; desc->name; desc++) {
         if (desc->num == signum) {
             name = desc->name;
@@ -508,7 +590,9 @@ static SCM_DEFINE_SUBR(default_sighandler_stub, 1, 0,
 #define DEFAULT_SIGHANDLER    SCM_OBJ(&default_sighandler_stub)
 
 /* For some signals, exits. */
-static ScmObj exit_sighandler(ScmObj *args, int nargs, void *data)
+static ScmObj exit_sighandler(ScmObj *args SCM_UNUSED,
+                              int nargs SCM_UNUSED,
+                              void *data SCM_UNUSED)
 {
     Scm_Exit(0);
     return SCM_UNDEFINED;       /* dummy */
@@ -523,22 +607,22 @@ static SCM_DEFINE_SUBR(exit_sighandler_stub, 1, 0,
 
 #define EXIT_SIGHANDLER    SCM_OBJ(&exit_sighandler_stub)
 
-#if 0                           /* not used for now */
 /* For some signals, gauche does nothing */
-static ScmObj through_sighandler(ScmObj *args, int nargs, void *data)
+static ScmObj indifferent_sighandler(ScmObj *args SCM_UNUSED,
+                                     int nargs SCM_UNUSED,
+                                     void *data SCM_UNUSED)
 {
     return SCM_UNDEFINED;
 }
 
-static SCM_DEFINE_STRING_CONST(through_sighandler_name,
-                               "%through-signal-handler", 20, 20);
-static SCM_DEFINE_SUBR(through_sighandler_stub, 1, 0,
-                       SCM_OBJ(&through_sighandler_name),
-                       through_sighandler,
+static SCM_DEFINE_STRING_CONST(indifferent_sighandler_name,
+                               "%indifferent-signal-handler", 27, 27);
+static SCM_DEFINE_SUBR(indifferent_sighandler_stub, 1, 0,
+                       SCM_OBJ(&indifferent_sighandler_name),
+                       indifferent_sighandler,
                        NULL, NULL);
 
-#define THROUGH_SIGHANDLER    SCM_OBJ(&through_sighandler_stub)
-#endif
+#define INDIFFERENT_SIGHANDLER    SCM_OBJ(&indifferent_sighandler_stub)
 
 /*
  * An emulation stub for Windows
@@ -554,19 +638,20 @@ static SCM_DEFINE_SUBR(through_sighandler_stub, 1, 0,
  */
 #if defined(GAUCHE_WINDOWS)
 int sigaction(int signum, const struct sigaction *act,
-	      struct sigaction *oact)
+              struct sigaction *oact)
 {
     if (oact != NULL) {
-	Scm_Panic("sigaction() with oldact != NULL isn't supported on MinGW port");
+        Scm_Panic("sigaction() with oldact != NULL isn't supported on MinGW port");
     }
     if (signal(signum, act->sa_handler) == SIG_ERR) {
-	return -1;
+        return -1;
     } else {
-	return 0;
+        return 0;
     }
 }
 
-int sigprocmask_win(int how, const sigset_t *set, sigset_t *oldset)
+int sigprocmask_win(int how SCM_UNUSED, const sigset_t *set SCM_UNUSED,
+                    sigset_t *oldset SCM_UNUSED)
 {
     return 0;
 }
@@ -577,17 +662,15 @@ int sigprocmask_win(int how, const sigset_t *set, sigset_t *oldset)
  */
 ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
 {
-    struct sigaction act;
-    struct sigdesc *desc;
     sigset_t sigset;
-    int badproc = FALSE, sigactionfailed = FALSE, passthrough = FALSE;
+    int badproc = FALSE, sigactionfailed = FALSE;
 
+    sigemptyset(&sigset);
     if (SCM_INTP(sigs)) {
         int signum = SCM_INT_VALUE(sigs);
-        if (signum < 0 || signum >= NSIG) {
+        if (signum < 0 || signum >= SCM_NSIG) {
             Scm_Error("bad signal number: %d", signum);
         }
-        sigemptyset(&sigset);
         sigaddset(&sigset, signum);
     } else if (SCM_SYS_SIGSET_P(sigs)) {
         sigset = SCM_SYS_SIGSET(sigs)->set;
@@ -597,6 +680,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
 
     if (SCM_UNDEFINEDP(handler)) return SCM_UNDEFINED;
 
+    struct sigaction act;
     if (SCM_TRUEP(handler)) {
         act.sa_handler = SIG_DFL;
     } else if (SCM_FALSEP(handler)) {
@@ -609,7 +693,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
     }
 
     if (mask == NULL) {
-        /* If no mask is specified, block singals in SIGS. */
+        /* If no mask is specified, block signals in SIGS. */
         mask = make_sigset();
         mask->set = sigset;
     }
@@ -618,7 +702,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
     if (!badproc) {
         sigfillset(&act.sa_mask); /* we should block all the signals */
         act.sa_flags = 0;
-        for (desc=sigDesc; desc->name; desc++) {
+        for (struct sigdesc *desc=sigDesc; desc->name; desc++) {
             if (!sigismember(&sigset, desc->num)) continue;
             if (!sigismember(&sigHandlers.masterSigset, desc->num)) continue;
             else if (sigaction(desc->num, &act, NULL) != 0) {
@@ -637,7 +721,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
 
 ScmObj Scm_GetSignalHandler(int signum)
 {
-    if (signum < 0 || signum >= NSIG) {
+    if (signum < 0 || signum >= SCM_NSIG) {
         Scm_Error("bad signal number: %d", signum);
     }
     /* No lock; atomic pointer access */
@@ -647,7 +731,7 @@ ScmObj Scm_GetSignalHandler(int signum)
 ScmObj Scm_GetSignalHandlerMask(int signum)
 {
     ScmSysSigset *r;
-    if (signum < 0 || signum >= NSIG) {
+    if (signum < 0 || signum >= SCM_NSIG) {
         Scm_Error("bad signal number: %d", signum);
     }
     /* No lock; atomic pointer access */
@@ -657,21 +741,19 @@ ScmObj Scm_GetSignalHandlerMask(int signum)
 
 ScmObj Scm_GetSignalHandlers(void)
 {
-    ScmObj h = SCM_NIL, hp;
-    ScmObj handlers[NSIG];
-    struct sigdesc *desc;
-    sigset_t masterSet;
-    int i;
+    ScmObj h = SCM_NIL;
+    ScmObj handlers[SCM_NSIG];
 
     /* copy handler vector and master sig set locally, so that we won't
        grab the lock for extensive time */
     (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
-    for (i=0; i<NSIG; i++) handlers[i] = sigHandlers.handlers[i];
-    masterSet = sigHandlers.masterSigset;
+    for (int i=0; i<SCM_NSIG; i++) handlers[i] = sigHandlers.handlers[i];
+    sigset_t masterSet = sigHandlers.masterSigset;
     (void)SCM_INTERNAL_MUTEX_UNLOCK(sigHandlers.mutex);
-        
-    for (desc=sigDesc; desc->name; desc++) {
+
+    for (struct sigdesc *desc=sigDesc; desc->name; desc++) {
         if (!sigismember(&masterSet, desc->num)) continue;
+        ScmObj hp;
         SCM_FOR_EACH(hp, h) {
             if (SCM_EQ(SCM_CDAR(hp), handlers[desc->num])) {
                 sigaddset(&(SCM_SYS_SIGSET(SCM_CAAR(hp))->set), desc->num);
@@ -707,7 +789,7 @@ void Scm_SetMasterSigmask(sigset_t *set)
     actoff.sa_handler = SIG_DFL;
     sigemptyset(&actoff.sa_mask);
     actoff.sa_flags = 0;
-    
+
     for (; desc->name; desc++) {
         if (sigismember(&sigHandlers.masterSigset, desc->num)
             && !sigismember(set, desc->num)) {
@@ -735,12 +817,26 @@ void Scm_SetMasterSigmask(sigset_t *set)
                 case SIGDEF_EXIT:
                     sigHandlers.handlers[desc->num] = EXIT_SIGHANDLER;
                     break;
+                case SIGDEF_INDIFFERENT:
+                    sigHandlers.handlers[desc->num] = INDIFFERENT_SIGHANDLER;
+                    break;
                 default:
                     Scm_Panic("Scm_SetMasterSigmask: can't be here");
                 }
             }
         }
     }
+#ifdef GAUCHE_PTHREAD_SIGNAL
+    /* On pthread and when available, we reserve one signal for inter-thread
+       communication.  See gauche/pthread.h for the definition of
+       GAUCHE_PTHREAD_SIGNAL.  In sigHandlers we set DEFAULT_SIGHANDLER,
+       but this signal is intercepted in Scm_SigCheck() so a you can't
+       set Scheme handler for this signal. */
+    if (sigaction(GAUCHE_PTHREAD_SIGNAL, &acton, NULL) != 0) {
+        Scm_SysError("sigaction on %d failed", GAUCHE_PTHREAD_SIGNAL);
+    }
+    sigHandlers.handlers[GAUCHE_PTHREAD_SIGNAL] = DEFAULT_SIGHANDLER;
+#endif  /* GAUCHE_PTHREAD_SIGNAL */
     sigHandlers.masterSigset = *set;
     Scm_VM()->sigMask = sigHandlers.masterSigset;
 }
@@ -751,6 +847,10 @@ void Scm_SetMasterSigmask(sigset_t *set)
 
 /*
  * Convenience routines hiding platform-dependent stuff
+ *
+ * TRANSIENT: These used to be used in vm.c, but no longer.  As of 0.9.3 these
+ * aren't used anywhere.  Scm_SysSigmask covers those functionalities,
+ * so we'll drop them by 1.0.
  */
 void Scm_GetSigmask(sigset_t *mask)
 {
@@ -766,7 +866,6 @@ void Scm_SetSigmask(sigset_t *mask)
     }
 }
 
-
 /*
  * set signal mask
  */
@@ -831,6 +930,7 @@ static void scm_sigsuspend(sigset_t *mask)
     SIGPROCMASK(SIG_SETMASK, &omask, NULL);
     SCM_SIGCHECK(vm);
 #else  /* GAUCHE_WINDOWS */
+    (void)mask; /* suppress unused var warning */
     Scm_Error("sigsuspend not supported on Windows port");
 #endif /* GAUCHE_WINDOWS */
 }
@@ -864,31 +964,31 @@ ScmObj Scm_Pause(void)
  * and sigwait makes havoc.  Since Gauche installs sig_handle()
  * implicitly to some signals, a casual user may be confused by the
  * unpredictable behavior when he doesn't reset signal handlers explicitly.
- * So we take care of them here. 
+ * So we take care of them here.
  *
  * We remove the signal handlers for the signals to be waited before calling
  * sigwait(), and restore them after its return.  We assume those signals
  * are blocked at this moment (if not, the behavior of sigwait() is
  * undefined), so we don't need to care about race condition.  If another
  * thread replaces signal handlers during this thread's waiting for a
- * signal, it would be reverted upon returning from this function, but 
+ * signal, it would be reverted upon returning from this function, but
  * such operation is inherently unsafe anyway, so we don't care.
  */
 int Scm_SigWait(ScmSysSigset *mask)
 {
 #if defined(HAVE_SIGWAIT)
-    int i, r = 0, sig = 0;
+    int r = 0, sig = 0;
     int failed_sig = -1;
     int sigwait_called = FALSE;
     int errno_save = 0;
     sigset_t to_wait;        /* real set of signals to wait */
     sigset_t saved;
-    struct sigaction act, oacts[NSIG];
+    struct sigaction act, oacts[SCM_NSIG];
 
     (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
     /* we can't wait for the signals Gauche doesn't handle. */
     to_wait = mask->set;
-    for (i=0; i<NSIG; i++) {
+    for (int i=0; i<SCM_NSIG; i++) {
         if (!sigismember(&sigHandlers.masterSigset, i)) {
             sigdelset(&to_wait, i);
         }
@@ -898,7 +998,7 @@ int Scm_SigWait(ScmSysSigset *mask)
     sigemptyset(&saved);
     act.sa_handler = SIG_DFL;
     act.sa_flags = 0;
-    for (i=1; i<NSIG; i++) {
+    for (int i=1; i<SCM_NSIG; i++) {
         if (!sigismember(&to_wait, i)) continue;
         if (sigaction(i, &act, &oacts[i]) < 0) {
             failed_sig = i;
@@ -907,7 +1007,7 @@ int Scm_SigWait(ScmSysSigset *mask)
         }
         sigaddset(&saved, i);
     }
-    
+
     if (failed_sig < 0) {
         (void)SCM_INTERNAL_MUTEX_UNLOCK(sigHandlers.mutex);
         sigwait_called = TRUE;
@@ -916,7 +1016,7 @@ int Scm_SigWait(ScmSysSigset *mask)
     }
 
     /* Restore C-level handlers */
-    for (i=1; i<NSIG; i++) {
+    for (int i=1; i<SCM_NSIG; i++) {
         if (!sigismember(&saved, i)) continue;
         if (sigaction(i, &oacts[i], NULL) < 0) {
             failed_sig = i;
@@ -939,6 +1039,7 @@ int Scm_SigWait(ScmSysSigset *mask)
     }
     return sig;
 #else  /* !HAVE_SIGWAIT */
+    (void)mask; /* suppress unused var warning */
     Scm_Error("sigwait not supported on this platform");
     return 0;
 #endif
@@ -953,17 +1054,15 @@ void Scm__InitSignal(void)
 {
     ScmModule *mod = Scm_GaucheModule();
     ScmObj defsigh_sym = Scm_Intern(&default_sighandler_name);
-    struct sigdesc *desc;
-    int i;
 
     (void)SCM_INTERNAL_MUTEX_INIT(sigHandlers.mutex);
     sigemptyset(&sigHandlers.masterSigset);
-    for (i=0; i<NSIG; i++) sigHandlers.handlers[i] = SCM_UNDEFINED;
-    
+    for (int i=0; i<SCM_NSIG; i++) sigHandlers.handlers[i] = SCM_UNDEFINED;
+
     Scm_InitStaticClass(&Scm_SysSigsetClass, "<sys-sigset>",
                         mod, NULL, 0);
 
-    for (desc = sigDesc; desc->name; desc++) {
+    for (struct sigdesc *desc = sigDesc; desc->name; desc++) {
         SCM_DEFINE(mod, desc->name, SCM_MAKE_INT(desc->num));
     }
     Scm_Define(mod, SCM_SYMBOL(defsigh_sym), DEFAULT_SIGHANDLER);
