@@ -1,12 +1,12 @@
 /*
  * treemap.c - tree map implementation
  *
- *   Copyright (c) 2007-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2007-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -46,7 +46,7 @@ typedef struct NodeRec {
     intptr_t     value;
     int          color;
     struct NodeRec *parent;
-    struct NodeRec *left; 
+    struct NodeRec *left;
     struct NodeRec *right;
 } Node;
 
@@ -57,7 +57,7 @@ enum TreeOp {
     TREE_DELETE = SCM_DICT_DELETE,
     TREE_NEAR
 };
-    
+
 #define BLACK 0
 #define RED   1
 
@@ -87,6 +87,7 @@ static Node *next_node(Node *n);
 static Node *prev_node(Node *n);
 static Node *delete_node(ScmTreeCore *tc, Node *n);
 static Node *copy_tree(Node *parent, Node *self);
+static int   node_cleared_p(Node *n);
 
 /*
  * Public API
@@ -132,8 +133,8 @@ ScmDictEntry *Scm_TreeCoreClosestEntries(ScmTreeCore *tc,
                                          ScmDictEntry **lo,
                                          ScmDictEntry **hi)
 {
-    Node *l, *h, *r;
-    r = core_ref(tc, key, TREE_NEAR, &l, &h);
+    Node *l, *h;
+    Node *r = core_ref(tc, key, TREE_NEAR, &l, &h);
     *lo = (ScmDictEntry*)l;
     *hi = (ScmDictEntry*)h;
     return (ScmDictEntry*)r;
@@ -186,13 +187,12 @@ int Scm_TreeCoreNumEntries(ScmTreeCore *tc)
 int Scm_TreeCoreEq(ScmTreeCore *a, ScmTreeCore *b)
 {
     ScmTreeIter ai, bi;
-    ScmDictEntry *ae, *be;
     if (a->num_entries != b->num_entries) return FALSE;
     Scm_TreeIterInit(&ai, a, NULL);
     Scm_TreeIterInit(&bi, b, NULL);
     for (;;) {
-        ae = Scm_TreeIterNext(&ai);
-        be = Scm_TreeIterNext(&bi);
+        ScmDictEntry *ae = Scm_TreeIterNext(&ai);
+        ScmDictEntry *be = Scm_TreeIterNext(&bi);
         if (ae == NULL) {
             if (be == NULL) return TRUE;
             else return FALSE;
@@ -202,6 +202,18 @@ int Scm_TreeCoreEq(ScmTreeCore *a, ScmTreeCore *b)
     }
 }
 
+static ScmDictEntry *advance_iter(ScmDictEntry *e)
+{
+    if (e) return (ScmDictEntry*)next_node((Node*)e);
+    else return NULL;
+}
+
+static ScmDictEntry *retrogress_iter(ScmDictEntry *e)
+{
+    if (e) return (ScmDictEntry*)prev_node((Node*)e);
+    else return NULL;
+}
+
 /* START can be NULL; in which case, if next call is TreeIterNext,
    it iterates from the minimum node; if next call is TreeIterPrev,
    it iterates from the maximum node. */
@@ -213,42 +225,52 @@ void Scm_TreeIterInit(ScmTreeIter *iter,
         Scm_Error("Scm_TreeIterInit: iteration start point is not a part of the tree.");
     }
     iter->t = tc;
-    iter->e = start;
-    iter->at_end = FALSE;
+    iter->c = start;
+    iter->n = (start
+               ? advance_iter(start)
+               : Scm_TreeCoreGetBound(iter->t, SCM_TREE_CORE_MIN));
+    iter->p = (start
+               ? retrogress_iter(start)
+               : Scm_TreeCoreGetBound(iter->t, SCM_TREE_CORE_MAX));
 }
 
+/* Mind that the 'current' node might be deleted. */
 ScmDictEntry *Scm_TreeIterNext(ScmTreeIter *iter)
 {
-    if (iter->at_end) return NULL;
-    if (iter->e) {
-        iter->e = (ScmDictEntry*)next_node((Node*)iter->e);
+    if (node_cleared_p((Node*)iter->c)) {
+        iter->c = iter->n;
+        iter->p = retrogress_iter(iter->c);
+        iter->n = advance_iter(iter->c);
     } else {
-        iter->e = Scm_TreeCoreGetBound(iter->t, SCM_TREE_CORE_MIN);
+        iter->p = iter->c;
+        iter->c = iter->n;
+        iter->n = advance_iter(iter->c);
     }
-    if (iter->e == NULL) iter->at_end = TRUE;
-    return iter->e;
+    return iter->c;
 }
 
 ScmDictEntry *Scm_TreeIterPrev(ScmTreeIter *iter)
 {
-    if (iter->at_end) return NULL;
-    if (iter->e) {
-        iter->e = (ScmDictEntry*)prev_node((Node*)iter->e);
+    if (node_cleared_p((Node*)iter->c)) {
+        iter->c = iter->p;
+        iter->n = advance_iter(iter->c);
+        iter->p = retrogress_iter(iter->c);
     } else {
-        iter->e = Scm_TreeCoreGetBound(iter->t, SCM_TREE_CORE_MAX);
+        iter->n = iter->c;
+        iter->c = iter->p;
+        iter->p = retrogress_iter(iter->c);
     }
-    if (iter->e == NULL) iter->at_end = TRUE;
-    return iter->e;
+    return iter->c;
 }
 
 ScmDictEntry *Scm_TreeIterCurrent(ScmTreeIter *iter)
 {
-    return iter->e;
+    return iter->c;
 }
 
 int Scm_TreeIterAtEnd(ScmTreeIter *iter)
 {
-    return iter->at_end;
+    return iter->c == NULL && (iter->p == NULL || iter->n == NULL);
 }
 
 /* consistency check */
@@ -258,10 +280,10 @@ int Scm_TreeIterAtEnd(ScmTreeIter *iter)
 static int check_traverse(Node *node, int depth, int *count)
 {
     int ld, rd;
-    
+
     (*count)++;                 /* entry count */
     if (BLACKP(node)) depth++;
-    
+
     if (node->left) {
         if (REDP(node) && REDP(node->left)) {
             Scm_Error("[internal] tree map has adjacent red nodes");
@@ -300,14 +322,16 @@ void Scm_TreeCoreCheckConsistency(ScmTreeCore *tc)
  * ScmTreeMap
  */
 
-static void treemap_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
+static void treemap_print(ScmObj obj, ScmPort *port, 
+                          ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmTreeMap *tm = SCM_TREE_MAP(obj);
     Scm_Printf(port, "#<tree-map %p (%d entries)>", tm,
                Scm_TreeCoreNumEntries(SCM_TREE_MAP_CORE(tm)));
 }
 
-SCM_DEFINE_BUILTIN_CLASS(Scm_TreeMapClass, treemap_print, NULL, NULL, NULL,
+SCM_DEFINE_BUILTIN_CLASS(Scm_TreeMapClass, treemap_print, Scm_ObjectCompare,
+                         NULL, NULL,
                          SCM_CLASS_ORDERED_DICTIONARY_CPL);
 
 /*
@@ -345,18 +369,19 @@ ScmObj Scm_TreeMapRef(ScmTreeMap *tm, ScmObj key, ScmObj fallback)
     }
 }
 
+/* Returns previous value; can return SCM_UNBOUND when the association hasn't
+   been there.  Be careful not to let SCM_UNBOUND leak out to Scheme! */
 ScmObj Scm_TreeMapSet(ScmTreeMap *tm, ScmObj key, ScmObj value, int flags)
 {
     ScmDictEntry *e = Scm_TreeCoreSearch(SCM_TREE_MAP_CORE(tm),
                                          (intptr_t)key,
                                          (flags&SCM_DICT_NO_CREATE)? SCM_DICT_GET : SCM_DICT_CREATE);
     if (!e) return SCM_UNBOUND;
-    if (e->value) {
-        if (flags&SCM_DICT_NO_OVERWRITE) return SCM_DICT_VALUE(e);
-        else return SCM_DICT_SET_VALUE(e, value);
-    } else {
-        return SCM_DICT_SET_VALUE(e, value);
+    ScmObj oldval = e->value? SCM_DICT_VALUE(e) : SCM_UNBOUND;
+    if (!(flags&SCM_DICT_NO_OVERWRITE) || SCM_UNBOUNDP(oldval)) {
+        (void)SCM_DICT_SET_VALUE(e, value);
     }
+    return oldval;
 }
 
 ScmObj Scm_TreeMapDelete(ScmTreeMap *tm, ScmObj key)
@@ -370,9 +395,8 @@ ScmObj Scm_TreeMapDelete(ScmTreeMap *tm, ScmObj key)
 /* for debug */
 static void dump_traverse(Node *node, int depth, ScmPort *out, int scmobj)
 {
-    int i;
     if (node->left) dump_traverse(node->left, depth+1, out, scmobj);
-    for (i=0; i<depth; i++) Scm_Printf(out, "  ");
+    for (int i=0; i<depth; i++) Scm_Printf(out, "  ");
     if (scmobj) {
         Scm_Printf(out, "%c:%S => %S\n", BLACKP(node)?'B':'R',
                    SCM_OBJ(node->key), SCM_OBJ(node->value));
@@ -459,6 +483,12 @@ static void clear_node(Node *node)
     node->parent = node->left = node->right = NULL;
 }
 
+static int node_cleared_p(Node *node)
+{
+    return (node && node->parent == NULL
+            && node->left == NULL && node->right == NULL);
+}
+
 /* replace N's position by M. M could be NULL. */
 static void replace_node(ScmTreeCore *tc, Node *n, Node *m)
 {
@@ -482,9 +512,9 @@ static void replace_node(ScmTreeCore *tc, Node *n, Node *m)
 */
 static void rotate_right(ScmTreeCore *tc, Node *n)
 {
-    Node *l = n->left, *gr;
+    Node *l = n->left;
     SCM_ASSERT(l != NULL);
-    gr = l->right;
+    Node *gr = l->right;
 
     replace_node(tc, n, l);
     l->right = n;  n->parent = l;
@@ -501,9 +531,9 @@ static void rotate_right(ScmTreeCore *tc, Node *n)
 */
 static void rotate_left(ScmTreeCore *tc, Node *n)
 {
-    Node *r = n->right, *gl;
+    Node *r = n->right;
     SCM_ASSERT(r != NULL);
-    gl = r->left;
+    Node *gl = r->left;
 
     replace_node(tc, n, r);
     r->left = n;   n->parent = r;
@@ -519,15 +549,15 @@ static void rotate_left(ScmTreeCore *tc, Node *n)
 /* balance tree after insertion of N */
 static void balance_tree(ScmTreeCore *tc, Node *n)
 {
-    Node *p = n->parent, *u, *g;
-    
+    Node *p = n->parent;
+
     if (!p) { BALANCE_CASE("1"); n->color = BLACK; return; }  /* root */
     if (BLACKP(p)) { BALANCE_CASE("2"); return; }      /* nothing to do */
 
     /* Here we're sure we have grandparent. */
-    g = p->parent;
+    Node *g = p->parent;
     SCM_ASSERT(g != NULL);
-    u = (g->left == p)? g->right : g->left;
+    Node *u = (g->left == p)? g->right : g->left;
 
     if (REDP(u)) {
         p->color = u->color = BLACK;
@@ -568,8 +598,8 @@ static void balance_tree(ScmTreeCore *tc, Node *n)
    Note that CHILD can be NULL (empty BLACK node) */
 static void delete_node1(ScmTreeCore *tc, Node *todie, Node *child)
 {
-    Node *sibling, *parent = todie->parent;
-    
+    Node *parent = todie->parent;
+
     replace_node(tc, todie, child);
     if (REDP(todie)) { DELETE_CASE("1"); return; }
     if (REDP(child)) { DELETE_CASE("2"); child->color = BLACK; return; }
@@ -577,7 +607,7 @@ static void delete_node1(ScmTreeCore *tc, Node *todie, Node *child)
   recur:
     /* At this point, child is BLACK. */
     if (parent == NULL) { DELETE_CASE("3"); return; }
-    sibling = SIBLING2(parent, child);
+    Node *sibling = SIBLING2(parent, child);
     /* sibling can't be NULL, since it would break the invariance of
        consistent # of black nodes for every path. */
     SCM_ASSERT(sibling != NULL);
@@ -644,8 +674,6 @@ static void delete_node1(ScmTreeCore *tc, Node *todie, Node *child)
 static void swap_node(ScmTreeCore *tc, Node *x, Node *y)
 {
 #define SWAP(x, y, tmp) do { tmp = x; x = y; y = tmp; } while (0)
-    Node *t;
-    int c;
     if (x->parent) {
         if (LEFTP(x)) x->parent->left = y;
         else          x->parent->right = y;
@@ -654,6 +682,7 @@ static void swap_node(ScmTreeCore *tc, Node *x, Node *y)
         if (LEFTP(y)) y->parent->left = x;
         else          y->parent->right = x;
     }
+    Node *t;
     SWAP(x->parent, y->parent, t);
 
     if (x->left) x->left->parent = y;
@@ -664,6 +693,7 @@ static void swap_node(ScmTreeCore *tc, Node *x, Node *y)
     if (y->right) y->right->parent = x;
     SWAP(x->right, y->right, t);
 
+    int c;
     SWAP(x->color, y->color, c);
     if (x == ROOT(tc)) SET_ROOT(tc, y);
     else if (y == ROOT(tc)) SET_ROOT(tc, x);
@@ -727,7 +757,7 @@ Node *core_ref(ScmTreeCore *tc, intptr_t key, enum TreeOp op,
             }
             return e;
         }
-        
+
         if (tc->cmp? (r < 0) : (e->key < key)) {
             /* Key is larger than E */
             if (e->right) {
@@ -778,4 +808,3 @@ static Node *copy_tree(Node *parent, Node *self)
     if (self->right) n->right = copy_tree(n, self->right);
     return n;
 }
-
