@@ -1,12 +1,12 @@
 ;;;
 ;;; vminsn.scm - Virtual machine instruction definition
 ;;;
-;;;   Copyright (c) 2000-2011  Shiro Kawai  <shiro@acm.org>
-;;; 
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
+;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
 ;;;   are met:
-;;; 
+;;;
 ;;;   1. Redistributions of source code must retain the above copyright
 ;;;      notice, this list of conditions and the following disclaimer.
 ;;;
@@ -35,22 +35,53 @@
 ;;; gauche/vminsn.h and vminsn.c, which are then included in vm.c.
 ;;; This file is also used by the compiler.
 ;;;
-;;; 
-;;; (define-insn <name> <num-params> <operand-type> [<combination>] [<body>])
+;;;
+;;; (define-insn <name> <num-params> <operand-type>
+;;;              :optional (<combination> '())
+;;;                        (<body> #f)
+;;;                        <flags> ...)
 ;;;
 ;;;   <name> - instruction name.  In C, an enum SCM_VM_<name> is defined.
 ;;;
 ;;;   <num-params> - # of parameters the instruction takes.
+;;;                  Can be (N M ...) - in this case, N is the proper
+;;;                  <num-params> of this insn, but vm-build-insn tolerates
+;;;                  the passed insn to have M parameters as well.  If N < M,
+;;;                  extra parameters are ignored.  If N > M, 0 is assumed
+;;;                  for missing parameters.  This is mainly to compile
+;;;                  Gauche itself with the previous version of Gauche.
 ;;;
 ;;;   <operand-type> - none : the insn doesn't take an operand.
 ;;;                    obj  : an ScmObj operand.
 ;;;                    addr : an address the next pc points.
 ;;;                    code : an ScmCompiledCode operand.
 ;;;                    codes: a list of ScmCompiledCodes.
+;;;                    obj+addr : an ScmObj, followed by an address
 ;;;
-;;;   If this insn is a combined insn, <combination> gives a list of
-;;;   insns that combines into this one.  It is used to generate a
-;;;   code-emitting table.
+;;;   <combination>  - If this is a comibined insn, list the ingredients
+;;;                    here.  It is used in multiple purposes:
+;;;                    * The instruction body is generated automatically,
+;;;                      fusing the ingredients' body.
+;;;                    * A DFA is set up in the instruction emitter
+;;;                      that replaces this specific sequence of
+;;;                      insns to the combined insn.  Because of this,
+;;;                      pass5 (instruction generation) doesn't need to
+;;;                      handle instruction combination explicitly.
+;;;
+;;;   <body>         - a CiSE expression that handles the instruction.
+;;;                    can be #f for combined insn.
+;;;
+;;;   <flags>
+;;;           :obsoleted  - mark obsoleted insn.  handled but won't be
+;;;                         generated.
+;;;           :fold-lref  - the insn must be a combination of LREF + something,
+;;;                         with depth and offset parameters.
+;;;                         this indicates that the combined insn should be
+;;;                         emitted even if preceding insn is 'shortcut'
+;;;                         LREFs such as LREF0 or LREF21; that is, instead
+;;;                         of having LREF0-SOMETHING or LREF21-SOMETHING
+;;;                         separately, we'll have LREF-SOMETHING(0,0) and
+;;;                         LREF-SOMETHING(2,1), respectively.
 
 ;;;==============================================================
 ;;; Common Cise macros
@@ -65,56 +96,51 @@
 
 (define-cise-stmt $result
   [(_ expr) `(begin ,@(case (result-type)
-                       [(reg)  `((set! VAL0 ,expr) NEXT1)]
-                       [(push) `((PUSH-ARG ,expr) NEXT1)]
-                       [(call) `((set! VAL0 ,expr))] ; no NEXT1
+                       [(reg)  `((set! VAL0 ,expr)
+                                 (set! (-> vm numVals) 1)
+                                 NEXT_PUSHCHECK)]
+                       [(push) `((PUSH-ARG ,expr) NEXT)]
+                       [(call) `((set! VAL0 ,expr))]
                        [(ret)  `((set! VAL0 ,expr)
                                  (set! (-> vm numVals) 1)
                                  (RETURN-OP)
                                  NEXT)]
-;                        [(w/push) `((if (SCM_VM_INSN_WITH_PUSH code)
-;                                      (PUSH-ARG ,expr)
-;                                      (set! VAL0 ,expr))
-;                                    NEXT1)]
-;                        [(w/ret)  `((set! VAL0 ,expr)
-;                                    (when (SCM_VM_INSN_WITH_RET code)
-;                                      (set! (-> vm numVals) 1)
-;                                      (RETURN-OP))
-;                                    NEXT1)]
                        ))])
-  
+
 ;; variations of $result with type coercion
-(define-cise-stmt $result:b 
+(define-cise-stmt $result:b
   [(_ expr) `($result (SCM_MAKE_BOOL ,expr))])
 (define-cise-stmt $result:i
   [(_ expr) (let1 r (gensym "cise__")
               `(let* ([,r :: long ,expr])
                  ($result (SCM_MAKE_INT ,r))))])
 (define-cise-stmt $result:n
-  [(_ expr) (let1 r (gensym "cise__")
-              `(let* ([,r :: long ,expr])
+  [(_ expr) (let ([r (gensym "cise__")]
+                  [v (gensym "cise__")])
+              `(let* ([,r :: long ,expr] [,v])
                  (if (SCM_SMALL_INT_FITS ,r)
-                   ($result (SCM_MAKE_INT ,r))
-                   ($result (Scm_MakeInteger ,r)))))])
+                   (set! ,v (SCM_MAKE_INT ,r))
+                   (set! ,v (Scm_MakeInteger ,r)))
+                 ($result ,v)))])
 (define-cise-stmt $result:u
-  [(_ expr) (let1 r (gensym "cise__")
-              `(let* ([,r :: u_long ,expr])
+  [(_ expr) (let ([r (gensym "cise__")]
+                  [v (gensym "cise__")])
+              `(let* ([,r :: u_long ,expr] [,v])
                  (if (SCM_SMALL_INT_FITS ,r)
-                   ($result (SCM_MAKE_INT ,r))
-                   ($result (Scm_MakeIntegerU ,r)))))])
+                   (set! ,v (SCM_MAKE_INT ,r))
+                   (set! ,v (Scm_MakeIntegerU ,r)))
+                 ($result ,v)))])
 (define-cise-stmt $result:f
   [(_ expr) (let1 r (gensym "cise__")
               `(let* ([,r :: double ,expr])
                  ($result (Scm_VMReturnFlonum ,r))))])
 
-;; Extract local value.  If it is boxed, we need dereference.
+;; Extract local value.
+;; If local var nees unboxing, necessary UNBOX insn will be generated
+;; by the compiler, so we don't need to worry about it.
 (define-cise-stmt $lref!
   [(_ var env off)
-   (let1 v (gensym)
-     `(let* ([,v (ENV-DATA ,env ,off)])
-        (if (SCM_BOXP ,v)
-          (set! ,var (SCM_BOX_VALUE ,v))
-          (set! ,var ,v))))])
+   `(set! ,var (ENV-DATA ,env ,off))])
 
 ;;
 ;; ($w/argr <val> <expr> ...)
@@ -149,7 +175,6 @@
                                                'ENV
                                                `(-> ,(loop (- d 1)) up)))
                                           ,o)))
-                     (if (SCM_BOXP ,val) (set! ,val (SCM_BOX_VALUE ,val)))
                      ,@body)])])
 
 (define-cise-stmt $w/argr               ; use VAL0 default
@@ -219,13 +244,14 @@
 ;;   Branch.  $branch* leaves the result in VAL0.
 ;;
 (define-cise-stmt $branch
-  [(_ expr) `(begin (if ,expr (FETCH-LOCATION PC) INCR-PC) NEXT)])
+  [(_ expr) `(begin (if ,expr (FETCH-LOCATION PC) INCR-PC) CHECK-INTR NEXT)])
 (define-cise-stmt $branch*
   [(_ expr)
    `(begin
       (if ,expr
         (begin (set! VAL0 SCM_FALSE) (FETCH-LOCATION PC))
         (begin (set! VAL0 SCM_TRUE) INCR-PC))
+      CHECK-INTR
       NEXT)])
 
 ;;
@@ -272,26 +298,63 @@
 ;; ($include var)
 ;;   Preprocessor directives.
 ;;
-(define-cise-stmt $undef   [(_ var) `("#undef " ,(x->string var) "\n")])
-(define-cise-stmt $define  [(_ var) `("#define " ,(x->string var) "\n")])
-(define-cise-stmt $include [(_ var) `("#include " ,(write-to-string var) "\n")])
+(define-cise-stmt $undef
+  [(_ var) `("\n" |#reset-line| "#undef " ,(x->string var) "\n")])
+(define-cise-stmt $define
+  [(_ var) `("\n" |#reset-line| "#define " ,(x->string var) "\n")])
+(define-cise-stmt $include
+  [(_ var) `("\n" |#reset-line| "#include " ,(write-to-string var) "\n")])
+
+;; ($lref depth offset)
+;;   Common code in LREF and XLREF.
+(define-cise-stmt $lref
+  [(_ depth offset)
+   (let ([dep (gensym)]
+         [off (gensym)]
+         [e (gensym)])
+     `(let* ([,dep ::int ,depth]
+             [,off ::int ,offset]
+             [,e ::ScmEnvFrame* ENV])
+        (for [() (> ,dep 0) (post-- ,dep)]
+          (set! ,e (-> ,e up)))
+        ($result (ENV-DATA ,e ,off))))])
+
+;; ($lset depth offset)
+;;   Common code in LSETL and XLSET.
+(define-cise-stmt $lset
+  [(_ depth offset)
+   (let ([dep (gensym)]
+         [off (gensym)]
+         [e (gensym)]
+         [box (gensym)])
+     `(let* ([,dep ::int ,depth]
+             [,off ::int ,offset]
+             [,e ::ScmEnvFrame* ENV])
+        (for [() (> ,dep 0) (post-- ,dep)]
+          (VM-ASSERT (!= ,e NULL))
+          (set! ,e (-> ,e up)))
+        (VM-ASSERT (!= ,e NULL))
+        (VM-ASSERT (> (-> ,e size) ,off))
+        (SCM_FLONUM_ENSURE_MEM VAL0)
+        (let* ([,box (ENV-DATA ,e ,off)])
+          (VM_ASSERT (SCM_BOXP ,box))
+          (SCM_BOX_SET ,box VAL0))
+        (set! (-> vm numVals) 1)
+        NEXT))])
 
 ;;
 ;; ($lrefNN depth offset)
 ;;   Generate an expr to refer to the local variable, starting from ENV.
-;;
+;;   DEPTH *must* be a literal integer.  We unroll the loop.
 (define-cise-stmt $lrefNN
   [(_ depth offset)
-   ;; TODO 0.9.2: Extra check for backward compatibility
    (let1 v (gensym)
      `(let* ([,v :: ScmObj (ENV_DATA ,(let loop ((d depth))
                                         (case d
                                           [(0) 'ENV]
                                           [else `(-> ,(loop (- d 1)) up)]))
                                      ,offset)])
-        (if (SCM_BOXP ,v)
-          ($result (SCM_BOX_VALUE ,v))
-          ($result ,v))))])
+        ($result ,v)))])
 
 ;;
 ;; ($values)
@@ -362,8 +425,9 @@
     (FETCH-LOCATION next)
     (PUSH-CONT next)
     INCR-PC
+    CHECK-INTR
     NEXT))
-         
+
 ;; combined insn
 (define-insn PUSH-PRE-CALL 1 addr (PUSH PRE-CALL))
 
@@ -375,7 +439,7 @@
     (CHECK-STACK reqstack)
     NEXT))
 
-;; CALL(NARGS)
+;; CALL(nargs)
 ;;  Call procedure in val0.  The continuation of this call is already
 ;;  pushed by PRE-CALL, so this instruction is always the end of a graph.
 ;;
@@ -388,24 +452,32 @@
     ($define APPLY_CALL)
     ($include "./vmcall.c")))
 
-;; TAIL-CALL(NARGS)
-;;  Call procedure in val0.  Same as CALL except this discards the 
-;;  caller's arugment frame and shift the callee's argument frame.
+;; TAIL-CALL(nargs)
+;;  Call procedure in val0.  Same as CALL except this discards the
+;;  caller's argument frame and shift the callee's argument frame.
 ;;
 (define-insn TAIL-CALL 1 none #f
-  (begin (DISCARD-ENV) ($goto-insn CALL)))
+  (begin
+    (let* ([ct::ScmCallTrace* (-> vm callTrace)])
+      (when (!= ct NULL)
+        (set! (ref (aref (-> ct entries) (-> ct top)) base) BASE)
+        (set! (ref (aref (-> ct entries) (-> ct top)) pc) PC)
+        (set! (-> ct top) (logand (+ (-> ct top) 1)
+                                  (- (-> ct size) 1)))))
+    (DISCARD-ENV)
+    ($goto-insn CALL)))
 
 ;; JUMP <addr>
-;;  Jump to <addr>.  
+;;  Jump to <addr>.
 ;;
 (define-insn JUMP      0 addr #f
-  (begin (FETCH-LOCATION PC) NEXT))
+  (begin (FETCH-LOCATION PC) CHECK-INTR NEXT))
 
 ;; RET
 ;;  Pop the continuation stack.
 ;;
 (define-insn RET       0 none #f
-  (begin (RETURN-OP) NEXT))
+  (begin (RETURN-OP) CHECK-INTR NEXT))
 
 ;; DEFINE(flag) <symbol>
 ;;  Defines global binding of SYMBOL in the current module.
@@ -421,8 +493,9 @@
     (VM_ASSERT (SCM_IDENTIFIERP var))
     (SCM_FLONUM_ENSURE_MEM val)
     INCR-PC
-    (let* ([mod::ScmModule* (-> (SCM_IDENTIFIER var) module)]
-           [name::ScmSymbol* (-> (SCM_IDENTIFIER var) name)])
+    (let* ([id::ScmIdentifier* (Scm_OutermostIdentifier (SCM_IDENTIFIER var))]
+           [mod::ScmModule* (-> id module)]
+           [name::ScmSymbol* (SCM_SYMBOL (-> id name))])
       (case (SCM_VM_INSN_ARG code)        ;flag
         [(0)                    (Scm_MakeBinding mod name val 0)]
         [(1 SCM_BINDING_CONST)  (Scm_MakeBinding mod name val SCM_BINDING_CONST)]
@@ -439,7 +512,7 @@
     INCR-PC
     ($result (Scm_MakeClosure body (get_env vm)))))
 
-;; LOCAL-ENV(NLOCALS)
+;; LOCAL-ENV(nlocals)
 ;;  Create a new environment frame from the current arg frame.
 ;;  Used for let.
 (define-insn LOCAL-ENV  1 none #f
@@ -491,51 +564,18 @@
 (define-insn POP-LOCAL-ENV 0 none #f
   (begin (set! ENV (-> ENV up)) NEXT))
 
-;; LOCAL-ENV-JUMP(DEPTH) <addr>
-;;  This instruction appears when local function call is optimized.
-;;  The stack already has NLOCALS values.  This instruction creates an
-;;  env frame with them (just like LOCAL-ENV), then jump to <addr>.
-;;  (# of arguments can be known by SP - ARGP).
+;; LOCAL-ENV-JUMP(depth) <addr>
+;;  Combination of LOCAL-ENV-SHIFT + JUMP.
+;;  We can use this when none of the new environment needs boxing.
 (define-insn LOCAL-ENV-JUMP 1 addr #f
-  (let* ([nargs::int (cast int (- SP ARGP))]
-         [env_depth::int (SCM_VM_INSN_ARG code)]
-         [to::ScmObj*] [tenv::ScmEnvFrame* ENV])
-    ;; We can discard env_depth environment frames.
-    ;; There are several cases:
-    ;;  - if the target env frame (TENV) is in stack:
-    ;;   -- if the current cont frame is over TENV
-    ;;      => shift argframe on top of the current cont frame
-    ;;   -- otherwise => shift argframe on top of TENV
-    ;;  - if TENV is in heap:
-    ;;   -- if the current cont frame is in stack
-    ;;      => shift argframe on top of the current cont frame
-    ;;   -- otherwise => shift argframe at the stack base
-    (while (> (post-- env_depth) 0)
-      (SCM_ASSERT tenv)
-      (set! tenv (-> tenv up)))
-    (cond [(IN-STACK-P (cast ScmObj* tenv))
-           (if (and (IN-STACK-P (cast ScmObj* CONT))
-                    (> (cast ScmObj* CONT) (cast ScmObj* tenv)))
-             (set! to (CONT-FRAME-END CONT))
-             (set! to (+ (cast ScmObj* tenv) ENV_HDR_SIZE)))]
-          [else
-           (if (IN-STACK-P (cast ScmObj* CONT))
-             (set! to (CONT-FRAME-END CONT))
-             (set! to (-> vm stackBase)))]) ; continuation has been saved
-    (when (and (> nargs 0) (!= to ARGP))
-      (let* ([t::ScmObj* to] [a::ScmObj* ARGP])
-        (dotimes [c nargs]
-          (set! (* (post++ t)) (* (post++ a))))))
-    (set! ARGP to)
-    (set! SP (+ to nargs))
-    (if (> nargs 0)
-      (FINISH-ENV SCM_FALSE tenv)
-      (set! ENV tenv))
+  (begin
+    (local_env_shift vm (SCM_VM_INSN_ARG code))
     (FETCH-LOCATION PC)
+    CHECK-INTR
     NEXT))
 
-;; LOCAL-ENV-CALL(DEPTH)
-;; LOCAL-ENV-TAIL-CALL(DEPTH)
+;; LOCAL-ENV-CALL(depth)
+;; LOCAL-ENV-TAIL-CALL(depth)
 ;;  This instruction appears when local function call is optimized.
 ;;  VAL0 has a closure to call, and the stack already has the arguments.
 ;;
@@ -557,9 +597,10 @@
     (set! (-> vm base) (SCM_COMPILED_CODE (-> (SCM_CLOSURE VAL0) code)))
     (set! PC (-> vm base code))
     (CHECK-STACK (-> vm base maxstack))
+    CHECK-INTR
     (SCM_PROF_COUNT_CALL vm (SCM_OBJ (-> vm base)))
     NEXT))
-         
+
 (define-insn LOCAL-ENV-TAIL-CALL 1 none #f
   (let* ([nargs::int (cast int (- SP ARGP))] [to::ScmObj*])
     (VM-ASSERT (SCM_CLOSUREP VAL0))
@@ -587,8 +628,8 @@
 ;;   Conditional branches.
 ;;   The combined operations leave the boolean value of the test result
 ;;   in VAL0.
-(define-insn BF      0 addr #f ($branch (SCM_FALSEP VAL0)))
-(define-insn BT      0 addr #f ($branch (not (SCM_FALSEP VAL0))))
+(define-insn BF      0 addr #f ($branch (SCM_CHECKED_FALSEP VAL0)))
+(define-insn BT      0 addr #f ($branch (not (SCM_CHECKED_FALSEP VAL0))))
 (define-insn BNEQ    0 addr #f ($w/argp z ($branch* (not (SCM_EQ VAL0 z)))))
 (define-insn BNEQV   0 addr #f ($w/argp z ($branch* (not (Scm_EqvP VAL0 z)))))
 (define-insn BNNULL  0 addr #f ($branch* (not (SCM_NULLP VAL0))))
@@ -604,7 +645,7 @@
 ;; of LREF + BNLT etc. (which would compare stack top and LREF).  These insns
 ;; save one stack operation.  The compiler recognizes the pattern and
 ;; emits these.  See pass5/if-numcmp and pass5/if-numeq.
-(define-insn LREF-VAL0-BNUMNE 2 addr #f ($arg-source lref ($insn-body BNUMNE))) 
+(define-insn LREF-VAL0-BNUMNE 2 addr #f ($arg-source lref ($insn-body BNUMNE)))
 (define-insn LREF-VAL0-BNLT 2 addr #f ($arg-source lref ($insn-body BNLT)))
 (define-insn LREF-VAL0-BNLE 2 addr #f ($arg-source lref ($insn-body BNLE)))
 (define-insn LREF-VAL0-BNGT 2 addr #f ($arg-source lref ($insn-body BNGT)))
@@ -634,8 +675,8 @@
 ;; RNEQ
 ;; RNEQV
 ;;   Conditional returns.
-(define-insn RF          0 none #f ($retc (SCM_FALSEP VAL0)))
-(define-insn RT          0 none #f ($retc (not (SCM_FALSEP VAL0))))
+(define-insn RF          0 none #f ($retc (SCM_CHECKED_FALSEP VAL0)))
+(define-insn RT          0 none #f ($retc (not (SCM_CHECKED_FALSEP VAL0))))
 (define-insn RNEQ        0 none #f ($w/argp v
                                      ($retc* (not (SCM_EQ VAL0 v)))))
 (define-insn RNEQV       0 none #f ($w/argp v
@@ -656,6 +697,9 @@
       (when (and (not restarg) (> (-> vm numVals) reqargs))
         ($vm-err "received more values than expected"))
       (set! argsize (+ reqargs (?: restarg 1 0)))
+      (cast void argsize) ; suppress unused variable warning
+      (cast void nextpc) ; suppress unused variable warning
+      (cast void size) ; suppress unused variable warning
       ,@stmts
       (cond [(> reqargs 0) (PUSH-ARG VAL0) (post++ i)]
             [(and restarg (> (-> vm numVals) 0))
@@ -668,7 +712,8 @@
              (SCM_APPEND1 rest tail (aref (-> vm vals) (- i 1))))
         (PUSH-ARG rest))
       (FINISH-ENV SCM_FALSE ENV)
-      NEXT1)])
+      (set! (-> vm numVals) 1) ; we already processed extra vals, so reset it
+      NEXT)])
 
 ;; RECEIVE(nargs,restarg) <cont-offset>
 ;;  Primitive operation for receive and call-with-values.
@@ -696,7 +741,7 @@
 ;;  into the stack and makes them into an environment.
 ;;  It is used primarily for the occasions that compiler knows it needs
 ;;  to save the values temporarily (e.g. for evaluate 'after' thunk of
-;;  dynamic-wind, the results of its body needs to be saved).  
+;;  dynamic-wind, the results of its body needs to be saved).
 ;;  This must be twined with VALUES-N, which reverses the effects, i.e.
 ;;  turn the values in the env frame into values.
 (define-insn RECEIVE-ALL 0 addr #f
@@ -707,8 +752,8 @@
     (PUSH-CONT nextpc)
     ($goto-insn TAIL-RECEIVE-ALL)))
 
-;; TAIL-RECEIVE-ALL 
-;;  Tail version of RECEIVE-ALL.  
+;; TAIL-RECEIVE-ALL
+;;  Tail version of RECEIVE-ALL.
 (define-insn TAIL-RECEIVE-ALL 0 none #f
   (begin (CHECK-STACK-PARANOIA (ENV-SIZE (+ (-> vm numVals) 1)))
          (PUSH-ARG VAL0)
@@ -724,7 +769,7 @@
 (define-insn VALUES-N 0 none #f
   (begin
     (VM-ASSERT ENV)
-    (let* ([nvals::int (cast int (-> ENV size))] [v])
+    (let* ([nvals::int (cast int (-> ENV size))])
       (set! (-> vm numVals) nvals)
       (for [() (> nvals 1) (post-- nvals)]
            (POP-ARG (aref (-> vm vals) (- nvals 1))))
@@ -735,21 +780,8 @@
 ;;  Local set
 ;;
 (define-insn LSET        2 none #f
-  (let* ([dep::int (SCM_VM_INSN_ARG0 code)]
-         [off::int (SCM_VM_INSN_ARG1 code)]
-         [e::ScmEnvFrame* ENV])
-    (for [() (> dep 0) (post-- dep)]
-         (VM-ASSERT (!= e NULL))
-         (set! e (-> e up)))
-    (VM-ASSERT (!= e NULL))
-    (VM-ASSERT (> (-> e size) off))
-    (SCM_FLONUM_ENSURE_MEM VAL0)
-    ;; TODO 0.9.2 : Extra check to maintain compatibility
-    (let* ([box::ScmObj (ENV-DATA e off)])
-      (if (SCM_BOXP box)
-        (SCM_BOX_SET box VAL0)
-        (set! (ENV-DATA e off) VAL0)))
-    NEXT1))
+  ($lset (SCM_VM_INSN_ARG0 code)        ; depth
+         (SCM_VM_INSN_ARG1 code)))      ; offset
 
 ;; GSET <location>
 ;;  LOCATION may be a symbol or gloc
@@ -757,22 +789,24 @@
 (define-insn GSET        0 obj #f
   (let* ((loc))
     (FETCH-OPERAND loc)
+    (SCM_FLONUM_ENSURE_MEM VAL0)
     (cond
      [(SCM_GLOCP loc) (SCM_GLOC_SET (SCM_GLOC loc) VAL0)]
      [(SCM_IDENTIFIERP loc)
       ;; If runtime flag LIMIT_MODULE_MUTATION is set,
       ;; we search only for the id's module, so that set! won't
       ;; mutate bindings in the other module.
-      (let* ([id::ScmIdentifier* (SCM_IDENTIFIER loc)]
+      (let* ([id::ScmIdentifier* (Scm_OutermostIdentifier (SCM_IDENTIFIER loc))]
+             [name::ScmSymbol* (SCM_SYMBOL (-> id name))]
              [limit::int
               (SCM_VM_RUNTIME_FLAG_IS_SET vm SCM_LIMIT_MODULE_MUTATION)]
              [gloc::ScmGloc*
-              (Scm_FindBinding (-> id module) (-> id name)
+              (Scm_FindBinding (-> id module) name
                                (?: limit SCM_BINDING_STAY_IN_MODULE 0))])
         (when (== gloc NULL)
           ;; Do search again for meaningful error message
           (when limit
-            (set! gloc (Scm_FindBinding (-> id module) (-> id name) 0))
+            (set! gloc (Scm_FindBinding (-> id module) name 0))
             (when (!= gloc NULL)
               ($vm-err "can't mutate binding of %S, \
                         which is in another module"
@@ -785,25 +819,15 @@
         (set! (* PC) (SCM_WORD gloc)))]
      [else ($vm-err "GSET: can't be here")])
     INCR-PC
-    NEXT1))
+    (set! (-> vm numVals) 1)
+    NEXT))
 
 ;; LREF(depth,offset)
 ;;  Retrieve local value.
 ;;
 (define-insn LREF        2 none #f
-  (let* ([dep::int (SCM_VM_INSN_ARG0 code)]
-         [off::int (SCM_VM_INSN_ARG1 code)]
-         [e::ScmEnvFrame* ENV])
-    (for [() (> dep 0) (post-- dep)]
-         (VM-ASSERT (!= e NULL))
-         (set! e (-> e up)))
-    (VM-ASSERT (!= e NULL))
-    (VM-ASSERT (> (-> e size) off))
-    ;; TODO 0.9.2: Extra test to keep compatiblity
-    (let* ((v::ScmObj (ENV-DATA e off)))
-      (if (SCM_BOXP v)
-        ($result (SCM_BOX_VALUE v))
-        ($result v)))))
+  ($lref (SCM_VM_INSN_ARG0 code)   ; depth
+         (SCM_VM_INSN_ARG1 code))) ; offset 
 
 ;; Shortcut for the frequent depth/offset.
 ;; From statistics, we found that the following depth/offset combinations
@@ -823,7 +847,7 @@
 (define-insn LREF21 0 none #f ($lrefNN 2 1))
 (define-insn LREF30 0 none #f ($lrefNN 3 0))
 
-;; combined instrction
+;; combined instruction
 (define-insn LREF-PUSH   2 none  (LREF PUSH))
 (define-insn LREF0-PUSH  0 none  (LREF0 PUSH))
 (define-insn LREF1-PUSH  0 none  (LREF1 PUSH))
@@ -836,6 +860,8 @@
 (define-insn LREF21-PUSH 0 none  (LREF21 PUSH))
 (define-insn LREF30-PUSH 0 none  (LREF30 PUSH))
 
+(define-insn-lref* LREF-RET 0 none (LREF RET))
+
 ;; GREF <location>
 ;;  LOCATION may be a symbol or GLOC object.
 ;;  Retrieve global value in the current module.
@@ -856,10 +882,6 @@
 (define-insn PUSH-GREF-CALL 1 obj   (PUSH GREF CALL))
 (define-insn PUSH-GREF-TAIL-CALL 1 obj (PUSH GREF TAIL-CALL))
 
-(define-insn LREF0-PUSH-GREF 0 obj  (LREF0 PUSH GREF) #f :obsoleted)
-(define-insn LREF0-PUSH-GREF-CALL 1 obj (LREF0 PUSH GREF CALL) #f :obsoleted)
-(define-insn LREF0-PUSH-GREF-TAIL-CALL 1 obj (LREF0 PUSH GREF TAIL-CALL) #f :obsoleted)
-
 ;; PROMISE
 ;;  Delay syntax emits this instruction.  Wrap a procedure into a promise
 ;;  object.
@@ -867,7 +889,7 @@
 (define-insn PROMISE  0 none #f
   ($result (Scm_MakePromise FALSE VAL0)))
 
-;; VALUES_APPLY(nargs) <args>
+;; VALUES-APPLY(nargs) <args>
 ;;  This instruction only appears in the code generated dynamically
 ;;  by Scm_Apply(Rec).  This is used to pass the application information
 ;;  across the boundary frame (see user_eval_inner() in vm.c).
@@ -937,19 +959,19 @@
 
 (define-insn LIST-STAR   1 none #f      ; list*
   (let* ([nargs::int (SCM_VM_INSN_ARG code)] [cp SCM_NIL] [arg])
-    (when (> nargs 0)
-      (SCM_FLONUM_ENSURE_MEM VAL0)
-      (set! cp VAL0)
-      (while (> (pre-- nargs) 0)
-        (POP-ARG arg)
-        (set! cp (Scm_Cons arg cp))))
+    (VM-ASSERT (>= nargs 1))
+    (SCM_FLONUM_ENSURE_MEM VAL0)
+    (set! cp VAL0)
+    (while (> (pre-- nargs) 0)
+      (POP-ARG arg)
+      (set! cp (Scm_Cons arg cp)))
     ($result cp)))
-          
+
 (define-insn LENGTH      0 none #f      ; length
   (let* ([len::int (Scm_Length VAL0)])
     (when (< len 0) ($vm-err "proper list required, but got %S" VAL0))
     ($result:i len)))
-    
+
 (define-insn MEMQ 0 none #f ($w/argp v ($result (Scm_Memq v VAL0))))
 (define-insn MEMV 0 none #f ($w/argp v ($result (Scm_Memv v VAL0))))
 (define-insn ASSQ 0 none #f ($w/argp v ($result (Scm_Assq v VAL0))))
@@ -958,17 +980,27 @@
 (define-insn EQV  0 none #f ($w/argp v ($result:b (Scm_EqvP v VAL0))))
 
 (define-insn APPEND      1 none #f
-  (let* ([nargs::int (SCM_VM_INSN_ARG code)] [cp SCM_NIL] [arg])
-    (when (> nargs 0)
-      (SCM_FLONUM_ENSURE_MEM VAL0)
-      (set! cp VAL0)
-      (while (> (pre-- nargs) 0)
-        (POP-ARG arg)
-        (when (< (Scm_Length arg) 0) ($vm-err "list required, but got %S" arg))
-        (set! cp (Scm_Append2 arg cp))))
+  (let* ([nargs::int (SCM_VM_INSN_ARG code)] [cp SCM_NIL] [args SCM_NIL] [a])
+    ;; We special-case up to 2 args to save allocation of extra pairs
+    ;; for argument lists.
+    (case nargs
+      [(0) (break)]
+      [(1) (set! cp VAL0) (break)]
+      [(2) (SCM_FLONUM_ENSURE_MEM VAL0)
+       (POP-ARG a)
+       (set! cp (Scm_Append2 a VAL0))
+       (break)]
+      [else
+       (set! args (Scm_Cons VAL0 SCM_NIL))
+       ;; We want to pop all args before doing works, for Scm_Append may cause
+       ;; forcing lazy-pair.
+       (while (> (pre-- nargs) 0)
+         (POP-ARG a)
+         (set! args (Scm_Cons a args)))
+       (set! cp (Scm_Append args))])
     ($result cp)))
 
-(define-insn NOT      0 none #f ($w/argr v ($result:b (SCM_FALSEP v))))
+(define-insn NOT      0 none #f ($w/argr v ($result:b (SCM_CHECKED_FALSEP v))))
 (define-insn REVERSE  0 none #f ($w/argr v ($result (Scm_Reverse v))))
 
 (define-insn APPLY       1 none #f
@@ -983,7 +1015,7 @@
     (POP-ARG VAL0)                      ; get proc
     (TAIL-CALL-INSTRUCTION)
     (set! VAL0 (Scm_VMApply VAL0 cp))
-    NEXT1))
+    NEXT))
 
 (define-insn TAIL-APPLY  1 none #f
   ;; Inlined apply.  Assumes the call is at the tail position.
@@ -1001,33 +1033,30 @@
   ;;
   ;; where N = SCMVM_INSN_ARG(code)-2.
   ;; rest contains a list of "all the rest arguments".
+  ;; We just push VAL0 onto stack, and move ARGP 
   ;; We "rotate" the stack and VAL0 and jump to the
-  ;; TAIL-CALL processing.  (The unfolding of rest
+  ;; tail_apply_entry.  (The unfolding of rest
   ;; argument will be done in ADJUST_ARGUMENT_FRAME later,
   ;; if necessary.)
-  ;; 
+  ;;
   ;;   SP  >|      |
-  ;;        | rest |
+  ;;        | rest |< original SP position
   ;;        | argN |
   ;;        |   :  |
   ;;   ARGP>| arg0 |        VAL0=proc
-  ;;
-  (let* ([rargc::int (Scm_Length VAL0)]
+  ;;        | proc |< original ARGP postiion (proc unused) 
+  (let* ([rest VAL0]
+         [rargc::int (check_arglist_tail_for_apply vm rest)]
          [nargc::int (- (SCM_VM_INSN_ARG code) 2)]
          [proc (* (- SP nargc 1))])
-    (when (< rargc 0) ($vm-err "improper list not allowed: %S" VAL0))
-    (while (> nargc 0)
-      (set! (* (- SP nargc 1)) (* (- SP nargc)))
-      (post-- nargc))
+    (when (< rargc 0) ($vm-err "improper list not allowed: %S" rest))
+    (set! VAL0 proc)
+    (post++ ARGP)
     ;; a micro-optimization: if VAL0 is (), we just omit it and
     ;; pretend this is a normal TAIL-CALL.
-    (when (== rargc 0)
-      (post-- SP)
-      (set! VAL0 proc)
-      ($goto-insn TAIL-CALL))
+    (when (== rargc 0) ($goto-insn TAIL-CALL))
     ;; normal path
-    (set! (* (- SP 1)) VAL0)
-    (set! VAL0 proc)
+    (set! (* (post++ SP)) rest)
     (DISCARD-ENV)
     (goto tail_apply_entry)))
 
@@ -1037,9 +1066,7 @@
     (let* ([c::ScmClass* (SCM_CLASS VAL0)])
       ;; be careful to handle class redifinition case
       (cond [(not (SCM_FALSEP (-> (Scm_ClassOf obj) redefined)))
-             (CHECK-STACK CONT_FRAME_SIZE)
-             (PUSH_CONT PC)
-             (set! PC PC_TO_RETURN)
+             (Scm__VMProtectStack vm)
              ($result (Scm_VMIsA obj c))]
             [else ($result:b (SCM_ISA obj c))]))))
 
@@ -1072,21 +1099,22 @@
         (SCM_FLONUM_ENSURE_MEM arg)
         (set! (SCM_VECTOR_ELEMENT vec 0) arg)))
     ($result vec)))
-    
+
 (define-insn LIST2VEC    0 none #f      ; list->vector
   ($w/argr v ($result (Scm_ListToVector v 0 -1))))
 
 (define-insn APP-VEC     1 none #f      ; (compose list->vector append)
-  (let* ([nargs::int (SCM_VM_INSN_ARG code)] [cp SCM_NIL] [arg])
+  (let* ([nargs::int (SCM_VM_INSN_ARG code)] [cp SCM_NIL] [args SCM_NIL] [a])
     (when (> nargs 0)
       (SCM_FLONUM_ENSURE_MEM VAL0)
       (set! cp VAL0)
       (while (> (pre-- nargs) 0)
-        (POP-ARG arg)
-        (when (< (Scm_Length arg) 0)
-          ($vm-err "list required, but got %S" arg))
-        (set! cp (Scm_Append2 arg cp)))
-      ($result (Scm_ListToVector cp 0 -1)))))
+        (POP-ARG a)
+        (set! args (Scm_Cons a args)))
+      (dolist [a (Scm_ReverseX args)]
+        (when (< (Scm_Length a) 0) ($vm-err "list required, but got %S" a))
+        (set! cp (Scm_Append2 a cp))))
+    ($result (Scm_ListToVector cp 0 -1))))
 
 (define-insn VEC-LEN     0 none #f      ; vector-length
   ($w/argr v
@@ -1115,7 +1143,7 @@
       (SCM_FLONUM_ENSURE_MEM v)
       (set! (SCM_VECTOR_ELEMENT vec k) v)
       ($result SCM_UNDEFINED))))
-  
+
 ;; VEC-REF and VEC-SET with immediate index.  VAL0 must be a vector.
 (define-insn VEC-REFI    1 none #f
   ($w/argr vec
@@ -1124,7 +1152,7 @@
       (when (or (< k 0) (>= k (SCM_VECTOR_SIZE vec)))
         ($vm-err "vector-ref index out of range: %d" k))
       ($result (SCM_VECTOR_ELEMENT vec k)))))
-  
+
 (define-insn VEC-SETI    1 none #f
   ($w/argp vec
     ($type-check vec SCM_VECTORP "vector")
@@ -1167,7 +1195,7 @@
      [(and (SCM_FLONUMP VAL0) (SCM_FLONUMP arg))
       ($result:b (== (SCM_FLONUM_VALUE VAL0) (SCM_FLONUM_VALUE arg)))]
      [else ($result:b (Scm_NumEq arg VAL0))])))
-  
+
 (define-insn NUMLT2  0 none #f ($w/numcmp r <  ($result:b r)))
 (define-insn NUMLE2  0 none #f ($w/numcmp r <= ($result:b r)))
 (define-insn NUMGT2  0 none #f ($w/numcmp r >  ($result:b r)))
@@ -1196,11 +1224,26 @@
     ;; we take a shortcut if either one is flonum and the
     ;; other is real.  (if both are integers, the overflow check
     ;; would be cumbersome so we just call Scm_Mul).
-    (if (or (and (SCM_FLONUMP arg) (SCM_REALP VAL0))
-            (and (SCM_FLONUMP VAL0) (SCM_REALP arg)))
-      ($result:f (* (Scm_GetDouble arg) (Scm_GetDouble VAL0)))
-      ($result (Scm_Mul arg VAL0)))))
-  
+    ;; NB: If one arg is inexact real and another arg is exact zero,
+    ;; the result should be an exact zero.
+    (if (SCM_FLONUMP arg)
+      (cond [(and (== (SCM_MAKE_INT 0) VAL0)
+                  (not (SCM_IS_INF (SCM_FLONUM_VALUE arg)))
+                  (not (SCM_IS_NAN (SCM_FLONUM_VALUE arg))))
+             ($result (SCM_MAKE_INT 0))]
+            [(SCM_REALP VAL0)
+             ($result:f (* (Scm_GetDouble arg) (Scm_GetDouble VAL0)))]
+            [else ($result (Scm_Mul arg VAL0))])
+      (if (SCM_FLONUMP VAL0)
+        (cond [(and (== (SCM_MAKE_INT 0) arg) 
+                    (not (SCM_IS_INF (SCM_FLONUM_VALUE VAL0)))
+                    (not (SCM_IS_NAN (SCM_FLONUM_VALUE VAL0))))
+               ($result (SCM_MAKE_INT 0))]
+              [(SCM_REALP arg)
+               ($result:f (* (Scm_GetDouble arg) (Scm_GetDouble VAL0)))]
+              [else ($result (Scm_Mul arg VAL0))])
+        ($result (Scm_Mul arg VAL0))))))
+
 (define-insn NUMDIV2 0 none #f          ; / (binary)
   ($w/argp arg
     (if (or (and (SCM_FLONUMP arg) (SCM_REALP VAL0))
@@ -1209,7 +1252,7 @@
       ($result (Scm_Div arg VAL0)))))
 
 (define-insn LREF-VAL0-NUMADD2 2 none #f ($arg-source lref ($insn-body NUMADD2)))
-  
+
 (define-insn NEGATE  0 none #f          ; -  (unary)
   ($w/argr v
     (cond
@@ -1223,24 +1266,24 @@
       ($result:f (+ (Scm_GetDouble arg) (Scm_GetDouble VAL0)))
       ($result (Scm_Add (Scm_Inexact arg)
                         (Scm_Inexact VAL0))))))
-      
+
 (define-insn NUMISUB2    0 none #f      ; -. (binary)
   ($w/argp arg
     (if (and (SCM_REALP arg) (SCM_REALP VAL0))
       ($result:f (- (Scm_GetDouble arg) (Scm_GetDouble VAL0)))
       ($result (Scm_Sub (Scm_Inexact arg) (Scm_Inexact VAL0))))))
-  
+
 (define-insn NUMIMUL2    0 none #f      ; *.
   ($w/argp arg
     (if (and (SCM_REALP arg) (SCM_REALP VAL0))
       ($result:f (* (Scm_GetDouble arg) (Scm_GetDouble VAL0)))
       ($result (Scm_Mul (Scm_Inexact arg) (Scm_Inexact VAL0))))))
-  
+
 (define-insn NUMIDIV2    0 none #f      ; /. (binary)
   ($w/argp arg
-    (if (and (SCM_REALP arg) (SCM_REALP VAL0))
+    (if (and (SCM_FLONUMP arg) (SCM_FLONUMP VAL0))
       ($result:f (/ (Scm_GetDouble arg) (Scm_GetDouble VAL0)))
-      ($result (Scm_Div (Scm_Inexact arg) (Scm_Inexact VAL0))))))
+      ($result (Scm_VMDivInexact arg VAL0)))))
 
 (define-insn NUMADDI     1 none #f      ; +, if one of op is small int
   (let* ([imm::long (SCM_VM_INSN_ARG code)])
@@ -1260,19 +1303,33 @@
             [(SCM_FLONUMP arg)
              ($result:f (- (cast double imm) (SCM_FLONUM_VALUE arg)))]
             [else           ($result (Scm_Sub (SCM_MAKE_INT imm) arg))]))))
-  
 
-;(define-insn NUMQUOT     0 none)        ; quotient
-;(define-insn NUMMOD      0 none)        ; modulo
-;(define-insn NUMREM      0 none)        ; remainder
-;(define-insn ASH         0 none)        ; ash
-;(define-insn LOGAND      0 none)        ; logand
-;(define-insn LOGIOR      0 none)        ; logior
-;(define-insn LOGXOR      0 none)        ; logxor
-;(define-insn LOGNOT      0 none)        ; lognot
-;(define-insn LOGANDI     1 none)        ; logand w/ immediate small int
-;(define-insn LOGIORI     1 none)        ; logior w/ immediate small int
-;(define-insn LOGXORI     1 none)        ; logxor w/ immediate small int
+
+(define-insn NUMMODI      1 none #f
+  (let* ([divisor::ScmSmallInt (SCM_VM_INSN_ARG code)])
+    ($w/argr arg ($result (Scm_Modulo arg (SCM_MAKE_INT divisor) FALSE)))))
+(define-insn NUMREMI      1 none #f
+  (let* ([divisor::ScmSmallInt (SCM_VM_INSN_ARG code)])
+    ($w/argr arg ($result (Scm_Modulo arg (SCM_MAKE_INT divisor) TRUE)))))
+
+(define-insn ASHI         1 none #f
+  (let* ([cnt::ScmSmallInt (SCM_VM_INSN_ARG code)])
+    ($w/argr arg ($result (Scm_Ash arg cnt)))))
+(define-insn LOGAND       0 none #f
+  ($w/argp x ($result (Scm_LogAnd x VAL0))))
+(define-insn LOGIOR       0 none #f
+  ($w/argp x ($result (Scm_LogIor x VAL0))))
+(define-insn LOGXOR       0 none #f
+  ($w/argp x ($result (Scm_LogXor x VAL0))))
+(define-insn LOGANDC      0 obj #f
+  (let* ([obj])
+    ($w/argr x (FETCH-OPERAND obj) INCR-PC ($result (Scm_LogAnd x obj)))))
+(define-insn LOGIORC      0 obj #f
+  (let* ([obj])
+    ($w/argr x (FETCH-OPERAND obj) INCR-PC ($result (Scm_LogIor x obj)))))
+(define-insn LOGXORC      0 obj #f
+  (let* ([obj])
+    ($w/argr x (FETCH-OPERAND obj) INCR-PC ($result (Scm_LogXor x obj)))))
 
 (define-insn READ-CHAR   1 none #f      ; read-char
   (let* ([nargs::int (SCM_VM_INSN_ARG code)] [ch::int 0] [port::ScmPort*])
@@ -1283,7 +1340,7 @@
            (set! port SCM_CURIN)])
     (set! ch (Scm_Getc port))
     ($result (?: (< ch 0) SCM_EOF (SCM_MAKE_CHAR ch)))))
-  
+
 (define-insn PEEK-CHAR   1 none #f      ; peek-char
   (let* ([nargs::int (SCM_VM_INSN_ARG code)] [ch::int 0] [port::ScmPort*])
     (cond [(== nargs 1)
@@ -1327,7 +1384,7 @@
       (SCM_FLONUM_ENSURE_MEM slot)
       (SCM_FLONUM_ENSURE_MEM VAL0)
       ($result (Scm_VMSlotSet obj slot VAL0)))))
-  
+
 (define-insn SLOT-REFC   0 obj #f       ; slot-ref with constant slot name
   (let* ((slot))
     (FETCH-OPERAND slot)
@@ -1335,7 +1392,7 @@
     (TAIL-CALL-INSTRUCTION)
     (SCM_FLONUM_ENSURE_MEM VAL0)
     ($result (Scm_VMSlotRef VAL0 slot FALSE))))
-    
+
 (define-insn SLOT-SETC   0 obj #f       ; slot-set! with constant slot name
   (let* ((slot))
     (FETCH-OPERAND slot)
@@ -1345,9 +1402,9 @@
       (SCM_FLONUM_ENSURE_MEM VAL0)
       ($result (Scm_VMSlotSet obj slot VAL0)))))
 
-;;;
-;;; Additional instructions
-;;;
+;;
+;; Dynamic handlers
+;;
 
 ;; PUSH-HANDLERS
 ;; POP-HANDLERS
@@ -1361,25 +1418,82 @@
     (SCM_FLONUM_ENSURE_MEM after)
     (set! (-> vm handlers) (Scm_Acons before after (-> vm handlers)))
     NEXT))
-  
+
 (define-insn POP-HANDLERS 0 none #f     ; pop dynamic handlers
   (begin
     (VM-ASSERT (SCM_PAIRP (-> vm handlers)))
     (set! (-> vm handlers) (SCM_CDR (-> vm handlers)))
     NEXT))
 
-
 ;;
-;; Experimental
+;; Environment
 ;;
 
-(define-insn-lref* LREF-RET 0 none (LREF RET))
-
-(define-insn BOX 0 none #f              ; replace VAL0 with box(VAL0)
-  (begin
-    (SCM_FLONUM_ENSURE_MEM VAL0)
-    (let* ((b::ScmBox* (Scm_MakeBox VAL0)))
-      (set! VAL0 (SCM_OBJ b)))
+;; if param == 0, VAL0 <- box(VAL0)
+;; else if param > 0,  ENV[param-1] <- box(ENV[param-1])
+;; The second case is for arguments that are mutated.
+(define-insn BOX 1 none #f
+  (let* ([param::int (SCM_VM_INSN_ARG code)])
+    (cond [(== param 0)
+           (SCM_FLONUM_ENSURE_MEM VAL0)
+           (let* ([b::ScmBox* (Scm_MakeBox VAL0)])
+             (set! VAL0 (SCM_OBJ b)))]
+          [(> param 0)
+           (let* ([off::int (- param 1)])
+             (VM-ASSERT (> (-> ENV size) off))
+             (let* ([v (ENV-DATA ENV off)])
+               (SCM_FLONUM_ENSURE_MEM v)
+               (let* ([b::ScmBox* (Scm_MakeBox v)])
+                 (set! (ENV-DATA ENV off) (SCM_OBJ b)))))])
     NEXT))
 
+;; ENV-SET(offset)
+;;  Mutate the top env's specified slot with VAL0
+;;  This is used with LOCAL-ENV-CLOSURES to initialize non-procedure
+;;  slots of the env.  We used to use LSET for this purpose, but now
+;;  LSET counts on that all mutable lvars are boxed, so we need a separete
+;;  insn that directly initialize the env frame.
+(define-insn ENV-SET 1 none #f
+  (let* ([off::int (SCM_VM_INSN_ARG code)])
+    (VM-ASSERT (> (-> ENV size) off))
+    (SCM_FLONUM_ENSURE_MEM VAL0)
+    (set! (ENV-DATA ENV off) VAL0)
+    NEXT))
 
+;; UNBOX
+;;  VAL0 <- unbox(VAL0)
+(define-insn UNBOX 0 none #f
+  ($w/argr v
+    (set! VAL0 (SCM_BOX_VALUE v))
+    NEXT))
+
+(define-insn LREF-UNBOX 2 none (LREF UNBOX) #f :fold-lref)
+
+;; LOCAL-ENV-SHIFT(depth)
+;;  This instruction appears when local function call is optimized.
+;;  The stack already has NLOCALS values.  We discard DEPTH env frames,
+;;  and creates a new local env from the stack value.
+(define-insn LOCAL-ENV-SHIFT 1 none #f
+  (begin
+    (local_env_shift vm (SCM_VM_INSN_ARG code))
+    NEXT))
+
+;; TRANSIENT: TODO: Move these up below LREF on 1.0 release
+
+;; XLREF(offset) depth
+;;   LREF, but used when depth and/or offset can't fit in the params field.
+(define-insn XLREF 1 obj #f
+  (let* ([dep_s])
+    (FETCH-OPERAND dep_s)
+    INCR-PC
+    ($lref (SCM_INT_VALUE dep_s)          ;depth
+           (SCM_VM_INSN_ARG code))))      ;offset
+
+;; XLSET(offset) depth
+;;   LSET, but used when depth and/or offset can't fit in the params field.
+(define-insn XLSET 1 obj #f
+  (let* ([dep_s])
+    (FETCH-OPERAND dep_s)
+    INCR-PC
+    ($lset (SCM_INT_VALUE dep_s)        ; depth
+           (SCM_VM_INSN_ARG code))))    ; offset
