@@ -1,12 +1,12 @@
 /*
  * vmstat.c - statistics gathering code for vm.c
  *
- *   Copyright (c) 2005-2011  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2005-2019  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -48,7 +48,7 @@ static ScmWord fetch_insn_counting(ScmVM *vm, ScmWord code)
         insn2_freq[SCM_VM_INSN_CODE(code)][SCM_VM_INSN_CODE(*vm->pc)]++;
     }
     code = *vm->pc++;
-    insn1_freq[SCM_VM_INSN_CODE(code)]++; 
+    insn1_freq[SCM_VM_INSN_CODE(code)]++;
     switch (SCM_VM_INSN_CODE(code)) {
     case SCM_VM_LREF0: lref_freq[0][0]++; break;
     case SCM_VM_LREF1: lref_freq[0][1]++; break;
@@ -89,27 +89,26 @@ static ScmWord fetch_insn_counting(ScmVM *vm, ScmWord code)
 
 static void dump_insn_frequency(void *data)
 {
-    int i, j;
     Scm_Printf(SCM_CUROUT, "(:instruction-frequencies (");
-    for (i=0; i<SCM_VM_NUM_INSNS; i++) {
+    for (int i=0; i<SCM_VM_NUM_INSNS; i++) {
         Scm_Printf(SCM_CUROUT, "(%s %d", Scm_VMInsnName(i), insn1_freq[i]);
-        for (j=0; j<SCM_VM_NUM_INSNS; j++) {
+        for (int j=0; j<SCM_VM_NUM_INSNS; j++) {
             Scm_Printf(SCM_CUROUT, " %d", insn2_freq[i][j]);
         }
         Scm_Printf(SCM_CUROUT, ")\n");
     }
     Scm_Printf(SCM_CUROUT, ")\n :lref-frequencies (");
-    for (i=0; i<LREF_FREQ_COUNT_MAX; i++) {
+    for (int i=0; i<LREF_FREQ_COUNT_MAX; i++) {
         Scm_Printf(SCM_CUROUT, "(");
-        for (j=0; j<LREF_FREQ_COUNT_MAX; j++) {
+        for (int j=0; j<LREF_FREQ_COUNT_MAX; j++) {
             Scm_Printf(SCM_CUROUT, "%d ", lref_freq[i][j]);
         }
         Scm_Printf(SCM_CUROUT, ")\n");
     }
     Scm_Printf(SCM_CUROUT, ")\n :lset-frequencies (");
-    for (i=0; i<LREF_FREQ_COUNT_MAX; i++) {
+    for (int i=0; i<LREF_FREQ_COUNT_MAX; i++) {
         Scm_Printf(SCM_CUROUT, "(");
-        for (j=0; j<LREF_FREQ_COUNT_MAX; j++) {
+        for (int j=0; j<LREF_FREQ_COUNT_MAX; j++) {
             Scm_Printf(SCM_CUROUT, "%d ", lset_freq[i][j]);
         }
         Scm_Printf(SCM_CUROUT, ")\n");
@@ -119,4 +118,3 @@ static void dump_insn_frequency(void *data)
 }
 
 #endif /*COUNT_INSN_FREQUENCY*/
-
