@@ -127,7 +127,7 @@
 (define r3 (make <r> :b 100 :a 20))
 
 (define-method slot-values ((obj <r>))
-  (map (lambda (s) (slot-ref obj s)) '(a b)))
+  (map (^s (slot-ref obj s)) '(a b)))
 
 (test* "make <r>" '(4 5) (slot-values r1))
 (test* "make <r> :a" '(9 5) (slot-values r2))
@@ -141,14 +141,14 @@
    (c :allocation :class         :init-keyword :c :init-value #\c)
    (s :allocation :each-subclass :init-keyword :s :init-value #\s)
    (v :allocation :virtual       :init-keyword :v
-      :slot-ref (lambda (o) (cons (slot-ref o 'i) (slot-ref o 'c)))
+      :slot-ref (^o (cons (slot-ref o 'i) (slot-ref o 'c)))
       :slot-set! (lambda (o v)
                    (slot-set! o 'i (car v))
                    (slot-set! o 'c (cdr v))))
    ))
 
 (define-method slot-values ((obj <s>))
-  (map (lambda (s) (slot-ref obj s)) '(i c s v)))
+  (map (^s (slot-ref obj s)) '(i c s v)))
 
 (define s1 (make <s>))
 (define s2 (make <s>))
@@ -226,9 +226,9 @@
 
 (define-class <sss> ()
   ((v :allocation :virtual
-      :slot-ref  (lambda (o) (slot-ref o 'vv))
+      :slot-ref  (^o (slot-ref o 'vv))
       :slot-set! (lambda (o v) (slot-set! o 'vv v))
-      :slot-bound? (lambda (o) (slot-bound? o 'vv)))
+      :slot-bound? (^o (slot-bound? o 'vv)))
    vv))
 
 (define s5 (make <sss>))
@@ -241,6 +241,28 @@
               (list (slot-bound? s5 'v)
                     (slot-ref s5 'v))))
 
+;; Direct instance-slot access is only for under-the-hood work, and
+;; shouldn't be casually used.  But that doesn't diminish the importance
+;; of tests!
+(define-class <ssss> () ((a :init-keyword :a) b))
+
+(let ((z (make <ssss> :a 1)))
+  (test* "instance-slot-ref a" 1 (instance-slot-ref z 0))
+  (test* "instance-slot-ref b" (test-error <error> #/unbound/)
+         (instance-slot-ref z 1))
+  (test* "instance-slot-ref b fallback" 2
+         (instance-slot-ref z 1 2))
+  (test* "instance-slot-set!" 3
+         (begin (instance-slot-set! z 1 3)
+                (instance-slot-ref z 1)))
+  (test* "instance-slot-ref oob" (test-error <error> #/out of bound/)
+         (instance-slot-ref z 2))
+  (test* "instance-slot-ref oob" (test-error <error> #/out of bound/)
+         (instance-slot-ref z -1))
+  (test* "instance-slot-set! oob" (test-error <error> #/out of bound/)
+         (instance-slot-set! z 2 0))
+  )
+
 ;;----------------------------------------------------------------
 (test-section "next method")
 
@@ -274,6 +296,14 @@
        '(y-in (y*-in (x-in (t*-in fallback t*-out) x-out) y*-out) y-out)
        (nm (make <y>)))
 
+;; check if method-leaf is properly set
+(define-method nm-1 (obj) #f)
+(define-method nm-2 (obj) (next-method))
+
+(test* "method-leaf" '(#t #f)
+       (list (method-leaf? (car (slot-ref nm-1 'methods)))
+             (method-leaf? (car (slot-ref nm-2 'methods)))))
+
 ;;----------------------------------------------------------------
 (test-section "method sorting")
 
@@ -319,6 +349,78 @@
 (test* "method optarg (optarg)" '(a b 2 ()) (optarg (make <y>) 'a 'b))
 (test* "method optarg (optarg)" '(a b 4 (:z 4)) (optarg (make <y>) 'a 'b :z 4))
 
+;;----------------------------------------------------------------
+(test-section "accelerated dispatch")
+
+(define-generic acc-dis-1)
+(define-generic acc-dis-2)
+((with-module gauche.object generic-build-dispatcher!) acc-dis-2 0)
+
+(define-method acc-dis-1 ((a <top>)) 'top)
+(define-method acc-dis-2 ((a <top>)) 'top)
+
+(define *acc-dis-count* 25)
+
+(define-macro (gen-acc-dis-classes&methods)
+  (define (gen-1 n)
+    (let1 cl (symbol-append '<acc-dis- n '>)
+      `(begin
+         (define-class ,cl () ())
+         (define-method acc-dis-1 ((a ,cl)) `(,',n 0))
+         (define-method acc-dis-1 ((a ,cl) x) `(,',n 1))
+         (define-method acc-dis-1 ((a ,cl) x y) `(,',n 2))
+         (define-method acc-dis-1 ((a ,cl) x y . z) `(,',n 3))
+         (define-method acc-dis-2 ((a ,cl)) `(,',n 0))
+         (define-method acc-dis-2 ((a ,cl) x) `(,',n 1))
+         (define-method acc-dis-2 ((a ,cl) x y) `(,',n 2))
+         (define-method acc-dis-2 ((a ,cl) x y . z) `(,',n 3))
+         )))
+  `(begin ,@(append-map gen-1 (iota *acc-dis-count*))))
+
+(gen-acc-dis-classes&methods)
+
+(define-macro (acc-dis-classes)
+  `(list ,@(map (^n (symbol-append '<acc-dis- n '>)) (iota *acc-dis-count*))))
+
+(define (test-acc-dis name gf :optional (expect '()))
+  (test* name expect
+         (let1 r '()
+           (do ([cs (acc-dis-classes) (cdr cs)]
+                [n  0       (+ n 1)])
+               [(null? cs) r]
+             (do ([obj (make (car cs))]
+                  [k 0 (+ k 1)]
+                  [args '() (cons #f args)])
+                 [(= k 4)]
+               (let* ([expected `(,n ,k)]
+                      [actual (apply gf obj args)])
+                 (unless (equal? expected actual)
+                   (push! r `(:expected ,expected :actual ,actual)))))))))
+
+((with-module gauche.object generic-build-dispatcher!) acc-dis-1 0)
+(test-acc-dis "batch build" acc-dis-1)
+
+(test-acc-dis "incremental build" acc-dis-2)
+
+(define-method acc-dis-1 ((c <acc-dis-0>)) (cons '(0 0) (next-method)))
+(define-method acc-dis-1 ((c <acc-dis-1>)) 'redefined)
+
+(test-acc-dis "redefinition 1" acc-dis-1
+              '((:expected (1 0) :actual redefined)
+                (:expected (0 0) :actual ((0 0) . top))))
+
+(define-method acc-dis-1 ((c <acc-dis-0>)) '(0 0))
+(define-method acc-dis-1 ((c <acc-dis-1>) (x <integer>)) (cons '(1 1) (next-method)))
+(test-acc-dis "redefinition 2" acc-dis-1
+              '((:expected (1 0) :actual redefined)))
+(test* "redefiniton 2'" '((1 1) 1 1)
+       (acc-dis-1 (make <acc-dis-1>) 2))
+(define-method acc-dis-1 ((c <acc-dis-1>) (x <integer>)) 'leaf)
+(test* "redefiniton 3" '((1 1) . leaf)
+       (cons (acc-dis-1 (make <acc-dis-1>) #f)
+             (acc-dis-1 (make <acc-dis-1>) 2)))
+
+
 ;;----------------------------------------------------------------
 (test-section "module and accessor")
 
@@ -364,6 +466,31 @@
            (ma-s m 'ei)
            (slot-ref m 'a))))
 
+;; Another tricky interference between modules and implicit setter.
+;; The :accessor slot option may create the setter generic function
+;; with the weird name |setter of <getter-name>|.  The name isn't supposed
+;; to be visible from the user ever.  However, until 0.9.6, when the 
+;; MB.base below does not export the setter name, the derived class 
+;; couldn't find the setter gf and created a separate setter gf, making it
+;; incompatible with base class and derived class.
+(define-module MB.base
+  (export <mb-base> mb-a)
+  (define-class <mb-base> ()
+    ((a :accessor mb-a :init-value 0))))
+
+(define-module MB.derived
+  (import MB.base)
+  (export <mb-base> <mb-derived> mb-a)
+  (define-class <mb-defived> (<mb-base>) ()))
+
+(test* "module and implicit setter" 3
+       ;; This caused 'no applicable method for #<generic |setter of mb-a|>
+       ;; in 0.9.6 and before.
+       (with-module MB.derived
+         (let1 m (make <mb-base>)
+           (set! (mb-a m) 3)
+           (~ m 'a))))
+
 ;;----------------------------------------------------------------
 (test-section "class redefinition (part 1)")
 
@@ -421,7 +548,7 @@
 
 (test* "subclass redefinition <y> (slots)"
        '((a b c) (a b c x) (c d e a b) (c d e a b x))
-       (map (lambda (c) (map (lambda (s) (car s)) (class-slots c)))
+       (map (^c (map (lambda (s) (car s)) (class-slots c)))
             (list <x>-orig <x> <y>-orig <y>)))
 
 (test* "subclass redefinition <w> (links)"
@@ -434,7 +561,7 @@
 
 (test* "subclass redefinition <w> (slots)"
        '((e f c d a b) (e f c d a b x) (e f c d a b) (e f c d a b x))
-       (map (lambda (c) (map (lambda (s) (car s)) (class-slots c)))
+       (map (^c (map (lambda (s) (car s)) (class-slots c)))
             (list <w>-orig <w> <w2>-orig <w2>)))
 
 (test* "subclass redefinition (hierarchy)"
@@ -518,7 +645,7 @@
 
 (test* "instance update (y2) - cascade"
        '(5 7 9 3)
-       (map (lambda (s) (slot-ref y2 s)) '(a c e x)))
+       (map (^s (slot-ref y2 s)) '(a c e x)))
 
 (test* "redefine <y> without inheriting <x>" '(a e)
        (begin
@@ -534,7 +661,7 @@
              (memq <x> (ref <y> 'direct-supers))))
 
 (test* "instance update (y1)" '(0 4)
-       (map (lambda (s) (slot-ref y1 s)) '(a e)))
+       (map (^s (slot-ref y1 s)) '(a e)))
 
 (test* "subclass redefinition <w>" '(e f a)
        (map car (class-slots <w>)))
@@ -599,9 +726,6 @@
 (test* "object-compare" -1 (compare #\a #\b))
 (test* "object-compare" 0  (compare #\a #\a))
 (test* "object-compare" 1  (compare #\b #\a))
-(test* "object-compare" (test-error) (compare #\b 4))
-(test* "object-compare" (test-error) (compare "zzz" 4))
-(test* "object-compare" (test-error) (compare 2+i 3+i))
 (test* "object-compare" -1 (compare (make <cmp> :x 3) (make <cmp> :x 4)))
 (test* "object-compare" 0 (compare (make <cmp> :x 3) (make <cmp> :x 3)))
 (test* "object-compare" 1 (compare (make <cmp> :x 4) (make <cmp> :x 3)))
@@ -774,7 +898,7 @@
          (set! (a-of zz) 1)
          (set! (b-of zz) 2)
          (set! (c-of zz) 3)
-         (map (lambda (s) (slot-ref zz s)) '(a b c))))
+         (map (^s (slot-ref zz s)) '(a b c))))
 
 (define-class <uu> (<zz>)
   (d e f))
@@ -787,7 +911,7 @@
          (set! (d-of uu) 4)
          (set! (e-of uu) 5)
          (set! (f-of uu) 6)
-         (map (lambda (s) (slot-ref uu s)) '(a b c d e f))))
+         (map (^s (slot-ref uu s)) '(a b c d e f))))
 
 (define-class <vv> (<zz> <xx>)
   ())
@@ -797,7 +921,7 @@
          (set! (a-of vv) 1)
          (set! (b-of vv) 2)
          (set! (c-of vv) 3)
-         (map (lambda (s) (slot-ref vv s)) '(a b c))))
+         (map (^s (slot-ref vv s)) '(a b c))))
 (test "metaclass" '(<vv> <yy> <xx>)
       (lambda () (class-slot-ref <listing-class> 'classes)))
 
@@ -814,7 +938,7 @@
          (set! (d-of ww) 4)
          (set! (e-of ww) 5)
          (set! (f-of ww) 6)
-         (map (lambda (s) (slot-ref ww s)) '(a b c d e f))))
+         (map (^s (slot-ref ww s)) '(a b c d e f))))
 (test* "metaclass" '(<ww> <vv> <yy> <xx>)
        (class-slot-ref <listing-class> 'classes))
 
@@ -873,7 +997,7 @@
    (doc2 :init-value "no doc")
    (sub  :allocation :virtual
          :slot-set! (lambda (o v) #f)
-         :slot-ref  (lambda (o) (slot-ref o 'doc2)))))
+         :slot-ref  (^o (slot-ref o 'doc2)))))
 
 (test* "redefinition of metaclass" '("Doc doc" "no doc" "no doc")
        (list (slot-ref <xxx> 'doc)
@@ -892,6 +1016,22 @@
        '(A b C x Y z)
        (map (cut slot-ref *docu-sub* <>) '(a b c x y z)))
 
+;;----------------------------------------------------------------
+(test-section "metaclass/bound-slot")
+
+(use gauche.mop.bound-slot)
+(test-module 'gauche.mop.bound-slot)
+
+(define-class <bound-slot-A> (<bound-slot-mixin>)
+  ((a :init-keyword :a)
+   (b :init-keyword :b :init-value 1)))
+
+(test* "<bound-slot-A> check" (test-error)
+       (make <bound-slot-A>))
+
+(test* "<bound-slot-A> check" #t
+       (is-a? (make <bound-slot-A> :a 3) <bound-slot-A>))
+
 ;;----------------------------------------------------------------
 (test-section "metaclass/singleton")
 
@@ -1067,7 +1207,7 @@
 
 (test* "instance-pool-for-each" (list pool-x1 pool-z1 pool-x2 pool-z2)
        (let ((r '()))
-         (instance-pool-for-each <pool-x> (lambda (p) (push! r p)))
+         (instance-pool-for-each <pool-x> (^p (push! r p)))
          r)
        (cut lset= eq? <> <>))
 
@@ -1395,7 +1535,63 @@
 (test* "apply special path (apply->normal)" '(z c b a)
        (app-sp-path-test2 (make <app-sp-path2>) 'a 'b 'c))
 
+;;----------------------------------------------------------------
+(test-section "applicable?")
 
+;; The procedure applicable? depends on libobj's method-applicable-for-clases,
+;; so we test it here.
 
+(define-syntax atest*
+  (syntax-rules (=>)
+    [(_ fn (args => expected) ...)
+     (begin (test* (format "applicable? ~a~s" 'fn 'args)
+                   expected (applicable? fn . args))
+            ...)]))
+
+(atest* cons [() => #f] [(<top>) => #f] [(<top> <top>) => #t]
+        [(<top> <top> <top>) => #f])
+(atest* list [() => #t] [(<top>) => #t] [(<top> <top>) => #t])
+(atest* apply [() => #f] [(<top>) => #f] [(<top> <top>) => #t])
+
+(atest* (case-lambda ([a] 1) ([a b c] 2) ([a b c d e . f] 3))
+        [() => #f] [(<top>) => #t] [(<top> <top>) => #f]
+        [(<top> <top> <top>) => #t] [(<top> <top> <top> <top>) => #f]
+        [(<top> <top> <top> <top> <top>) => #t]
+        [(<top> <top> <top> <top> <top> <top>) => #t]
+        [(<top> <top> <top> <top> <top> <top> <top>) => #t])
+
+;; NB: The result may change if loaded module adds new method to 'ref'.
+(atest* ref [() => #f] [(<boolean>) => #f] [(<boolean> <top>) => #f]
+        [(<top>) => #f] [(<top> <symbol>) => #t] [(<top> <symbol> <top>) => #t]
+        [(<vector>) => #f] [(<vector> <integer>) => #t]
+        [(<vector> <vector> <integer> <integer>) => #f])
+
+(atest* #/a/ [() => #f] [(<string>) => #t] [(<integer>) => #f])
+
+;;----------------------------------------------------------------
+(test-section "expansion hygiene")
+
+;; define-class etc. are expanded in libobj.scm.  we should make sure that
+;; the expansion won't break even user env doesn't import gauche native stuff.
+
+(define-module object-hygiene-1
+  (define-syntax defclass (with-module gauche define-class))
+  (define-syntax defmethod (with-module gauche define-method))
+  (define ref (with-module gauche ref))
+  (define-syntax quote (with-module gauche quote))
+  (export <object-hygiene-1> object-hygiene-1-c)
+  (extend)
+  (defclass <object-hygiene-1> ()
+    (a b (c :init-keyword ':c)))
+  (defmethod object-hygiene-1-c ((self <object-hygiene-1>))
+    (ref self 'c)))
+
+(import object-hygiene-1)
+(test* "object-hygiene-1" '(#t (a) (b) (c :init-keyword :c))
+       (cons (is-a? <object-hygiene-1> <class>)
+             (class-slots <object-hygiene-1>)))
+(test* "object-hygiene-1-c" 4
+       (let1 z (make <object-hygiene-1> :c 4)
+         (object-hygiene-1-c z)))
+  
 (test-end)
-
