@@ -26,7 +26,9 @@
 
 (prim-test "when" 4          (lambda ()  (when 3 5 4)))
 (prim-test "when" (undefined)    (lambda ()  (when #f 5 4)))
+(prim-test "when" (undefined)    (lambda ()  (when #f unbound-var)))
 (prim-test "unless" (undefined)  (lambda ()  (unless 3 5 4)))
+(prim-test "unless" (undefined)  (lambda ()  (unless #t unbound-var)))
 (prim-test "unless" 4        (lambda ()  (unless #f 5 4)))
 
 (prim-test "cond" (undefined)  (lambda ()  (cond (#f 2))))
@@ -43,6 +45,9 @@
 
 (prim-test "case" #t (lambda ()  (case (+ 2 3) ((1 3 5 7 9) #t) ((0 2 4 6 8) #f))))
 (prim-test "case" #t (lambda () (undefined? (case 1 ((2 3) #t)))))
+(prim-test "case" #t (lambda () (case 1 (() #f) ((1) #t))))
+(prim-test "case" #t (lambda () (case 1 (() #f) (else #t))))
+(prim-test "case" #t (lambda () (undefined? (case 1 (() #t)))))
 (prim-test "case (srfi-87)" 0 (lambda () (case (+ 2 3) ((1 3 5) 0) (else => values))))
 (prim-test "case (srfi-87)" 6 (lambda () (case (+ 2 3) ((1 3 5) => (cut + 1 <>)) (else => values))))
 (prim-test "case (srfi-87)" 5 (lambda () (case (+ 2 3) ((2 4 6) 0) (else => values))))
@@ -91,26 +96,28 @@
 ;;----------------------------------------------------------------
 (test-section "application")
 
-(prim-test "apply" '(1 2 3) (lambda ()  (apply list '(1 2 3))))
-(prim-test "apply" '(2 3 4) (lambda ()  (apply list 2 '(3 4))))
-(prim-test "apply" '(3 4 5) (lambda ()  (apply list 3 4 '(5))))
-(prim-test "apply" '(4 5 6) (lambda ()  (apply list 4 5 6 '())))
+(define Apply apply) ; avoid inline expansion
 
-(prim-test "apply^2" '() (lambda () (apply apply list '() '())))
-(prim-test "apply^2" '() (lambda () (apply apply list '(()))))
-(prim-test "apply^2" '(1 . 2) (lambda () (apply apply cons '((1 2)))))
-(prim-test "apply^2" '(3 . 4) (lambda () (apply apply cons 3 '((4)))))
-(prim-test "apply^2" '(5 . 6) (lambda () (apply apply (list cons 5 '(6)))))
+(prim-test "apply" '(1 2 3) (lambda ()  (Apply list '(1 2 3))))
+(prim-test "apply" '(2 3 4) (lambda ()  (Apply list 2 '(3 4))))
+(prim-test "apply" '(3 4 5) (lambda ()  (Apply list 3 4 '(5))))
+(prim-test "apply" '(4 5 6) (lambda ()  (Apply list 4 5 6 '())))
+
+(prim-test "apply^2" '() (lambda () (Apply Apply list '() '())))
+(prim-test "Apply^2" '() (lambda () (Apply Apply list '(()))))
+(prim-test "apply^2" '(1 . 2) (lambda () (Apply Apply cons '((1 2)))))
+(prim-test "apply^2" '(3 . 4) (lambda () (Apply Apply cons 3 '((4)))))
+(prim-test "apply^2" '(5 . 6) (lambda () (Apply Apply (list cons 5 '(6)))))
                                           
 
-(prim-test "apply" '(6 7 8) (lambda ()  (apply apply (list list 6 7 '(8)))))
+(prim-test "apply" '(6 7 8) (lambda ()  (Apply Apply (list list 6 7 '(8)))))
 
 
 ;; This tests 'unfolding' path in ADJUST_ARGUMENT_FRAME.
 (prim-test "apply, copying args" '(1 2 3)
            (lambda ()
              (let ((orig (list 1 2 3)))
-               (let ((new (apply list orig)))
+               (let ((new (Apply list orig)))
                  (set-car! (cdr new) '100)
                  orig))))
 
@@ -118,7 +125,7 @@
 (prim-test "apply, copying args" '(2 3)
            (lambda ()
              (let ((orig (list 2 3)))
-               (let ((new (apply list 1 orig)))
+               (let ((new (Apply list 1 orig)))
                  (set-car! (cdr new) '100)
                  orig))))
 
@@ -128,8 +135,8 @@
     (define (a . args)
       (receive x args
         (cons x x)
-        (apply values x))
-      (apply format args))
+        (Apply values x))
+      (Apply format args))
     (define (b bar)
       (a "~a" bar))
     (b 1)
@@ -285,6 +292,13 @@
 (prim-test "qqq#"  '(1 `#(1 ,(a b x) ,(y c d)))
            (lambda () `(1 `#(1 ,(,@quasi2 x) ,(y ,@quasi3)))))
 
+(prim-test "qq-hygiene 0" '(2 1)
+           (lambda () (let ((quasiquote reverse)) `(list 1 2))))
+(prim-test "qq-hygiene 1" '(,(+ 1 2))
+           (lambda () (let ((unquote 3)) `(,(+ 1 2)))))
+(prim-test "qq-hygiene 2" '(,@(+ 1 2))
+           (lambda () (let ((unquote-splicing 3)) `(,@(+ 1 2)))))
+
 ;;----------------------------------------------------------------
 (test-section "multiple values")
 
@@ -348,7 +362,24 @@
         (with-error-handler
             (lambda (e) foo)
           (lambda ()
-            (eval '(apply car foo '()) (find-module 'primsyn.test))))))
+            (eval '(Apply car foo '()) (find-module 'primsyn.test))))))
+
+;;----------------------------------------------------------------
+(test-section "max literal arguments")
+
+;; Fix this after we have separate compile-error condition.
+(define (test-max-literal-args msg expr)
+  (prim-test (string-append "max literal arguments for " msg)
+             'caught 
+             (lambda ()
+               (with-error-handler (lambda (e) 'caught)
+                 (lambda () (eval expr (interaction-environment)))))))
+
+(test-max-literal-args "inliner" `(list ,@(make-list 10000 #f)))
+(test-max-literal-args "global proc" `(make ,@(make-list 10000 #f)))
+(test-max-literal-args "local proc"
+                       `(let ((foo (lambda x x)))
+                          (foo ,@(make-list 10000 #f))))
 
 ;;----------------------------------------------------------------
 (test-section "local procedure optimization")
@@ -406,135 +437,6 @@
                        (error "zz"))
                     (interaction-environment)))))
 
-;;----------------------------------------------------------------
-(test-section "lazy, delay & force")
-
-(prim-test "simple delay" 3
-      (lambda () (force (delay (+ 1 2)))))
-
-(prim-test "delay w/state" 3
-      (lambda ()
-        (let ((x 9))
-          (let ((d (delay (/ x 3))))
-            (force d)
-            (set! x 99)
-            (force d)))))
-
-(prim-test "delay recursive" 6  ;; R5RS 6.4
-      (lambda ()
-        (letrec ((count 0)
-                 (x 5)
-                 (p (delay (begin (set! count (+ count 1))
-                                  (if (> count x)
-                                      count
-                                      (force p))))))
-          (force p)
-          (set! x 10)
-          (force p))))
-
-;; check to see the compiler does the right thing about expanding
-;; (delay x) to (lazy (eager x)).
-(prim-test "delay compilation" 3
-           (lambda ()
-             (force 
-              (let ((lazy list)
-                    (eager list))
-               (delay (force 3))))))
-
-;; srfi-45 test suite
-(prim-test "memoize 1" 1
-           (lambda ()
-             (let ((count 0))
-               (define s (delay (begin (set! count (+ count 1)) 1)))
-               (force s)
-               (force s)
-               count)))
-
-(prim-test "memoize 2" 1
-           (lambda ()
-             (let ((count 0))
-               (define s (delay (begin (set! count (+ count 1)) 1)))
-               (+ (force s) (force s))
-               count)))
-
-(prim-test "memoize 3" 1  ;; (Alejandro Forero Cuervo)
-           (lambda ()
-             (let ((count 0))
-               (let* ((r (delay (begin (set! count (+ count 1)) 1)))
-                      (s (lazy r))
-                      (t (lazy s)))
-                 (force t)
-                 (force r)
-                 count))))
-
-(prim-test "memoize 4" 5  ;; stream memoization
-           (lambda ()
-             (let ((count 0))
-               (define (stream-drop s index)
-                 (lazy
-                  (if (zero? index)
-                    s
-                    (stream-drop (cdr (force s)) (- index 1)))))
-               (define (ones)
-                 (delay (begin
-                          (set! count (+ count 1))
-                          (cons 1 (ones)))))
-               (let ((s (ones)))
-                 (car (force (stream-drop s 4)))
-                 (car (force (stream-drop s 4)))
-                 count))))
-
-(prim-test "reentrancy 1" 'second       ;; see srfi-40 post-discussion
-           (lambda ()
-             (define f
-               (let ((first? #t))
-                 (delay
-                   (if first?
-                     (begin
-                       (set! first? #f)
-                       (force f))
-                     'second))))
-             (force f)))
-
-(prim-test "reentrancy 2" '(5 0 10) ;; (John Shutt)
-           (lambda ()
-             (define q
-               (let ((count 5))
-                 (define (get-count) count)
-                 (define p (delay (if (<= count 0)
-                                    count
-                                    (begin (set! count (- count 1))
-                                           (force p)
-                                           (set! count (+ count 2))
-                                           count))))
-                 (list get-count p)))
-             (let* ((get-count (car q))
-                    (p (cadr q))
-                    (a (get-count))
-                    (b (force p))
-                    (c (get-count)))
-               (list a b c))))
-
-;; This leak test takes long time, so we exclude it by default.
-;; If you run this on pre-0.8.6 Gauche (with replacing lazy by delay+force),
-;; it'll bust the memory.
-;(prim-test "leak 1" 10000000
-;           (lambda ()
-;             (define (stream-filter p? s)
-;               (lazy
-;                (let ((lis (force s)))
-;                  (if (null? lis)
-;                    (delay '())
-;                    (let ((h (car lis))
-;                          (t (cdr lis)))
-;                      (if (p? h)
-;                        (delay (cons h (stream-filter p? t)))
-;                        (stream-filter p? t)))))))
-;             (define (from n)
-;               (delay (cons n (from (+ n 1)))))
-;             (car (force (stream-filter (lambda (n) (= n 10000000))
-;                                        (from 0))))))
-
 ;;----------------------------------------------------------------
 (test-section "optimized frames")
 
@@ -605,5 +507,27 @@
               ((> x 3) x)
             #f))))
 
+;;----------------------------------------------------------------
+(test-section "letrec and letrec*")
+
+(prim-test "letrec reordering" '((1 3) . (2 3 1))
+           (lambda ()
+             (let ((r '()))
+               (cons (letrec ((a (begin (set! r (cons 1 r)) 1))
+                              (b (begin (set! r (cons 2 r)) 2))
+                              (c (begin (set! r (cons 3 r)) 3)))
+                       (list a c))
+                     r))))
+
+(prim-test "letrec* non-reordering" '((1 3) . (3 2 1))
+           (lambda ()
+             (let ((r '()))
+               (cons (letrec* ((a (begin (set! r (cons 1 r)) 1))
+                               (b (begin (set! r (cons 2 r)) 2))
+                               (c (begin (set! r (cons 3 r)) 3)))
+                       (list a c))
+                     r))))
+
+
 (test-end)
 
