@@ -331,7 +331,7 @@
 
 ;; let-keywords* combined with syntax rules
 (define-syntax lambda++
-  (syntax-rules ()
+  (syntax-rules (:key)
     ((lambda++ "sub" () (margs ...) kargs . body)
      (lambda (margs ... . rest)
        (let-keywords* rest kargs
@@ -457,6 +457,21 @@
                (1 ,(undefined) ,(undefined) 0 1 ())
                ,(test-error)))
 
+;; If keyword is symbol, we should allow them as formals.
+(test-section "rnrs lambda formals")
+
+(define-syntax rnrs-lambda (with-module null lambda))
+(when (symbol? :a)
+  (test* "rnrs lambda formals" '(x y)
+         ((rnrs-lambda (:a :b) (list :a :b)) 'x 'y))
+  (test* "rnrs lambda formals" '(x z)
+         ((rnrs-lambda (a :key b) (list a b)) 'x 'y 'z))
+  (test* "rnrs lambda formals" '(x (z))
+         ((rnrs-lambda (a :optional . b) (list a b)) 'x 'y 'z))
+  (test* "rnrs lambda formals" '(x y z)
+         ((rnrs-lambda :rest :rest) 'x 'y 'z))
+  )
+
 ;;-----------------------------------------------------------------------
 ;; case-lambda
 (test-section "case-lambda")
@@ -465,7 +480,7 @@
   (for-each (lambda (arg&exp)
               (let [(args (car arg&exp))
                     (exp  (cdr arg&exp))]
-                (test* #`"case-lambda ,sig ,(length args)" exp
+                (test* #"case-lambda ~sig ~(length args)" exp
                        (apply fn args))))
             arg&exp))
 
@@ -524,4 +539,87 @@
                   `[(1) . (foo 1 ())]
                   `[(1 2) . (foo 1 (2))])
 
+;;-----------------------------------------------------------------------
+;; generator-*
+(test-section "generator-*")
+
+(define (gen-1arg-test name input expect thunk :optional (wrap identity))
+  (test* (format "~a, 1 arg" name) expect
+         (wrap (with-input-from-string input thunk))))
+
+(gen-1arg-test "generator-fold" "a b c d e" '(e d c b a . z)
+               (cut generator-fold cons 'z read))
+(gen-1arg-test "generator-fold-right" "a b c d e" '(a b c d e . z)
+               (cut generator-fold-right cons 'z read))
+(gen-1arg-test "generator-map" "a b c d e" '("a" "b" "c" "d" "e")
+               (cut generator-map symbol->string read))
+(let1 r '()
+  (gen-1arg-test "generator-for-each" "a b c d e" '(e d c b a)
+                 (cut generator-for-each (^v (push! r v)) read)
+                 (^_ r)))
+(gen-1arg-test "generator-find" "2 4 0 8 3 6 9" 3
+               (cut generator-find odd? read))
+
+(define (gen-2arg-test name input1 input2 expect proc :optional (wrap identity))
+  (test* (format "~a, 2 args" name) expect
+         (wrap (call-with-input-string input1
+                 (^x (call-with-input-string input2
+                       (^y (proc x y))))))))
+
+(gen-2arg-test "generator-fold" "a b c d e f" "1 2 3 4 5"
+               '(e 5 (d 4 (c 3 (b 2 (a 1 z)))))
+               (^[x y] (generator-fold list 'z (cut read x) (cut read y))))
+(gen-2arg-test "generator-fold-right" "a b c d e f" "1 2 3 4 5"
+               '(a 1 (b 2 (c 3 (d 4 (e 5 z)))))
+               (^[x y] (generator-fold-right list 'z (cut read x) (cut read y))))
+(gen-2arg-test "generator-map" "1 2 3 4 5" "6 7 8 9"
+               '(7 9 11 13)
+               (^[x y] (generator-map + (cut read x) (cut read y))))
+(let1 r '()
+  (gen-2arg-test "generator-for-each" "1 2 3 4 5" "6 7 8 9"
+                 '(13 11 9 7)
+                 (^[x y] (generator-for-each (^[a b] (push! r (+ a b)))
+                                             (cut read x) (cut read y)))
+                 (^_ r)))
+
+;;-----------------------------------------------------------------------
+;; large env frame
+;; https://github.com/shirok/Gauche/issues/487
+(test-section "large env frame")
+
+(define-macro (biglet n)
+  (let1 syms (map (^_ (gensym)) (iota n))
+    `(let (,@(map (^s `(,s #f)) syms))
+       ,@(map (^[s i] `(set! ,s ,i)) syms (iota n))
+       (list ,@syms))))
+
+(test* "bigenv 1030" (iota 1030)
+       (biglet 1030))
+
+;; Test LREF-VAL0-NUMADD2 path
+(define-macro (bigsum n)
+  (let1 syms (map (^_ (gensym)) (iota n))
+    `(lambda ,syms (+ ,@syms))))
+
+(define bigsum-proc (bigsum 1029))
+
+(test* "bigsum 1029" (apply + (iota 1029))
+       (apply bigsum-proc (iota 1029)))
+
+;; Test LREF-VAL0-BNGT etc. path
+(define-macro (bigcond op n)
+  (let1 syms (map (^_ (gensym)) (iota n))
+    `(lambda (k ,@syms)
+       (cond ,@(map (^s `((,op k ,s) ,s)) syms)))))
+
+(define bigcond<-proc (bigcond < 1028))
+
+(test* "bigcond< 1028" 1025
+       (apply bigcond<-proc 1024 (iota 1028)))
+
+(define bigcond=-proc (bigcond = 1028))
+
+(test* "bigcond= 1028" 1027
+       (apply bigcond=-proc 1027 (iota 1028)))
+
 (test-end)
